
/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/**
 *  @class
 *  @name ADUtils
 *
 *  @since AdLib 0.1
 */

/**
 *  Indicates whether the platform is an iPad.
 *  @constant
 *  @type bool
 */
const iPad = (navigator.platform == 'iPad');
/**
 *  Indicates whether the platform supports touches.
 *  @constant
 *  @type bool
 */
const ADSupportsTouches = ('createTouch' in document);
/**
 *  The interaction start event name, returns <code>touchstart</code> if the platform supports touches
 *  and <code>mousedown</code> if it does not.
 *  @constant
 *  @type String
 */
const ADStartEvent = ADSupportsTouches ? 'touchstart' : 'mousedown';
/**
 *  The interaction move event name, returns <code>touchmove</code> if the platform supports touches
 *  and <code>mousemove</code> if it does not.
 *  @constant
 *  @type String
 */
const ADMoveEvent = ADSupportsTouches ? 'touchmove' : 'mousemove';
/**
 *  The interaction end event name, returns <code>touchend</code> if the platform supports touches
 *  and <code>mouseup</code> if it does not.
 *  @constant
 *  @type String
 */
const ADEndEvent = ADSupportsTouches ? 'touchend' : 'mouseup';

function ADUtils () {
};

ADUtils.assetsPath = '';

/* ==================== TRANSFORMS SHORTHANDS ==================== */

/**
 *  Prints a <code>translate3d()</code> command that can be used as input for a <code>-webkit-transform</code> property.
 *  
 *  @param {int} tx The x coordinate for the translation.
 *  @param {int} ty The y coordinate for the translation
 *
 *  @returns {String} The <code>translate3d()</code> command
 */
ADUtils.t = function (tx, ty) {
  return ADUtils.t3d(tx, ty, 0);
};

/**
 *  Prints a <code>translate3d()</code> command that can be used as input for a <code>-webkit-transform</code> property.
 *  
 *  @param {int} tx The x coordinate for the translation.
 *  @param {int} ty The y coordinate for the translation
 *  @param {int} tz The z coordinate for the translation
 *
 *  @returns {String} The <code>translate3d()</code> command
 */
ADUtils.t3d = function (tx, ty, tz) {
  return 'translate3d(' + tx + 'px, ' + ty + 'px, ' + tz + 'px)';
};

/**
 *  Prints a <code>rotate3d()</code> command that can be used as input for a <code>-webkit-transform</code> property.
 *  
 *  @param {int} x The x component of the rotation vector
 *  @param {int} y The y component of the rotation vector
 *  @param {int} z The z component of the rotation vector
 *  @param {int} angle The angle in radians for the rotation
 *
 *  @returns {String} The <code>rotate3d()</code> command
 */
ADUtils.r3d = function (x, y, z, angle) {
  return 'rotate3d(' + x + ', ' + y + ', ' + z + ', ' + angle + 'rad)';
};

/**
 *  Creates a CSS string representation for a number in pixels.
 *  
 *  @param {number} value The value to be converted.
 *
 *  @returns {String} A CSS string representation for <code>value</code> in pixels.
 */
ADUtils.px = function (value) {
  return value + 'px';
};

/* ==================== Angles ==================== */

/**
 *  Converts an angle from degrees to radians.
 *  
 *  @param {number} value The angle in degrees to be converted.
 *
 *  @returns {numver} The converted angle in radians.
 */
ADUtils.degreesToRadians = function (degreesAngle) {
  return (degreesAngle / 360) * (Math.PI * 2);
};

/**
 *  Converts an angle from radians to degrees.
 *  
 *  @param {number} value The angle in radians to be converted.
 *
 *  @returns {numver} The converted angle in degrees.
 */
ADUtils.radiansToDegrees = function (radiansAngle) {
  return (radiansAngle / (Math.PI * 2)) * 360;
};

/* ==================== Array ==================== */

/**
 *  Copies all properties from one object onto another.
 *  
 *  @param {Object} sourceObject The object from which we will copy properties.
 *  @param {Object} targetObject The array onto which we will copy properties.
 */
ADUtils.copyPropertiesFromSourceToTarget = function (source, target) {
  for (var property in source) {
    target[property] = source[property];
  }
};

/* ==================== Delegates ==================== */

/**
 *  Indicates whether an object is a <code>Function</code>.
 *  
 *  @param {Object} object The object purported to be a <code>Function</code>.
 *
 *  @returns {bool} Whether the object is a <code>Function</code>.
 */
ADUtils.objectIsFunction = function (object) {
  return (typeof object == 'function');
};

/**
 *  Indicates whether an object is <code>undefined</code>.
 *  
 *  @param {Object} object The object purported to be <code>undefined</code>.
 *
 *  @returns {bool} Whether the object is <code>undefined</code>.
 */
ADUtils.objectIsUndefined = function (object) {
  return (object === undefined);
};

/**
 *  Indicates whether an object is a string literal or a <code>String</code> instance.
 *  
 *  @param {Object} object The object purported to be a string literal or a <code>String</code> instance.
 *
 *  @returns {bool} Whether the object is a string literal or a <code>String</code> instance.
 */
ADUtils.objectIsString = function (object) {
  return (typeof object == 'string' || object instanceof String);
};

/**
 *  Indicates whether an object is an <code>Array</code>.
 *  
 *  @param {Object} object The object purported to be an <code>Array</code>.
 *
 *  @returns {bool} Whether the object is an <code>Array</code>.
 */
ADUtils.objectIsArray = function (object) {
  return (object instanceof Array);
};

/**
 *  Indicates whether an object implements a given method, useful to check if a delegate
 *  object implements a given delegate method.
 *  
 *  @param {Object} object The object purported to implement a given method.
 *  @param {String} methodNameAsString The method name as a <code>String</code>.
 *
 *  @returns {bool} Whether the object implements the given method.
 */
ADUtils.objectHasMethod = function (object, methodNameAsString) {
  return (  object !== null &&
            !this.objectIsUndefined(object[methodNameAsString]) &&
            this.objectIsFunction(object[methodNameAsString])
         );
};

/* ==================== SCROLLING OVERRIDE ==================== */

/**
 *  Allows to over-ride scrolling when the active element can handle touches on its own,
 *  for instance a native HTML form control. Calling this method will override the default
 *  <code>UIWebView</code> scrolling as well as the custom scrolling provided by the
 *  {@link ADScrollView}, if the <code>event</code> provided is at a processing phase prior
 *  to bubbling to the parent scroll view. The scrolling behavior will be restored
 *  automatically upon a <code>touchend</code> or <code>touchcancel</code> event, or
 *  manually by calling the {@link ADUtils.restoreScrolling} method.
 *  
 *  @param {TouchEvent} event The event.
 */
ADUtils.disableScrolling = function (event) {
  event.stopPropagation();
  //
  window.addEventListener('touchmove', ADUtils.preventEventDefault, true);
  window.addEventListener('touchend', ADUtils.restoreScrollingBehavior, true);
  window.addEventListener('touchcancel', ADUtils.restoreScrollingBehavior, true);
};

ADUtils.preventEventDefault = function (event) {
  event.preventDefault();
};

/**
 *  Restores default scrolling behavior following a call to {@link ADUtils.disableScrolling}.
 */
ADUtils.restoreScrolling = function () {
  window.removeEventListener('touchmove', ADUtils.preventEventDefault, true);
  window.removeEventListener('touchend', ADUtils.restoreScrollingBehavior, true);
  window.removeEventListener('touchcancel', ADUtils.restoreScrollingBehavior, true);
};

/* ==================== Event ==================== */

ADUtils.createUIEvent = function (eventType, event) {
  return ADSupportsTouches ? this.createEventWithTouch(eventType, event) : this.createEventWithMouse(eventType, event);
};

ADUtils.createEventWithTouch = function (eventType, touchEvent) {
  // create a new type of touch event
  var event = document.createEvent('TouchEvent');
  // init it with our custom type and the cloned properties of the original touch
  event.initTouchEvent( eventType, touchEvent.bubbles, touchEvent.cancelable, window,
                        touchEvent.detail, touchEvent.screenX, touchEvent.screenY,
                        touchEvent.clientX, touchEvent.clientY, touchEvent.ctrlKey,
                        touchEvent.altKey, touchEvent.shiftKey, touchEvent.metaKey,
                        touchEvent.touches, touchEvent.targetTouches, touchEvent.changedTouches,
                        touchEvent.scale, touchEvent.rotation);
  // return our custom event
  return event;
};

ADUtils.createEventWithMouse = function (eventType, mouseEvent) {
  // create a new type of touch event
  var event = document.createEvent('MouseEvent');
  // init it with our custom type and the cloned properties of the original touch
  event.initMouseEvent( eventType, mouseEvent.bubbles, mouseEvent.cancelable, document.defaultView,
                        mouseEvent.detail, mouseEvent.screenX, mouseEvent.screenY,
                        mouseEvent.clientX, mouseEvent.clientY, mouseEvent.ctrlKey,
                        mouseEvent.altKey, mouseEvent.shiftKey, mouseEvent.metaKey,
                        mouseEvent.metaKey, mouseEvent.button, mouseEvent.relatedTarget);
  // return our custom event
  return event;
};

/* ==================== INIT ==================== */

ADUtils.init = function () {
  // turn on the iPad class if we detect it
  if (iPad) {
    document.body.addClassName('iPad');
  }
  // locate the AdLib CSS file to get the path for assets
  for (var i = 0; i < document.styleSheets.length; i++) {
    var sheet = document.styleSheets[i];
    var index = sheet.href ? sheet.href.indexOf('dist/AdLib') : -1;
    if (index != -1) {
      ADUtils.assetsPath = sheet.href.substring(0, index) + 'assets/';
      break;
    }
  }
};

/**
 *  Preloads an image located in the AdLib <code>assets/</code> directory.
 *  
 *  @param {String} path The path relative to the <code>assets/</code> directory.
 */
ADUtils.preloadImageAsset = function (path) {
  new Image().src = ADUtils.assetsPath + path;
};

/**
 *  Sets up the .displayNames for all functions defined on the specified class, including its prototype.
 *  
 *  @param {Object} class The class.
 *  @param {String} className The class name as a string, in case it can not be derived from <code>class</code>. Optional.
 */
ADUtils.setupDisplayNames = function (object, className) {
  var class_name = className || object.name;
  for (var i in object) {
    // make sure we don't touch properties that were synthetized
    if (object.__lookupGetter__(i)) {
      continue;
    }
    var prop = object[i];
    if (ADUtils.objectIsFunction(prop)) {
      prop.displayName = ADUtils.createDisplayName(class_name, i);
    }
  }
  for (var i in object.prototype) {
    // make sure we don't touch properties that were synthetized
    if (object.prototype.__lookupGetter__(i)) {
      continue;
    }
    var prop = object.prototype[i];
    if (ADUtils.objectIsFunction(prop)) {
      prop.displayName = ADUtils.createDisplayName(class_name, i);
    }
  }
};

ADUtils.createDisplayName = function (className, methodName) {
  return className + '.' + methodName + '()';
};

/**
 *  Creates a DOM node object from an HTML fragment as a string.
 *  
 *  @param {String} text The HTML markup as a string.
 *
 *  @returns {Node} The DOM node.
 */
ADUtils.createNodeFromString = function (text) {
  var new_document = document.implementation.createHTMLDocument();
  new_document.write(text);
  return document.importNode(new_document.body.firstElementChild, true);
};

window.addEventListener('DOMContentLoaded', ADUtils.init, false);

ADUtils.setupDisplayNames(ADUtils, 'ADUtils');

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

var ADEventTriage = {};

ADEventTriage.handleEvent = function (event) {
  // first call the super class method that we may be overriding otherwise
  if (this instanceof ADObject) {
    this.callSuper(event);
  }
  //
  var type = event.type;
  var methodName = 'handle' + type.charAt(0).toUpperCase() + type.substr(1);
  if (ADUtils.objectHasMethod(this, methodName)) {
    this[methodName](event);
  }
};

ADUtils.setupDisplayNames(ADEventTriage, 'ADEventTriage');

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

var ADEventTarget = {};

/* ==================== Public Methods ==================== */

/**
 *  Registers an event listener, depending on the <code>useCapture</code> parameter, on the capture phase
 *  of the DOM event flow or its target and bubbling phases.
 *
 *  @param {String} type The name of the event.
 *  @param {EventListener} listener The listener parameter takes an object implemented by the user which implements
 *  the <a href="http://www.w3.org/TR/DOM-Level-3-Events/events.html#Events-EventListener"><code>EventListener</code></a>
 *  interface and contains the method to be called when the event occurs. Can be a simple JavaScript <code>Function</code> object.
 *  @param {bool} useCapture If <code>true</code>, indicates that the user wishes to add the event listener for the capture phase
 *  only, i.e. this event listener will not be triggered during the target and bubbling phases. If <code>false</code>, the event
 *  listener will only be triggered during the target and bubbling phases.
 */
ADEventTarget.addEventListener = function (type, listener, useCapture) {
  this.eventTarget.addEventListener(type, listener, useCapture);
};

/**
 *  Removes an event listener. Calling <code>removeEventListener</code> with arguments which do not identify any
 *  currently registered <code>EventListener</code> has no effect.
 *
 *  @param {String} type The name of the event.
 *  @param {EventListener} listener The <a href="http://www.w3.org/TR/DOM-Level-3-Events/events.html#Events-EventListener"><code>EventListener</code></a>
 *  to be removed. Can be a simple JavaScript <code>Function</code> object.
 *  @param {bool} useCapture Specifies whether the <code>EventListener</code> being removed was registered for the
 *  capture phase or not. If a listener was registered twice, once for the capture phase and once for the target and
 *  bubbling phases, each must be removed separately. Removal of an event listener registered for the capture phase
 *  does not affect the same event listener registered for the target and bubbling phases, and vice versa.
 */
ADEventTarget.removeEventListener = function (type, listener, useCapture) {
  this.eventTarget.removeEventListener(type, listener, useCapture);
};

/**
 *  Dispatches an event into the implementation's event model.
 *
 *  @param {Event} event The event to be dispatched.
 */
ADEventTarget.dispatchEvent = function (event) {
  this.eventTarget.dispatchEvent(event);
};

ADUtils.setupDisplayNames(ADEventTarget, 'ADEventTarget');

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

var ADPropertyTriage = {};

ADPropertyTriage.handlePropertyChange = function (object, propertyName) {
  var methodName = 'handle' + propertyName.charAt(0).toUpperCase() + propertyName.substr(1) + 'Change';
  if (ADUtils.objectHasMethod(this, methodName)) {
    this[methodName](object);
  }
};

ADUtils.setupDisplayNames(ADPropertyTriage, 'ADPropertyTriage');

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/**
 *  @class
 *  @name Element
 *  @description Extensions to the DOM Core <a href="http://www.w3.org/TR/DOM-Level-3-Core/core.html#ID-745549614"><code>Element</code></a> interface.
 *  @since AdLib 0.1
 */

/* ==================== Element Extensions ==================== */

/**
 *  Indicates whether the element has a given class name within its <code>class</code> attribute.
 *
 *  @param {String} className The CSS class name.
 *  @returns {bool} Whether the element has this class name within its <code>class</code> attribute.
 *  @memberOf Element.prototype
 */
Element.prototype.hasClassName = function (className) {
  return new RegExp('(?:^|\\s+)' + className + '(?:\\s+|$)').test(this.className);
};

/**
 *  Adds the given class name to the element's <code>class</code> attribute if it's not already there.
 *
 *  @param {String} className The CSS class name.
 *  @memberOf Element.prototype
 */
Element.prototype.addClassName = function (className) {
  if (!this.hasClassName(className)) {
    this.className = [this.className, className].join(' ');
  }
};

/**
 *  Removes the given class name from the element's <code>class</code> attribute if it's there.
 *
 *  @param {String} className The CSS class name.
 *  @memberOf Element.prototype
 */
Element.prototype.removeClassName = function (className) {
  if (this.hasClassName(className)) {
    var curClasses = this.className;
    this.className = curClasses.replace(new RegExp('(?:^|\\s+)' + className + '(?:\\s+|$)', 'g'), ' ');
  }
};

/**
 *  Adds the given class name to the element's <code>class</code> attribute if it's not there, or removes it if it's already set.
 *
 *  @param {String} className The CSS class name.
 *  @memberOf Element.prototype
 */
Element.prototype.toggleClassName = function (className) {
  this[this.hasClassName(className) ? 'removeClassName' : 'addClassName'](className);
};

ADUtils.setupDisplayNames(Element, 'Element');

/**
 *  @class
 *  @name Node
 *  @description Extensions to the DOM Core <a href="http://www.w3.org/TR/DOM-Level-3-Core/core.html#ID-1950641247"><code>Node</code></a> interface.
 *  @since AdLib 0.1
 */

/**
 *  Finds the nearest parent view for this node.
 *
 *  @returns {ADView} The nearest parent view, or <code>null</code> if the node is outside of the view hierarchy.
 */
Node.prototype.getNearestView = function () {
  var element = this;
  while (ADUtils.objectIsUndefined(element._view) && element.parentNode) {
    element = element.parentNode;
  }
  return (ADUtils.objectIsUndefined(element._view)) ? null : element._view;
};

ADUtils.setupDisplayNames(Node, 'Node');

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/**
 *  @class
 *
 *  <p>The ADClass function will process a constructor function and set up its inheritance chain, synthetize
 *  a number of its instance properties and mix in additional capabilities based properties defined on
 *  that function. The supported properties are:</p>
 *
 *  <ul>
 *    <li><code>inherits</code>  the superclass</li>
 *    <li><code>includes</code>  an Array of modules to pull in</li>
 *    <li><code>synthetizes</code>  an Array of properties to synthetize</li>
 *  </ul>
 *
 *  <p>For instance, consider the following class declaration:</p>
 *
 *  <pre class="example">
// extends MySuperClass
MyClass.inherits = MySuperClass;

// mixes in the ADEventTriage and ADPropertyTriage modules
MyClass.includes = [ADEventTriage, ADPropertyTriage]; 

// synthetizes the .foo and .bar properties
MyClass.synthetizes = ['foo', 'bar'];

function MyClass () {
  // constructor code goes here
};

// process properties set up on the MyClass constructor
ADClass(MyClass);
 *  </pre>
 *
 *  <p>Synthetization is fully automated as long as the class that wishes to offer synthetized properties follows
 *  the following rules:</p>
 *
 *  <ul>
 *    <li>say the class wishes to synthetize the property <code>.foo</code></li>
 *    <li>if the class chooses to define a getter method, it must be defined as <code>getFoo</code> on its <code>prototype</code>
 *    <li>if the class chooses to define a setter method, it must be defined as <code>setFoo</code> on its <code>prototype</code>
 *    <li>the class must define an instance variable <code>._foo</code> which is the internal object that the class is responsible to update if either a setter is specified</li>
 *  </ul>
 *
 *  <p>So our previous class declaration could be extended as follows:</p>
 *
 *  <pre class="example">
function MyClass () {
  this._foo = '';
};

// custom setter for .foo, the getter is not specified
// and ADClass() will automatically create it
MyClass.prototype.setFoo = function (newFooValue) {
  this._foo = newFooValue;
};

// custom getter for .foo, the setter is not specified
// and ADClass() will automatically create it
MyClass.prototype.getBar = function (newFooValue) {
  return 'Hello ' + this._foo;
};
 *  </pre>
 *
 *  @since AdLib 0.1
 *
 *  @constructor
 *  @param theClass {Object} The class.
 */

function ADClass (theClass) {
  // check out if we have inheritance set up, otherwise, make ADObject the default superclass
  if (ADUtils.objectIsUndefined(theClass.inherits) && theClass !== ADObject) {
    theClass.inherits = ADObject;
  }
  
  // check out if we have mixins
  if (theClass.includes) {
    ADClass.mixin(theClass.prototype, theClass.includes);
  }

  // synthetizes properties defined locally
  var properties = (theClass.synthetizes) ? theClass.synthetizes : [];
  // now synthetize
  for (var i = 0; i < properties.length; i++) {
    ADClass.synthetizeProperty(theClass.prototype, properties[i], true);
  }

  // synthetizes properties by compiling them up the inheritance chain
  var aClass = theClass;
  var properties = [];
  while (aClass.inherits) {
    aClass = aClass.inherits;
    if (aClass.synthetizes) {
      properties = aClass.synthetizes.concat(properties);
    }            
  }
  // now synthetize
  for (var i = 0; i < properties.length; i++) {
    ADClass.synthetizeProperty(theClass.prototype, properties[i], false);
  }

  // go through each method and save its name as a custom property
  // that we'll use later in ADObject.callSuper()
  for (var i in theClass.prototype) {
    // make sure we don't touch properties that were synthetized
    if (theClass.prototype.__lookupGetter__(i)) {
      continue;
    }
    var prop = theClass.prototype[i];
    if (ADUtils.objectIsFunction(prop)) {
      prop._class = theClass;
      prop._name = i;
      prop.displayName = ADUtils.createDisplayName(theClass.name, i);
    }
  }

  // inherit from the superclass
  // default to ADObject if nothing is specified
  if (theClass !== ADObject) {
    theClass.prototype.__proto__ = theClass.inherits.prototype;
  }
};

ADClass.synthetizeProperty = function (object, propertyName, isPropertyInherited) {
  var camel_ready = propertyName.charAt(0).toUpperCase() + propertyName.substr(1);
  var getter_name = 'get' + camel_ready;
  var setter_name = 'set' + camel_ready;
  // check on function availability
  var has_getter = ADUtils.objectHasMethod(object, getter_name);
  var has_setter = ADUtils.objectHasMethod(object, setter_name);
  
  // if we have neither a getter or a setter, then do nothing
  // unless the property is defined locally
  if (!isPropertyInherited && !(has_getter || has_setter)) {
    return;
  }
  
  // assign the setter function if we have one
  if (has_setter) {
    var specified_setter_function = function (newValue) {
      object[setter_name].call(this, newValue);
      this.notifyPropertyChange(propertyName);
    };
    specified_setter_function.displayName = 'Specified setter for .' + propertyName + ' on ' + object.constructor.name;
    object.__defineSetter__(propertyName, specified_setter_function);
  }
  // otherwise just assign to _propertyName
  else {
    var default_setter_function = function (newValue) {
      this['_' + propertyName] = newValue;
      this.notifyPropertyChange(propertyName);
    };
    default_setter_function.displayName = 'Default setter for .' + propertyName + ' on ' + object.constructor.name;
    object.__defineSetter__(propertyName, default_setter_function);
  }
  
  // assign the getter function if we have one
  if (has_getter) {
    object.__defineGetter__(propertyName, object[getter_name]);
  }
  // otherwise just return _propertyName
  else {
    var default_getter_function = function () {
      return this['_' + propertyName];
    };
    default_getter_function.displayName = 'Default getter for .' + propertyName + ' on ' + object.constructor.name;
    object.__defineGetter__(propertyName, default_getter_function);
  }
};

ADClass.mixin = function (target, sources) {
  for (var i = 0; i < sources.length; i++) {
    ADUtils.copyPropertiesFromSourceToTarget(sources[i], target);
  }
};

ADUtils.setupDisplayNames(ADClass, 'ADClass');

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

const ADObjectPropertyChanged = 'handlePropertyChange';

/* ==================== ADObject ==================== */

/**
 *  @class
 *
 *  <p>The ADObject class is the base class for all AdLib objects and provides a way to receive and trigger
 *  notifications when a JavaScript property changes.</p>
 *
 *  <p>Observing a property on an object is done through the
 *  {@link #addPropertyObserver} method, specifying the property to observe as well as the object that will
 *  receive the notifications of the property change. Likewise, an object can stop observing such changes using
 *  the {@link #removePropertyObserver} method. When a property value changes, the author needs to call the 
 *  {@link #notifyPropertyChange} method so that registered observers will be notified. Note that synthetized
 *  properties will automatically trigger notifications when changed, so in this case, setters need not 
 *  manually call {@link #notifyPropertyChange}.</p>
 *
 *  <p>Finally, ADObject also provides the {@link #callSuper} method to call the superclass's implementation
 *  of an overload method.</p>
 *  
 *  @since AdLib 0.1
 */
function ADObject () {
  this.observedProperties = {};
};

/* ==================== Method overriding ==================== */

/**
 *  This method calls the object's superclass implementation of the current method
 */
ADObject.prototype.callSuper = function () {
  // figure out what function called us
  var caller = ADObject.prototype.callSuper.caller;
  // if that function is a constructor, we call the parent class's constructor
  if (ADUtils.objectHasMethod(caller, 'inherits')) {
    caller.inherits.apply(this, arguments);
  }
  // otherwise we look at that function name in the parent prototype
  else {
    var parent = caller._class.inherits.prototype;
    var method_name = caller._name;
    if (ADUtils.objectHasMethod(parent, method_name)) {
      return parent[method_name].apply(this, arguments);
    }
  }
};

/* ==================== Observing ==================== */

/**
 *  Indicates whether this object has registered observers for this property
 *
 *  @param {String} propertyName The property that may be observed.
 *
 *  @returns {bool} Whether this object has registered observers for this property
 *  @private
 */
ADObject.prototype.isPropertyObserved = function (propertyName) {
  return !ADUtils.objectIsUndefined(this.observedProperties[propertyName]);
};

/**
 *  Adds an observer for the named property on this object. In case an existing
 *  combination of the observer and property is already registered, this method
 *  is a no-op.
 *
 *  @param {String} propertyName The property to observe.
 *  @param {Object} observer The object that will be notified when the property is changed.
 *  @param {String} methodName The optional method name that will be called on the observer when the property is changed. 
 *  If this property is not provided, the <code>observer</code> must implement the {@link ADPropertyValueChange} protocol.
 */
ADObject.prototype.addPropertyObserver = function (propertyName, observer, methodName) {
  // create the array for this property if it's not already there
  if (!this.isPropertyObserved(propertyName)) {
    this.observedProperties[propertyName] = new Array();
  }
  // do nothing if we already have this observer registered
  else if (this.observedProperties[propertyName].indexOf(observer) > -1) {
    return;
  }
  // now, add the observer to the observer array for this property if valid
  var methodName = methodName || ADObjectPropertyChanged;
  if (ADUtils.objectHasMethod(observer, methodName)) {
    this.observedProperties[propertyName].push({
      observer: observer,
      methodName : methodName
    });
  }
};

/**
 *  Removes the observer for the named property on this object. In case an existing
 *  combination of the observer and property is not already registered, this method
 *  is a no-op.
 *
 *  @param {String} propertyName The observed property.
 *  @param {Object} observer The object that was notified when the property changed.
 *
 *  @returns {bool} Whether an observer was removed
 */
ADObject.prototype.removePropertyObserver = function (propertyName, observer) {
  // do nothing if this property is not observed
  if (!this.isPropertyObserved(propertyName)) {
    return false;
  }
  // now get the observer's index in the array
  var observers = this.observedProperties[propertyName];
  var observer_index = observers.indexOf(observer);
  // remove the observer if it was registered
  if (observer_index > -1) {
    observers.splice(observer_index, 1);
  }
  // let the user know whether we succeeded
  return (observer_index > -1);
};

/**
 *  Triggers a notification that the given property on this object has changed.
 *  For synthesized properties, this is called automatically upon setting the
 *  property. For methods that update an instance variable that is not synthesized,
 *  {@link #notifyPropertyChange} has to be called manually so that observers are notified.
 *
 *  @param {String} propertyName The observed property.
 */
ADObject.prototype.notifyPropertyChange = function (propertyName) {
  // do nothing if this property is not observed
  if (!this.isPropertyObserved(propertyName)) {
    return;
  }
  // now, go through each observer for this property and notify them
  var observers = this.observedProperties[propertyName];
  for (var i = 0; i < observers.length; i++) {
    var observer = observers[i];
    observer.observer[observer.methodName](this, propertyName);
  }
};

/**
 *  Calls a method on this object after a delay, allowing any number of optional extra arguments to be
 *  passed after the first two mandatory ones.
 *
 *  @param {String} methodName The method name to be called.
 *  @param {int} delay The delay in miliseconds.
 *
 *  @returns {int} The timeout that can be used to call <code>clearTimeout</code>.
 */
ADObject.prototype.callMethodNameAfterDelay = function (methodName, delay) {
  var _this = this;
  var args = Array.prototype.slice.call(arguments, 2);
  var generated_function = function () {
    _this[methodName].apply(_this, args);
  }
  generated_function.displayName = ADUtils.createDisplayName(this.constructor.name, methodName);
  return setTimeout(generated_function, delay);
};

/**
 *  @class
 *  @name ADPropertyValueChange
 *  @since AdLib 0.1
 */

/**
 *  Invoked when a property on an observed object has been changed.
 *
 *  @name handlePropertyChange
 *  @function
 *
 *  @param {Object} observedObject The observed object
 *  @param {String} propertyName The observed property's name as a string
 *  @memberOf ADPropertyValueChange.prototype
 */

ADClass(ADObject, 'ADObject');

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== ADPoint ==================== */

/**
 *  @class
 *
 *  @since AdLib 0.1
 *
 *  @param {int} x The horizontal coordinate
 *  @param {int} y The vertical coordinate
 */
function ADPoint (x, y) {
  this.x = (x != null && !isNaN(x)) ? x : 0;
  this.y = (y != null && !isNaN(y)) ? y : 0;
};

/**
 *  @name ADPoint.prototype
 *  @property {int} x The horizontal coordinate in pixels.
 *  @property {int} y The vertical coordinate in pixels.
 */

/**
 *  Creates a point from an Event reflecting the page coordinates contained in the first Touch object if it's a touch
 *  event, or directly on the event for any other type of event, such as a mouse event.
 *
 *  @param {UIEvent} The event object containing the coordinates to be retrieved as a ADPoint.
 *
 *  @returns {ADPoint} the event location
 */
ADPoint.fromEvent = function (event) {
  var event = (event.touches && event.touches.length > 0) ? event.touches[0] : event;
  return new ADPoint(event.pageX, event.pageY);
};

/**
 *  Creates a point from an Event reflecting the page coordinates contained in the first Touch object if it's a touch
 *  event, or directly on the event for any other type of event, such as a mouse event, relative to an element.
 *
 *  @param {UIEvent} The event object containing the coordinates to be retrieved as a ADPoint.
 *  @param {HTMLElement} The element of which coordinate system the returned point will be relative to.
 *
 *  @returns {ADPoint} the event location relative to the given element
 */
ADPoint.fromEventInElement = function (event, element) {
  var event = (event.touches && event.touches.length > 0) ? event.touches[0] : event;
  return window.webkitConvertPointFromPageToNode(element, new WebKitPoint(event.pageX, event.pageY));
};

/**
 *  Prints out a point, for debugging purposes
 *  @private
 */
ADPoint.prototype.toString = function () {
  return 'ADPoint[' + this.x + ',' + this.y + ']';
};

/**
 *  Returns an exact copy of this point
 *  @returns {ADPoint} An exact copy of this point
 */
ADPoint.prototype.copy = function () {
  return new ADPoint(this.x, this.y);
};

/**
 *  Tests whether this point is equal to another point.
 *  @param {ADPoint} anotherPoint A point.
 *  @returns {bool} Whether this point is equal to another point.
 */
ADPoint.prototype.equals = function (anotherPoint) {
  return (this.x == anotherPoint.x && this.y == anotherPoint.y);
};

ADUtils.setupDisplayNames(ADPoint, 'ADPoint');

/* ==================== ADSize ==================== */

/**
 *  @class
 *
 *  @since AdLib 0.1
 *
 *  @param {int} width The width
 *  @param {int} height The height
 */
function ADSize (width, height) {
  this.width = (width != null && !isNaN(width)) ? width : 0;
  this.height = (height != null && !isNaN(height)) ? height : 0;
};

/**
 *  @name ADSize.prototype
 *  @property {int} width The width in pixels.
 *  @property {int} height The height in pixels.
 */

/**
 *  Prints out a size, for debugging purposes
 *  @private
 */
ADSize.prototype.toString = function () {
  return 'ADSize[' + this.width + ',' + this.height + ']';
};

/**
 *  Returns an exact copy of this size
 *  @returns {ADSize} An exact copy of this size
 */
ADSize.prototype.copy = function () {
  return new ADSize(this.width, this.height);
};

/**
 *  Tests whether this size is equal to another size.
 *  @param {ADPoint} anotherSize A size.
 *  @returns {bool} Whether this size is equal to another size.
 */
ADSize.prototype.equals = function (anotherSize) {
  return (this.width == anotherSize.width && this.height == anotherSize.height);
};

ADUtils.setupDisplayNames(ADSize);

/* ==================== ADEdgeInsets ==================== */

/**
 *  @class
 *
 *  @since AdLib 0.2
 *
 *  @param {int} top The top inset.
 *  @param {int} right The right inset.
 *  @param {int} bottom The bottom inset.
 *  @param {int} left The left inset.
 */
function ADEdgeInsets (top, right, bottom, left) {
  this.top = top;
  this.right = right;
  this.bottom = bottom;
  this.left = left;
};

/**
 *  @name ADEdgeInsets.prototype
 *  @property {int} top The top inset.
 *  @property {int} right The right inset.
 *  @property {int} bottom The bottom inset.
 *  @property {int} left The left inset.
 */

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== ADImage ==================== */

ADImage.inherits = ADObject;
ADImage.synthetizes = ['width', 'height'];

/**
 *  @class
 *
 *  @since AdLib 0.1
 *
 *  @param {String} url The url for this image.
 */

function ADImage (url) {
  this.callSuper();
  //
  this.url = url;
  this.loaded = false;
  //
  this.element = new Image();
  this.element.src = url;
  this.element.addEventListener('load', this, false);
  //
  this._width = 0;
  this._height = 0;
};

/* ==================== Synthetized APIs ==================== */

/**
*  @name ADImage.prototype
*  @property {int} width The image's width.
*  @property {int} height The image's height.
*  @property {String} url The image's URL.
*  @property {HTMLImageElement} element The image's DOM element.
*  @property {bool} loaded Indicates if the image was fully loaded from the network yet.
*/

ADImage.prototype.getWidth = function () {
  return this.element.width;
};

ADImage.prototype.getHeight = function () {
  return this.element.height;
};

ADImage.prototype.handleEvent = function (event) {
  this.loaded = true;
  this.notifyPropertyChange('loaded');
};

ADClass(ADImage);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

const ADAnimatorLinearType = 0;
const ADAnimatorSplinesType = 1;
const ADAnimatorInvalidArgsException = 2;

const ADAnimatorAnimationDidIterate = 'animationDidIterate';
const ADAnimatorAnimationDidEnd = 'animationDidEnd';

/* ==================== ADAnimator ==================== */

function ADAnimator (duration, delegate, spline) {
  // check validity of arguments number
  if (arguments.length != 2 && arguments.length != 3 && arguments.length != 7) {  
    throw ADAnimatorInvalidArgsException;
    return false;
  }
  // init some base properties
  this.ready = false;
  this.animating = false;
  this.timer = null;
  // handle arguments
  this.duration = duration; // ms
  this.delegate = delegate;
  // check we have the required delegation method
  if (!ADUtils.objectHasMethod(this.delegate, ADAnimatorAnimationDidIterate)) {
    return;
  }
  // handle splines arguments, if available
  if (arguments.length >= 2) {
    this.type = ADAnimatorSplinesType;
    this.x1 = spline[0];
    this.y1 = spline[1];
    this.x2 = spline[2];
    this.y2 = spline[3];
    this.init();
  }
  else { // linear animation
    this.type = ADAnimatorLinearType;
  }
  this.ready = true;
};

ADAnimator.prototype.init = function () {
  // calculate the polynomial coefficients
  this.cx = 3 * this.x1;
  this.bx = 3 * (this.x2 - this.x1) - this.cx;
  this.ax = 1 - this.cx - this.bx;
  this.cy = 3 * this.y1;
  this.by = 3 * (this.y2 - this.y1) - this.cy;
  this.ay = 1 - this.cy - this.by;
  // compute points
  var numberOfPoints = (this.duration / 1000) * 240;
  this.curve = new Array(numberOfPoints);
  var dt = 1.0 / ( numberOfPoints - 1 );
  for (var i = 0; i < numberOfPoints; i++) {
    var t = i * dt;
    this.curve[i] = {
      x : (this.ax * Math.pow(t, 3)) + (this.bx * Math.pow(t, 2)) + (this.cx * t),
      y : (this.ay * Math.pow(t, 3)) + (this.by * Math.pow(t, 2)) + (this.cy * t)
    };
  }
};

ADAnimator.prototype.start = function () {
  if (!this.ready) {
    var _this = this;
    this.timer = setTimeout(function () { _this.start() }, 0);
    return;
  }
  this.animating = true;
  this.lastIndex = 0;
  this.startTime = (new Date()).getTime();
  this.iterate();
};

ADAnimator.prototype.stop = function () {
  this.animating = false;
  clearTimeout(this.timer);
};

ADAnimator.prototype.iterate = function () {
  var ellapsedTime = (new Date()).getTime() - this.startTime;
  if (ellapsedTime < this.duration) {
    var x = ellapsedTime / this.duration;
    // handle splines case
    if (this.type == ADAnimatorSplinesType) {
      var y = 0;
      for (var i = this.lastIndex; i < this.curve.length; i++) {
        var point = this.curve[i];
        if (point.x >= x && i > 0) {
          var previous_point = this.curve[i - 1];
          if ((x - previous_point.x) < (point.x - x)) {
            this.lastIndex = i - 1;
            y = previous_point.y;
          }
          else {
            this.lastIndex = i;
            y = point.y;
          }
          break;
        }
      }
    }
    this.delegate[ADAnimatorAnimationDidIterate]((this.type == ADAnimatorSplinesType) ? y : x);
    var _this = this;
    this.timer = setTimeout(function () { _this.iterate() }, 0);
  }
  else {
    this.delegate[ADAnimatorAnimationDidIterate](1);
    if (ADUtils.objectHasMethod(this.delegate, ADAnimatorAnimationDidEnd)) {
      this.delegate[ADAnimatorAnimationDidEnd]();
    }
    this.animating = false;    
  }
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADAnimator);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/**
 *  @class
 *  @name ADTransitionDelegate
 *  @since AdLib 0.1
 */

/**
 *  Invoked when a transition has completed.
 *
 *  @name transitionDidComplete
 *  @function
 *
 *  @param {ADTransition} theTransition The transition that just completed.
 *  @memberOf ADTransitionDelegate.prototype
 */

const ADTransitionDidCompleteDelegate = 'transitionDidComplete';

const ADTransitionDefaults = {
  duration  : 0.5,
  delay     : 0,
  removesTargetUponCompletion : false,
  revertsToOriginalValues : false
};

const ADTransitionStyles = ['-webkit-transition-property', '-webkit-transition-duration', '-webkit-transition-timing-function', '-webkit-transition-delay', '-webkit-transition'];

/* ==================== ADTransition ==================== */

/**
 *  @class
 *
 *  <p>The ADTransition class allows to create a synchronized change of one or more property values on a
 *  {@link ADView} instance or any DOM element.</p>
 *
 *  <p>First, the user sets the {@link #target} for the transition,
 *  identifying the target view or element, and second the set of {@link #properties}, in the form of an <code>Array</code>
 *  of either JavaScript properties of the view or CSS properties for the element. Then the {@link #to} and optional
 *  {@link #from} value arrays are provided to define what values will be used for the transition. Each item in the
 *  {@link #from} and {@link #to} arrays matches the item at the same index in the {@link #properties} array, and the same
 *  applies to the {@link #duration} property, which can also be a single value shared for all transitioned properties.</p>
 *
 *  <p>Finally, once the properties on the transition are all set, the {@link #start} method must be called so that the
 *  transition is started. Note that it is possible to group transitions within a {@link ADTransaction} and set a default
 *  set of transition properties within a transaction using the {@link ADTransaction.defaults} static property.</p>
 *  
 *  <p>The following example shows how to make a transition that fades a view in as it animates from the right side of the screen:</p>
 *
<pre class="example">
new ADTransition({
  target : aView,
  properties : ['opacity', 'position'];
  from : [0, new ADPoint(320, 0)],
  to : [1, new ADPoint(0, 0)]
  duration : 0.5;
}).start();
</pre>
 *
 *  <p>Note that properties of a {@link ADTransition} object can be passed as a batch to the {@link ADTransition} constructor
 *  as an anonymous object. This also allows reuse of a set of parameters across several transitions. Also, a set of pre-baked
 *  transitions exist can be easily applied to a view with the {@link ADView#applyTransition} method.</p>
 *
 *  @since AdLib 0.1
 */
function ADTransition (params) {
  // set up defaults
  /**
   *  The transition target, either a {@link ADView} instance or an <code>HTMLElement</code>.
   *  @type Object 
   */
  this.target = null;
  /**
   *  The set of properties that will be transitioned. The properties are either JavaScript properties
   *  of the target {@link ADView} instance or CSS properties of the targeted <code>HTMLElement</code>.
   *  @type Array 
   */
  this.properties = null;
  /**
   *  The set of durations in seconds, each duration matching a property in the {@link #properties} array.
   *  Note that it's possible to specify a single value instead of an array to use the same duration for 
   *  all properties.
   *  @type Object 
   */
  this.duration = null;
  /**
   *  The set of delays in seconds, each delay matching a property in the {@link #properties} array.
   *  Note that it's possible to specify a single delay instead of an array to use the same delay for 
   *  all properties.
   *  @type Object 
   */
  this.delay = null;
  /**
   *  Optional list of values to start the transition from. Each value in this array must match the property
   *  at the same index in the {@link #properties} array.
   *  @type Array
   */
  this.from = null;
  /**
   *  Required list of values to transition to. Each value in this array must match the property
   *  at the same index in the {@link #properties} array.
   *  @type Array
   */
  this.to = null;
  /**
   *  The set of timing functions, each timing function matching a property in the {@link #properties}
   *  array. Note that it's possible to specify a single timing function instead of an array to use the
   *  same timing function for all properties.
   *  @type Object 
   */
  this.timingFunction = null;
  /**
   *  The delegate object that implements the {@link ADTransitionDelegate} protocol.
   *  @type Object
   */
  this.delegate = null;
  /**
   *  Indicates whether the target needs to be removed once the transition completes. If the target is a
   *  view, {@link ADView#removeFromSuperview} is called, while <a href="http://www.w3.org/TR/DOM-Level-3-Core/core.html#ID-1734834066">
   *  removeChild</a> is called if the target is an element. Defaults to <code>false</code>.
   *  @type bool
   */
  this.removesTargetUponCompletion = null;
  /**
   *  Indicates whether the target needs to reset the property that are transitioned to their original values
   *  once the transition completes. Defaults to <code>false</code>.
   *  @type bool
   */
  this.revertsToOriginalValues = null;
  //
  this.defaultsApplied = false;
  this.archivedStyles = null;
  this.archivedValues = [];
  // import params
  ADUtils.copyPropertiesFromSourceToTarget(params, this);
};

/* ==================== Dealing with defaults ==================== */

ADTransition.prototype.applyDefaults = function () {
  if (this.defaultsApplied) {
    return;
  }
  //
  for (var i in ADTransitionDefaults) {
    if (this[i] === null) {
      this[i] = ADTransitionDefaults[i];
    }
  }
  //
  this.defaultsApplied = true;
};

/* ==================== Archiving the transition styles ==================== */

ADTransition.prototype.archiveTransitionStyles = function () {
  // do nothing if we already have archived styles in this run session
  if (this.archivedStyles !== null) {
    return;
  }
  // get element we'll be animating
  var element = (this.target instanceof ADView) ? this.target.layer : this.target;
  // iterate all ADTransitionStyles and archive them
  this.archivedStyles = [];
  for (var i = 0; i < ADTransitionStyles.length; i++) {
    this.archivedStyles.push(element.style.getPropertyValue(ADTransitionStyles[i]));
  }
};

ADTransition.prototype.restoreTransitionStyles = function () {
  // iterate all ADTransitionStyles and restore them
  for (var i = 0; i < ADTransitionStyles.length; i++) {
    this.element.style.setProperty(ADTransitionStyles[i], this.archivedStyles[i], '');
  }
  // reset archived styles
  this.archivedStyles = null;
};

/* ==================== Archiving the base values ==================== */

ADTransition.prototype.archiveBaseValues = function () {
  // do nothing if we don't need to archive base values
  if (!this.revertsToOriginalValues) {
    return;
  }
  // store properties if it's a view
  if (this.target instanceof ADView) {
    for (var i = 0; i < this.properties.length; i++) {
      this.archivedValues.push(this.target[this.properties[i]]);
    }
  }
  // or an element
  else {
    for (var i = 0; i < this.properties.length; i++) {
      this.archivedValues.push(this.target.layer.style.getPropertyValue(this.properties[i]));
    }
  }
};

ADTransition.prototype.restoreBaseValues = function () {
  // store properties if it's a view
  if (this.target instanceof ADView) {
    for (var i = 0; i < this.properties.length; i++) {
      this.target[this.properties[i]] = this.archivedValues[i];
    }
  }
  // or an element
  else {
    for (var i = 0; i < this.properties.length; i++) {
      this.target.layer.style.setProperty(this.properties[i], this.archivedValues[i], null);
    }
  }
};

/* ==================== Starting the transition ==================== */

/**
 *  Starts the transition.
 */
ADTransition.prototype.start = function () {
  // if we have an active transaction, just add to it
  if (ADTransaction.openTransactions > 0) {
    ADTransaction.addTransition(this);
    return;
  }  
  // otherwise, we'll just get it going
  this.applyDefaults();
  if (this.from === null) {
    this.applyToState();
  }
  else {
    this.applyFromState();
    var _this = this;
    setTimeout(function () {
      _this.applyToState();
    }, 0);  
  }
};

/* ==================== Applying the "from" state ==================== */

ADTransition.prototype.applyFromState = function () {
  // do nothing if we have no "from" state specified
  if (this.from === null) {
    return;
  }
  //
  this.applyDefaults();
  this.archiveTransitionStyles();
  // we're dealing with a ADView
  if (this.target instanceof ADView) {
    this.target.layer.style.webkitTransitionDuration = 0;
    for (var i = 0; i < this.properties.length; i++) {
      this.target[this.properties[i]] = this.processTransitionValue(this.from[i]);
    }
  }
  // we're dealing with an Element
  else {
    this.target.style.webkitTransitionDuration = 0;
    for (var i = 0; i < this.properties.length; i++) {
      this.target.style.setProperty(this.properties[i], this.from[i], '');
    }
  }
};

/* ==================== Applying the "to" state ==================== */

ADTransition.prototype.applyToState = function () {
  // first, make sure we have defaults applied if some
  // properties are not explicitely set on our transition
  this.applyDefaults();

  // and that we archive the transition styles to be reverted
  this.archiveTransitionStyles();

  // and that we archive the original values
  this.archiveBaseValues();
  
  // track wether we are animating a view
  var is_view = (this.target instanceof ADView);

  // now compile the styles needed for this transition
  this.cssProperties = [];
  var transition_styles = [];
  for (var i = 0; i < this.properties.length; i++) {
    // get the name of the property we are trying to transition
    var property = (is_view) ? this.target.cssPropertyNameForJSProperty(this.properties[i]) : this.properties[i];
    // do nothing if we already have an animation defined for this
    if (this.cssProperties.indexOf(property) > -1) {
      continue;
    }
    // else, set up the CSS style for this transition
    var duration = (ADUtils.objectIsArray(this.duration)) ? this.duration[i] : this.duration;
    var timing = (ADUtils.objectIsArray(this.timingFunction)) ? this.timingFunction[i] : this.timingFunction;
    var delay = (ADUtils.objectIsArray(this.delay)) ? this.delay[i] : this.delay;
    transition_styles.push([property, duration + 's', timing, delay + 's'].join(' '));
    // and remember we are animating this property
    this.cssProperties.push(property);
  }

  // if we are dealing with a view, set the properties on the view
  if (is_view) {
    this.element = this.target.layer;
    for (var i = 0; i < this.properties.length; i++) {
      this.target[this.properties[i]] = this.processTransitionValue(this.to[i]);
    }
  }

  else {
    this.element = this.target;
    for (var i = 0; i < this.properties.length; i++) {
      this.target.style.setProperty(this.properties[i], this.to[i], '');
    }
  }

  // set up the transition styles
  this.element.style.webkitTransition = transition_styles.join(', ');

  // register for events to track transition completions
  this.element.addEventListener('webkitTransitionEnd', this, false);
  this.completedTransitions = 0;
};

/* ==================== Tracking transition completion ==================== */

// XXX: we won't be getting an event for properties that have the same value in the to
// state, so we'll need to do a little work to track css properties that won't really transition
ADTransition.prototype.handleEvent = function (event) {
  // do nothing if that event just bubbled from our target's sub-tree
  if (event.currentTarget !== this.element) {
    return;
  }

  // update the completion counter
  this.completedTransitions++;

  // do nothing if we still have running transitions
  if (this.completedTransitions != this.cssProperties.length) {
    return;
  }

  // the counter reached our properties count, fire up the delegate
  if (ADUtils.objectHasMethod(this.delegate, ADTransitionDidCompleteDelegate)) {
    this.delegate[ADTransitionDidCompleteDelegate](this);
  }

  // remove from the tree if instructed to do so
  if (this.removesTargetUponCompletion) {
    var target = this.target;
    if (this.target instanceof ADView) {
      target.removeFromSuperview();
    }
    else {
      target.parentNode.removeChild(target);
    }
  }
  // otherwise, restore transition styles
  else {
    this.restoreTransitionStyles();
  }
  
  // restore original values if instructed to do so
  if (this.revertsToOriginalValues) {
    this.restoreBaseValues();
  }
};

/* ==================== Processing transition values ==================== */

const ADTransitionWidthRegExp = new RegExp(/\$width/g);
const ADTransitionHeightRegExp = new RegExp(/\$height/g);

// translates $width and $height
ADTransition.prototype.processTransitionValue = function (value) {
  if (!ADUtils.objectIsString(value)) {
    return value;
  }
  value = value.replace(ADTransitionWidthRegExp, ADUtils.px(this.target.size.width));
  return value.replace(ADTransitionHeightRegExp, ADUtils.px(this.target.size.height));
};

ADClass(ADTransition);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== ADTransaction ==================== */

/**
 *  @class
 *
 *  <p>A transaction allows to run a variety of transitions in sync, no matter in what function or JavaScript run loop
 *  the transition itself was applied. Transactions can be nested, and a transaction is only truly commited when there
 *  are no other open transactions. In other words, there must be as many calls to {@link ADTransaction.begin} as there
 *  are calls to {@link ADTransaction.commit} before any transition in one of those transactions can be applied.
 *
 *  @since AdLib 0.1
 */
var ADTransaction = {
  transitions : [],
  openTransactions : 0,
  /**
   *  The set of default properties that will be applied to any {@link ADTransition} until the next call to {@link ADTransaction.commit}.
   *  Any of the properties that can be applied to a {@link ADTransition} can be applied to this object.
   *  @type Object
   */
  defaults : {},
  defaultsStates : []
};

/**
 *  Begins a new transaction and makes it the current transaction.
 */
ADTransaction.begin = function () {
  // reset the group if we're starting fresh
  if (this.openTransactions == 0) {
    this.transitions = [];
    this.defaults = {};
  }
  // otherwise, archive the current state of defaults
  else {
    this.defaultsStates.push(this.defaults);
    // XXX: should we restore defaults here as well?
  }
  // increase the number of open transactions
  this.openTransactions++;
};

/**
 *  Add Transition
 *
 *  @private
 */
ADTransaction.addTransition = function (transition) {
  // first, apply the transaction defaults to the transitions
  for (var i in this.defaults) {
    if (transition[i] === null) {
      transition[i] = this.defaults[i];
    }
  }
  // and add the transition to our array
  this.transitions.push(transition);
};

/**
 *  Closes the current transaction.
 */
ADTransaction.commit = function () {
  // do nothing if we have no open transactions
  if (this.openTransactions == 0) {
    return;
  }
  // decrease the number of open transactions
  this.openTransactions--;
  // if we still have open transactions, just
  // restore the previous defaults state
  if (this.openTransactions != 0) {
    this.defaults = this.defaultsStates.pop();
    return;
  }
  // otherwise, it's time to shake and bake, we'll apply the
  // "from" states directly and the "to" states immediately
  // after in a new run loop so that the "from" styles have
  // been resolved first
  var transitions = this.transitions;
  for (var i = 0; i < transitions.length; i++) {
    transitions[i].applyFromState();
  }
  setTimeout(function () {
    for (var i = 0; i < transitions.length; i++) {
      transitions[i].applyToState();
    }
  }, 0);
};

ADUtils.setupDisplayNames(ADTransaction, 'ADTransaction');

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

const ADAnimationDidCompleteDelegate = 'animationDidComplete';

const ADAnimationDefaults = {
  duration  : 1,
  delay : 0,
  removesTargetUponCompletion : false,
  timingFunction : 'ease'
};

const ADAnimationStyles = ['-webkit-animation-name', '-webkit-animation-timing-function', '-webkit-animation-duration', '-webkit-animation-iteration-count', '-webkit-animation-direction', '-webkit-animation-delay', '-webkit-animation'];

/* ==================== ADAnimation ==================== */

function ADAnimation (params) {
  this.rule = ADAnimation.createRule();
  // set up defaults
  this.target = null;
  this.duration = null;
  this.delay = null;
  this.keyframes = [];
  this.timingFunction = null;
  this.delegate = null;
  this.removesTargetUponCompletion = null;
  // import params
  ADUtils.copyPropertiesFromSourceToTarget(params, this);
  //
  this.init();
};

/* ==================== Starting the animation ==================== */

ADAnimation.prototype.init = function () {
  // apply defaults
  for (var i in ADAnimationDefaults) {
    if (this[i] === null) {
      this[i] = ADAnimationDefaults[i];
    }
  }
  // archive styles
  this.archiveAnimationStyles();
  // clear all animation styles
  var style = this.target.layer.style;
  for (var i = 0; i < ADAnimationStyles; i++) {
    style.setProperty(ADAnimationStyles[i], '', '');
  }
  // add keyframes to rule
  for (var i = 0; i < this.keyframes.length; i++) {
    var keyframe = this.keyframes[i];
    var rule = (keyframe.offset * 100) + '% {';
    for (var j = 0; j < keyframe.values.length; j+=2) {
      rule += this.target.cssPropertyNameForJSProperty(keyframe.values[j]) + ': ' + keyframe.values[j + 1] + ';';
    }
    rule += '}';
    this.rule.insertRule(rule);
  }
};

ADAnimation.prototype.start = function () {
  // start the animation
  var style = this.target.layer.style;
  style.webkitAnimationName = this.rule.name;
  style.webkitAnimationDuration = this.duration + 's';
  style.webkitAnimationDelay = this.delay + 's';
  style.webkitAnimationTimingFunction = this.timingFunction;
  // track events on this element
  this.target.layer.addEventListener('webkitAnimationEnd', this, false);
};

/* ==================== Archiving the transition styles ==================== */

ADAnimation.prototype.archiveAnimationStyles = function () {
  // iterate all ADAnimationStyles and archive them
  this.archivedStyles = [];
  for (var i = 0; i < ADAnimationStyles.length; i++) {
    this.archivedStyles.push(this.target.layer.style.getPropertyValue(ADAnimationStyles[i]));
  }
};

ADAnimation.prototype.restoreAnimationStyles = function () {
  // iterate all ADAnimationStyles and restore them
  for (var i = 0; i < ADAnimationStyles.length; i++) {
    this.target.layer.style.setProperty(ADAnimationStyles[i], this.archivedStyles[i], '');
  }
};

/* ==================== Chaining transitions ==================== */

ADAnimation.prototype.handleEvent = function (event) {
  // do nothing if this event is not coming from our element directly
  if (event.currentTarget !== this.target.layer) {
    return;
  }
  // update the style to match the last frame
  var values = this.keyframes[this.keyframes.length - 1].values;
  for (var i = 0; i < values.length; i+=2) {
    this.target[values[i]] = values[i+1];
  }
  // the counter reached our properties count, fire up the delegate
  if (ADUtils.objectHasMethod(this.delegate, ADAnimationDidCompleteDelegate)) {
    this.delegate[ADAnimationDidCompleteDelegate](this);
  }
  // remove from the tree if instructed to do so
  if (this.removesTargetUponCompletion) {
    this.target.removeFromSuperview();
  }
  // otherwise, restore transition styles
  else {
    this.restoreAnimationStyles();
  }
};

ADClass(ADAnimation);

/* ==================== Creating Keyframes Rules ==================== */

ADAnimation.counter = 0;

ADAnimation.createRule = function () {
  var sheet = document.styleSheets[0];
  var name = 'AD-Animation-' + this.counter++;
  var index = sheet.insertRule('@-webkit-keyframes ' + name + ' { }', sheet.rules.length);
  return sheet.rules.item(index);
};

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/**
 *  @class
 *
 *  A virtual class that allows the definition of pre-baked transitions that can be used as parameter
 *  for the {@link ADView#applyTransition} method. In reality, any object that provides the type of
 *  properties instances of this class offers can be used with that method.
 *
 *  @name ADTransitionDefinition
 *  @since AdLib 0.1
 */

/**
 *  @name ADTransitionDefinition.prototype
 *  @property {Array} base A set of property / value pairs that are applied to the view as the transition starts. The properties
 *  listed there are not supposed to be transitioned. If you wish to set a from state for a property that gets transitioned, use
 *  the {@link #properties} and {@link from} properties instead.
 *  @property {Array} properties The set of properties that will be transitioned. The properties are JavaScript properties
 *  of the {@link ADView} instance that will be transitioned using this definition.
 *  @property {Array} from Optional list of values to start the transition from. Each value in this array must match the property
 *  at the same index in the {@link #properties} array.
 *  @property {Array} to Required list of values to transition to. Each value in this array must match the property
 *  at the same index in the {@link #properties} array.
 */

/**
 *  Fades out.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionDissolveOut = {
  properties : ['opacity'],
  from : [1],
  to : [0]
};

/**
 *  Fades in.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionDissolveIn = {
  properties : ['opacity'],
  from : [0],
  to : [1]
};

/**
 *  Fades in while scaling up to identity scale.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionZoomIn = {
  properties : ['opacity', 'transform'],
  from : [0, 'scale(0.2)'],
  to : [1, 'scale(1)']
};

/**
 *  Fades out while scaling down to identity scale.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionZoomOut = {
  properties : ['opacity', 'transform'],
  from : [0, 'scale(1.2)'],
  to : [1, 'scale(1)']
};

/**
 *  Fades in while rotating from the right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionCrossSpinRight = {
  properties : ['opacity', 'transform'],
  from : [0, 'rotate(20deg)'],
  to : [1, 'rotate(0)']
};

/**
 *  Fades in while rotating from the left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionCrossSpinLeft = {
  properties : ['opacity', 'transform'],
  from : [0, 'rotate(-20deg)'],
  to : [1, 'rotate(0)']
};

/**
 *  Flips towards the back side from right to left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionFlipLeftOut = {
  properties : ['transform'],
  from : ['perspective(800) rotateY(0deg)'],
  to : ['perspective(800) rotateY(-180deg)']
};

/**
 *  Flips towards the front side from right to left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionFlipLeftIn = {
  properties : ['transform'],
  from : ['perspective(800) rotateY(180deg)'],
  to : ['perspective(800) rotateY(0deg)']
};

/**
 *  Flips towards the back side from left to right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionFlipRightOut = {
  properties : ['transform'],
  from : ['perspective(800) rotateY(0deg)'],
  to : ['perspective(800) rotateY(180deg)']
};

/**
 *  Flips towards the front side from left to right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionFlipRightIn = {
  properties : ['transform'],
  from : ['perspective(800) rotateY(-180deg)'],
  to : ['perspective(800) rotateY(0deg)']
};

/**
 *  Cube transition going out from right to left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionCubeLeftOut = {
  base : ['anchorPoint', new ADPoint(1, 0.5)],
  properties : ['transform'],
  from : ['perspective(800) rotateY(0deg) translateZ(0)'],
  to : ['perspective(800) rotateY(-90deg) translateZ($width)']
};

/**
 *  Cube transition coming in from right to left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionCubeLeftIn = {
  base : ['anchorPoint', new ADPoint(0, 0.5)],
  properties : ['transform'],
  from : ['perspective(800) rotateY(90deg) translateZ($width)'],
  to : ['perspective(800) rotateY(0deg) translateZ(0)']
};

/**
 *  Cube transition going out from left to right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionCubeRightOut = {
  base : ['anchorPoint', new ADPoint(0, 0.5)],
  properties : ['transform'],
  from : ['perspective(800) rotateY(0deg) translateZ(0)'],
  to : ['perspective(800) rotateY(90deg) translateZ($width)']
};

/**
 *  Cube transition coming in from left to right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionCubeRightIn = {
  base : ['anchorPoint', new ADPoint(1, 0.5)],
  properties : ['transform'],
  from : ['perspective(800) rotateY(-90deg) translateZ($width)'],
  to : ['perspective(800) rotateY(0deg) translateZ(0)']
};

/**
 *  Door transition going out from left to right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionDoorOpenLeftOut = {
  base : ['anchorPoint', new ADPoint(0, 0.5), 'zIndex', 1],
  properties : ['transform'],
  from : ['perspective(800) rotateY(0deg)'],
  to : ['perspective(800) rotateY(-90deg)']
};

/**
 *  Door transition coming in from left to right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionDoorCloseLeftIn = {
  base : ['anchorPoint', new ADPoint(0, 0.5), 'zIndex', 2],
  properties : ['transform'],
  from : ['perspective(800) rotateY(-90deg)'],
  to : ['perspective(800) rotateY(0deg)']
};

/**
 *  Door transition going out from right to left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionDoorOpenRightOut = {
  base : ['anchorPoint', new ADPoint(1, 0.5), 'zIndex', 1],
  properties : ['transform'],
  from : ['perspective(800) rotateY(0deg)'],
  to : ['perspective(800) rotateY(90deg)']
};

/**
 *  Door transition coming in from right to left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionDoorCloseRightIn = {
  base : ['anchorPoint', new ADPoint(1, 0.5), 'zIndex', 2],
  properties : ['transform'],
  from : ['perspective(800) rotateY(90deg)'],
  to : ['perspective(800) rotateY(0deg)']
};

/**
 *  Revolve transition fading out to the front on the left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionRevolveTowardsLeftOut = {
  base : ['anchorPoint', new ADPoint(0, 0.5)],
  properties : ['transform', 'opacity'],
  from : ['perspective(800) rotateY(0deg)', 1],
  to : ['perspective(800) rotateY(-90deg)', 0]
};

/**
 *  Revolve transition coming in from the front on the left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionRevolveTowardsLeftIn = {
  base : ['anchorPoint', new ADPoint(0, 0.5)],
  properties : ['transform'],
  from : ['perspective(800) rotateY(90deg)'],
  to : ['perspective(800) rotateY(0deg)']
};

/**
 *  Revolve transition going out to the back on the left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionRevolveAwayLeftOut = {
  base : ['anchorPoint', new ADPoint(0, 0.5)],
  properties : ['transform'],
  from : ['perspective(800) rotateY(0deg)'],
  to : ['perspective(800) rotateY(90deg)']
};

/**
 *  Revolve transition fading in from the front on the left.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionRevolveAwayLeftIn = {
  base : ['anchorPoint', new ADPoint(0, 0.5)],
  properties : ['transform', 'opacity'],
  from : ['perspective(800) rotateY(-90deg)', 0],
  to : ['perspective(800) rotateY(0deg)', 1]
};

/**
 *  Revolve transition fading out to the front on the right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionRevolveTowardsRightOut = {
  base : ['anchorPoint', new ADPoint(1, 0.5)],
  properties : ['transform', 'opacity'],
  from : ['perspective(800) rotateY(0deg)', 1],
  to : ['perspective(800) rotateY(90deg)', 0]
};

/**
 *  Revolve transition coming in from the front on the right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionRevolveTowardsRightIn = {
  base : ['anchorPoint', new ADPoint(1, 0.5)],
  properties : ['transform'],
  from : ['perspective(800) rotateY(-90deg)'],
  to : ['perspective(800) rotateY(0deg)']
};

/**
 *  Revolve transition going out from the front on the right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionRevolveAwayRightOut = {
  base : ['anchorPoint', new ADPoint(1, 0.5)],
  properties : ['transform'],
  from : ['perspective(800) rotateY(0deg)'],
  to : ['perspective(800) rotateY(-90deg)']
};

/**
 *  Revolve transition fading in from the front on the right.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionRevolveAwayRightIn = {
  base : ['anchorPoint', new ADPoint(1, 0.5)],
  properties : ['transform', 'opacity'],
  from : ['perspective(800) rotateY(90deg)', 0],
  to : ['perspective(800) rotateY(0deg)', 1]
};

/**
 *  Spin transition fading out.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionSpinOut = {
  properties : ['transform', 'opacity'],
  from : ['perspective(800) rotate(0)', 1],
  to : ['perspective(800) rotate(-180deg)', 0]
};

/**
 *  Spin transition fading in.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionSpinIn = {
  base : ['zIndex', 1],
  properties : ['transform', 'opacity'],
  from : ['perspective(800) rotate(-180deg)', 0],
  to : ['perspective(800) rotate(0)', 1]
};

/**
 *  Scale transition scaling in.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionScaleIn = {
  base : ['zIndex', 1],
  properties : ['transform'],
  from : ['scale(0.01)'],
  to : ['scale(1)']
};

/**
 *  Scale transition scaling out.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewTransitionScaleOut = {
  base : ['zIndex', 1],
  properties : ['transform'],
  from : ['scale(1)'],
  to : ['scale(0.01)']
};

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== ADViewLayerInsertionNotificationHelper ==================== */

function ADViewLayerInsertionNotificationHelper () {
  this.callSuper();
  this.views = [];
  // listen to the DOMSubtreeModified event, which gets called after a node is attached to
  // the render tree, which is the trigger to processing of the views
  document.addEventListener('DOMSubtreeModified', this, true);
};

ADViewLayerInsertionNotificationHelper.prototype.considerView = function (view) {
  // only add the view if we don't already have it
  // and set the _ignoreView flag to false
  if (this.views.indexOf(view) == -1) {
    view._ignoreView = false;
    this.views.push(view);
  }
};

ADViewLayerInsertionNotificationHelper.prototype.handleEvent = function (event) {
  // we've received the DOMSubtreeModified event, so start processing the views
  this.processViews();
};

ADViewLayerInsertionNotificationHelper.prototype.processViews = function () {
  // nothing to do if we have no views to notify at all
  if (this.views.length < 1) {
    return;
  }
  // first, filter the amount of views so that any view that is a descendent of another
  // one in the list is removed given it'll get the notification anyways
  var view;
  for (var i = 0; i < this.views.length; ++i) {
    view = this.views[i];
    for (var ancestor = view.superview; ancestor && !view._ignoreView; ancestor = ancestor.superview) {
      if ('_ignoreView' in ancestor) {
        view._ignoreView = true;
      }
    }
  }
  // now dispatch the notifications to all the views that made the cut and ensure we removed our _ignoreView flags
  while (this.views.length > 0) {
    view = this.views.pop();
    if (view._ignoreView) {
      delete view._ignoreView;
      continue;
    }
    delete view._ignoreView;
    view.dispatchNotificationOfLayerInsertionIntoDocument();
  }
};

ADClass(ADViewLayerInsertionNotificationHelper);

/* ==================== ADView ==================== */

ADView.inherits = ADObject;
ADView.synthetizes = ['id', 'position', 'size', 'transform', 'anchorPoint', 'doubleSided', 'zIndex', 'opacity', 'clipsToBounds', 'transitionsEnabled', 'transitionsDuration', 'hostingLayer'];

ADView.layerInsertionNotificationHelper = new ADViewLayerInsertionNotificationHelper();

/**
 *  @class
 *
 *  <p>The ADView class is the base class for all the AdLib views. In DOM terms, a ADView is backed by a {@link #layer},
 *  which really is an HTMLElement with that is subject to hardware-accelerated transforms, using the <code>-webkit-transform</code>
 *  CSS property. Most views have additional structure underneath this containing layer, and, because of this, a
 *  separate hosting layer can be provided by ADView subclasses by setting the {@link #hostingLayer} property, the
 *  hostingLayer accessor returning this layer if available, or the top container layer otherwise. Using this
 *  hosting layer, developers can add contents freely.</p>
 *
 *  <p>A view has a position and a size, and the visibility of contents overflowing the size of the view can be
 *  controlled using the {@link #clipsToBounds} property. The contents of a view can be provided in two ways. The first
 *  way is to simply add a AdLib view as a subview using the {@link #addSubview} method. The other way is to provide
 *  contents using any other DOM-based API and use the <code><a href="http://www.w3.org/TR/DOM-Level-3-Core/core.html#ID-184E7107">
 *  appendChild</a></code> method to add content to the view's hosting layer. This allows AdLib to be used not only with
 *  AdLib views, but also content created by other AJAX framework or technique.</p>
 *
 *  <p>Because the position of the view is controlled with the <code>-webkit-transform</code> CSS property, the underlying CSS Transitions
 *  mechanism can be directly leveraged by turning transitions on and off using the {@link #transitionsEnabled} property.
 *  Having transitions turned on will make any change of position or opacity happen with a smooth, hardware-accelerated animation.</p>
 *
 *  <p>The root of the ADView hierarchy is the {@link ADRootView.sharedRoot} property.</p>
 *
 *  @extends ADObject
 *  @since AdLib 0.1
 */
function ADView () {
  this.callSuper();
  //
  this.superview = null;
  this.subviews = [];
  /**
   *  Controls whether the view reacts to touch events following a <code>touchstart</code> event, therefore letting the
   *  {@link #touchesMoved} and {@link #touchesEnded} method be called, or disregards them. Should you want all touch events
   *  tracking to be disabled, use the {@link #userInteractionEnabled} property. The default is <code>false</code>.
   *  @type bool
   */
  this.tracksTouchesOnceTouchesBegan = false;
  /**
   *  A Boolean value that determines whether user events are taken into account or ignored. The default is <code>true</code>.
   *  @type bool
   */
  this.userInteractionEnabled = true;
  /**
   *  A Boolean value that determines whether the view automatically resizes its subviews when its size changes. The default is <code>true</code>.
   *  @type bool
   */
  this.autoresizesSubviews = true;
  /**
   *  An integer bit mask that determines how the receiver resizes its subviews when its bounds change. This mask can be specified by combining,
   *  using the C bitwise OR operator, any of the constants prefixed with <code>ADViewAutoresizing</code>. Where more than one option along an axis
   *  is set, the default behavior is to distribute the size difference as evenly as possible among the flexible portions. If the autoresizing mask
   *  is equal to {@link ADViewAutoresizingNone}, then the view doesn't resize at all when its bounds changes. The default value is
   *  {@link ADViewAutoresizingNone}.
   *  @type int
   */
  this.autoresizingMask = ADViewAutoresizingNone;
  /**
   *  Indicates whether the view's {@link #layer} is currently part of the document. A view can be notified of when its {@link #layer} is added or removed
   *  from the document with the {@link #layerWasInsertedIntoDocument} and {@link #layerWasRemovedFromDocument} methods.
   *  @type bool
   */
  this.layerIsInDocument = false;
  // synthetized properties
  this._position = new ADPoint();
  this._size = new ADSize();
  this._anchorPoint = new ADPoint(0.5, 0.5);
  this._doubleSided = true;
  this._zIndex = 0;
  this._transform = ADUtils.t(0,0);
  this._clipsToBounds = false;
  this._transitionsEnabled = false;
  this._transitionsDuration = 0.5;
  this._hostingLayer = null;
  // get layer content
  if (ADUtils.objectIsUndefined(this.layer) || this.layer === null) {
    this.createLayer();
  }
  // now track touches on our layer
  this.layer.addEventListener(ADStartEvent, this, false);
  // and also track for when we get added or removed from the document
  this.layer.addEventListener('DOMNodeInsertedIntoDocument', this, true);
  this.layer.addEventListener('DOMNodeRemovedFromDocument', this, true);
  // finally, attach a pointer to the view from the layer
  // so we can easily go from the DOM to our ADView instance
  this.layer._view = this;
};

/* ==================== toString ==================== */

ADView.prototype.toString = function () {
  return [this.constructor.name, '[', this._size.width, 'x', this._size.height, '@', this._position.x, ',', this._position.y, ']'].join('');
};

/* ==================== Synthetized APIs ==================== */

/**
*  @name ADView.prototype
*  @property {String} id The unique identifier for this view, maps to the id of the {@link layer}.
*  @property {Array} subviews The views for which this view is the parent view.
*  @property {ADView} superview The view's parent view.
*  @property {ADPoint} position The view's position relative to its parent view.
*  @property {ADSize} size The view's size.
*  @property {ADPoint} anchorPoint The point around which transforms will be applied, uses float coordinates.
*  @property {bool} doubleSided Whether the view shows its backface when rotated.
*  @property {int} zIndex The view's index in the z axis.
*  @property {String} transform The view's transform, any string that can be used to set the transform CSS property.
*  @property {bool} clipsToBounds Controls whether the layer's contents are clipped to fit the view's size. Maps to the CSS overflow property.
*  @property {bool} transitionsEnabled Whether transitions are enabled for further property changes.
*  @property {float} transitionsDuration The duration to use for transitions.
*  @property {DOMElement} layer The top container for the view. This can be provided by subclasses in the {@link #createLayer} method.
*  @property {DOMElement} hostingLayer The optional contents container for the view, defaulting to the {@link #layer} if not provided by subclasses within the {@link #createLayer} method.
*  @property {float} opacity The view's opacity, a number between 0 and 1.
*/

ADView.prototype.getId = function () {
  return this.layer.id;
};

ADView.prototype.setId = function (id) {
  this.layer.id = id;
};

ADView.prototype.setPosition = function (aPoint) {
  // do nothing if we're not changing position
  if (this._position.equals(aPoint)) {
    return;
  }
  //
  this._position = aPoint;
  this.updateLayerTransform();
};

ADView.prototype.setSize = function (aSize) {
  // do nothing if we're not changing size
  if (this._size.equals(aSize)) {
    return;
  }
  //
  var old_size = this._size.copy();
  //
  this._size = aSize;
  this.layer.style.width = aSize.width + 'px';
  this.layer.style.height = aSize.height + 'px';
  // auto-resize subviews if need be
  if (this.autoresizesSubviews) {
    this.resizeSubviewsWithOldSize(old_size);
  }
};

ADView.prototype.setTransform = function (newTransform) {
  this._transform = newTransform;
  this.updateLayerTransform();
};

ADView.prototype.setAnchorPoint = function (aPoint) {
  this._anchorPoint = aPoint;
  this.layer.style.webkitTransformOrigin = Math.round(aPoint.x * 100) + '% ' + Math.round(aPoint.y * 100) + '% 0';
};

ADView.prototype.setDoubleSided = function (doubleSided) {
  this._doubleSided = doubleSided;
  this.layer.style.webkitBackfaceVisibility = doubleSided ? 'visible' : 'hidden';
};

ADView.prototype.setZIndex = function (zIndex) {
  this._zIndex = zIndex;
  this.layer.style.zIndex = zIndex;
};

/**
 *  Updates the underlying <code>-webkit-transform</code> CSS property following a change to either
 *  {@link #position} or {@link #transform}
 *
 *  @private
 */
ADView.prototype.updateLayerTransform = function () {
  this.layer.style.webkitTransform = ADUtils.t(this._position.x, this._position.y) + this._transform;
};

ADView.prototype.getOpacity = function () {
  return Number(window.getComputedStyle(this.layer).opacity);
};

ADView.prototype.setOpacity = function (opacity) {
  this.layer.style.opacity = opacity;
};

ADView.prototype.setTransitionsEnabled = function (enabled) {
  if (enabled) {
    this.layer.style.webkitTransitionProperty = '-webkit-transform, opacity';
    this.layer.style.webkitTransitionDuration = this._transitionsDuration + 's';
  }
  else {
    this.layer.style.webkitTransitionDuration = '0s';
  }
  this._transitionsEnabled = enabled;
};

ADView.prototype.setTransitionsDuration = function (duration) {
  this.layer.style.webkitTransitionDuration = duration + 's';
  this._transitionsDuration = duration;
};

ADView.prototype.setClipsToBounds = function (clipsToBounds) {
  this._clipsToBounds = clipsToBounds;
  this.layer.style.overflow = clipsToBounds ? 'hidden' : 'visible';
};

/* ==================== SUBVIEWS ==================== */

ADView.prototype.getHostingLayer = function () {
  return (this._hostingLayer != null) ? this._hostingLayer : this.layer;
};

/**
 *  Adds a new subview to this view, placing at the end of the {@link #subviews} list.
 *
 *  @param {ADView} aView The new subview to be added.
 *
 *  @returns {ADView} The view that is being added as a subview to this view.
 */
ADView.prototype.addSubview = function (aView) {
  return this.insertSubviewAtIndex(aView, this.subviews.length);
};

/**
 *  Detaches this view from its {@link #superview}, also removing it from display.
 */
ADView.prototype.removeFromSuperview = function () {
  if (this.superview == null) {
    return;
  }
  // send notifications
  this.willMoveToSuperview(null);
  this.superview.willRemoveSubview(this);
  // remove the view from the superview's subviews array
  // and update the indices of the other views in the array
  var view_index = this._indexInSuperviewSubviews;
  this.superview.subviews.splice(view_index, 1);
  for (var i = view_index; i < this.superview.subviews.length; i++) {
    this.superview.subviews[i]._indexInSuperviewSubviews = i;
  }
  // remove the view's layer from its parent layer
  this.layer.parentNode.removeChild(this.layer);
  // track that we have no more superview
  this.superview = null;
  //
  this.didMoveToSuperview();
};

/**
 *  Adds a new subview to this view, placing at the index provided in the subviews list. This method
 *  has no effect if the index goes beyond the length of the list.
 *
 *  @param {ADView} aView The new subview to be added.
 *  @param {int} index The index where to add the subview.
 *
 *  @returns {ADView} The view that is being added as a subview to this view.
 */
ADView.prototype.insertSubviewAtIndex = function (aView, index) {
  // do nothing if index is too high
  if (index > this.subviews.length) {
    return;
  }
  // in case the subview was previously in this view already
  // decrease the index
  if (aView.superview === this) {
    index--;
  }
  // make sure that if that view was already somewhere else,
  // we remove it from its previous superview
  aView.removeFromSuperview();
  // notify the view it's being moved to a new superview
  aView.willMoveToSuperview(this);
  // adds the view to the subviews array
  this.subviews.splice(index, 0, aView);
  aView._indexInSuperviewSubviews = index;
  // update the index of the new subviews
  for (var i = index + 1; i < this.subviews.length; i++) {
    this.subviews[i]._indexInSuperviewSubviews = i;
  }
  // add the view to the DOM
  var host = this.hostingLayer;
  if (index == this.subviews.length - 1) {
    host.appendChild(aView.layer);
  }
  else {
    host.insertBefore(aView.layer, this.subviews[index + 1].layer);
  }
  // notify the view the move to a new superview succeeded
  aView.superview = this;
  aView.didMoveToSuperview();
  // notify that a new subview was added
  this.didAddSubview(aView);
  //
  return aView;
};

/**
 *  Adds a new subview to this view, placing it right after an existing subview. If the view being
 *  inserted was the subview of another view, it is removed from that subview.
 *
 *  @param {ADView} aView The new subview to be added.
 *  @param {ADView} siblingSubview The view that will be the new subview's previous sibling.
 *
 *  @returns {ADView} The view that is being added as a subview to this view.
 */
ADView.prototype.insertSubviewAfterSubview = function (aView, siblingSubview) {
  // do nothing if the siblingSubview is not actually a subview
  if (siblingSubview.superview !== this) {
    return;
  }
  // insert before siblingView's next sibling if it exists
  var insertion_index = siblingSubview._indexInSuperviewSubviews + 1;
  if (insertion_index < this.subviews.length) {
    this.insertSubviewAtIndex(aView, insertion_index);
  }
  else {
    this.addSubview(aView);
  }
  //
  return aView;
};

/**
 *  Adds a new subview to this view, placing it right before an existing subview. If the view being
 *  inserted was the subview of another view, it is removed from that subview.
 *
 *  @param {ADView} aView The new subview to be added.
 *  @param {ADView} siblingSubview The view that will be the new subview's next sibling.
 *
 *  @returns {ADView} The view that is being added as a subview to this view.
 */
ADView.prototype.insertSubviewBeforeSubview = function (aView, siblingSubview) {
  // do nothing if the siblingSubview is not actually a subview
  if (siblingSubview.superview !== this) {
    return;
  }
  // simply call insertSubviewAtIndex with the index of the sibling view
  return this.insertSubviewAtIndex(aView, siblingSubview._indexInSuperviewSubviews);
};

/**
 *  Switches two subviews in the subviews list. If either of the provided indices do not match
 *  a subview, this method has no effect.
 *
 *  @param {int} firstIndex The index of the first subview.
 *  @param {int} secondIndex The index of the second subview.
 */
ADView.prototype.exchangeSubviewsAtIndices = function (firstIndex, secondIndex) {
  // do nothing if there are no such subviews
  if (firstIndex >= this.subviews.length || secondIndex >= this.subviews.length) {
    return;
  }
  // re-order views in the subviews array
  var firstView = this.subviews[firstIndex];
  var secondView = this.subviews[secondIndex];
  this.subviews[firstIndex] = secondView;
  this.subviews[secondIndex] = firstView;
  // make sure we update the indices attached to the layer as well
  firstView._indexInSuperviewSubviews = secondIndex;
  secondView._indexInSuperviewSubviews = firstIndex;
  // now, re-order elements in the DOM
  var firstElement = firstView.layer;
  var secondElement = secondView.layer;
  var host = this.hostingLayer;
  var firstElementNextSibling = firstElement.nextSibling;
  var secondElementNextSibling = secondElement.nextSibling;
  // now, make sure we have a sibling to insertBefore, otherwise
  // that means one of the views provided was the last child
  // and thus we need to appendChild
  if (firstElementNextSibling != null) {
    host.insertBefore(secondElement, firstElementNextSibling);
  }
  else {
    host.appendChild(secondElement)
  }
  if (secondElementNextSibling != null) {
    host.insertBefore(firstElement, secondElementNextSibling);
  }
  else {
    host.appendChild(firstElement)
  }
};

/**
 *  Indicates whether this view is a descendant of another view.
 *
 *  @param {ADView} aParentView The supposed parent of this view.
 *
 *  @returns {bool} A boolean indicating whether this view is a descendant of the provided view.
 */
ADView.prototype.isDescendantOfView = function (aParentView) {
  var found = false;
  var view = this;
  while (view.superview != null) {
    if (view.superview === aParentView) {
      found = true;
      break;
    }
    view = view.superview;
  }
  return found;
};

/* ==================== Notification Methods ==================== */

/**
 *  Called within the {@link ADView} constructor and allows to set the {@link #layer} property.
 */
ADView.prototype.createLayer = function () {
  this.layer = document.createElement('div');
  this.layer.className = 'ad-view';
};

/**
 *  Notifies the view is about to be added to a new superview.
 *
 *  @param {ADView} newSuperview The new superview for this view, null if the view is being removed from the view hierarchy.
 */
ADView.prototype.willMoveToSuperview = function (newSuperview) {
};

/**
 *  Notifies the view was moved to a new superview, which can then be accessed directly with the {@link #superview} property.
 */
ADView.prototype.didMoveToSuperview = function () {
};

/**
 *  Notifies the view that a new view was added to the {@link #subviews} array.
 *
 *  @param {ADView} subview The new subview that was added.
 */
ADView.prototype.didAddSubview = function (subview) {};

/**
 *  Notifies the view that a subview is about to be removed.
 *
 *  @param {ADView} subview The subview to be removed.
 */
ADView.prototype.willRemoveSubview = function (subview) {};

/**
 *  Notifies the view that its layer was added to the document.
 */
ADView.prototype.layerWasInsertedIntoDocument = function () {
  this.layerIsInDocument = true;
};

/**
 *  Notifies the view that its layer was removed from the document.
 */
ADView.prototype.layerWasRemovedFromDocument = function () {
};

ADView.prototype.handleLayerInsertionIntoDocument = function (event) {
  // we're only interested in the element that _actually_ was inserted
  if (event.target !== this.layer || this.layer.ownerDocument !== document || this.layerIsInDocument) {
    return;
  }
  // make sure the event won't go further down the tree as each layer gets its own DOMNodeInsertedIntoDocument event
  event.stopPropagation();
  // and let the notification helper know about this view
  ADView.layerInsertionNotificationHelper.considerView(this);
};

ADView.prototype.dispatchNotificationOfLayerInsertionIntoDocument = function () {
  // sort all subviews by depth
  var subviews_sorted_by_depth = [];
  var queue = [].concat(this.subviews);
  var depth = 0;
  while (queue.length > 0) {
    var n = queue.shift();
    if (typeof(n) === "number") {
      depth = n;
    } else {
      if (subviews_sorted_by_depth[depth] === undefined) {
        subviews_sorted_by_depth[depth] = [];
      }
      subviews_sorted_by_depth[depth].push(n);
      queue = queue.concat(depth + 1, n.subviews);
    }
  }
  
  // now call .layerWasInsertedIntoDocument() on all of those started with the deepest and going back up
  var subviews;
  while (subviews_sorted_by_depth.length > 0) {
    subviews = subviews_sorted_by_depth.pop();
    while (subviews.length > 0) {
      subviews.pop().layerWasInsertedIntoDocument();
    }
  }

  // finally, notify ourselves that we were added to the document
  this.layerWasInsertedIntoDocument();
};

ADView.prototype.handleLayerRemovalFromDocument = function (event) {
  if (event.target === this.layer && this.layer.ownerDocument === document) {
    // make sure the event won't go further down the tree as each layer gets its own DOMNodeRemovedFromDocument event
    event.stopPropagation();
    // trigger the notification method
    this.layerIsInDocument = false;
    this.layerWasRemovedFromDocument();
  }
};

/* ==================== Touch Tracking ==================== */

/**
 *  Called whenever an event that was tracked with this view as the <code>listener</code> parameter to <code>addEventListener</code>
 *  is dispatched. This method performs the event routing to methods such as {@link #touchesBegan} and is meant to be extended by ADView subclasses.
 *
 *  @param {TouchEvent} event The event.
 */ 
ADView.prototype.handleEvent = function (event) {
  switch (event.type) {
    case ADStartEvent :
      this.touchesBegan(event);
      break;
    case ADMoveEvent :
      this.touchesMoved(event);
      break;
    case ADEndEvent :
      this.touchesEnded(event);
      break;
    case 'touchcancel' :
      this.touchesCancelled(event);
      break;
    case 'DOMNodeInsertedIntoDocument' :
      this.handleLayerInsertionIntoDocument(event);
      break;
    case 'DOMNodeRemovedFromDocument' :
      this.handleLayerRemovalFromDocument(event);
      break;
  }
};

/**
 *  Called when a <code>touchstart</code> event was received on the view. Set {@link #tracksTouchesOnceTouchesBegan} to <code>false</code>
 *  if you're not interested in tracking events past this method call. Should you want all touch events tracking to be disabled, use the
 *  {@link #userInteractionEnabled} property. This method is meant to be extended by ADView subclasses.
 *
 *  @param {TouchEvent} event The event.
 */ 
ADView.prototype.touchesBegan = function (event) {
  // nothing to do if we don't have user interaction supported
  if (!this.userInteractionEnabled) {
    return;
  }
  // add move event handlers if view cares about touch tracking
  if (this.tracksTouchesOnceTouchesBegan) {
    window.addEventListener(ADMoveEvent, this, true);
    window.addEventListener(ADEndEvent, this, true);
    window.addEventListener('touchcancel', this, true);
  }
};

/**
 *  Called when a <code>touchmove</code> event was received on the view. This method is meant to be extended by ADView subclasses.
 *
 *  @param {TouchEvent} event The event.
 */
ADView.prototype.touchesMoved = function (event) {
  // nothing to do if we don't have user interaction supported
  if (!this.userInteractionEnabled) {
    return;
  }
  event.preventDefault();
};

/**
 *  Called when a <code>touchend</code> event was received on the view. This method is meant to be extended by ADView subclasses.
 *
 *  @param {TouchEvent} event The event.
 */
ADView.prototype.touchesEnded = function (event) {
  // nothing to do if we don't have user interaction supported
  if (!this.userInteractionEnabled) {
    return;
  }
  // remove event handlers
  window.removeEventListener(ADMoveEvent, this, true);
  window.removeEventListener(ADEndEvent, this, true);
  window.removeEventListener('touchcancel', this, true);
};

/**
 *  Called when a <code>touchcancel</code> event was received on the view. This method is meant to be extended by ADView subclasses.
 *
 *  @param {TouchEvent} event The event.
 */
ADView.prototype.touchesCancelled = function (event) {
  // nothing to do if we don't have user interaction supported
  if (!this.userInteractionEnabled) {
    return;
  }
  // remove event handlers
  window.removeEventListener(ADMoveEvent, this, true);
  window.removeEventListener(ADEndEvent, this, true);
  window.removeEventListener('touchcancel', this, true);
};

/**
 *  Returns a Boolean value indicating whether the receiver contains the specified point.
 *
 *  @param {ADPoint} point A point that is in the receiver's coordinate system.
 *
 *  @returns {bool} <code>true</code> if <code>point</code> is inside the view's bounds; otherwise, <code>false</code>.
 */
ADView.prototype.pointInside = function (point) {
  return (point.x >= 0 && point.x <= this.size.width && point.y >= 0 && point.y <= this.size.height);
};

/* ==================== Autosizing ==================== */

/**
 *  The view does not resize.
 *  @constant
 *  @type int
 */
const ADViewAutoresizingNone                 = 0;
/**
 *  The view resizes by expanding or shrinking in the direction of the left margin.
 *  @constant
 *  @type int
 */
const ADViewAutoresizingFlexibleLeftMargin   = 1 << 0;
/**
 *  The view resizes by expanding or shrinking its width.
 *  @constant
 *  @type int
 */
const ADViewAutoresizingFlexibleWidth        = 1 << 1;
/**
 *  The view resizes by expanding or shrinking in the direction of the right margin.
 *  @constant
 *  @type int
 */
const ADViewAutoresizingFlexibleRightMargin  = 1 << 2;
/**
 *  The view resizes by expanding or shrinking in the direction of the top margin.
 *  @constant
 *  @type int
 */
const ADViewAutoresizingFlexibleTopMargin    = 1 << 3;
/**
 *  The view resizes by expanding or shrinking its height.
 *  @constant
 *  @type int
 */
const ADViewAutoresizingFlexibleHeight       = 1 << 4;
/**
 *  The view resizes by expanding or shrinking in the direction of the bottom margin.
 *  @constant
 *  @type int
 */
const ADViewAutoresizingFlexibleBottomMargin = 1 << 5;

ADView.prototype.resizeSubviewsWithOldSize = function (oldSize) {
  for (var i = 0; i < this.subviews.length; i++) {
    this.subviews[i].resizeWithOldSuperviewSize(oldSize);
  }
}

ADView.prototype.resizeWithOldSuperviewSize = function (oldSize) {
  // set up the base position and sizes to be the same as current
  var position = this._position.copy();
  var size = this._size.copy();
  // get a local pointer to our mask
  var mask = this.autoresizingMask;
  // handle x axis
  var x_mask = (mask & ADViewAutoresizingFlexibleLeftMargin) + (mask & ADViewAutoresizingFlexibleWidth) + (mask & ADViewAutoresizingFlexibleRightMargin);
  switch (x_mask) {
    case ADViewAutoresizingNone:
      // nothing to do here
      break;
    case ADViewAutoresizingFlexibleLeftMargin:
      // offset left by amount of change
      position.x += this.superview._size.width - oldSize.width;
      break;
    case ADViewAutoresizingFlexibleWidth:
      // keep left and right margin, scale width
      size.width = this.superview._size.width - (oldSize.width - this._size.width);
      break;
    case ADViewAutoresizingFlexibleLeftMargin | ADViewAutoresizingFlexibleWidth:
      // keep right margin, scale left and width
      var right_margin = (oldSize.width - this._size.width - this._position.x);
      position.x = (this._position.x / (oldSize.width - right_margin)) * (this.superview._size.width - right_margin);
      size.width = this.superview._size.width - position.x - right_margin;
      break;
    case ADViewAutoresizingFlexibleRightMargin:
      // nothing to do here
      break;
    case ADViewAutoresizingFlexibleLeftMargin | ADViewAutoresizingFlexibleRightMargin:
      // offset left by amount of change weighted on the respective margin sizes
      var right_margin = (oldSize.width - this._size.width - this._position.x);
      position.x += (this.superview._size.width - oldSize.width) * (this.position.x / (this.position.x + right_margin));
      break;
    case ADViewAutoresizingFlexibleRightMargin | ADViewAutoresizingFlexibleWidth:
      // keep left margin, scale width and right
      var right_margin = (oldSize.width - this._size.width - this._position.x);
      scaled_right_margin = (right_margin / (oldSize.width - this._position.x)) * (this.superview._size.width - this._position.x);
      size.width = this.superview._size.width - position.x - scaled_right_margin;
      break;
    case ADViewAutoresizingFlexibleLeftMargin | ADViewAutoresizingFlexibleWidth | ADViewAutoresizingFlexibleRightMargin:
      // scale everything proportionally
      position.x = (this._position.x / oldSize.width) * this.superview._size.width;
      size.width = (this._size.width / oldSize.width) * this.superview._size.width;
      break;
  }
  // handle y axis
  var y_mask = (mask & ADViewAutoresizingFlexibleTopMargin) + (mask & ADViewAutoresizingFlexibleHeight) + (mask & ADViewAutoresizingFlexibleBottomMargin);
  switch (y_mask) {
    case ADViewAutoresizingNone:
      // nothing to do here
      break;
    case ADViewAutoresizingFlexibleTopMargin:
      // offset top by amount of change
      position.y += this.superview._size.height - oldSize.height;
      break;
    case ADViewAutoresizingFlexibleHeight:
      // keep top and bottom margin, scale height
      size.height = this.superview._size.height - (oldSize.height - this._size.height);
      break;
    case ADViewAutoresizingFlexibleTopMargin | ADViewAutoresizingFlexibleHeight:
      // keep bottom margin, scale top and height
      var bottom_margin = (oldSize.height - this._size.height - this._position.y);
      position.y = (this._position.y / (oldSize.height - bottom_margin)) * (this.superview._size.height - bottom_margin);
      size.height = this.superview._size.height - position.y - bottom_margin;
      break;
    case ADViewAutoresizingFlexibleBottomMargin:
      // nothing to do here
      break;
    case ADViewAutoresizingFlexibleTopMargin | ADViewAutoresizingFlexibleBottomMargin:
      // offset top by amount of change weighted on the respective margin sizes
      var bottom_margin = (oldSize.height - this._size.height - this._position.y);
      position.y += (this.superview._size.height - oldSize.height) * (this.position.y / (this.position.y + bottom_margin));
      break;
    case ADViewAutoresizingFlexibleBottomMargin | ADViewAutoresizingFlexibleHeight:
      // keep top margin, scale height and bottom
      var bottom_margin = (oldSize.height - this._size.height - this._position.y);
      scaled_bottom_margin = (bottom_margin / (oldSize.height - this._position.y)) * (this.superview._size.height - this._position.y);
      size.height = this.superview._size.height - position.y - scaled_bottom_margin;
      break;
    case ADViewAutoresizingFlexibleTopMargin | ADViewAutoresizingFlexibleHeight | ADViewAutoresizingFlexibleBottomMargin:
      // scale everything proportionally
      position.y = (this._position.y / oldSize.height) * this.superview._size.height;
      size.height = (this._size.height / oldSize.height) * this.superview._size.height;
      break;
  }
  // assign new metrics to our view
  this.position = position;
  this.size = size;
};

/* ==================== CSS Properties Mapping for Transitions ==================== */

const ADViewPropertyMapping = {
  'opacity' : 'opacity',
  'transform' : '-webkit-transform',
  'position' : '-webkit-transform',
  'anchorPoint' : '-webkit-transform-origin',
  'doubleSided' : '-webkit-backface-visibility',
  'zIndex' : 'z-index'
};

/**
 *  Provides the mapping from a ADView property to a CSS property, used by {@link ADTransition}
 *
 *  @private
 */
ADView.prototype.cssPropertyNameForJSProperty = function (jsPropertyName) {
  return ADViewPropertyMapping[jsPropertyName];
};

/* ==================== Transitions ==================== */

/**
 *  Applies a transition definition to the view, allowing the transition to be reversed. If this method is called
 *  within a {@link ADTransaction}, the transition will only be commited when the transaction is completed.
 *
 *  @param {ADTransitionDefinition} transitionDefinition The transition applied to the view.
 *  @param {bool} reversed Indicates whether the transition is to be applied in reverse.
 */
ADView.prototype.applyTransition = function (definition, reversed) {
  // nothing to do if we have no definition
  if (definition === null) {
    return;
  }
  // create a ADTransition from the definition
  var transition = new ADTransition(definition);
  // this view will be the target
  transition.target = this;
  // revert from / to values as instructed
  if (reversed) {
    var from = transition.from;
    transition.from = transition.to;
    transition.to = from;
  }
  // set up base properties, if any
  if (definition.base) {
    for (var i = 0; i < definition.base.length; i += 2) {
      this[definition.base[i]] = definition.base[i+1];
    }
  }
  // start the transition
  transition.start();
};

/* ==================== CLASS METHODS ==================== */

/**
 *  Returns the view with the given {@link #id}.
 *
 *  @param {String} id The view's id.
 *  @returns {ADView} The view with the given {@link #id}, or <code>null</code> if it does not exist.
 */
ADView.getViewById = function (id) {
  var element = document.getElementById(id);
  return (element && !ADUtils.objectIsUndefined(element._view)) ? element._view : null;
};

ADClass(ADView);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== ADContentView ==================== */

ADContentView.inherits = ADView;

/**
 *  @class
 *
 *  <p>The ADContentView class is a very simple ADView subclass that simply allows to create a view based on an existing element in the DOM.
 *  Note that if this element was already contained in a DOM tree that is managed by a ADView instance, it is still necessary to add your
 *  content view as a subview to that view if you want the content view to be listed as a subview of that view.</p>
 *
 *  <p>Note that content views do not automatically update their {@link ADView#size} as content changes. In order to explicitely update the
 *  size of a content view, use the {@link #refreshSize} method.</p>
 *
 *  @extends ADView
 *  @since AdLib 0.1
 *
 *  @param {Object} elementOrSelector The element to use as the {@link ADView#layer} for the view. Can be either a CSS selector string
 *  or a <code>HTMLElement</code> object.
 */
function ADContentView (layerOrSelector) {
  // retrieve layer
  var layer = layerOrSelector;
  if (ADUtils.objectIsString(layerOrSelector)) {
    layer = document.querySelector(layerOrSelector);
  }
  //
  this.layer = layer;
  this.callSuper();
  this.layer.addClassName('ad-view');
  // auto-size to window size if <body>
  if (layer === document.body) {
    this.size = new ADSize(window.innerWidth, window.innerHeight);
  }
};

ADContentView.prototype.layerWasInsertedIntoDocument = function () {
  this.callSuper();
  // update internal size if we moved somewhere else in the tree
  this.refreshSize();
};

/**
 *  Updates the internal {@link ADView#size} of the view. This method should be called when content within the view changes.
 */
ADContentView.prototype.refreshSize = function () {
  var style = window.getComputedStyle(this.layer);
  this._size = new ADSize(parseInt(style.width), parseInt(style.height));
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADContentView);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== ADContentView ==================== */

ADRootView.inherits = ADContentView;
ADRootView.synthetizes = ['disablesDefaultScrolling'];

/**
 *  @class
 *
 *  <p>The <code>ADRootView</code> class represents a special type of view that is the root of a <code>ADView</code> hierarchy.
 *  There should be a single instance of this class stored in {@link ADRootView.sharedRoot} and it's usually created automatically
 *  by the AdLib runtime using the <code>&lt;body></code> element has the supporting layer. The user can however choose to change
 *  the root view by setting the {@link ADRootView.sharedRoot} property manually.
 *
 *  @extends ADContentView
 *  @since AdLib 0.1
 *
 *  @param {Object} elementOrSelector The element to use as the {@link ADView#layer} for the view. Can be either a CSS selector string
 *  or a <code>HTMLElement</code> object.
 *
 *  @property {ADRootView} sharedRoot The shared root of the AdLib view hierarchy.
 */
function ADRootView (layerOrSelector) {
  this.callSuper(layerOrSelector);
  //
  this._disablesDefaultScrolling = true;
  // now turn off automated scrolling on this view
  this.disablesDefaultScrolling = true;
  // track rotation changes if we're the <body>
  // and remove the ad-view class as it ought not
  // to be positioned or layered
  if (this.layer === document.body) {
    window.addEventListener('orientationchange', this, false);
    this.layer.removeClassName('ad-view');
  }
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADRootView.prototype
 *  @property {bool} disablesDefaultScrolling Indicates whether the built-in <code>UIWebView</code> page scrolling behavior should be
 *  overriden in this view. The default value is <code>true</code>.
 */

ADRootView.prototype.setDisablesDefaultScrolling = function (disabled) {
  this.layer[disabled ? 'addEventListener' : 'removeEventListener'](ADMoveEvent, ADUtils.preventEventDefault, false);
  this._disablesDefaultScrolling = disabled;
};

/* ==================== Dealing with orientation change ==================== */

ADRootView.prototype.handleEvent = function (event) {
  this.callSuper(event);
  //
  if (event.type == 'orientationchange') {
    var _this = this;
    setTimeout(function () {
      _this.size = new ADSize(window.innerWidth, window.innerHeight);
      window.scrollTo(0,0);
    }, 0);
  }
};

/* ==================== Shared Instance ==================== */

ADRootView._sharedRoot = null;

ADRootView.__defineGetter__('sharedRoot', function () {
  // set up the shared instance if it's not been done before
  if (ADRootView._sharedRoot === null) {
    ADRootView._sharedRoot = new ADRootView(document.body);
  }
  return ADRootView._sharedRoot;
});

ADRootView.__defineSetter__('sharedRoot', function (newInstance) {
  ADRootView._sharedRoot = newInstance;
});

/* ==================== CLASS CREATION ==================== */

ADClass(ADRootView);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

const ADHTMLFragmentLoaderDidFail = 'htmlFragmentLoaderDidFail';
const ADHTMLFragmentLoaderDidLoad = 'htmlFragmentLoaderDidLoad';

/* ==================== ADHTMLFragmentLoader ==================== */

ADViewController.inherits = ADObject;
ADHTMLFragmentLoader.includes = [ADEventTriage];

function ADHTMLFragmentLoader (url, delegate) {
  this.callSuper();
  //
  this.url = url;
  this.delegate = delegate;
  //
  this.fragment = null;
  this.request = new XMLHttpRequest();
  // register event handlers on our request
  this.request.addEventListener('load', this, false);
  this.request.addEventListener('error', this, false);
  // send the request right away if we have a URL
  if (this.url !== undefined) {
    this.load();
  }
};

/**
 *  Starts loading the fragment located at the receiver's URL.
 *  
 *  @returns {bool} Indicates whether the request is performed synchronously or asynchronously.
 */
ADHTMLFragmentLoader.prototype.load = function () {
  // ensure that any running request is canceled
  this.fragment = null;
  this.request.abort();
  // figure out if we're loading async or not, assuming that
  // anything that's not HTTP is local
  // XXX: check if we don't have a protocol if we're currently running local
  var loads_async = (this.url.indexOf('http') == 0);
  // set up the request
  this.request.open('GET', this.url, loads_async);
  // launch the request
  try {
    this.request.send();
  } catch (err) {
    this.requestDidFail();
  }
  return loads_async;
};

// automatically routed through ADEventTriage
ADHTMLFragmentLoader.prototype.handleLoad = function (event) {
  this.requestDidLoad();
};

// automatically routed through ADEventTriage
ADHTMLFragmentLoader.prototype.handleError = function (event) {
  this.requestDidFail();
};

ADHTMLFragmentLoader.prototype.requestDidLoad = function () {
  // get fragment from string
  this.fragment = ADUtils.createNodeFromString(this.request.responseText);
  // inform our delegate
  if (ADUtils.objectHasMethod(this.delegate, ADHTMLFragmentLoaderDidLoad)) {
    this.delegate[ADHTMLFragmentLoaderDidLoad](this);
  }
};

ADHTMLFragmentLoader.prototype.requestDidFail = function () {
  if (ADUtils.objectHasMethod(this.delegate, ADHTMLFragmentLoaderDidFail)) {
    this.delegate[ADHTMLFragmentLoaderDidFail](this);
  }
};

ADClass(ADHTMLFragmentLoader);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

const ADScrollIndicatorThickness = 7;
const ADScrollIndicatorEndSize = 3;

const ADScrollIndicatorTypeHorizontal = 'horizontal';
const ADScrollIndicatorTypeVertical = 'vertical';

/* ==================== ADScrollIndicator ==================== */

ADScrollIndicator.inherits = ADView;
ADScrollIndicator.synthetizes = ['visible', 'width', 'height', 'style'];

function ADScrollIndicator (type) {
  this.callSuper();
  //
  this.type = type;
  this.layer.addClassName(type);
  //
  this._visible = false;
  this._width = ADScrollIndicatorThickness;
  this._height = ADScrollIndicatorThickness;
  //
  this.position = new ADPoint(-ADScrollIndicatorThickness, -ADScrollIndicatorThickness);
  this.positionBeforeHide = this.position;
  this.lastPositionUpdateInHide = false;
  //
  this._style = ADScrollViewIndicatorStyleDefault;
  //
  this.visible = false;
};

/* ==================== ADView APIs ==================== */

ADScrollIndicator.prototype.createLayer = function () {
  this.layer = document.createElement('div');
  this.layer.addClassName('ad-scroll-indicator');
  this.layer.addEventListener('webkitTransitionEnd', this, false);
  // create the three sections
  this.start = this.layer.appendChild(document.createElement('div'));
  this.middle = this.layer.appendChild(document.createElement('img'));
  this.end = this.layer.appendChild(document.createElement('div'));
};

ADScrollIndicator.prototype.setPosition = function (aPoint) {
  aPoint.x = Math.round(aPoint.x);
  aPoint.y = Math.round(aPoint.y);
  //
  this.callSuper(aPoint);
  //
  this.lastPositionUpdateInHide = false;
};

ADScrollIndicator.prototype.setSize = function (aSize) {
  this.width = aSize.width;
  this.height = aSize.height;
  this._size = aSize;
};

ADScrollIndicator.prototype.setStyle = function (aStyle) {
  this._style = aStyle;
  // update CSS class name
  this.layer.removeClassName(this._style);
  this.layer.addClassName(this._style);
  // update image
  var orientation = (this.type === ADScrollIndicatorTypeHorizontal) ? 'Horizontal' : 'Vertical';
  var style = 'Default';
  switch (aStyle) {
    case ADScrollViewIndicatorStyleWhite : 
      style = 'White';
      break;
    case ADScrollViewIndicatorStyleBlack : 
      style = 'Black';
      break;
  }
  this.middle.src = ADUtils.assetsPath + 'scrollindicator/UIScrollerIndicator' + style + orientation + 'Middle.png';
};

/* ==================== Synthetized Properties / APIs ==================== */

ADScrollIndicator.prototype.setWidth = function (width) {
  this.middle.style.webkitTransform = 'translate3d(0,0,0) scale(' + (width - ADScrollIndicatorEndSize * 2) + ',1)';
  this.end.style.webkitTransform = 'translate3d(' + (width - ADScrollIndicatorEndSize) + 'px,0,0)';
  this._width = width;
};

ADScrollIndicator.prototype.setHeight = function (height) {
  this.middle.style.webkitTransform = 'translate3d(0,0,0) scale(1,' + (height - ADScrollIndicatorEndSize * 2) + ')';
  this.end.style.webkitTransform = 'translate3d(0,' + (height - ADScrollIndicatorEndSize) + 'px,0)';
  this._height = height;
};

ADScrollIndicator.prototype.setVisible = function (visible) {
  if (visible) {
    this.fading = false;
    this.opacity = 1;
    this.position = this.lastPositionUpdateInHide ? this.positionBeforeHide : this.position;
  }
  else if (!this.fading) {
    this.fading = true;
    // XXX: we have a bug here where opacity is indeed
    // set to 0 and the engine does not update the view
    this.opacity = 0;
    // record previous position
    this.lastPositionUpdateInHide = true;
    this.positionBeforeHide = this.position;
  }
  this._visible = visible;
};

ADScrollIndicator.prototype.flash = function () {
  this.flashing = true;
  // XXX: this method isn't implemented anymore, where did it go?
  // this.show();
};

ADScrollIndicator.prototype.handleEvent = function (event) {
  if (event.type != 'webkitTransitionEnd') {
    return;
  }
  //
  this.callSuper(event);
  //
  if (this.flashing) {
    // XXX: this method isn't implemented anymore, where did it go?
    // this.hide();
    this.flashing = false;
  }
  else if (this.fading) {
    this.position = new ADPoint(-ADScrollIndicatorThickness, -ADScrollIndicatorThickness);
    this.fading = false;
  }
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADScrollIndicator);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  @class
 *  @name ADScrollViewDelegate
 *  @since AdLib 0.1
 */

/**
 *  Invoked when a {@link ADScrollView} is about to be dragged.
 *
 *  @name scrollViewWillBeginDragging
 *  @function
 *
 *  @param {ADScrollView} theScrollView The scroll view.
 *  @memberOf ADScrollViewDelegate.prototype
 */

/**
 *  Invoked when a {@link ADScrollView} is done animating to a new page.
 *
 *  @name scrollViewDidEndScrollingAnimation
 *  @function
 *
 *  @param {ADScrollView} theScrollView The scroll view.
 *  @memberOf ADScrollViewDelegate.prototype
 */

/**
 *  Invoked when a {@link ADScrollView} has scrolled.
 *
 *  @name scrollViewDidScroll
 *  @function
 *
 *  @param {ADScrollView} theScrollView The scroll view.
 *  @memberOf ADScrollViewDelegate.prototype
 */

/**
*  Invoked when the user has finished dragging a {@link ADScrollView}. Optional.
*
*  @name scrollViewDidEndDragging
*  @function
*
*  @param {ADScrollView} theScrollView The scroll view.
*  @memberOf ADScrollViewDelegate.prototype
*/

/**
*  Invoked when the scroll view is starting to decelerate the scrolling movement. Optional.
*
*  @name scrollViewWillBeginDecelerating
*  @function
*
*  @param {ADScrollView} theScrollView The scroll view.
*  @memberOf ADScrollViewDelegate.prototype
*/

/**
*  Invoked when the scroll view has ended decelerating the scrolling movement. Optional.
*
*  @name scrollViewDidEndDecelerating
*  @function
*
*  @param {ADScrollView} theScrollView The scroll view.
*  @memberOf ADScrollViewDelegate.prototype
*/

// delegate methods
const ADScrollViewWillBeginDragging = 'scrollViewWillBeginDragging';
const ADScrollViewDidEndScrollingAnimation = 'scrollViewDidEndScrollingAnimation';
const ADScrollViewDidScroll = 'scrollViewDidScroll';
const ADScrollViewDidEndDragging = 'scrollViewDidEndDragging';
const ADScrollViewWillBeginDecelerating = 'scrollViewWillBeginDecelerating';
const ADScrollViewDidEndDecelerating = 'scrollViewDidEndDecelerating';

// class constants
const ADScrollViewMinimumTrackingForDrag = 5;
const ADScrollViewPagingTransitionDuration = '0.25s';

// minimal length of a scroll indicator
const ADScrollViewMinIndicatorLength = 34;

// deceleration animation parameters
const ADScrollViewAcceleration = 15;
const ADScrollViewMaxTimeForTrackingDataPoints = 100;
const ADScrollViewDecelerationFrictionFactor = 0.950;
const ADScrollViewDesiredAnimationFrameRate = 1000 / 60; // 60fps
const ADScrollViewMinimumVelocity = 0.05;

// the bigger, the faster the deceleration
const ADScrollViewPenetrationDeceleration = 0.08;
// the bigger, the faster the snapping to the edge
const ADScrollViewPenetrationAcceleration = 0.15;

const ADScrollViewMinVelocityForDeceleration = 1;
const ADScrollViewMinVelocityForDecelerationWithPaging = 4;
const ADScrollViewMaxVelocityForBouncingWithPaging = 20;

// list of clickable elements
const ADScrollViewClickableElementNames = ['a', 'button', 'input', 'select'];
const ADScrollViewContentTouchesDelay = 150;

/**
 *  Used as a value for {@link ADScrollView#contentSize} to indicate that it should infer the content size automatically.
 *  @constant
 *  @type ADSize
 */
const ADScrollViewAutomatedContentSize = -1;

/**
 *  The default style of scroll indicator, which is black with a white border. This style is good against any content background.
 *  @constant
 *  @type String
 */
const ADScrollViewIndicatorStyleDefault = 'indicator-default';
/**
 *  A style of indicator which is black smaller than the default style. This style is good against a white content background.
 *  @constant
 *  @type String
 */
const ADScrollViewIndicatorStyleBlack = 'indicator-black';
/**
 *  A style of indicator is white and smaller than the default style. This style is good against a black content background.
 *  @constant
 *  @type String
 */
const ADScrollViewIndicatorStyleWhite = 'indicator-white';

/* ==================== ADScrollView ==================== */

ADScrollView.inherits = ADView;
ADScrollView.synthetizes = ['contentOffset', 'contentSize', 'indicatorStyle', 'scrollEnabled', 'scrollIndicatorInsets'];

/**
 *  @class
 *
 *  <p>The scroll view allows developers to embed contents in a scrollable container. The first
 *  step for the developer is to set the {@link #contentSize} property in order to indicate the size of
 *  the scrollable contents. Scrolling can then be done on both horizontal and vertical axes at
 *  the same time, and the {@link #horizontalScrollEnabled} and {@link #verticalScrollEnabled} properties control
 *  which way the content can scroll. Scroll indicators appear automatically as the user scrolls to
 *  indicate how far in the total size of the contents the user has scrolled. The
 *  {@link #showsHorizontalScrollIndicator} and {@link #showsVerticalScrollIndicator} properties tell
 *  the view whether these should show during the scroll interaction. Furthermore, the appearance of the
 *  indicators can be customized with the {@link #indicatorStyle} property.</p>
 *
 *  <p>While by default a scroll interaction will directly reflect where the user drags the content
 *  across the screen, it is sometimes desirable to snap the ensure the content always shows a given
 *  section of the content. To allow for this, the {@link #pagingEnabled} property when set to true will
 *  snap to the closest page, where a page size is the size of the viewable area. This allows, for
 *  instance, the creation of a slide view as exhibited in the Wheather iPhone application.</p>
 *
 *  <p>While the content offset position of the scroll view will generally be updated automatically
 *  following user interaction, it is sometimes desirable to manually update the content offset and
 *  the {@link #setContentOffsetWithAnimation} method allows to do just that, also allowing to animate to the
 *  new offset. Developers may also want to give the user an idea of the content offset by flashing
 *  the scroll indicators with the {@link #flashScrollIndicators} method, typically used when a view is
 *  transitioned back on screen.</p>

 *  <p>The scroll view, like most specific views, is not meant to be subclassed but rather used with a
 *  delegate in order to be notified of certain interactions. Using the {@link ADScrollViewDelegate} protocol,
 *  developers can be notified of scroll preparation or finalization. Should the developer want to
 *  create a ADScrollView subclass, the {@link #touchesBegan} and related methods will allow to extend the
 *  basic scrolling behavior, which the {@link ADTableView} class does for instance.</p>
 *
 *  <p>Should content within a <code>ADScrollView</code> need to override the default scrolling behavior,
 *  the {@link ADUtils.disableScrolling} method can be called within a <code>touchstart</code> event handler,
 *  or the {@link #touchesBegan} method of a <code>ADScrollView</code> subclass.</p>
 *  
 *  @extends ADView
 *  @since AdLib 0.1
 */
function ADScrollView () {
  this.callSuper();
  //
  this._contentOffset = new ADPoint();
  this._contentSize = ADScrollViewAutomatedContentSize;
  this.adjustedContentSize = new ADSize();
  /**
   *  Returns whether the user has touched the content to initiate scrolling. The property holds <code>true</code> if the user has touched
   *  the content view but might not have yet have started dragging it.
   *  @type bool
   */
  this.tracking = false;
  /**
   *  A Boolean value that indicates whether the user has begun scrolling the content. The value held by this property might require some time
   *  or distance of scrolling before it is set to <code>true</code>.
   *  @type bool
   */
  this.dragging = false;
  this.horizontalScrollEnabled = true;
  this.verticalScrollEnabled = true;
  /**
   *  A Boolean value that indicates whether the content is moving in the scroll view after the user lifted his or her finger.
   *  @type bool
   */
  this.decelerating = false;
  this.decelerationTimer = null;
  //
  this._indicatorStyle = '';
  this.showsHorizontalScrollIndicator = true; 
  this.showsVerticalScrollIndicator = true;
  this.scrollIndicatorsNeedFlashing = false;
  this._scrollIndicatorInsets = new ADEdgeInsets(0, 0, 0, 0);
  //
  this.pagingEnabled = false;
  this.bounces = true;
  //
  this.clipsToBounds = true;
  // object implementing our protocol
  this.delegate = null;
  // track transition end events
  this.layer.addEventListener('webkitTransitionEnd', this, false);
  this.hostingLayer.addEventListener('webkitTransitionEnd', this, false);
  // set up default indicator style
  this.indicatorStyle = ADScrollViewIndicatorStyleDefault;
  // make sure we get all touch events
  this.tracksTouchesOnceTouchesBegan = true;
  // add a capture event so we can cancel touches to begin with
  this.layer.addEventListener(ADStartEvent, this, true);
  /**
   *  A Boolean value that determines whether the scroll view delays the handling of touch-down gestures. If the value of this
   *  property is <code>true</code> (the default), the scroll view delays handling the touch-down gesture until it can determine
   *  if scrolling is the intent.
   *  @type bool
   */
  this.delaysContentTouches = true;
  /**
   *  A Boolean value that controls whether touches in the content view always lead to tracking. If the value of the property is <code>true</code>
   *  (the default) and a view in the content has begun tracking a finger touching it, and if the user drags the finger enough to initiate a scroll,
   *  the scroll view handles the touch as a scroll. If the value of the property is <code>false</code>, the scroll view does not scroll regardless
   *  of finger movement once the content view starts tracking.
   *  @type bool
   */
  this.canCancelContentTouches = true;
  //
  this._scrollEnabled = true;
  this._setContentOffsetWithAnimationCalledFromSetter = false;
};

/* ==================== ADView APIs ==================== */

ADScrollView.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-scroll-view');
  // add the scroll indicators
  this.horizontalScrollIndicator = new ADScrollIndicator(ADScrollIndicatorTypeHorizontal);
  this.verticalScrollIndicator = new ADScrollIndicator(ADScrollIndicatorTypeVertical);
  this.layer.appendChild(this.horizontalScrollIndicator.layer);
  this.layer.appendChild(this.verticalScrollIndicator.layer);
  // the hostingLayer is the layer to which scrolling is applied
  this.hostingLayer = this.layer.insertBefore(document.createElement('div'), this.horizontalScrollIndicator.layer);
  this.hostingLayer.className = 'hosting-layer';
};

ADScrollView.prototype.setSize = function (aSize) {
  this.callSuper(aSize);
  this.adjustContentSize(true);
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADScrollView.prototype
 *  @property {ADPoint} contentOffset The point at which the origin of the content view is offset from the origin of the scroll view. Use the {@link #setContentOffsetWithAnimation} to animate to a new value for this property.
 *  @property {ADSize} contentSize The size of the content view, {@link ADScrollViewAutomatedContentSize} by default.
 *  @property {bool} horizontalScrollEnabled Sets whether horizontal scrolling is possible. Defaults to <code>true</code>.
 *  @property {bool} verticalScrollEnabled Sets whether vertical scrolling is possible. Defaults to <code>true</code>.
 *  @property {bool} showsHorizontalScrollIndicator Sets whether the horizontal scroll indicator shows during dragging interaction. Only relevant if {@link #horizontalScrollEnabled} is set to <code>true</code>. Defaults to <code>true</code>.
 *  @property {bool} showsVerticalScrollIndicator Sets whether the vertical scroll indicator shows during dragging interaction. Only relevant if {@link #verticalScrollEnabled} is set to <code>true</code>. Defaults to <code>true</code>.
 *  @property {bool} pagingEnabled Controls whether scrolling should snap to the nearest content page, where a page is the size of the view. Defaults to <code>false</code>.
 *  @property {bool} bounces Controls whether the scroll view bounces past the edge of content and back again. Defaults to <code>true</code>.
 *  @property {Object} delegate Object implementing the {@link ADScrollViewDelegate} protocol.
 *  @property {String} indicatorStyle The style of the scroll indicators. The default style is {@link ADScrollViewIndicatorStyleDefault}.
 *  @property {bool} scrollEnabled A Boolean value that determines whether scrolling is enabled, <code>true</code> by default.
 *  @property {ADEdgeInsets} scrollIndicatorInsets The distance the scroll indicators are inset from the edge of the scroll view.
 */
 
ADScrollView.prototype.setScrollEnabled = function (scrollEnabled) {
  // update ivar
  this._scrollEnabled = scrollEnabled;
  // make sure we're not tracking anymore
  if (!scrollEnabled) {
    this.stopTrackingTouches();
  }
};

ADScrollView.prototype.setContentOffset = function (aPoint) {
  this._setContentOffsetWithAnimationCalledFromSetter = true;
  this.setContentOffsetWithAnimation(aPoint, false);
};

/**
 *  Sets the offset from the content view's origin with an optional animation.
 *
 *  @param {ADPoint} aPoint The new offset for this view's content.
 *  @param {bool} animated Specifies whether to animate to the new offset point.
 */
ADScrollView.prototype.setContentOffsetWithAnimation = function (aPoint, animated) {
  // check that we're scrolling to a different point
  if (aPoint.equals(this._contentOffset)) {
    return;
  }
  // update position
  this._contentOffset = aPoint;
  // make sure we scroll within the scrollable region, unless we're dragging or decelerating which allows bouncing
  if (!this.dragging && !this.decelerating) {
    this.adjustContentSize(false);
    this._contentOffset.x = Math.max(Math.min(this.maxPoint.x, this._contentOffset.x), 0);
    this._contentOffset.y = Math.max(Math.min(this.maxPoint.y, this._contentOffset.y), 0);
  }
  // apply new position to the hosting layer
  this.hostingLayer.style.webkitTransform = ADUtils.t(-this._contentOffset.x, -this._contentOffset.y);
  // turn on transitions if we are animating
  if (animated) {
    this.scrollTransitionsNeedRemoval = true;
    this.hostingLayer.style.webkitTransitionDuration = ADScrollViewPagingTransitionDuration;
  }
  // otherwise, do a direct notification of scrolling
  else {
    this.didScroll(false);
  }
  // update indicators if we're not animating
  if (!animated) {
    if (this.horizontalScrollEnabled && this.showsHorizontalScrollIndicator) {
      this.updateHorizontalScrollIndicator();
    }
    if (this.verticalScrollEnabled && this.showsVerticalScrollIndicator) {
      this.updateVerticalScrollIndicator();
    }
  }
  // notify property change if the method was called directly
  if (!this._setContentOffsetWithAnimationCalledFromSetter) {
    this.notifyPropertyChange('contentOffset');
  }
  this._setContentOffsetWithAnimationCalledFromSetter = false;
};

ADScrollView.prototype.snapContentOffsetToBounds = function (animated) {
  // handle special modes
  var should_move_to_point = false;
  var new_point = new ADPoint();
  // take into account paging
  if (this.pagingEnabled) {
    new_point.x = Math.round(this._contentOffset.x / this._size.width) * this._size.width;
    new_point.y = Math.round(this._contentOffset.y / this._size.height) * this._size.height;
    should_move_to_point = true;
  }
  // clip to not go out of bounds
  else if (this.bounces) {
    new_point.x = Math.max(Math.min(this.maxPoint.x, this._contentOffset.x), 0);
    new_point.y = Math.max(Math.min(this.maxPoint.y, this._contentOffset.y), 0);
    should_move_to_point = (new_point.x != this._contentOffset.x || new_point.y != this._contentOffset.y);
  }
  // update to the new point if we need to
  if (should_move_to_point) {
    this.setContentOffsetWithAnimation(new_point, animated);
  }
};

ADScrollView.prototype.getContentSize = function () {
  var size = this._contentSize;
  // we have automated sizing turned on
  if (size === ADScrollViewAutomatedContentSize) {
    size = new ADSize(this._hostingLayer.offsetWidth, this._hostingLayer.offsetHeight);
    // handle case where we have subviews
    if (this.subviews.length) {
      for (var i = 0; i < this.subviews.length; i++) {
        var subview = this.subviews[i];
        size.width = Math.max(size.width, subview.position.x + subview.size.width);
        size.height = Math.max(size.height, subview.position.y + subview.size.height);
      }
    }
  }
  return size;
};

ADScrollView.prototype.setContentSize = function (contentSize) {
  this._contentSize = contentSize;
  this.adjustContentSize(false);
};

ADScrollView.prototype.adjustContentSize = function (adjustsScrollOffset) {
  // track what the content offset was prior to adjusting the content size
  if (adjustsScrollOffset) {
    var content_offset_in_percent = new ADPoint();
    if (this.adjustedContentSize.width != 0) {
      content_offset_in_percent.x = this._contentOffset.x / this.adjustedContentSize.width;
    }
    if (this.adjustedContentSize.height != 0) {
      content_offset_in_percent.y = this._contentOffset.y / this.adjustedContentSize.height;
    }
  }
  // update the content size based on the new metrics
  this.adjustedContentSize.width = Math.max(this._size.width, this.contentSize.width);
  this.adjustedContentSize.height = Math.max(this._size.height, this.contentSize.height);
  // get max positions for the hosting layer
  this.maxPoint = new ADPoint(this.adjustedContentSize.width - this._size.width, this.adjustedContentSize.height - this._size.height);
  // update the content offset based on the old proportions we had
  if (adjustsScrollOffset) {
    this.contentOffset = new ADPoint(
      Math.min(content_offset_in_percent.x * this.adjustedContentSize.width, this.maxPoint.x),
      Math.min(content_offset_in_percent.y * this.adjustedContentSize.height, this.maxPoint.y)
    );
  }
};

ADScrollView.prototype.setIndicatorStyle = function (indicatorStyle) {
  this._indicatorStyle = indicatorStyle;
  this.horizontalScrollIndicator.style = indicatorStyle;
  this.verticalScrollIndicator.style = indicatorStyle;
};

ADScrollView.prototype.setScrollIndicatorInsets = function (insets) {
  this._scrollIndicatorInsets = insets;
  if (this.horizontalScrollIndicator.visible) {
    this.updateHorizontalScrollIndicator();
  }
  if (this.verticalScrollIndicator.visible) {
    this.updateVerticalScrollIndicator();
  }
};

/* ==================== SCROLL INDICATORS ==================== */

ADScrollView.prototype.updateHorizontalScrollIndicator = function () {
  var margins = (this.verticalScrollEnabled && this.showsVerticalScrollIndicator) ? ADScrollIndicatorEndSize * 2 : 1;
  //
  var available_width = this._size.width - this._scrollIndicatorInsets.left - this._scrollIndicatorInsets.right - margins;
  var width = Math.max(ADScrollViewMinIndicatorLength, Math.round((this._size.width / this.adjustedContentSize.width) * available_width));
  var x = (this._contentOffset.x / (this.adjustedContentSize.width - this._size.width)) * (available_width - margins - width) + this._scrollIndicatorInsets.left;
  var y = this._size.height - ADScrollIndicatorThickness - 1 - this._scrollIndicatorInsets.bottom;
  // handle bouncing case
  if (this._contentOffset.x < 0) {
    width = Math.round(Math.max(width + this._contentOffset.x, ADScrollIndicatorThickness));
    x = 1 + this._scrollIndicatorInsets.left;
  }
  else if (this._contentOffset.x > this.maxPoint.x) {
    width = Math.round(Math.max(width + this.adjustedContentSize.width - this._size.width - this._contentOffset.x, ADScrollIndicatorThickness));
    x = this._size.width - width - margins - this._scrollIndicatorInsets.right;
  }
  // update the indicator's properties
  this.horizontalScrollIndicator.position = new ADPoint(x, y);
  this.horizontalScrollIndicator.width = width;
};

ADScrollView.prototype.updateVerticalScrollIndicator = function () {
  var margins = (this.horizontalScrollEnabled && this.showsHorizontalScrollIndicator) ? ADScrollIndicatorEndSize * 2 : 1;
  //
  var available_height = this._size.height - this._scrollIndicatorInsets.top - this._scrollIndicatorInsets.bottom - margins;
  var height = Math.max(ADScrollViewMinIndicatorLength, Math.round((this._size.height / this.adjustedContentSize.height) * available_height));
  var x = this._size.width - ADScrollIndicatorThickness - 1 - this._scrollIndicatorInsets.right;
  var y = (this._contentOffset.y / (this.adjustedContentSize.height - this._size.height)) * (available_height - margins - height) + this._scrollIndicatorInsets.top;
  // handle bouncing case
  if (this._contentOffset.y < 0) {
    height = Math.round(Math.max(height + this._contentOffset.y, ADScrollIndicatorThickness));
    y = 1 + this._scrollIndicatorInsets.top;
  }
  else if (this._contentOffset.y > this.maxPoint.y) {
    height = Math.round(Math.max(height + this.adjustedContentSize.height - this._size.height - this._contentOffset.y, ADScrollIndicatorThickness));
    y = this._size.height - height - margins - this._scrollIndicatorInsets.bottom;
  }
  // update the indicator's properties
  this.verticalScrollIndicator.position = new ADPoint(x, y);
  this.verticalScrollIndicator.height = height;
};

/**
 *  Displays the scroll indicators momentarily. You should call this method whenever you bring the scroll view to front.
 *
 *  @param {bool} uponTransitionCompletion Specifies whether to delay the call until the end of any current transition running for this view, for instance if the view is being transitioned onto the screen.
 */
ADScrollView.prototype.flashScrollIndicators = function (upon_transition_completion) {
  if (upon_transition_completion) {
    this.scrollIndicatorsNeedFlashing = true;
    return;
  }
  //
  if (this.horizontalScrollEnabled && this.showsHorizontalScrollIndicator && (this.adjustedContentSize.width > this._size.width)) {
    this.updateHorizontalScrollIndicator();
    this.horizontalScrollIndicator.flash();
  }
  if (this.verticalScrollEnabled && this.showsVerticalScrollIndicator && (this.adjustedContentSize.height > this._size.height)) {
    this.updateVerticalScrollIndicator();
    this.verticalScrollIndicator.flash();
  }
};

ADScrollView.prototype.hideScrollIndicators = function () {
  this.horizontalScrollIndicator.visible = false;
  this.verticalScrollIndicator.visible = false;
};

/* ==================== Touch Tracking ==================== */

ADScrollView.prototype.handleEvent = function (event) {
  this.callSuper(event);
  //
  if (event.type == 'webkitTransitionEnd') {
    this.transitionEnded(event);
  }
};

ADScrollView.prototype.touchesBegan = function (event) {
  // nothing to do if we have scrolling turned off
  if (!this._scrollEnabled) {
    return;
  }
  // the event is still going down the tree, we need
  // to decide if we let it through to content or not
  if (event.eventPhase == Event.CAPTURING_PHASE) {
    // we don't care about manufactured events
    if (event._manufactured) {
      return;
    }
    // record original hit
    this.originalTarget = (ADSupportsTouches ? event.touches[0] : event).target;
    // if we delay content touches
    if (this.delaysContentTouches) {
      // start by making sure content will not see this touchstart event
      event.stopPropagation();
      // and propagate a touchstart in content after a delay
      this.callMethodNameAfterDelay('beginTouchesInContent', ADScrollViewContentTouchesDelay, event);
      //
      if (!this.tracking) {
        this.beginTracking(event);
      }
    }
  }
  // the event is going back up and has not been cancelled, means we can start tracking
  else {
    if (!this.tracking) {
      this.beginTracking(event);
    }
  }
};

// this is called after the ADScrollViewContentTouchesDelay delay has elapsed
ADScrollView.prototype.beginTouchesInContent = function (touchEvent) {
  // dispatch the touchstart event to children if we haven't dragged yet
  if (this.tracking && !this.dragging) {
    // create a new touchstart event with the cloned properties of the original touch
    var event = ADUtils.createUIEvent(ADStartEvent, touchEvent);
    // track this event was manufactured
    event._manufactured = true;
    //
    this.originalTarget.dispatchEvent(event);
    // stop tracking if we can't cancel the content touches
    if (!this.canCancelContentTouches) {
      this.touchesEnded(ADUtils.createUIEvent(ADEndEvent, touchEvent));
    }
  }
};

ADScrollView.prototype.beginTracking = function (event) {
  // first off, we make sure we won't be getting click events once we get a touchend
  // this will majorly help with performance when decelerating
  // note that this break click events going through to content
  event.preventDefault();
  // stop any running animation
  this.stopDecelerationAnimation();
  // ensure transitions are off on the scrolling layer
  this.hostingLayer.style.webkitTransitionDuration = 0;
  // adjust the content size
  this.adjustContentSize();
  // snap content offset to bounds in case we were out of bounds (could happen if we start tracking during deceleration)
  this.snapContentOffsetToBounds(false);
  // get useful data points about this interaction
  var content_offset = this._contentOffset.copy();
  // purge tracking data points and add a new one
  this.trackingDataPoints = [];
  this.addTrackingDataPoint(event.timeStamp, content_offset);
  // keep the very initial information about this touch around
  this.startContentOffset = content_offset;
  this.startTouchPosition = ADPoint.fromEventInElement(event, this.layer);
  // we're tracking
  this.tracking = true;
  this.dragging = false;
  this.touchesHaveMoved = false;
  // add move event handlers
  window.addEventListener(ADMoveEvent, this, true);
  window.addEventListener(ADEndEvent, this, true);
  window.addEventListener('touchcancel', this, true);
  // use an extra event listener for bubbling phase to give content a chance to pursue
  // another action that over-rides the selection by stopping propagation
  window.addEventListener(ADEndEvent, this, false);
};

/**
 *  Called when a <code>touchmove</code> event was received on the view. This method is meant to be extended by ADScrollView extensions.
 *
 *  @param {TouchEvent} event The event.
 */
ADScrollView.prototype.touchesMoved = function (event) {
  this.touchesHaveMoved = true;
  //
  this.callSuper(event);
  //
  var touch_position = ADPoint.fromEventInElement(event, this.layer);
  var delta_x = touch_position.x - this.startTouchPosition.x;
  var delta_y = touch_position.y - this.startTouchPosition.y;
  // check if dragging needs to be turned on
  if (!this.dragging) {
    if ((Math.abs(delta_x) >= ADScrollViewMinimumTrackingForDrag && this.horizontalScrollEnabled) ||
        (Math.abs(delta_y) >= ADScrollViewMinimumTrackingForDrag && this.verticalScrollEnabled)) {
      // let our delegate know we will be scrolling
      if (ADUtils.objectHasMethod(this.delegate, ADScrollViewWillBeginDragging)) {
        this.delegate[ADScrollViewWillBeginDragging](this);
      }
      // flag that we will be dragging from now on
      this.dragging = true;
      this.firstDrag = true;
      // show indicators
      if (this.horizontalScrollEnabled && this.showsHorizontalScrollIndicator && (this.adjustedContentSize.width > this._size.width)) {
        this.horizontalScrollIndicator.visible = true;
      }
      if (this.verticalScrollEnabled && this.showsVerticalScrollIndicator && (this.adjustedContentSize.height > this._size.height)) {
        this.verticalScrollIndicator.visible = true;
      }
    }
  }
  // now drag the content view if dragging is on
  if (this.dragging) {
    // prevent the event from getting inside subviews
    event.stopPropagation();
    // update scroll offset
    var new_x = this.horizontalScrollEnabled ? (this.startContentOffset.x - delta_x) : this._contentOffset.x;
    var new_y = this.verticalScrollEnabled ? (this.startContentOffset.y - delta_y) : this._contentOffset.y;
    // divide the distance outside of bounds if we allow bouncing
    if (this.bounces) {
      new_x -= ((new_x > this.maxPoint.x) ? (new_x - this.maxPoint.x) : ((new_x < 0) ? new_x : 0)) / 2;
      new_y -= ((new_y > this.maxPoint.y) ? (new_y - this.maxPoint.y) : ((new_y < 0) ? new_y : 0)) / 2;
    }
    // otherwise hard-clip to bounds
    else {
      new_x = Math.max(Math.min(this.maxPoint.x, new_x), 0);
      new_y = Math.max(Math.min(this.maxPoint.y, new_y), 0);
    }
    // in case it's out first drag, simply update the .startTouchPosition and wait for the next drag
    // to actually change the .contentOffset
    if (this.firstDrag) {
      this.firstDrag = false;
      this.startTouchPosition = touch_position;
      return;
    }
    // update content offset with new drag coordinates
    this.contentOffset = new ADPoint(new_x, new_y);
  }
  // track information about this touch
  this.addTrackingDataPoint(event.timeStamp, this._contentOffset.copy());
};

ADScrollView.prototype.touchesEnded = function (event) {
  this.callSuper(event);
  // say that we're not tracking anymore
  this.tracking = false;
  // clean up our dragging interaction if we were
  if (this.dragging) {
    this.dragging = false;
    // prevent the event from getting inside subviews
    event.stopPropagation();
    // make sure we've purged data points with our current time
    this.purgeTrackingDataPointsWithTime(event.timeStamp);
    // start smooth scrolling if need be
    if (this.trackingDataPoints.length > 1) {
      this._contentOffsetBeforeDeceleration = this._contentOffset.copy();
      this.startDecelerationAnimation();
    }
    // remove the bubble phase handler as it won't be triggered
    window.removeEventListener(ADEndEvent, this, false);
    // let our delegate know we ended dragging
    if (ADUtils.objectHasMethod(this.delegate, ADScrollViewDidEndDragging)) {
      this.delegate[ADScrollViewDidEndDragging](this);
    }
  }
  // if we're not animating, ensure the scroll indicators are hidden, as they may have been up
  // if the interaction was started while a deceleration animation was happening, and also
  // check that our content offset is well within the bounds
  if (!this.decelerating) {
    this.snapContentOffsetToBounds(true);
    this.hideScrollIndicators();
  }
  // also remove the bubbling phase event handler if we got all the way down there
  if (event.eventPhase == Event.BUBBLING_PHASE) {
    window.removeEventListener(ADEndEvent, this, false);
    // send a click event to the original target if we did not move
    if (!this.touchesHaveMoved && this.originalTarget !== null && event.type == ADEndEvent) {
      this.activateOriginalTarget();
    }
  }
};

ADScrollView.prototype.touchesCancelled = function (event) {
  this.callSuper(event);
  this.touchesEnded(event);
};

ADScrollView.prototype.stopTrackingTouches = function () {
  // nothing to do if we're not tracking touches already
  if (!this.tracking) {
    return;
  }
  // say that we're not tracking anymore
  this.tracking = false;
  // clean up our dragging interaction if we were
  if (this.dragging) {
    this.dragging = false;
    // snap content offset to the bounds of view
    this.snapContentOffsetToBounds(true);
    // let our delegate know we ended dragging
    if (ADUtils.objectHasMethod(this.delegate, ADScrollViewDidEndDragging)) {
      this.delegate[ADScrollViewDidEndDragging](this);
    }
    // hide scroll indicators immediately
    this.hideScrollIndicators();
  }
  // remove event handlers
  window.removeEventListener(ADMoveEvent, this, true);
  window.removeEventListener(ADEndEvent, this, true);
  window.removeEventListener(ADEndEvent, this, false);
  window.removeEventListener('touchcancel', this, true);
};

ADScrollView.prototype.purgeTrackingDataPointsWithTime = function (time) {
  // remove any stale data point
  while (this.trackingDataPoints.length > 0) {
    if (time - this.trackingDataPoints[0].time <= ADScrollViewMaxTimeForTrackingDataPoints) {
      break;
    }
    this.trackingDataPoints.shift();
  }
};

ADScrollView.prototype.addTrackingDataPoint = function (time, contentOffset) {
  this.purgeTrackingDataPointsWithTime(time);
  // add the new data point
  this.trackingDataPoints.push({
    time: time,
    contentOffset: contentOffset
  });
};

/* ==================== TRANSITION HOOKUPS ==================== */

ADScrollView.prototype.transitionEnded = function (event) {
  if (this.scrollIndicatorsNeedFlashing && event.currentTarget === this.layer) {
    this.scrollIndicatorsNeedFlashing = false;
    this.flashScrollIndicators();
  }
  if (this.scrollTransitionsNeedRemoval && event.currentTarget === this.hostingLayer) {
    this.scrollTransitionsNeedRemoval = false;
    this.hostingLayer.style.webkitTransitionDuration = 0;
    this.didScroll(true);
  }
};

/* ==================== Notifications ==================== */

ADScrollView.prototype.didScroll = function (animated) {
  // notify of animated scroll
  if (animated && ADUtils.objectHasMethod(this.delegate, ADScrollViewDidEndScrollingAnimation)) {
    this.delegate[ADScrollViewDidEndScrollingAnimation](this);
  }
  // notify of direct scroll
  if (ADUtils.objectHasMethod(this.delegate, ADScrollViewDidScroll)) {
    this.delegate[ADScrollViewDidScroll](this);
  }
};

/* ==================== Deceleration Animation and ADAnimator Delegation ==================== */

ADScrollView.prototype.startDecelerationAnimation = function () {
  // first, check that we're not currently outside of the extreme scrollable points, in which case we do not want to animate
  if (this.bounces && (this._contentOffset.x > this.maxPoint.x || this._contentOffset.y > this.maxPoint.y ||
      this._contentOffset.x < 0 || this._contentOffset.y < 0)) {
    return;
  }
  // might need to filter out direction changes there
  var first_data_point = this.trackingDataPoints[0];
  var last_data_point = this.trackingDataPoints[this.trackingDataPoints.length - 1];
  // get the distance
  var scroll_distance = new ADPoint(
    last_data_point.contentOffset.x - first_data_point.contentOffset.x,
    last_data_point.contentOffset.y - first_data_point.contentOffset.y
  );
  // compute the duration of the scroll, doing a little magic math
  // to get a duration on the right scale
  var scroll_duration = (last_data_point.time - first_data_point.time) / ADScrollViewAcceleration;
  // compute initial smooth scrolling velocity
  this.decelerationVelocity = new ADPoint(
    scroll_distance.x / scroll_duration,
    scroll_distance.y / scroll_duration
  );
  // compute min and max scrolling points, taking into account paging mode
  this.minDecelerationPoint = new ADPoint(0, 0);
  this.maxDecelerationPoint = this.maxPoint.copy();
  if (this.pagingEnabled) {
    this.minDecelerationPoint.x = Math.max(0, Math.floor(this._contentOffsetBeforeDeceleration.x / this._size.width) * this._size.width);
    this.minDecelerationPoint.y = Math.max(0, Math.floor(this._contentOffsetBeforeDeceleration.y / this._size.height) * this._size.height);
    this.maxDecelerationPoint.x = Math.min(this.maxPoint.x, Math.ceil(this._contentOffsetBeforeDeceleration.x / this._size.width) * this._size.width);
    this.maxDecelerationPoint.y = Math.min(this.maxPoint.y, Math.ceil(this._contentOffsetBeforeDeceleration.y / this._size.height) * this._size.height);
  }
  // compute the bounce factors
  this.penetrationDeceleration = ADScrollViewPenetrationDeceleration;
  this.penetrationAcceleration = ADScrollViewPenetrationAcceleration;
  if (this.pagingEnabled) {
    this.penetrationDeceleration *= 5;
  }
  // animate if we're beyond the threshold
  var min_velocity = this.pagingEnabled ? ADScrollViewMinVelocityForDecelerationWithPaging : ADScrollViewMinVelocityForDeceleration;
  if (Math.abs(this.decelerationVelocity.x) > min_velocity || Math.abs(this.decelerationVelocity.y) > min_velocity) {
    this.decelerating = true;
    this.decelerationTimer = this.callMethodNameAfterDelay('stepThroughDecelerationAnimation', ADScrollViewDesiredAnimationFrameRate);
    // record time we started
    this.lastFrame = new Date();
    // let our delegate know we're about to decelerate
    if (ADUtils.objectHasMethod(this.delegate, ADScrollViewWillBeginDecelerating)) {
      this.delegate[ADScrollViewWillBeginDecelerating](this);
    }
  }
};

ADScrollView.prototype.stopDecelerationAnimation = function () {
  this.decelerating = false;
  clearTimeout(this.decelerationTimer);
};

ADScrollView.prototype.stepThroughDecelerationAnimation = function (passthrough) {

  // nothing to do if we're not supposed to be decelerating
  if (!this.decelerating) {
    return;
  }

  // figure out how many frames we dropped since the last iteration
  var now = new Date();
  var elapsed_time = now - this.lastFrame;
  var missed_frames = passthrough ? 0 : (Math.round(elapsed_time / ADScrollViewDesiredAnimationFrameRate) - 1);
  // make up for lost time on the deceleration velocity
  for (var i = 0; i < missed_frames; i++) {
    this.stepThroughDecelerationAnimation(true);
  }

  // compute new values for x and y
  var new_x = this._contentOffset.x + this.decelerationVelocity.x;
  var new_y = this._contentOffset.y + this.decelerationVelocity.y;
  // hard-clip to bounds in case bouncing is turned off
  if (!this.bounces) {
    var clipped_x = Math.max(Math.min(this.maxPoint.x, new_x), 0);
    if (clipped_x != new_x) {
      new_x = clipped_x;
      this.decelerationVelocity.x = 0;
    }
    var clipped_y = Math.max(Math.min(this.maxPoint.y, new_y), 0);
    if (clipped_y != new_y) {
      new_y = clipped_y;
      this.decelerationVelocity.y = 0;
    }
  }

  // update only the properties of content offset so as not to trigger
  // any scrolling, but remember the values for the next passthrough loop
  if (passthrough) {
    this._contentOffset.x = new_x;
    this._contentOffset.y = new_y;
  }
  // otherwise, really update the content offset with the new positions
  else {
    this.contentOffset = new ADPoint(new_x, new_y);
  }

  // and update the scroll velocity to take into account deceleration
  // if we're not in paging mode in which case we want constant velocity
  // until we start bouncing
  if (!this.pagingEnabled) {
    this.decelerationVelocity.x *= ADScrollViewDecelerationFrictionFactor;
    this.decelerationVelocity.y *= ADScrollViewDecelerationFrictionFactor;
  }


  var abs_x_velocity = Math.abs(this.decelerationVelocity.x);
  var abs_y_velocity = Math.abs(this.decelerationVelocity.y);
  
  // hide the scroll indicators if we're below the velocity threshold
  if (!passthrough && abs_x_velocity <= ADScrollViewMinimumVelocity && abs_y_velocity <= ADScrollViewMinimumVelocity) {
    // hide the scroll indicators
    this.hideScrollIndicators();
    //
    this.decelerationAnimationCompleted();
    return;
  }

  // otherwise, let's just move to the next step
  if (!passthrough) {
    this.decelerationTimer = this.callMethodNameAfterDelay('stepThroughDecelerationAnimation', ADScrollViewDesiredAnimationFrameRate);
  }

  // check if we're beyond bounds if we have bouncing enabled
  if (this.bounces) {
    // figure out bounce penetration factor
    var penetration_factor = new ADPoint(0, 0);
    if (new_x < this.minDecelerationPoint.x) {
      penetration_factor.x = this.minDecelerationPoint.x - new_x;
    }
    else if (new_x > this.maxDecelerationPoint.x) {
      penetration_factor.x = this.maxDecelerationPoint.x - new_x;
    }
    if (new_y < this.minDecelerationPoint.y) {
      penetration_factor.y = this.minDecelerationPoint.y - new_y;
    }
    else if (new_y > this.maxDecelerationPoint.y) {
      penetration_factor.y = this.maxDecelerationPoint.y - new_y;
    }
    // update velocity based on penetration factor
    if (penetration_factor.x != 0) {
      // stop decelerating if the velocity is too high if we're paging
      if (this.pagingEnabled && Math.abs(this.decelerationVelocity.x) >= ADScrollViewMaxVelocityForBouncingWithPaging) {
        this.decelerationAnimationCompleted();
        return;
      }
      if (penetration_factor.x * this.decelerationVelocity.x <= 0) {
        this.decelerationVelocity.x += penetration_factor.x * this.penetrationDeceleration;
      }
      else {
        this.decelerationVelocity.x = penetration_factor.x * this.penetrationAcceleration;
      }
    }
    if (penetration_factor.y != 0) {
      // stop decelerating if the velocity is too high if we're paging
      if (this.pagingEnabled && Math.abs(this.decelerationVelocity.y) >= ADScrollViewMaxVelocityForBouncingWithPaging) {
        this.decelerationAnimationCompleted();
        return;
      }
      if (penetration_factor.y * this.decelerationVelocity.y <= 0) {
        this.decelerationVelocity.y += penetration_factor.y * this.penetrationDeceleration;
      }
      else {
        this.decelerationVelocity.y = penetration_factor.y * this.penetrationAcceleration;
      }
    }
  }

  // record the last time we actually updated the content offset
  // making sure we avoid the passthrough case
  if (!passthrough) {
    this.lastFrame = now;
  }

};

ADScrollView.prototype.decelerationAnimationCompleted = function () {
  // track that we're not decelerating anymore
  this.stopDecelerationAnimation();
  // make sure we snap to an increment if we have paging enabled
  if (this.pagingEnabled) {
    this.setContentOffsetWithAnimation(new ADPoint(
      Math.round(this._contentOffset.x / this._size.width) * this._size.width,
      Math.round(this._contentOffset.y / this._size.height) * this._size.height
    ), false);
  }
  // let our delegate know we're done decelerating
  if (ADUtils.objectHasMethod(this.delegate, ADScrollViewDidEndDecelerating)) {
    this.delegate[ADScrollViewDidEndDecelerating](this);
  }
};

/* ==================== Manually activating the original target ==================== */

// XXX: <select> does not react to click events to activate
ADScrollView.prototype.activateOriginalTarget = function () {
  // look for the closest parent element that would be listed in ADScrollViewClickableElementNames
  var node = this.originalTarget;
  while (node.parentNode && node !== this.hostingLayer) {
    if (node.nodeType == Node.ELEMENT_NODE) {
      if (ADScrollViewClickableElementNames.indexOf(node.localName) != -1) {
        break;
      }
    }
    node = node.parentNode;
  }
  // nothing to do here if we don't have touches support as the
  // original click event will be sent through
  if (!ADSupportsTouches) {
    return;
  }
  // create a new click event
  var event = document.createEvent('MouseEvent');
  event.initMouseEvent( 'click', true, true, document.defaultView,
                        0, 0, 0,
                        0, 0, 0,
                        0, 0, 0,
                        0, 0, null);
  // track that this is a manufactured event
  event._manufactured = true;
  // dispatch the event to the target
  node.dispatchEvent(event);
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADScrollView);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  @class
 *  @name ADTableViewDataSource
 *  @since AdLib 0.1
 */

/**
 *  Asks the data source for a cell to insert in a particular location of the table view. Required for non-custom styles,
 *  table views with a custom style will call the {@link #tableViewCustomCellForRowAtPath} method on its data source.
 *
 *  @name tableViewCellForRowAtPath
 *  @function
 *
 *  @param {ADTableView} tableView The table view object requesting the cell.
 *  @param {ADCellPath} path The index path locating the row in <code>tableView</code>.
 *  @returns {ADTableViewCell} The cell for the given path.
 *  @memberOf ADTableViewDataSource.prototype
 */

/**
 *  Tells the data source to return the number of rows in a given section of a table view. Required.
 *
 *  @name tableViewNumberOfRowsInSection
 *  @function
 *
 *  @param {ADTableView} tableView The table view object requesting this information.
 *  @param {int} section The index number identifying the section in <code>tableView</code>.
 *  @returns {int} The number of rows in <code>section</code>.
 *  @memberOf ADTableViewDataSource.prototype
 */

/**
 *  Asks the data source to return the number of sections in the table view. Optional.
 *
 *  @name numberOfSectionsInTableView
 *  @function
 *
 *  @param {ADTableView} tableView The table view object requesting this information.
 *  @returns {int} The number of sections in <code>tableView</code>. The default value is <code>1</code>.
 *  Values less than 1 are not valid and will fail immediately.
 *  @memberOf ADTableViewDataSource.prototype
 */

/**
 *  Asks the data source for the title of the header of the specified section of the table view. Optional.
 *
 *  @name tableViewTitleForHeaderInSection
 *  @function
 *
 *  @param {ADTableView} tableView The table view object requesting this information.
 *  @param {int} section An index number identifying a section of <code>tableView</code>.
 *  @returns {String} A string to use as the title of the section header. If you return <code>null</code>, the section will have no header.
 *  @memberOf ADTableViewDataSource.prototype
 */

/**
 *  Asks the data source for the title of the footer of the specified section of the table view. Optional.
 *
 *  @name tableViewTitleForFooterInSection
 *  @function
 *
 *  @param {ADTableView} tableView The table view object requesting this information.
 *  @param {int} section An index number identifying a section of <code>tableView</code>.
 *  @returns {String} A string to use as the title of the section footer. If you return <code>null</code>, the section will have no footer.
 *  @memberOf ADTableViewDataSource.prototype
 */

/**
 *  Asks the data source for an element to insert in a particular location of the custom-style table view. Required for custom styles,
 *  table views with non-custom style will call the {@link #tableViewCellForRowAtPath} method on its data source.
 *
 *  @name tableViewCustomCellForRowAtPath
 *  @function
 *
 *  @param {ADTableView} tableView The table view object requesting the cell.
 *  @param {ADCellPath} path The index path locating the row in <code>tableView</code>.
 *  @returns {HTMLDivElement} customCell The <code>&lt;div></code> element for the given path.
 *  @memberOf ADTableViewDataSource.prototype
 */

const ADTableViewCellForRowAtPath = 'tableViewCellForRowAtPath';
const ADTableViewNumberOfSectionsInTableView = 'numberOfSectionsInTableView';
const ADTableViewNumberOfRowsInSection = 'tableViewNumberOfRowsInSection';
const ADTableViewTitleForHeaderInSection = 'tableViewTitleForHeaderInSection';
const ADTableViewTitleForFooterInSection = 'tableViewTitleForFooterInSection';
const ADTableViewCustomCellForRowAtPath = 'tableViewCustomCellForRowAtPath';

/**
 *  @class
 *  @name ADTableViewDelegate
 *  @since AdLib 0.1
 */

/**
 *  Invoked when the {@link ADTableView} row at the given path was selected. Optional.
 *
 *  @name tableViewDidSelectRowAtPath
 *  @function
 *
 *  @param {ADTableView} tableView The table view.
 *  @param {ADCellPath} path The cell path.
 *  @memberOf ADTableViewDelegate.prototype
 */
const ADTableViewDidSelectRowAtPath = 'tableViewDidSelectRowAtPath';
/**
 *  Invoked when the {@link ADTableView} row at the given path had its accessory view selected. Optional.
 *
 *  @name tableViewDidSelectAccessoryForRowAtPath
 *  @function
 *
 *  @param {ADTableView} tableView The table view.
 *  @param {ADCellPath} path The cell path.
 *  @memberOf ADTableViewDelegate.prototype
 */
const ADTableViewDidSelectAccessoryForRowAtPath = 'tableViewDidSelectAccessoryForRowAtPath';

// CSS protocol
const ADTableViewCustomCellCSS = 'ad-custom-table-view-cell';

/**
 *  A plain table view. Any section headers or footers are displayed as inline separators and float when the table view is scrolled.
 *  @constant
 *  @type String
 */
const ADTableViewStylePlain = 'plain';

/**
 *  A table view whose sections present distinct groups of rows. The section headers and footers do not float.
 *  @constant
 *  @type String
 */
const ADTableViewStyleCustom = 'custom';

/**
 *  A table view whose sections present distinct groups of rows. The section headers and footers do not float.
 *  @constant
 *  @type String
 */
const ADTableViewStyleGrouped = 'grouped';

const ADTableViewMinTouchDurationForCellSelection = 150;

/* ==================== ADTableView ==================== */

ADTableView.inherits = ADScrollView;
ADTableView.synthetizes = ['style', 'separatorStyle'];
ADTableView.includes = [ADPropertyTriage];

/**
 *  @class
 *
 *  <p>The ADTableView class is a direct extension of the {@link ADScrollView} class and is tailor-made to
 *  display overflowing list of contents onto the screen. Using the ADTableView class amounts to creating a
 *  new instance, setting its size and position, implementing the required methods from the {@link ADTableViewDataSource}
 *  protocol and finally tell the view that data is ready for it to display using the {@link #reloadData}
 *  method. Should a controller wish to know when cells in the list are selected, the {@link ADTableViewDelegate}
 *  protocol's methods should be implemented. As this view is an extension of the {@link ADScrollView},
 *  it is critical to ensure that the {@link ADScrollView#contentSize} is set to reflect the total size
 *  of the list content.</p>
 *  
 *  @extends ADScrollView
 *  @since AdLib 0.1
 */
function ADTableView () {
  this.callSuper();
  //
  this._style = ADTableViewStyleCustom;
  this._separatorStyle = ADTableViewCellSeparatorStyleSingleLine;
  // customize scrolling behavior
  this.horizontalScrollEnabled = false;
  // object implementing our protocol
  this.delegate = null;
  this.dataSource = null;
  //
  this.touchedCell = null;
  this.touchedAccessory = null;
  this.shouldPreventScrolling = false;
  /**
   *  The number of sections for this table view, the cached value returned by the
   *  {@link ADTableViewDataSource#numberOfSectionsInTableView} method from this table view's {@link #dataSource}.
   *  @type int 
   */
  this.numberOfSections = 1;
  // private properties
  this.numberOfRows = [];
  this.sections = [];
  this.headers = [];
  this.sectionMetrics = [];
  this.selectedCell = null;
  // private property to track status of population
  this.populated = false;
};

/* ==================== ADView APIs ==================== */

ADTableView.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-table-view');
};

ADTableView.prototype.layerWasInsertedIntoDocument = function () {
  this.callSuper();
  //
  this.notifyCellsOfInsertionIntoDocument();
  // update section metrics now that we can reliably have some
  this.updateSectionMetrics();
};

/* ==================== ADScrollView APIs ==================== */

ADTableView.prototype.setContentOffsetWithAnimation = function (aPoint, animated) {
  this.callSuper(aPoint, animated);
  this.updateSectionHeaders();
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADTableView.prototype
 *  @property {Object} dataSource The required object acting as the data source for the view and implementing the {@link ADTableViewDataSource} protocol.
 *  @property {Object} delegate The object acting as a delegate for the view and implementing the {@link ADTableViewDelegate} protocol.
 *  @property {String} style The list style, {@link ADTableViewStyleCustom} by default.
 *  @property {String} separatorStyle The style for table cells used as separators. The value of this property is one of the separator-style constants
 *  with the <code>ADTableViewCell</code> prefix. <code>ADTableView</code> uses this property to set the separator style on the cell returned from the
 *  delegate in {@link ADTableViewDataSource#tableViewCellForRowAtPath}.
 */

ADTableView.prototype.setStyle = function (style) {
  this.layer.removeClassName(this._style);
  this.layer.addClassName(style);
  this._style = style;
};

ADTableView.prototype.setSeparatorStyle = function (separatorStyle) {
  this.layer.removeClassName(this._separatorStyle);
  this.layer.addClassName(separatorStyle);
  this._separatorStyle = separatorStyle;
};

/* ==================== Public APIs ==================== */

/**
 *  Returns the number of rows (table cells) in a specified section.
 *  @param {int} section An index number that identifies a section of the table, <code>0</code> being the first index.
 *  @returns {int} An index number that identifies a row in the given section.
 */
ADTableView.prototype.numberOfRowsInSection = function (section) {
  if (section > this.numberOfSections - 1) {
    return;
  }
  //
  return this.numberOfRows[section];
};

/**
 *  Returns the table cell at the specified path.
 *  @param {ADCellPath} path The path locating the row in the receiver.
 *  @returns {ADTableViewCell} An object representing a cell of the table or <code>null</code> if the cell is not visible or <code>path</code> is out of range.
 */
ADTableView.prototype.cellForRowAtPath = function (path) {
  if (this._style === ADTableViewStyleCustom || path.section > this.numberOfSections - 1 || path.row > this.numberOfRows[path.section]) {
    return null;
  }
  return this.hostingLayer.querySelector('.section:nth-of-type(' + (path.section + 1) + ') .cells > div:nth-of-type(' + (path.row + 1) + ')')._view;
};

/**
 *  Returns the custom cell at the specified path.
 *  @param {ADCellPath} path The path locating the row in the receiver.
 *  @returns {HTMLDivElement} An object representing a cell of the table or <code>null</code> if the cell is not visible or <code>path</code> is out of range.
 */
ADTableView.prototype.customCellForRowAtPath = function (path) {
  if (this._style !== ADTableViewStyleCustom || path.section > this.numberOfSections - 1 || path.row > this.numberOfRows[path.section]) {
    return null;
  }
  return this.hostingLayer.querySelector('.section:nth-of-type(' + (path.section + 1) + ') .cells > .' + ADTableViewCustomCellCSS + ':nth-of-type(' + (path.row + 1) + ')');
};

/**
 *  Returns a path representing the row and section of a given table view cell.
 *  @param {ADTableViewCell} cell A cell object of the table view.
 *  @returns {ADCellPath} An index path representing the row and section of the cell or nil if the index path is invalid.
 */
ADTableView.prototype.pathForCell = function (cell) {
  if (this._style === ADTableViewStyleCustom) {
    return null;
  }
  return cell._tableViewDataSourcePath;
};

/**
 *  Returns a path representing the row and section of a given table view custom cell.
 *  @param {HTMLDivElement} customCell A custom cell object of the table view.
 *  @returns {ADCellPath} An index path representing the row and section of the cell or nil if the index path is invalid.
 */
ADTableView.prototype.pathForCustomCell = function (customCell) {
  if (this._style !== ADTableViewStyleCustom) {
    return null;
  }
  return customCell._tableViewDataSourcePath;
};

/* ==================== Population ==================== */

// We need to manually tell cells that they have been added to the tree
// as we do not actually use .addSubview() to add the cells to the display.
ADTableView.prototype.notifyCellsOfInsertionIntoDocument = function () {
  // only act upon this when we're both in tree and populated
  if (!this.layerIsInDocument || !this.populated) {
    return;
  }
  // go through each cell and notify
  var cells = this.hostingLayer.querySelectorAll('.ad-table-view-cell');
  for (var i = 0; i < cells.length; i++) {
    cells[i]._view.layerWasInsertedIntoDocument();
  }
};

/**
 *  Notifies the view that there is new data to be drawn. This method will clear any existing content
 *  in the view and trigger the relevant {@link ADTableViewDataSource} protocol methods.
 */
ADTableView.prototype.reloadData = function () {
  // see if we're dealing with a custom table view
  var is_custom = (this._style === ADTableViewStyleCustom);
  // check validity of data source, custom style
  if (is_custom) {
    if (!ADUtils.objectHasMethod(this.dataSource, ADTableViewCustomCellForRowAtPath) || !ADUtils.objectHasMethod(this.dataSource, ADTableViewNumberOfRowsInSection)) {
      console.error("An ADTableView's dataSource must implement all required methods");
      return;
    }
  }
  // native styles
  else if (!ADUtils.objectHasMethod(this.dataSource, ADTableViewCellForRowAtPath) || !ADUtils.objectHasMethod(this.dataSource, ADTableViewNumberOfRowsInSection)) {
    console.error("An ADTableView's dataSource must implement all required methods");
    return;
  }
  // empty the list container
  this._hostingLayer.innerText = '';
  this.sections = [];
  this.headers = [];
  // get the number of sections, an optional delegate method, and cache it
  if (ADUtils.objectHasMethod(this.dataSource, ADTableViewNumberOfSectionsInTableView)) {
    this.numberOfSections = this.dataSource[ADTableViewNumberOfSectionsInTableView](this);
    if (this.numberOfSections < 1) {
      console.error('An ADTableView must have at least one section');
      return;
    }
  }
  // populate the list
  for (var section = 0; section < this.numberOfSections; section++) {
    var container = document.createElement('div');
    container.className = 'section';
    this.sections[section] = container;
    // create the title if there's one
    if (ADUtils.objectHasMethod(this.dataSource, ADTableViewTitleForHeaderInSection)) {
      var header = this.dataSource[ADTableViewTitleForHeaderInSection](this, section);
      if (header !== null) {
        var header_element = container.appendChild(document.createElement('h1'));
        header_element.innerText = header;
        this.headers[section] = header_element;
      }
    }
    // now get the number of rows and populate each of those, if we have any
    var number_of_rows = this.dataSource[ADTableViewNumberOfRowsInSection](this, section);
    if (number_of_rows > 0) {
      var cell_container = container.appendChild(document.createElement('div'));
      cell_container.className = 'cells';
      for (var row = 0; row < number_of_rows; row++) {
        var path = new ADCellPath(section, row);
        if (is_custom) {
          var cell = this.dataSource[ADTableViewCustomCellForRowAtPath](this, path);
          cell.addClassName(ADTableViewCustomCellCSS);
          cell._tableViewDataSourcePath = path;
          cell_container.appendChild(cell);
        }
        else {
          var cell = this.dataSource[ADTableViewCellForRowAtPath](this, path);
          cell._tableViewDataSourcePath = path;
          cell._table = this;
          cell_container.appendChild(cell.layer);
        }
      }
    }
    // create the footer if there's one
    if (ADUtils.objectHasMethod(this.dataSource, ADTableViewTitleForFooterInSection)) {
      var footer = this.dataSource[ADTableViewTitleForFooterInSection](this, section);
      if (footer !== null) {
        container.appendChild(document.createElement('span')).innerText = footer;
      }
    }
    // cache the number of rows
    this.numberOfRows[section] = number_of_rows;
    // add the section to the hosting layer
    this._hostingLayer.appendChild(container);
  }
  // track that we've populated the tree
  this.populated = true;
  // now get section positions if we have a plain-style table view
  this.updateSectionMetrics();
  // and notify cells if the table view is in the tree
  this.notifyCellsOfInsertionIntoDocument();
};

ADTableView.prototype.updateSectionMetrics = function () {
  // we can only do something useful if the layer is currently in the document,
  // the table is plain-styled and populated
  if (!this.layerIsInDocument || this._style !== ADTableViewStylePlain || !this.populated) {
    return;
  }
  // update metrics
  this.sectionMetrics = [];
  for (var i = 0; i < this.sections.length; i++) {
    this.sectionMetrics[i] = {
      y : this.sections[i].offsetTop,
      height : this.sections[i].offsetHeight
    };
  }
};

/* ==================== INTERACTION ==================== */

ADTableView.prototype.touchesBegan = function (event) {
  // nothing to do if we had a manufactured event
  if (event._manufactured) {
    return;
  }
  // see if we were decelerating
  this.wasDeceleratingWhenTouchesBegan = this.decelerating;
  //
  this.callSuper(event);
  //
  this.touchedCell = null;
  this.touchedAccessory = null;
  this.shouldPreventScrolling = false;
  // nothing to do if we were decelerating
  // or we did not start tracking (content prevented that)
  if (this.wasDeceleratingWhenTouchesBegan || !this.tracking) {
    return;
  }
  // get touched element
  var event_position_in_page = ADPoint.fromEvent(event);
  var touched_element = document.elementFromPoint(event_position_in_page.x, event_position_in_page.y);
  // see if we're dealing with a custom table view
  if (this._style === ADTableViewStyleCustom) {
    // look for a cell container
    var element = touched_element;
    while (element.parentNode) {
      if (element.hasClassName(ADTableViewCustomCellCSS)) {
        this.touchedCell = element;
        this.touchedCellWasSelected = this.touchedCell.hasClassName(ADControlStateSelectedCSS);
        break;
      }
      element = element.parentNode;
    }
  }
  // otherwise, we're dealing with a ADTableViewCell, maybe
  else {
    var touched_view = (touched_element._view !== undefined) ? touched_element._view : touched_element.getNearestView();
    if (touched_view instanceof ADTableViewCell) {
      // track what was touched
      this.touchedCell = touched_view;
      this.touchedCellWasSelected = this.touchedCell.selected;
    }
    else if (touched_view instanceof ADButton) {
      this.touchedAccessory = touched_view;
    }
  }
  // show cell as active with a delay
  if (this.touchedCell !== null || this.touchedAccessory !== null) {
    this.callMethodNameAfterDelay('detectedStationaryTouch', ADTableViewMinTouchDurationForCellSelection);
  }
};

ADTableView.prototype.touchesMoved = function (event) {
  if (this.shouldPreventScrolling) {
    return;
  }
  // see if we were dragging
  var was_dragging = this.dragging;
  //
  this.callSuper(event);
  // nothing to do if we started the interaction by stopping a deceleration animation
  if (this.wasDeceleratingWhenTouchesBegan) {
    return;
  }
  //
  if (was_dragging != this.dragging && this.touchedCell !== null && !this.touchedCellWasSelected) {
    if (this.touchedCell instanceof ADTableViewCell) {
      this.touchedCell.selected = false;
    }
    else {
      this.touchedCell.removeClassName(ADControlStateSelectedCSS);
    }
  }
};

ADTableView.prototype.touchesEnded = function (event) {
  //
  var was_dragging = this.dragging;
  //
  this.callSuper(event);
  // nothing to do if we started the interaction by stopping a deceleration animation
  if (this.wasDeceleratingWhenTouchesBegan) {
    return;
  }
  // nothing to do if we had cancelled touches and not a touchend
  if (event.type != ADEndEvent) {
    return;
  }
  // if we did not drag, and we're in bubbling phase so that cell contents
  // had a chance to pursue another action that over-rides the selection by stopping propagation
  // we select the row or the accessory if one was hit and it hasn't began tracking touches all by itself
  if (event.eventPhase == Event.BUBBLING_PHASE && !was_dragging) {
    if (this.touchedAccessory !== null && !this.shouldPreventScrolling) {
      this.disclosureButtonWasSelectedAtPath(this.touchedAccessory.superview._tableViewDataSourcePath);
    }
    else if (this.touchedCell !== null) {
      this.selectRowAtPath(this.touchedCell._tableViewDataSourcePath);
    }
  }
};

/* ==================== Managing Selection ==================== */

/**
 *  Returns an index path identifying the row and section of the selected row.
 *
 *  @returns {ADCellPath} An index path identifying the row and section indexes of the selected row or <code>null</code> if the index path is invalid.
 */
ADTableView.prototype.pathForSelectedRow = function () {
  if (this.selectedCell === null) {
    return null;
  }
  return (this._style === ADTableViewStyleCustom) ? this.pathForCustomCell(this.selectedCell) : this.pathForCell(this.selectedCell);
};

/**
 *  Deselects a given row identified by index path, with an option to animate the deselection. Note that animations are not currently supported.
 *
 *  @param {ADCellPath} path A cell path identifying a row in the receiver.
 *  @param {bool} animated <code>true</code> if you want to animate the deselection and <code>false</code> if the change should be immediate.
 */
ADTableView.prototype.deselectRowAtPathAnimated = function (path, animated) {
  // nothing to do if we don't have a path
  if (path === null) {
    return;
  }
  // get the cell
  var cell = (this._style === ADTableViewStyleCustom) ? this.customCellForRowAtPath(path) : this.cellForRowAtPath(path);
  // update selected state
  if (cell !== null) {
    this.markCellAsSelectedAnimated(cell, false, animated);
  }
};

/**
 *  Selects a row in the receiver identified by <code>path</code>.
 *
 *  @param {ADCellPath} path A cell path identifying a row in the receiver.
 */
ADTableView.prototype.selectRowAtPath = function (path) {
  // get the purported selected cell
  var selected_cell = (this._style === ADTableViewStyleCustom) ? this.customCellForRowAtPath(path) : this.cellForRowAtPath(path);
  // nothing to do if there's no such cell
  if (selected_cell === null) {
    throw(new Error('No cell at ' + path.toString()));
    return;
  }
  // deselect the previous cell
  this.deselectRowAtPathAnimated(this.pathForSelectedRow(), false);
  // track the cell
  this.selectedCell = selected_cell;
  // and mark as selected
  this.markCellAsSelectedAnimated(this.selectedCell, true, false);
  // notify the delegate
  if (ADUtils.objectHasMethod(this.delegate, ADTableViewDidSelectRowAtPath)) {
    this.delegate[ADTableViewDidSelectRowAtPath](this, path);
  }
};

ADTableView.prototype.detectedStationaryTouch = function () {
  if (!this.dragging && this.tracking) {
    if (this.touchedCell !== null) {
      this.markCellAsSelectedAnimated(this.touchedCell, true, false);
    }
    else if (this.touchedAccessory !== null) {
      this.shouldPreventScrolling = true;
    }
  }
};

ADTableView.prototype.markCellAsSelectedAnimated = function (cell, selected, animated) {
  if (cell instanceof ADTableViewCell) {
    cell.setSelectedAnimated(selected, animated);
  }
  else {
    cell[selected ? 'addClassName' : 'removeClassName'](ADControlStateSelectedCSS);
  }
};

/* ==================== SELECTION ==================== */

ADTableView.prototype.disclosureButtonWasSelectedAtPath = function (path) {
  var cell = this.cellForRowAtPath(path);
  if (cell.accessoryType === ADTableViewCellAccessoryDetailDisclosureButton &&
      ADUtils.objectHasMethod(this.delegate, ADTableViewDidSelectAccessoryForRowAtPath)) {
    this.delegate[ADTableViewDidSelectAccessoryForRowAtPath](this, path);
  }
};

/* ==================== Content Offset Tracking ==================== */

const ADTableViewPlainHeaderHeight = 23;

ADTableView.prototype.updateSectionHeaders = function () {
  // nothing to do if we don't have information about section metrics
  if (this.sectionMetrics.length != this.numberOfSections || this.style !== ADTableViewStylePlain) {
    return;
  }
  //
  var internal_y = this.contentOffset.y;
  for (var section = 0; section < this.numberOfSections; section++) {
    var header = this.headers[section];
    // nothing to do if there's no header for that section
    if (header === undefined) {
      continue;
    }
    // get the header's bottom y
    var metrics = this.sectionMetrics[section];
    var top_y = metrics.y;
    var bottom_y = top_y + metrics.height;
    var visible_tail_height = bottom_y - internal_y;
    // otherwise, check on appropriate location for this header
    var ty = 0;
    // we're showing the very end of our section
    if (visible_tail_height > 0 && visible_tail_height < (ADTableViewPlainHeaderHeight - 1)) {
      ty = metrics.height - ADTableViewPlainHeaderHeight;
    }
    // we're the top section
    else if (top_y <= internal_y && bottom_y > internal_y) {
      ty = Math.abs(top_y - internal_y) - 1;
    }
    header.style.webkitTransform = ADUtils.t(0, ty);
  }
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADTableView);

/* ==================== IMAGE PRELOAD ==================== */

ADTableView.init = function () {
  ADUtils.preloadImageAsset('tableview/UITableSelection.png');
};

window.addEventListener('load', ADTableView.init, false);

/* ==================== ADCellPath ==================== */

/**
 *  @class
 *
 *  <p>The ADCellPath class provides the path to a {@link ADTableView} cell.</p>
 *
 *  @since AdLib 0.1
 *
 *  @param {int} section The cell's section.
 *  @param {int} row The cell's row.
 */

function ADCellPath (section, row) {
  /**
   *  The cell's section
   *  @type int 
   */
  this.section = section || 0;
  /**
   *  The cell's row
   *  @type int 
   */
  this.row = row || 0;
};

ADCellPath.prototype.toString = function () {
  return 'ADCellPath with section ' + this.section + ' and row ' + this.row;
};

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  The {@link ADTableViewCell} does not have any accessory view. This is the default value.
 *  @constant
 *  @type String
 */
const ADTableViewCellAccessoryNone = 'no-accessory';
/**
 *  The {@link ADTableViewCell} has an accessory control shaped like a regular chevron. It is
 *  intended as a disclosure indicator. The control doesn't track touches.
 *  @constant
 *  @type String
 */
const ADTableViewCellAccessoryDisclosureIndicator = 'disclosure-accessory';
/**
 *  The {@link ADTableViewCell} has an accessory control that is a blue button with a chevron
 *  image as content. It is intended for configuration purposes. The control tracks touches.
 *  @constant
 *  @type String
 */
const ADTableViewCellAccessoryDetailDisclosureButton = 'detail-accessory';

// const ADTableViewCellSeparatorStyleNone = 'no-separator';
// const ADTableViewCellSeparatorStyleSingleLine = 'single-line-separator';

/**
 *  The {@link ADTableViewCell} has no distinct style for when it is selected.
 *  @constant
 *  @type String
 */
const ADTableViewCellSelectionStyleNone = 'no-selection';
/**
 *  The {@link ADTableViewCell} when selected has a blue background. This is the default value.
 *  @constant
 *  @type String
 */
const ADTableViewCellSelectionStyleBlue = 'blue-selection';
/**
 *  The {@link ADTableViewCell} when selected has a gray background.
 *  @constant
 *  @type String
 */
const ADTableViewCellSelectionStyleGray = 'gray-selection';

/**
 *  A simple style for a {@link ADTableViewCell} with a text label (black and left-aligned) and an optional image view.
 *  @constant
 *  @type String
 */
const ADTableViewCellStyleDefault = 'style-default';
/**
 *  A style for a {@link ADTableViewCell} with a label on the left side of the cell with left-aligned and black text; on
 *  the right side is a label that has smaller blue text and is right-aligned. The Settings application uses cells in this style.
 *  @constant
 *  @type String
 */
const ADTableViewCellStyleValue1 = 'style-value-1';
/**
 *  A style for a {@link ADTableViewCell} with a label on the left side of the cell with text that is right-aligned and blue;
 *  on the right side of the cell is another label with smaller text that is left-aligned and black. The Phone/Contacts
 *  application uses cells in this style.
 *  @constant
 *  @type String
 */
const ADTableViewCellStyleValue2 = 'style-value-2';
/**
 *  A style for a {@link ADTableViewCell} with a left-aligned label across the top and a left-aligned label below it in smaller
 *  gray text. The iPod application uses cells in this style.
 *  @constant
 *  @type String
 */
const ADTableViewCellStyleSubtitle = 'style-subtitle';

/**
 *  The separator cell has no distinct style.
 *  @constant
 *  @type String
 */
const ADTableViewCellSeparatorStyleNone = 'separator-none';
/**
 *  The separator cell has a single line running across its width. This is the default value for {@link #separatorStyle}.
 *  @constant
 *  @type String
 */
const ADTableViewCellSeparatorStyleSingleLine = 'separator-single-line';
/**
 *  The separator cell has double lines running across its width, giving it an etched look. This style is currently only supported for
 *  grouped-style table views.
 *  @constant
 *  @type String
 */
const ADTableViewCellSeparatorStyleSingleLineEtched = 'separator-single-line-etched';

/* ==================== ADTableViewCell ==================== */

ADTableViewCell.inherits = ADView;
ADTableViewCell.synthetizes = ['text', 'detailedText', 'selectionStyle', 'accessoryType', 'selected'];

/**
 *  @class
 *
 *  <p>The ADTableViewCell class defines the attributes and behavior of the cells that appear in ADTableView objects.
 *  A ADTableViewCell object (or table cell) includes properties and methods for managing selection, accessory and separator styles.</p>
 *  
 *  @extends ADView
 *  @since AdLib 0.1
 *
 *  @param {String} style A constant indicating a cell style. See the constants prefixed <code>ADTableViewCellStyle</code> for a
 *  description of these constants.
 */
function ADTableViewCell (style) {
  // record style
  this.style = style || ADTableViewCellStyleDefault;
  //
  this.callSuper();
  // private property backing synthetization
  this._selectionStyle = ADTableViewCellSelectionStyleBlue;
  this._accessoryType = ADTableViewCellAccessoryNone;
  this._selected = false;
  // make sure we don't track touches for this view, the ADTableView
  // will deal with touches
  this.layer.removeEventListener(ADStartEvent, this, false);
};

/* ==================== ADView APIs ==================== */

ADTableViewCell.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-table-view-cell ' + this.style);
  this.layer.setAttribute('role', 'button');
  //
  this.accessory = new ADButton(ADButtonTypeDetailDisclosure);
  this.accessory.addEventListener(ADControlTouchUpInsideEvent, this, false);
  this.addSubview(this.accessory);
  //
  this.textLabel = this.layer.appendChild(document.createElement('span'));
  this.textLabel.addClassName('text-label');
  this.detailedTextLabel = this.layer.appendChild(document.createElement('span'));
  this.detailedTextLabel.addClassName('detailed-text-label');
};

ADTableViewCell.prototype.handleEvent = function (event) {
  // the disclosure button was hit
  if (event.currentTarget === this.accessory.layer) {
    var table = this._table;
    var path = this._tableViewDataSourcePath;
    if (table !== undefined && path !== undefined) {
      table.disclosureButtonWasSelectedAtPath(path);
    }
  }
  else {
    this.callSuper(event);
  }
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADTableViewCell.prototype
 *  @property {String} text Holds the main label of the cell. <code>ADTableViewCell</code> adds an appropriate label when you create
 *  the cell in a given cell style. See the constants prefixed <code>ADTableViewCellStyle</code> for descriptions of the main label
 *  in currently defined cell styles.
 *  @property {String} detailedText Holds the secondary (or detail) label of the cell. <code>ADTableViewCell</code> adds an appropriate
 *  label when you create the cell in a style that supports secondary labels. If the style doesn't support detail labels, <code>null</code>
 *  is returned. See the constants prefixed <code>ADTableViewCellStyle</code> for descriptions of the detailed label in currently defined cell styles.
 *  @property {String} selectionStyle The selection style is a constant that determines the color of a cell when it is selected.
 *  The default value is {@link ADTableViewCellSelectionStyleBlue}.
 *  @property {String} accessoryType The accessory view appears in the the right side of the cell. The standard accessory views
 *  include the disclosure chevron. The default is {@link ADTableViewCellAccessoryNone}. If the accessory type is
 *  {@link ADTableViewCellAccessoryDetailDisclosureButton} then the accessory view tracks touches and, if tapped, invokes the
 *  {@link ADTableViewDelegate#tableViewDidSelectAccessoryForRowAtPath} delegate method on the host table view.
 *  @property {bool} selected A Boolean value that indicates whether the cell is selected. When the the selected state of a cell
 *  to <code>true</code>, it draws the background for selected cells with its title in white. The default value is is <code>false</code>.
 *  If you set the selection state to <code>true</code> through this property, the transition to the new state appearance is not animated.
 *  For animated selected-state transitions, see the {@link #setSelectedAnimated} method.
 */

ADTableViewCell.prototype.getText = function () {
  return this.textLabel.innerText;
};

ADTableViewCell.prototype.setText = function (text) {
  this.textLabel.innerText = text;
  this.updateTextLayout();
};

ADTableViewCell.prototype.getDetailedText = function () {
  return this.detailedTextLabel.innerText;
};

ADTableViewCell.prototype.setDetailedText = function (detailedText) {
  this.detailedTextLabel.innerText = detailedText;
  this.updateTextLayout();
};

ADTableViewCell.prototype.setSelectionStyle = function (selectionStyle) {
  this.layer.removeClassName(this._selectionStyle);
  this.layer.addClassName(selectionStyle);
  this._selectionStyle = selectionStyle;
};

const ADTableViewCellAccessoryDisclosureIndicatorWidth = 10;
const ADTableViewCellAccessoryDetailDisclosureButtonWidth = 44;

ADTableViewCell.prototype.setAccessoryType = function (accessoryType) {
  this.layer.removeClassName(this._accessoryType);
  this.layer.addClassName(accessoryType);
  this._accessoryType = accessoryType;
};

ADTableViewCell.prototype.setSelected = function (selected) {
  this.setSelectedAnimated(selected, false);
};

/**
 *  Sets the selected state of the cell, optionally animating the transition between states. When the the selected state of a cell to <code>true</code>,
 *  it draws the background for selected cells with its title in white. Note that animations are not currently supported.
 *
 *  @param {bool} selected <code>true</code> to set the cell as selected, <code>false</code> to set it as unselected.
 *  @param {bool} animated <code>true</code> to animate the transition between selected states, <code>false</code> to make the transition immediate.
 */
ADTableViewCell.prototype.setSelectedAnimated = function (selected, animated) {
  // do nothing if we already were in that selection state
  if (this._selected == selected) {
    return;
  }
  // remember selection in ivar
  this._selected = selected;
  // update style
  this.layer[selected ? 'addClassName' : 'removeClassName'](ADControlStateSelectedCSS);
};

/* ==================== Text Layout ==================== */

ADTableViewCell.prototype.layerWasInsertedIntoDocument = function () {
  this.callSuper();
  this.updateTextLayout();
};

const ADTableViewCellStyleValue1Margin = 10;

ADTableViewCell.prototype.updateTextLayout = function () {
  // only need to do something for ADTableViewCellStyleValue1
  if (this.style != ADTableViewCellStyleValue1 || !this.layerIsInDocument) {
    return;
  }

  // get the max text width minus 10 pixels to have between labels
  var max_text_width = this.textLabel.offsetWidth - 2 * ADTableViewCellStyleValue1Margin;

  // revert to auto-sizing
  this.textLabel.style.right = 'auto !important';
  this.detailedTextLabel.style.right = 'auto !important';

  // get metrics
  var text_width = Math.min(this.textLabel.offsetWidth, max_text_width);
  var detail_text_width = Math.min(this.detailedTextLabel.offsetWidth, max_text_width);

  // reset styles
  this.textLabel.setAttribute('style', '');
  this.detailedTextLabel.setAttribute('style', '');

  // see if we need to adjust the layout
  if (text_width + detail_text_width > max_text_width) {
    var adjusted_text_width = Math.floor((text_width / (text_width + detail_text_width)) * max_text_width);
    //
    if (text_width > detail_text_width) {
      this.textLabel.style.width = ADUtils.px(adjusted_text_width);
      this.detailedTextLabel.style.left = ADUtils.px(adjusted_text_width + ADTableViewCellStyleValue1Margin * 2);
    }
    else {
      this.textLabel.style.width = ADUtils.px(adjusted_text_width + ADTableViewCellStyleValue1Margin);
      this.detailedTextLabel.style.left = ADUtils.px(adjusted_text_width + ADTableViewCellStyleValue1Margin * 3);
    }
  }
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADTableViewCell);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  Vertical orientation for a {@link ADCarouselView}.
 *  @constant
 *  @type int
 */
const ADCarouselViewVerticalOrientation = 0;
/**
 *  Horizontal orientation for a {@link ADCarouselView}.
 *  @constant
 *  @type int
 */
const ADCarouselViewHorizontalOrientation = 1;
const ADCarouselViewSelectionRotationDuration = 0.5; // time in seconds for the animated transition upon selection

/**
 *  @class
 *  @name ADCarouselViewDataSource
 *  @since AdLib 0.1
 */

/**
 *  Provides a {@link ADCarouselView} with the number of items in the data source. Required.
 *
 *  @name carouselViewNumberOfCells
 *  @function
 *
 *  @param {ADCarouselView} theCarouselView The list view.
 *  @returns {int} The number of items in the data source.
 *  @memberOf ADCarouselViewDataSource.prototype
 */

/**
 *  Provides a {@link ADCarouselView} with a cell for a given index. The cell must be an <code>&lt;li></code> element. Required.
 *
 *  @name carouselViewCellForIndex
 *  @function
 *
 *  @param {ADCarouselView} theCarouselView The list view.
 *  @param {int} index The cell index.
 *  @returns {HTMLLiElement} The cell for the given index.
 *  @memberOf ADCarouselViewDataSource.prototype
 */

const ADCarouselViewNumberOfCells = 'carouselViewNumberOfCells';
const ADCarouselViewCellForIndex = 'carouselViewCellForIndex';

/**
 *  @class
 *  @name ADCarouselViewDelegate
 *  @since AdLib 0.1
 */

/**
 *  Notifies that the {@link ADCarouselView} cell at the given index has just been selected and that further
 *  selection-related processing is about to happen. For instance, if {@link ADCarouselView#centersSelection}
 *  is set to <code>true</code>, this delegate method is called prior to the centering animation. Optional.
 *
 *  @name carouselViewWillSelectCellAtIndex
 *  @function
 *
 *  @param {ADCarouselView} theCarouselView The carousel view.
 *  @param {int} index The cell index.
 *  @param {bool} isInstantaneous Indicates whether the carousel will rotate to center the newly-selected cell.
 *  @memberOf ADCarouselViewDelegate.prototype
 */

/**
 *  Invoked when the {@link ADCarouselView} cell at the given index was selected. Optional.
 *
 *  @name carouselViewDidSelectCellAtIndex
 *  @function
 *
 *  @param {ADCarouselView} theCarouselView The carousel view.
 *  @param {int} index The cell index.
 *  @param {bool} wasInstantaneous Indicates whether the carousel rotated to center the newly-selected cell.
 *  @memberOf ADCarouselViewDelegate.prototype
 */

const ADCarouselViewWillSelectCellAtIndex = 'carouselViewWillSelectCellAtIndex';  
const ADCarouselViewDidSelectCellAtIndex = 'carouselViewDidSelectCellAtIndex';

// CSS protocol
const ADCarouselViewCellActiveCSS = 'active';

/* ==================== ADCarouselView Constructor ==================== */

ADCarouselView.inherits = ADView;
ADCarouselView.synthetizes = ['rotation'];

/**
 *  @class
 *
 *  <p>The ADCarouselView class resembles very much the {@link ADTableView} class, and offers very similar
 *  protocols, but instead of presenting its cells as a flat list, it maps its contents onto a rotating
 *  3D carousel. Contrary to other {@link ADView} extensions, the {@link ADCarouselView} has no explicit
 *  size and setting its <code>size</code> property is a no-op. Instead, the developer provides a
 *  {@link #cellSize} and an optional {@link cellPadding} and the carousel automatically sizes itself to
 *  fit the specified number of items onto the carousel to match the provided metrics.</p>
 *
 *  <p>While, by default, carousels are displayed vertically, the {@link #orientation} property also allows
 *  to set a carousel in the horizontal orientation. However, this property is not live and the developer
 *  would have to call {@link #reloadData} to update the rendering of the carousel following a change of
 *  orientation.</p>
 *
 *  <p>The selection behavior of a carousel can be customized by the developer with the {@link #centersSelection}
 *  and the {@link centerSelectionDuration} properties to specify whether the carousel spins to center the item
 *  that was selected, and how long this animation lasts. When a selection happens, the developer is notified
 *  of the various states by implementing the {@link ADCarouselViewDelegate} protocol.</p>
 *  
 *  @extends ADView
 *  @since AdLib 0.1
 */
function ADCarouselView () {
  this.callSuper();
  // objects implementing our protocols
  this.dataSource = null;
  this.delegate = null;
  // public properties
  this.centersSelection = true;
  this.centerSelectionDuration = -1;
  this.orientation = ADCarouselViewVerticalOrientation;
  this.cellSize = new ADSize();
  this.cellPadding = 0;
  this.numberOfCells = 0;
  //
  this.touchedCell = null;
  this.busySelecting = false;
  this.showingHighlight = false;
  this.currentHighlightCell = null;
  this.trackingMoved = false;
  //
  this._rotation = 0;
  // store the last two points for velocity computation
  this.trackingPoints = [];
  // setting up the animator
  this.animator = new ADAnimator(2000, this, [0.211196, 0.811224, 0.641221, 0.979592]);
  // register event handler for transition end on the list container
  // so that we can trigger the selection callback when done
  this.listContainer.addEventListener('webkitTransitionEnd', this, false);
  // make sure we get all touch events
  this.tracksTouchesOnceTouchesBegan = true;
};

/**
 *  @name ADCarouselView.prototype
 *  @property {bool} busySelecting Indicates whether the carousel is spinning to center a newly selected item.
 *  @property {int} cellPadding The distance in pixels between each cell. Defaults to 0.
 *  @property {ADSize} cellSize The size of each of the cells in the ADCarouselView. Required.
 *  @property {float} centerSelectionDuration The duration of the centering animation when a new cell was selected, in seconds. Defaults to 0.5.
 *  @property {bool} centersSelection Indicates whether the carousel should spin to center the selected cell when a new cell is selected, defaults to <code>true.</code>
 *  @property {Object} dataSource The object acting as a data source for the view and implementing the {@link ADCarouselViewDataSource} protocol.
 *  @property {Object} delegate The object acting as a delegate for the view and implementing the {@link ADCarouselViewDelegate} protocol.
 *  @property {int} orientation The carousel orientation, which can be either horizontal or vertical, defaults to <code>ADCarouselViewVerticalOrientation</code>.
 *  @property {float} rotation The rotation angle (in degrees) of the carousel, normalized in the 0-360 range.
 */

/* ==================== ADView APIs ==================== */

ADCarouselView.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-carousel-view');
  //
  this.aligner = this.layer.appendChild(document.createElement('div'));
  this.aligner.className = 'aligner';
  //
  this.listContainer = this.aligner.appendChild(document.createElement('div'));
  this.listContainer.className = 'ad-carousel-view-list-container';
};

ADCarouselView.prototype.setSize = function (aSize) {
  // we need to override View.setSize() as the backing layer needs to have no metrics
};

/* ==================== Populating ==================== */

ADCarouselView.prototype.isVertical = function () {
  return (this.orientation == ADCarouselViewVerticalOrientation);
};

/**
 *  Notifies the view that there is new data to be drawn. This method will clear any existing content
 *  in the view and trigger the relevant {@link ADCarouselViewDataSource} protocol methods.
 */
ADCarouselView.prototype.reloadData = function () {
  // check validity of data source
  if (!ADUtils.objectHasMethod(this.dataSource, ADCarouselViewNumberOfCells) || !ADUtils.objectHasMethod(this.dataSource, ADCarouselViewCellForIndex)) {
    return;
  }
  // empty the list container
  this.listContainer.textContent = '';
  // get how many people we'll be fitting in the ring
  this.numberOfCells = this.dataSource[ADCarouselViewNumberOfCells](this);
  // deduce the radius from it
  var is_vertical = this.isVertical();
  this.radius = Math.ceil(this.numberOfCells * ((is_vertical ? this.cellSize.height : this.cellSize.width) + this.cellPadding)) / (2 * Math.PI);
  // update the aligner with the new radius
  this.aligner.style.webkitTransform = 'translateZ(' + (-this.radius) + 'px)';
  // get the rotation vector
  var rx = is_vertical ? 1 : 0;
  var ry = is_vertical ? 0 : 1;
  var angle_increment = 360 / this.numberOfCells;
  // get position for each item
  var top = is_vertical ? (-this.cellSize.height / 2) : 0;
  var left = is_vertical ? 0 : (-this.cellSize.width / 2);
  // populate the list container
  for (var i = 0; i < this.numberOfCells; i++) {
    var cell = this.dataSource[ADCarouselViewCellForIndex](this, i);
    //
    cell._list3DViewDataSourceIndex = i;
    // size and locate the cell
    cell.style.top = top + 'px';
    cell.style.left = left + 'px';
    // rotate the cell
    cell.style.webkitTransform = 'rotate3d(' + rx + ', ' + ry + ', 0, ' + (-angle_increment * i) +  'deg) translateZ(' + this.radius + 'px)';
    // listen to touch events on the cell
    cell.addEventListener(ADStartEvent, this, false);
    //
    this.listContainer.appendChild(cell);
  }
  // locate the list container to the center
  this.listContainer.style.top = (is_vertical ? this.radius : 0) + 'px';
  this.listContainer.style.left = (is_vertical ? 0 : this.radius) + 'px';
  // update the layer's perspective origin in absolute values as our backing layer has no metrics
  var px = (is_vertical ? this.cellSize.width / 2 : this.radius) + this.position.x;
  var py = (is_vertical ? this.radius : this.cellSize.height / 2) + this.position.y;
  this.layer.style.webkitPerspectiveOrigin = px + 'px ' + py + 'px';
  // reset any flicking that might have happened
  this.rotation = 0;
};

/* ============================== Synthetized APIs ============================== */

ADCarouselView.prototype.setRotation = function (aRotation) {
  // do nothing if we have a bad value
  if (isNaN(aRotation)) {
    return;
  }
  // normalize the angle
  this._rotation = aRotation % 360;
  // get the rotation vector
  var is_vertical = this.isVertical();
  var rx = is_vertical ? 1 : 0;
  var ry = is_vertical ? 0 : 1;
  //
  this.listContainer.style.webkitTransform = 'rotate3d(' + rx + ', ' + ry + ', 0, ' + this._rotation + 'deg)';
};

/* ============================== EVENT HANDLING ============================== */

ADCarouselView.prototype.handleEvent = function (event) {
  // do nothing if we are busy rotating following a selection
  if (this.busySelecting && event.type != 'webkitTransitionEnd') {
    return;
  }
  //
  this.callSuper(event);
  // conduct actions based on the event type
  if (event.type == 'webkitTransitionEnd') {
    this.transitionEnded(event);
  }
};

ADCarouselView.prototype.touchesBegan = function (event) {
  // do not do anything if we've fired the default ADView event handler
  if (event.currentTarget === this.layer) {
    return;
  }
  //
  this.callSuper(event);
  //
  this.touchedCell = event.currentTarget;
  // ensure no transitions will affect setting the rotation anymore
  this.listContainer.style.webkitTransitionDuration = 0;
  // stop animation if we were animating
  this.wasStopppedDuringAnimation = this.animator.animating;
  if (this.wasStopppedDuringAnimation) {
    this.stopAnimation();
  }
  else {
    this.highlightCurrentCell();
  }
  // track base information about this interaction
  this.startRotation = this.rotation;
  this.startTouchPosition = ADPoint.fromEvent(event);
  this.startXAngle = Math.acos((this.startTouchPosition.x - this.position.x - this.radius) / this.radius);
  this.startYAngle = Math.acos((this.startTouchPosition.y - this.position.y - this.radius) / this.radius);
  // set up default values for flags
  this.trackingMoved = false;
  // set up array to track interaction positions
  this.trackingPoints = [];
  this.storeEventLocation(this.startTouchPosition);
};

ADCarouselView.prototype.touchesMoved = function (event) {
  this.callSuper(event);
  //
  this.removeCellHighlight();
  // we actually moved our finger, useful to track selections
  var touch_position = ADPoint.fromEvent(event);
  this.storeEventLocation(touch_position);
  this.trackingMoved = true;  
  //  
  var angle_delta;
  if (this.isVertical()) {
    angle_delta = -(this.startYAngle - Math.acos((touch_position.y - this.position.y - this.radius) / this.radius));
  }
  else {
    angle_delta = this.startXAngle - Math.acos((touch_position.x - this.position.x - this.radius) / this.radius);
  }
  // update the rotation angle
  this.rotation = this.startRotation + ADUtils.radiansToDegrees(angle_delta);
};

ADCarouselView.prototype.touchesEnded = function (event) {
  this.callSuper(event);
  //
  // if we had not moved the finger and weren't in the middle of an animation,
  // that means we selected something and we want to place it in the middle of
  // the screen now with a nice animated rotation
  if (this.touchedCell != null && !this.trackingMoved && !this.wasStopppedDuringAnimation) {
    // first, track that we're busy selecting
    this.busySelecting = true;
    //
    var new_rotation = this.touchedCell._list3DViewDataSourceIndex * (360 / this.numberOfCells);
    //
    var instant_selection = (new_rotation == this.rotation);
    // tell any observer that we're initiating selection
    this.cellIsBeingSelected(instant_selection);
    // no need to animate there
    if (instant_selection || !this.centersSelection) {
      // trigger callback if one is available
      this.cellWasSelected(instant_selection);
      // track that the selection process is over
      this.busySelecting = false;
      // remove highlight as selection process is over
      this.removeCellHighlight();
    }
    // let's rotate to it
    else  {
      // figure out the absolute distance between the two angles
      var angle_delta = Math.abs(new_rotation - this.rotation);
      if (angle_delta > 90) {
        if (new_rotation < this.rotation) {
          angle_delta = new_rotation + 360 - this.rotation;
        }
        else {
          angle_delta = this.rotation + 360 - new_rotation;
        }
      }
      // compute a duration for the transition based on the delta so that we
      // always have the same speed when rotating, or using a user-defined setting
      var duration = this.centerSelectionDuration;
      if (duration < 0) {
        duration = (Math.min(60, angle_delta) / 60 * ADCarouselViewSelectionRotationDuration);
      }
      // set up the transition
      this.listContainer.style.webkitTransitionDuration = Math.max(0.1, duration) + 's';
      // rotate to center
      this.rotation = new_rotation;
    }
  }
  else {
    this.startAnimating();
  }
};

ADCarouselView.prototype.transitionEnded = function (event) {
  this.busySelecting = false;
  // we got called following an animated rotation to center
  // following a selection, so trigger selection callback if any
  this.cellWasSelected(this.touchedCell, false);
  // remove the highlight
  this.removeCellHighlight();
};

/* ============================== POINT TRACKING ============================== */

ADCarouselView.prototype.storeEventLocation = function (aPoint) {
  var point = { pos : (this.isVertical() ? aPoint.y : aPoint.x), date : new Date() };
  // push the new one
  this.trackingPoints.push(point);
  // dump the old
  if (this.trackingPoints.length > 10) {
    this.trackingPoints.shift();
  }
};

/* ============================== ANIMATION ============================== */

ADCarouselView.prototype.startAnimating = function() {
  // get the first interesting index
  var first_index = -1;
  for (var i = this.trackingPoints.length - 1; i > 0; --i) {
    // XXX also check max y bounds
    if (this.trackingPoints[i].pos != 0) {
      first_index = i - 1;
      break;
    }
  }
  // do nothing if we have no useful data to process
  if (first_index < 0) {
    return false;
  }
  // otherwise keep going
  var second_index = first_index + 1;
  // the last item is the touchend event, we just want to look at its date
  var release_delay = this.trackingPoints[second_index].date - this.trackingPoints[first_index].date;
  var delta = (this.trackingPoints[second_index].pos - this.trackingPoints[first_index].pos);
  // do nothing if the interaction delta is too small or the interaction was too slow
  if (Math.abs(delta) < 3 || release_delay > 35) {
    return false;
  }
  // get the interaction angle delta
  var angle_delta = -Math.atan(delta / this.radius);
  var absolute_angle = Math.abs(angle_delta);
  var total_angle;
  if (absolute_angle < 0.075) {
    total_angle = 360 / 4;
  }
  else if (absolute_angle < 0.15) {
    total_angle = 360 / 2;
  }
  else if (absolute_angle < 0.225) {
    total_angle = 360 * 0.75;
  }
  else {
    total_angle = 360;
  }
  // FIXME: rewrite this
  var direction = (angle_delta < 0) ? (this.isVertical() ? 'top-' : 'right-') : (this.isVertical() ? 'bottom-' : 'left-');
  if (direction == 'left-' || direction == 'top-') {
    total_angle = -total_angle;
  }
  // track the rotation before the animation started
  this.rotationBeforeAnimation = this.rotation;
  // set the rotation for when the animation completes so that the animation seems to have freeze fill
  this.animationAngle = total_angle;
  //
  this.animator.start();
  //
  return true;
};

// AdLib.Animator delegate method
ADCarouselView.prototype.animationDidIterate = function (progress) {
  this.rotation = this.rotationBeforeAnimation + progress * this.animationAngle;
};

ADCarouselView.prototype.stopAnimation = function () {
  this.animator.stop();
};

/* ============================== HIGHLIGHT ============================== */

ADCarouselView.prototype.highlightCurrentCell = function () {
  // do nothing if we have not identified what the touch target was
  if (this.touchedCell == null) {
    return;
  }
  //
  if (this.currentHighlightCell != null) {
    this.removeCellHighlight();
  }
  this.touchedCell.addClassName(ADCarouselViewCellActiveCSS);
  this.currentHighlightCell = this.touchedCell;
  //
  this.showingHighlight = true;
};

ADCarouselView.prototype.removeCellHighlight = function () {
  if (!this.showingHighlight || this.touchedCell == null) {
    return;
  }
  //
  this.currentHighlightCell.removeClassName(ADCarouselViewCellActiveCSS);
  this.currentHighlightCell = null;
  //
  this.showingHighlight = false;
};

/* ============================== DELEGATES TRIGGER ============================== */

ADCarouselView.prototype.cellIsBeingSelected = function (is_instant) {
  if (ADUtils.objectHasMethod(this.delegate, ADCarouselViewWillSelectCellAtIndex)) {
    this.delegate[ADCarouselViewWillSelectCellAtIndex](this, this.touchedCell._list3DViewDataSourceIndex, is_instant);
  }
};

ADCarouselView.prototype.cellWasSelected = function (is_instant) {
  if (ADUtils.objectHasMethod(this.delegate, ADCarouselViewDidSelectCellAtIndex)) {
    this.delegate[ADCarouselViewDidSelectCellAtIndex](this, this.touchedCell._list3DViewDataSourceIndex, is_instant);
  }
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADCarouselView);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  Default bar style for a {@link ADNavigationBar}.
 *  @constant
 *  @type String
 */
const ADNavigationBarStyleDefault = 'default';
/**
 *  Black bar style for a {@link ADNavigationBar}.
 *  @constant
 *  @type String
 */
const ADNavigationBarStyleBlack = 'black';
/**
 *  Translucent black bar style for a {@link ADNavigationBar}.
 *  @constant
 *  @type String
 */
const ADNavigationBarStyleBlackTranslucent = 'black-translucent';

// some metrics constants
const ADNavigationBarButtonMarginLeft = 5;
const ADNavigationBarButtonMarginRight = 8;
/**
 *  Height of the bar at the top of the {@link ADNavigationBar}.
 *  @constant
 *  @type int
 */
const ADNavigationBarHeight = 44;

/**
 *  Duration in seconds of the animation as items are popped or pushed onto the navigation stack.
 *  @constant
 *  @type float
 */
const ADNavigationBarAnimationDuration = 0.35;

/**
 *  @class
 *  @name ADNavigationBarDelegate
 *  @since AdLib 0.1
 */

/**
 *  Notifies that an item is about to be pushed onto the navigation stack, allowing a controller
 *  object to cancel this action by returning <code>false</code>.
 *
 *  @name navigationBarShouldPushItem
 *  @function
 *
 *  @param {ADNavigationBar} theNavigationBar The navigation bar.
 *  @param {ADNavigationItem} theNavigationItem The item that is proposed to be pushed onto the navigation stack.
 *  @returns {bool} Indicates whether the proposed item should indeed be pushed onto the navigation stack.
 *  @memberOf ADNavigationBarDelegate.prototype
 */

/**
 *  Notifies that an item has been pushed onto the navigation stack. In the case where there was an animated
 *  transition to display this item being pushed onto the stack, this notification happens only when the
 *  animation has completed.
 *
 *  @name navigationBarDidPushItem
 *  @function
 *
 *  @param {ADNavigationBar} theNavigationBar The navigation bar.
 *  @param {ADNavigationItem} theNavigationItem The item that was pushed onto the navigation stack.
 *  @memberOf ADNavigationBarDelegate.prototype
 */

/**
 *  Notifies that an item is about to be popped off the navigation stack, allowing a controller object to
 *  cancel this action by returning <code>false</code>.
 *
 *  @name navigationBarShouldPopItem
 *  @function
 *
 *  @param {ADNavigationBar} theNavigationBar The navigation bar.
 *  @param {ADNavigationItem} theNavigationItem The item that is proposed to be popped off the navigation stack.
 *  @returns {bool} Indicates whether the proposed item should indeed be popped off the navigation stack.
 *  @memberOf ADNavigationBarDelegate.prototype
 */

/**
 *  Notifies that an item has been popped off the navigation stack. In the case where there was an animated
 *  transition to display this item being pushed off the stack, this notification happens only when the
 *  animation has completed.
 *
 *  @name navigationBarDidPopItem
 *  @function
 *
 *  @param {ADNavigationBar} theNavigationBar The navigation bar.
 *  @param {ADNavigationItem} theNavigationItem The item that was popped off the navigation stack.
 *  @memberOf ADNavigationBarDelegate.prototype
 */

const ADNavigationBarShouldPushItem = 'navigationBarShouldPushItem';  
const ADNavigationBarDidPushItem = 'navigationBarDidPushItem';  
const ADNavigationBarShouldPopItem = 'navigationBarShouldPopItem';
const ADNavigationBarDidPopItem = 'navigationBarDidPopItem';

/* ==================== ADNavigationBar Constructor ==================== */

ADNavigationBar.inherits = ADView;
// XXX: need to make ._items a synthetized property
ADNavigationBar.synthetizes = ['barStyle', 'items', 'topItem', 'backItem'];

/**
 *  @class
 *
 *  <p>The ADNavigationBar class offers a view for navigating hierarchical content. This view contains a bar,
 *  displayed at the top of the screen, containing buttons for navigating up and down a hierarchy that is
 *  modeled as a stack of {@link ADNavigationItem} objects. In order to bring a new item onto the screen,
 *  simply use the {@link #pushNavigationItemAnimated} method and the screen will animate to reflect the new state
 *  of the {@link #items} stack. Likewise, you can move back to the previous item on the stack with the
 *  {@link #popNavigationItemAnimated} method, which is automatically called when the user hits the back button
 *  displayed in the left part of the top bar. Navigation items can specify a title and a custom view to
 *  display into the navigation view. Finally, any change to the state of the {@link #items} stack can be
 *  monitored with the delegate methods in the {@link ADNavigationBarDelegate} protocol.</p>
 *  
 *  @extends ADView
 *  @since AdLib 0.1
 */
function ADNavigationBar () {
  this.callSuper();
  // public ivars
  /**
   *  The object acting as a delegate for the view and implementing the {@link ADNavigationBarDelegate} protocol.
   *  @type Object 
   */
  this.delegate = null;
  this._barStyle = '';
  this._items = [];
  // private ivars
  this.busy = false;
  this.itemsToSetupAfterAnimatedChange = null;
  // set up base type
  this.barStyle = ADNavigationBarStyleDefault;
};

/* ==================== ADView APIs ==================== */

ADNavigationBar.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-navigation-bar');
};

ADNavigationBar.prototype.setSize = function (aSize) {
  this.callSuper(new ADSize(aSize.width, ADNavigationBarHeight));
  // update the top item's layout
  this.updateTopItemLayout();
};

ADNavigationBar.prototype.willMoveToSuperview = function (superview) {
  // force metrics to be that of superview if not hard set before
  if (superview !== null && this._size.width == 0) {
    this.size = new ADSize(superview._size.width, ADNavigationBarHeight);
  }
};

ADNavigationBar.prototype.layerWasInsertedIntoDocument = function (superview) {
  this.callSuper();
  this.updateTopItemLayout();
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADNavigationBar.prototype
 *  @property {String} barStyle The appearance of the navigation bar. The default value is <code>ADNavigationBarStyleDefault</code>.
 *  @property {Array} items An array of navigation items managed by the navigation bar. The bottom item is at index 0, the back item is
 *  at index n-2, and the top item is at index n-1, where n is the number of items in the array.
 *  @property {ADNavigationItem} topItem The item at the top of the {@link #items} stack, ie. the item that is currently visible.
 *  @property {ADNavigationItem} backItem The navigation item that is next on the {@link #items} stack. If the {@link #topItem} doesn't
 *  have a left custom button, the back item's title is displayed on a back button on the left side of the navigation bar.
 */

ADNavigationBar.prototype.setBarStyle = function (barStyle) {
  this.layer.removeClassName(this._barStyle);
  this.layer.addClassName(barStyle);
  this._barStyle = barStyle;
};

ADNavigationBar.prototype.getTopItem = function () {
  return (this._items.length > 0) ? this._items[this._items.length - 1] : null;
};

ADNavigationBar.prototype.getBackItem = function () {
  return (this._items.length > 1) ? this._items[this._items.length - 2] : null;
};

/* ==================== EVENT HANDLING ==================== */

ADNavigationBar.prototype.handleEvent = function (event) {
  this.callSuper(event);
  //
  if (this.busy && event.type == 'webkitTransitionEnd') {
    this.transitionsEnded();
  }
};

/* ==================== ITEMS ==================== */

// synthetized setter
ADNavigationBar.prototype.setItems = function (items) {
  this.setItemsAnimated(items, false);
};

/**
 *  <p>Replaces the navigation items currently managed by the navigation bar with the specified items.</p>
 *
 *  <p>You can use this method to update or replace the navigation items in the stack without pushing or popping each item explicitly. In addition,
 *  this method lets you update the stack without animating the changes, which might be appropriate at launch time when you want to restore the
 *  state of the navigation stack to some previous state.</p>
 *
 *  <p>If animations are enabled, this method decides which type of transition to perform based on whether the last item in the items array is already
 *  on the current navigation stack. If the item is currently on the stack, but is not the topmost item, this method uses a pop transition; if it is
 *  the topmost item, no transition is performed. If the item is not on the stack, this method uses a push transition. Only one transition is performed,
 *  but when that transition finishes, the entire contents of the stack are replaced with the new items. For example, if items A, B, and C are on the
 *  stack and you set items D, A, and B, this method uses a pop transition and the resulting stack contains the items D, A, and B.</p>
 *
 *  @param {Array} items The {@link ADNavigationItem} objects to place in the stack. The front-to-back order of the items in this array represents
 *  the new bottom-to-top order of the items in the navigation stack. Thus, the last item added to the array becomes the top item of the navigation stack.
 *  @param {bool} animated If <code>true</code>, animate the pushing or popping of the top stack item. If <code>false</code>, replace the stack items
 *  without any animations.
 */
ADNavigationBar.prototype.setItemsAnimated = function (items, animated) {

  // do nothing if we're busy or are served an empty array
  if (this.busy || items.length == 0) {
    return;
  }

  // start a new transaction
  ADTransaction.begin();

  // get pointers to the relevant items
  var previous_item = this.topItem;
  var back_item = (items.length > 1) ? items[items.length - 2] : null;
  var next_item = items[items.length - 1];

  // dissociate the navigation bar from items in the current stack 
  for (var i = 0; i < this._items.length; i++) {
    this._items[i].navigationBar = null;
  }
  // and associate the navigation bar to items in the new stack 
  for (var i = 0; i < items.length; i++) {
    items[i].navigationBar = this;
  }

  // there was no previous item, in this case we don't use a transition
  if (previous_item === null) {
    // check that we can push this item
    if (!this.shouldPushItem(next_item)) {
      ADTransaction.commit();
      return;
    }
    // lay out the items in the new view
    this.addItemViews(next_item, null);
    next_item.sizeItemsAndComputePositionsWithBackItem(null);
    next_item.updateLayout();
    // record new stack
    this._items = items;
    // tell the delegate we pushed an item
    if (ADUtils.objectHasMethod(this.delegate, ADNavigationBarDidPushItem)) {
      this.delegate[ADNavigationBarDidPushItem](this, next_item);
    }
    // close the transaction immediately
    ADTransaction.commit();
  }
  // we're sharing the topItem, so don't animate but update the subviews in the bar
  else if (previous_item === next_item) {
    // remove subviews for the item
    this.removeViewsForItem(previous_item);
    // make sure nothing is left
    while (this.subviews.length) {
      this.subviews[0].removeFromSuperview();
    }
    // lay out the items in the new view
    this.addItemViews(next_item, back_item);
    next_item.sizeItemsAndComputePositionsWithBackItem(back_item);
    next_item.updateLayout();
    // record new stack
    this._items = items;
    // close the transaction immediately
    ADTransaction.commit();
  }
  // otherwise, we'll need to transition items
  else {
    // check in what direction we're going
    var pushing = (this._items.indexOf(next_item) == -1);
    // check we can go forward with the transition
    if ((pushing && !this.shouldPushItem(next_item)) || (!pushing && !this.shouldPopItem(previous_item))) {
      ADTransaction.commit();
      return;
    }
    // add the views for the new item and compute its views' positions
    this.addItemViews(next_item, back_item);
    next_item.sizeItemsAndComputePositionsWithBackItem(back_item);
    // keep track of the stack to recall once the transition is over
    this.itemsAfterTransition = items;
    // trigger the transition
    this.transitionToItem(next_item, previous_item, pushing, animated);
  }
};

/**
 *  Pushes a new item to the top of the {@link #items} stack and transitions it onto the screen.
 *  A controller object can use the {@link ADNavigationBarDelegate#navigationBarShouldPushItem}
 *  delegate method to prevent items being pushed onto the stack.
 *
 *  @param {ADNavigationItem} item The navigation item to push on the stack.
 *  @param {bool} animated <code>true</code> if the navigation bar should be animated; otherwise, <code>false</code>.
 */
ADNavigationBar.prototype.pushNavigationItemAnimated = function (item, animated) {
  // do nothing if we are busy transitioning already
  if (this.busy) {
    return;
  }
  // start a new transaction to allow transitions to be triggered within the
  // delegate call to be executed in sync with the nav bar animation
  ADTransaction.begin();
  // do nothing if we're not supposed to push
  if (!this.shouldPushItem(item)) {
    ADTransaction.commit();
    return;
  }
  // set items
  this.setItemsAnimated(this._items.concat([item]), animated);
  // trigger the transaction
  ADTransaction.commit();
};

/**
 *  Removes {@link #topItem} navigation item from the top of the stack and transitions it off screen,
 *  moving in the previous item. A controller object can use the {@link ADNavigationBarDelegate#navigationBarShouldPopItem}
 *  delegate method to prevent items being popped off the stack.
 *
 *  @param {bool} animated <code>true</code> if the navigation bar should be animated; otherwise, <code>false</code>.
 */
ADNavigationBar.prototype.popNavigationItemAnimated = function (animated) {
  // do nothing if we are busy transitioning already or don't have a back item
  if (this.busy || this._items.length < 2) {
    return;
  }
  // start a new transaction to allow transitions to be triggered within the
  // delegate call to be executed in sync with the nav bar animation
  ADTransaction.begin();
  // do nothing if we're not supposed to pop
  if (!this.shouldPopItem(this.topItem)) {
    ADTransaction.commit();
    return;
  }
  // set items
  this.setItemsAnimated(this._items.slice(0, this._items.length - 1), animated);
  // trigger the transaction
  ADTransaction.commit();
};

/* ==================== VIEW CHANGE ==================== */

ADNavigationBar.prototype.addItemViews = function (item, backItem) {
  var left_bar_item = item._leftBarButtonItem || ((backItem !== null) ? backItem.backBarButtonItem : null) || null;
  var items = [left_bar_item, item.rightBarButtonItem, item.titleView];
  for (var i = 0; i < items.length; i++) {
    var item = items[i];
    if (ADUtils.objectIsUndefined(item) || item === null) {
      continue;
    }
    var view = item.view;
    // only add if it's not already a subview and it actually exists
    if (view.superview === null || view.superview !== this) {
      this.addSubview(view);
    }
  }
};

ADNavigationBar.prototype.transitionToItem = function (top_item, previous_item, going_forward, animated) {

  // mark that a transition is now in progress
  this.busy = animated;
  
  // record some parameters that we will need at the end of the transition
  this.previousItem = previous_item;
  this.transitionWentForward = going_forward;

  // use custom defaults
  ADTransaction.defaults.duration = animated ? ADNavigationBarAnimationDuration : 0;
  ADTransaction.defaults.properties = ['opacity', 'position'];

  // add transition for the left button leaving the screen: the old button will fade out
  // and slide to the left of the screen or to where the new title is if going backwards
  // unless it's a custom button when it just fades out
  if (previous_item._leftBarButtonItem !== null) {
    if (previous_item._leftBarButtonItem !== top_item._leftBarButtonItem) {
      new ADTransition({
        target : previous_item._leftBarButtonItem.view,
        properties : ['opacity'],
        to : [0]
      }).start();
    }
  }
  else if (previous_item.leftButton !== null) {
    var end_x = (going_forward) ? (-previous_item.leftButton.view.size.width - ADNavigationItemLeftButtonLeftMargin) : top_item.positions.title;
    new ADTransition({
      target : previous_item.leftButton.view,
      to : [0, new ADPoint(end_x, 0)]
    }).start();
  }

  // add transition for the title leaving the screen: the old title will fade out and slide
  // to where the next button is located or to the right of the screen if going backwards
  var end_x = (going_forward) ? ADNavigationItemLeftButtonLeftMargin : this.size.width;
  new ADTransition({
    target : previous_item.titleView.view,
    to : [0, new ADPoint(end_x, 0)]
  }).start();

  // add fade out transition for the right button leaving the screen
  if (previous_item._rightBarButtonItem !== null && previous_item._rightBarButtonItem !== top_item._rightBarButtonItem) {
    new ADTransition({
      target : previous_item._rightBarButtonItem.view,
      properties : ['opacity'],
      to : [0]
    }).start();
  }

  // add transition for the left button entering the screen: the new button will fade in and slide
  // from the current position of the title or the left of the screen if going backwards, unless it's
  // a custom button, in which case it just fades in
  if (top_item._leftBarButtonItem !== null) {
    if (top_item._leftBarButtonItem !== previous_item._leftBarButtonItem) {
      top_item._leftBarButtonItem.view.position = new ADPoint(top_item.positions.leftButton, 0);
      new ADTransition({
        target : top_item._leftBarButtonItem.view,
        properties : ['opacity'],
        from : [0],
        to : [1]
      }).start();
    }
  }
  else if (top_item.leftButton !== null) {
    var start_x = (going_forward) ? previous_item.positions.title : (-previous_item.leftButton.view.size.width - ADNavigationItemLeftButtonLeftMargin);
    new ADTransition({
      target : top_item.leftButton.view,
      from : [0, new ADPoint(start_x, 0)],
      to : [top_item.hidesBackButton ? 0 : 1, new ADPoint(top_item.positions.leftButton, 0)]
    }).start();
  }

  // add transition for the title entering the screen: the new title will fade in and slide
  // from the right of the screen or where the previous button was located if going backwards
  var start_x = (going_forward) ? this.size.width : ADNavigationItemLeftButtonLeftMargin;
  new ADTransition({
    target : top_item.titleView.view,
    from : [0, new ADPoint(start_x, 0)],
    to : [1, new ADPoint(top_item.positions.title, 0)]
  }).start();

  // add fade in transition for the right button entering the screen
  if (top_item._rightBarButtonItem !== null && top_item._rightBarButtonItem !== previous_item._rightBarButtonItem) {
    top_item._rightBarButtonItem.view.position = new ADPoint(top_item.positions.rightButton, 0);
    new ADTransition({
      target : top_item._rightBarButtonItem.view,
      properties : ['opacity'],
      from : [0],
      to : [1]
    }).start();
  }

  // and if we are dealing with an animated transition, we need to get a notification
  // for when the transition is done so that delegate methods can be called
  if (animated) {
    previous_item.titleView.view.layer.addEventListener('webkitTransitionEnd', this, false);
  }

  // and, finally, run the transitions and animations
  ADTransaction.commit();

  // in case we don't have animated transitions, let's do post-op cleanups
  if (!animated) {
    this.transitionsEnded();
  }
  
};

ADNavigationBar.prototype.transitionsEnded = function () {
  // update the stack
  this._items = this.itemsAfterTransition;
  // mark that we're not busy anymore
  this.busy = false;
  // we pushed an item
  if (this.transitionWentForward) {
    // tell the delegate we pushed an item
    if (ADUtils.objectHasMethod(this.delegate, ADNavigationBarDidPushItem)) {
      this.delegate[ADNavigationBarDidPushItem](this, this.topItem);
    }
  }
  else {
    // tell the delegate we popped an item
    if (ADUtils.objectHasMethod(this.delegate, ADNavigationBarDidPopItem)) {
      this.delegate[ADNavigationBarDidPopItem](this, this.previousItem);
    }
  }
  // remove the buttons from the previous item that aren't here anymore
  this.removeViewsForItem(this.previousItem);
  // tell the NavigationController, if there is one, that we're done transitioning
  if (this.superview instanceof ADNavigationView) {
    this.superview.viewController.transitionsEnded();
  }
};

ADNavigationBar.prototype.removeViewsForItem = function (item) {
  // remove the buttons from the previous item that aren't here anymore
  if (item.leftButton !== null && item.leftButton !== this.topItem.leftButton) {
    item.leftButton.view.removeFromSuperview();
  }
  item.titleView.view.removeFromSuperview();
  if (item._rightBarButtonItem !== null && item._rightBarButtonItem !== item._rightBarButtonItem) {
    item._rightBarButtonItem.view.removeFromSuperview();
  }
  // make sure the exiting back button, if it existed, has no event listener
  if (item.leftButton !== null) {
    item.leftButton.removeEventListener(ADControlTouchUpInsideEvent, this, false);
  }
};

ADNavigationBar.prototype.updateTopItemLayout = function () {
  if (this._items.length > 0) {
    this.topItem.updateLayoutIfTopItem();
  }
};

/* ==================== Delegate Methods Wrappers ==================== */

ADNavigationBar.prototype.shouldPushItem = function (item) {
  if (ADUtils.objectHasMethod(this.delegate, ADNavigationBarShouldPushItem)) {
    return this.delegate[ADNavigationBarShouldPushItem](this, item);
  }
  return true;
};

ADNavigationBar.prototype.shouldPopItem = function (item) {
  if (ADUtils.objectHasMethod(this.delegate, ADNavigationBarShouldPopItem)) {
    return this.delegate[ADNavigationBarShouldPopItem](this, item);
  }
  return true;
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADNavigationBar);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

const ADNavigationItemLeftButtonLeftMargin = 5;
const ADNavigationItemLeftButtonRightMargin = 8;
const ADNavigationItemRightButtonLeftMargin = 11;
const ADNavigationItemRightButtonRightMargin = 5;

/* ==================== ADNavigationItem ==================== */

ADNavigationItem.inherits = ADObject;
ADNavigationItem.synthetizes = ['title', 'backBarButtonItem', 'leftBarButtonItem', 'rightBarButtonItem', 'hidesBackButton'];
ADNavigationItem.includes = [ADEventTriage];

/**
 *  @class
 *
 *  <p>The ADNavigationItem class encapsulates information about a navigation item pushed on a {@link ADNavigationBar}
 *  {@link ADNavigationBar#items} stack. A navigation view is a view used to navigate hierarchical content.
 *  A ADNavigationItem specifies what is displayed on the navigation bar when it is the {@link ADNavigationBar#topItem}
 *  and also how it is represented when it is the {@link ADNavigationBar#backItem}.</p>
 *
 *  <p>The item cannot be represented on the navigation bar without a {@link #title}. Use the {@link #backBarButtonItem}
 *  property if you want to use a different title when this item is the back item. The {@link #backBarButtonItem} property
 *  is displayed as the back button unless a custom left view is specified.
 *
 *  <p>The navigation bar displays a back button on the left and the title in the center by default. You can change this
 *  behavior by specifying either a custom left or right view. Use the {@link #setLeftBarButtonItemWithAnimation} and
 *  {@link #setRightBarButtonItemWithAnimation} methods to change the left and right views; you can specify that the
 *  change be animated.</p>
 *
 *  <p>Use the {@link ADBarButtonItem} class to create custom views to add to navigation items.</p>
 *  
 *  @since AdLib 0.1
 *
 *  @param {String} title The item's title.
 */
function ADNavigationItem (title) {
  this.callSuper();
  this._title = '';
  this._backBarButtonItem = new ADBarButtonItem(ADBarButtonItemTypeBack);
  this._hidesBackButton = false;
  this._leftBarButtonItem = null;
  this._rightBarButtonItem = null;
  //
  this.titleView = new ADBarButtonItem(ADBarButtonItemTypePlain);
  this.titleView.view.layer.setAttribute('role', 'header');
  this.buttons = null;
  this.positions = null;
  this.navigationBar = null;
  this.viewController = null;
  //
  this.title = title || '';
  // add event listener onto the back button
  this._backBarButtonItem.addEventListener(ADControlTouchUpInsideEvent, this, false);
};

/* ==================== Back Button Event Handling ==================== */

ADNavigationItem.prototype.handleControlTouchUpInside = function (event) {
  // do nothing if the button is actually hidden
  if (this._hidesBackButton) {
    return;
  }
  // pop the view controller represented by this navigation item if there is one
  if (this.viewController !== null && this.viewController.parentViewController !== null) {
    this.viewController.parentViewController.popViewControllerAnimated(true);
  }
  // pop this item from the navigation bar if we're in one
  else if (this.navigationBar !== null) {
    this.navigationBar.popNavigationItemAnimated(true);
  }
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADNavigationItem.prototype
 *
 *  @property {String} title The item's title, appears in the navigation view's top bar in the center when it's the
 *  {@link ADNavigationBar#topItem} or in the back button if this item is the second item to the top of the
 *  {@link ADNavigationBar#items} stack.
 *
 *  @property {ADBarButtonItem} backBarButtonItem The bar button item to use when this item is represented by
 *  a back button on the navigation bar. When this item is the back item of the navigation bar  when it is the
 *  next item below the top item  it may be represented as a back button on the navigation bar. Use this property
 *  to specify the back button. The default value is a bar button item displaying the navigation item's title.
 *
 *  @property {bool} hidesBackButton <code>true</code> if the back button is hidden when this navigation item is the
 *  top item; otherwise, <code>false</code>. The default value is <code>false</code>.
 *
 *  @property {ADBarButtonItem} leftBarButtonItem A custom bar item displayed on the left of the navigation bar when
 *  this item is the top item.
 *
 *  @property {ADBarButtonItem} rightBarButtonItem A custom bar item displayed on the right of the navigation bar when
 *  this item is the top item.
 *
 */

ADNavigationItem.prototype.setTitle = function (title) {
  this._title = title;
  // update layout if we're top item
  this.updateLayoutIfTopItem();
};

ADNavigationItem.prototype.setBackBarButtonItem = function (item) {
  // now, update the top item's back button display if we're the back item
  if (this.navigationBar !== null && this.navigationBar.backItem === this) {
    if (this._backBarButtonItem !== null) {
      this._backBarButtonItem.view.removeFromSuperview();
    }
    this._backBarButtonItem = item;
    this.navigationBar.addSubview(this._backBarButtonItem.view);
    this.navigationBar.topItem.updateLayoutIfTopItem();
  }
  else {
    this._backBarButtonItem = item;
  }
  // ensure it is hooked up to have the right event listener
  if (this._backBarButtonItem !== null) {
    this._backBarButtonItem.addEventListener(ADControlTouchUpInsideEvent, this, false);
  }
};

ADNavigationItem.prototype.setHidesBackButton = function (hidesBackButton) {
  this.setHidesBackButtonWithAnimation(hidesBackButton, false);
};

ADNavigationItem.prototype.setLeftBarButtonItem = function (item) {
  if (this.navigationBar !== null && this.navigationBar.topItem === this) {
    // get the default back button in case we need it
    var back_button = this.getDefaultBackButton();
    // first, remove whatever was here
    if (this.leftButton !== null) {
      this.leftButton.view.removeFromSuperview();
    }
    // record what the new left bar button item is
    this._leftBarButtonItem = item;
    var new_button = this._leftBarButtonItem || this.getDefaultBackButton();
    // add the new button to the tree if we have one
    if (new_button !== null) {
      this.navigationBar.addSubview(new_button.view);
      // ensure we have it visible if it's the default back button and it should show
      if (this._leftBarButtonItem === null && !this.hidesBackButton) {
        new_button.view.opacity = 1;
      }
    }
    // update the display
    this.updateLayoutIfTopItem();
  }
  else {
    this._leftBarButtonItem = item;
  }
};

ADNavigationItem.prototype.setRightBarButtonItem = function (item) {
  if (this.navigationBar !== null && this.navigationBar.topItem === this) {
    if (this._rightBarButtonItem !== null) {
      this._rightBarButtonItem.view.removeFromSuperview();
    }
    this._rightBarButtonItem = item;
    this.navigationBar.addSubview(this._rightBarButtonItem.view);
    this.updateLayoutIfTopItem();
  }
  else {
    this._rightBarButtonItem = item;
  }
};

/* ==================== Public Methods ==================== */

/**
 *  Sets whether the back button is hidden, optionally animating the transition.
 *
 *  @param {bool} hidesBackButton <code>true</code> if the back button is hidden when this navigation item is the top item;
 *  otherwise, <code>false</code>.
 *  @param {bool} animated <code>true</code> to animate the transition; otherwise, <code>false</code>.
 */
ADNavigationItem.prototype.setHidesBackButtonWithAnimation = function (hidesBackButton, animated) {
  var back_button = this.getDefaultBackButton();
  // do nothing if we already had this value
  if (this._hidesBackButton == hidesBackButton) {
    return;
  }
  // record new value
  this._hidesBackButton = hidesBackButton;
  // do nothing if we don't display a back button
  if (back_button === null) {
    return;
  }
  back_button.view.transitionsEnabled = animated;
  back_button.view.opacity = hidesBackButton ? 0 : 1;
};

/**
 *  Sets the custom bar item, optionally animating the transition to the view. If two navigation
 *  items have the same custom left or right bar items, the views remain stationary during the
 *  transition when an item is pushed or popped.
 *
 *  @param {ADBarButtonItem} item A custom bar item to display on the left of the navigation bar.
 *  @param {bool} animated <code>true</code> to animate the transition to the custom bar item when
 *  this item becomes the top item; otherwise, <code>false</code>.
 */
ADNavigationItem.prototype.setLeftBarButtonItemWithAnimation = function (item, animated) {
  // regular set if we're not animated or we're not the top item
  if (!animated || this.navigationBar === null || this.navigationBar.topItem !== this) {
    this.leftBarButtonItem = item;
    return;
  }
  // start transaction
  ADTransaction.begin();
  ADTransaction.defaults.duration = 0.5;
  ADTransaction.defaults.properties = ['opacity'];
  // deal with the old button if present
  var old_button = this.leftButton;
  if (old_button !== null) {
    new ADTransition({
      target : old_button.view,
      to : [0],
      removesTargetUponCompletion : true
    }).start();
  }
  // deal with the new button if present
  this._leftBarButtonItem = item;
  var new_button = this._leftBarButtonItem || this.getDefaultBackButton();
  if (new_button !== null) {
    this.navigationBar.addSubview(new_button.view);
    // now apply the fade if we don't deal with the back button
    if (this._leftBarButtonItem !== null || !this._hidesBackButton) {
      new ADTransition({
        target : new_button.view,
        from : [0],
        to : [1]
      }).start();
    }
  }
  // refresh the layout of items
  this.updateLayoutIfTopItem();
  //
  ADTransaction.commit();
};

/**
 *  Sets the custom right bar item, optionally animating the transition to the view. If two navigation
 *  items have the same custom left or right bar items, the bar items remain stationary during the
 *  transition when an item is pushed or popped.
 *
 *  @param {ADBarButtonItem} item A custom bar item to display on the right of the navigation bar.
 *  @param {bool} animated <code>true</code> to animate the transition to the custom bar item when this
 *  item becomes the top item; otherwise, <code>false</code>.
 */
ADNavigationItem.prototype.setRightBarButtonItemWithAnimation = function (item, animated) {
  // regular set if we're not animated or we're not the top item
  if (!animated || this.navigationBar === null || this.navigationBar.topItem !== this) {
    this.rightBarButtonItem = item;
    return;
  }
  // start transaction
  ADTransaction.begin();
  ADTransaction.defaults.duration = 0.5;
  ADTransaction.defaults.properties = ['opacity'];
  // deal with the old button if present
  var old_button = this._rightBarButtonItem;
  if (old_button !== null) {
    new ADTransition({
      target : old_button.view,
      to : [0],
      removesTargetUponCompletion : true
    }).start();
  }
  // deal with the new button if present
  this._rightBarButtonItem = item;
  if (this._rightBarButtonItem !== null) {
    this.navigationBar.addSubview(this._rightBarButtonItem.view);
    // now apply the fade
    new ADTransition({
      target : this._rightBarButtonItem.view,
      from : [0],
      to : [1]
    }).start();
  }
  // refresh the layout of items
  this.updateLayoutIfTopItem();
  //
  ADTransaction.commit();
};

/* ==================== Private APIs ==================== */

/**
 *  This method returns the default back button for the back item if there is one
 *  and we can identify it.
 *
 *  @private
 */
ADNavigationItem.prototype.getDefaultBackButton = function () {
  return (this.navigationBar !== null && this.navigationBar.backItem !== null) ? this.navigationBar.backItem.backBarButtonItem : null;
};

/**
 *  This method is where all the layout constraints for displaying a navigation item
 *  in a navigation bar are. Once this method is called, the various buttons and the
 *  title view will be sized according to those constraints, and the .positions property
 *  will be updated with the position for each item. Actually positioning items is up
 *  to the navigation bar, or possibly the navigation item if it's the buttons or title
 *  are mutated.
 *
 *  @private
 */
ADNavigationItem.prototype.sizeItemsAndComputePositionsWithBackItem = function (back_item) {

  // do nothing if we're not yet hosted in a navigation bar
  if (this.navigationBar === null) {
    return;
  }

  // figure out what the left button will be since it can come from different places
  var left_button = this._leftBarButtonItem || ((back_item !== null) ? back_item.backBarButtonItem : null);

  // get the right button width
  var right_button_width = (this._rightBarButtonItem !== null) ? this._rightBarButtonItem.view.size.width : 0;

  // first let's see how much space we have once we have the right bar button item in there
  var available_width = this.navigationBar.size.width - ADNavigationItemLeftButtonLeftMargin - ADNavigationItemRightButtonRightMargin;
  // remove the left button right margin if we have one
  if (left_button !== null) {
    available_width -= ADNavigationItemLeftButtonRightMargin;
  }
  // remove the right button width and left margin if we have one, rule being that we always fit the right button, no matter what
  if (this._rightBarButtonItem !== null) {
    available_width -= ADNavigationItemRightButtonLeftMargin + right_button_width;
  }

  // figure out the maximum space taken by the left button
  var left_button_width = 0;
  if (left_button !== null) {
    left_button.maxWidth = this.navigationBar.size.width / 2;
    // update the title from the back item if we have an implicit left button
    if (left_button !== this._leftBarButtonItem && left_button.title == '' && left_button.image === null) {
      left_button.title = back_item.title;
    }
    left_button_width = left_button.view.size.width;
  }

  // update the title and get its ideal width when not trimmed
  this.titleView.maxWidth = 0;
  this.titleView.title = this._title;
  var title_width = this.titleView.view.size.width;

  // handle the case where we can't fit both without some trimming
  // the rule being that in case of trimming, the left button has
  // a max width of a third of the available width
  if (title_width + left_button_width > available_width) {
    if (left_button !== null) {
      left_button.maxWidth = Math.min(Math.max(available_width / 3, available_width - title_width), left_button.maxWidth);
      left_button_width = left_button.view.size.width;
    }
    this.titleView.maxWidth = available_width - left_button_width;
    title_width = this.titleView.view.size.width;
  }

  // now everything is sized, we can figure out the position for both buttons
  var left_button_x = ADNavigationItemLeftButtonLeftMargin;
  var right_button_x = this.navigationBar.size.width - ADNavigationItemRightButtonRightMargin - right_button_width;

  // the position of the title is a bit trickier. Ideally we fit display it in the center of the bar. If this would overlap
  // the left button, we align it to the right of the left button, and if it would overlap the right button, we align it to
  // the left of the right button.
  var left_most_title_x = left_button_x + ((left_button != null) ? left_button_width + ADNavigationItemLeftButtonRightMargin : 0);
  var right_most_title_x = right_button_x - title_width - ((right_button_width > 0) ? ADNavigationItemRightButtonLeftMargin : 0);
  var centered_title_x = (this.navigationBar.size.width - title_width) / 2;

  // default the title to being centered
  var title_x = centered_title_x;
  // if that overlaps a button, use the coordinate that's closest to center
  if (centered_title_x > right_most_title_x || centered_title_x < left_most_title_x) {
    var left_distance = Math.abs(centered_title_x - left_most_title_x);
    var right_distance = Math.abs(centered_title_x - right_most_title_x);
    title_x = (left_distance < right_distance) ? left_most_title_x : right_most_title_x;
  }

  // track what we use as your left button
  this.leftButton = left_button;

  // all done, save those metrics in the .positions ad-hoc object
  this.positions = {
    leftButton : left_button_x,
    title : title_x,
    rightButton : right_button_x
  };

};

ADNavigationItem.prototype.updateLayout = function () {
  // do nothin if we don't have positions
  if (this.positions === null || this.button === null) {
    return;
  }
  // position the left button, if any
  if (this.leftButton != null) {
    this.leftButton.view.position = new ADPoint(this.positions.leftButton, 0);
  }
  // position the right button, if any
  if (this._rightBarButtonItem != null) {
    this._rightBarButtonItem.view.position = new ADPoint(this.positions.rightButton, 0);
  }
  // position the title
  this.titleView.view.position = new ADPoint(this.positions.title, 0);
};

ADNavigationItem.prototype.updateLayoutIfTopItem = function (backItem) {
  // only go forward if we're the top item of the navigation view
  if (this.navigationBar === null || this.navigationBar.topItem !== this) {
    return;
  }
  // update ourselves
  this.sizeItemsAndComputePositionsWithBackItem(this.navigationBar.backItem);
  this.updateLayout();
};

/* ==================== Class Creation ==================== */

ADClass(ADNavigationItem);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/*** 

  How customization works
  
  - show a static grid of items
    - when the user drags from the grid, scale the drag image up
    - if the dragged item hovers something on the bar, show the glow
    - if the dragged item hovers itself on the bar, don't show the glow
    
  - when an item is dropped onto the bar
    - swap with the item in the bar, but don't update the grid
    - if the dropped item was already in the bar, animate swap the items in the bar (setItemsAnimated?)

  - check if the item is customizable with delegate call

  - dragging from the bar just swaps items

***/

/* ==================== Constants ==================== */

/**
 *  Height of a {@link ADTabBar}.
 *  @constant
 *  @type int
 */
const ADTabBarHeight = 49;

const ADTabBarCustomizeGridCellSize = 80;

/**
 *  @class
 *  @name ADTabBarDelegate
 *  @since AdLib 0.1
 */

/**
 *  Invoked when the user selects a tab bar item within a {@link ADTabBar}. This method is optional.
 *
 *  @name tabBarDidSelectItem
 *  @function
 *
 *  @param {ADTabBar} theTabBar The tab bar.
 *  @param {ADTabBarItem} theSelectedItem The selected tab bar item.
 *  @memberOf ADTabBarDelegate.prototype
 */
const ADTabBarDidSelectItem = 'tabBarDidSelectItem';

/**
 *  Invoked before the customizing modal view is displayed. Use the {@link ADTabBar#beginCustomizingItems} method of {@link ADTabBar}
 *  to display the customizing modal view and begin the customizing mode. This method is optional.
 *
 *  @name tabBarWillBeginCustomizingItems
 *  @function
 *
 *  @param {ADTabBar} theTabBar The tab bar that is being customized.
 *  @param {Array} items The items on the customizing modal view.
 *  @memberOf ADTabBarDelegate.prototype
 */
const ADTabBarWillBeginCustomizingItems = 'tabBarWillBeginCustomizingItems';

/**
 *  Invoked after the customizing modal view is displayed. This method is optional.
 *
 *  @name tabBarDidBeginCustomizingItems
 *  @function
 *
 *  @param {ADTabBar} theTabBar The tab bar that is being customized.
 *  @param {Array} items The items on the customizing modal view.
 *  @memberOf ADTabBarDelegate.prototype
 */
const ADTabBarDidBeginCustomizingItems = 'tabBarDidBeginCustomizingItems';  

/**
 *  Invoked before the customizing modal view is dismissed. This method is optional.
 *
 *  @name tabBarWillEndCustomizingItemsChanged
 *  @function
 *
 *  @param {ADTabBar} theTabBar The tab bar that is being customized.
 *  @param {Array} items The items on the customizing modal view.
 *  @param {bool} changed <code>true</code> if the visible set of items on the tab bar changed; otherwise, <code>false</code>.
 *  @memberOf ADTabBarDelegate.prototype
 */
const ADTabBarWillEndCustomizingItemsChanged = 'tabBarWillEndCustomizingItemsChanged';  

/**
 *  Invoked after the customizing modal view is dismissed. This method is optional.
 *
 *  @name tabBarDidEndCustomizingItemsChanged
 *  @function
 *
 *  @param {ADTabBar} theTabBar The tab bar that is being customized.
 *  @param {Array} items The items on the customizing modal view.
 *  @param {bool} changed <code>true</code> if the visible set of items on the tab bar changed; otherwise, <code>false</code>.
 *  @memberOf ADTabBarDelegate.prototype
 */
const ADTabBarDidEndCustomizingItemsChanged = 'tabBarDidEndCustomizingItemsChanged';

/* ==================== ADTabBar ==================== */

ADTabBar.inherits = ADView;
ADTabBar.synthetizes = ['items', 'selectedItem'];

/**
 *  @class
 *
 *  <p>The ADTabBar class implements a control for selecting one of two or more buttons, called {@link #items}.
 *  The most common use of a tab bar is to implement a modal interface where tapping an item changes
 *  the selection. Use a {@link ADToolbar} object if you want to momentarily highlight or not change the
 *  appearance of an item when tapped. The ADTabBar class provides the ability for the user to customize
 *  the tab bar by reordering, removing, and adding items to the bar. You can use a tab bar {@link #delegate}
 *  to augment this behavior.</p>
 *
 *  <p>Use the {@link ADTabBarItem} class to create items and the {@link #items} array to display them
 *  within the tab bar. Note that you can use the {@link #setItemsAnimated} method to add items to a
 *  tab bar with an optional animation. Use the selectedItem property to access the current item.</p>
 *
 *  @extends ADView
 *  @since AdLib 0.1
 */
function ADTabBar () {
  /**
   *  The delegate should conform to the {@link ADTabBarDelegate} protocol. Set this property to further modify
   *  the customizing behavior. The default value is <code>null</code>.
   *  @type Object 
   */
  this.delegate = null;
  this.isCustomizing = false;
  // synthesize read-write properties
  this._items = [];
  this._selectedItem = null;
  //
  this.callSuper();
  // track transition completions
  this.layer.addEventListener('webkitTransitionEnd', this, false);
};

/* ==================== ADView APIs ==================== */

ADTabBar.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-tab-bar');
};

ADTabBar.prototype.setSize = function (aSize) {
  aSize.height = ADTabBarHeight;
  this.callSuper(aSize);
  this.updateLayout();
};

ADTabBar.prototype.willMoveToSuperview = function (superview) {
  // do nothing if we're not moving to a new superview
  if (superview === null) {
    return;
  }
  // force bar width to be that of superview if not hard set before
  if (this._size.width == 0) {
    this.size = new ADSize(superview.size.width, ADTabBarHeight);
  }
  // force bar y to be that bottom superview if not hard set before
  if (this._position.y == 0) {
    this.position = new ADPoint(0, superview.size.height - ADTabBarHeight);
  }
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADTabBar.prototype
 *  @property {ADTabBarItem} selectedItem The currently selected item on the tab bar.
 *  @property {Array} items The items, instances of {@link ADTabBarItem}, that are visible on the tab bar in the order
 *  they appear in this array. Any changes to this property are not animated. Use the {@link #setItemsAnimated} method to animate changes.
 */

ADTabBar.prototype.setSelectedItem = function (selectedItem) {
  // do nothing if this item was already selected
  if (this._selectedItem === selectedItem) {
    return;
  }
  // de-select the previous item
  if (this._selectedItem != null) {
    this._selectedItem.selected = false;
  }
  // select the new one
  selectedItem.selected = true;
  // track selection
  this._selectedItem = selectedItem;
  // tell our delegate
  if (ADUtils.objectHasMethod(this.delegate, ADTabBarDidSelectItem)) {
    this.delegate[ADTabBarDidSelectItem](this, selectedItem);
  }
};

ADTabBar.prototype.setItems = function (items) {
  this.setItemsAnimated(items, false);
};

/**
 *  If animated is <code>true</code>, the changes are dissolved or the reordering is animated. For example,
 *  removed items fade out and new items fade in. This method also adjusts the spacing between items.
 *
 *  @param {Array} items The items to display on the tab bar.
 *  @param {bool} animated If <code>true</code>, animates the transition to the items; otherwise, does not.
 */
ADTabBar.prototype.setItemsAnimated = function (items, animated) {

  // first, go through current items and fade out any that
  // won't be needed anymore
  for (var i = 0; i < this._items.length; i++) {
    var item = this._items[i];
    if (items.indexOf(item) == -1) {
      item.transitionsEnabled = animated;
      if (!animated) {
        item.removeFromSuperview();
      }
      else {
        item.needsRemoval = true;
        item.transform = 'scale(0.001)';
        item.opacity = 0;
      }
    }
  }

  // now go through new items, adding those that weren't here before
  var item_width = (this.size.width - 4) / items.length;
  for (var i = 0; i < items.length; i++) {
    var item = items[i];
    item.position = new ADPoint(2 + Math.round(i * item_width), 0);
    item.size = new ADSize(Math.round(item_width), ADTabBarItemHeight);
    item.transitionsEnabled = animated;
    // add the ones that weren't already here
    if (item.superview !== this) {
      this.addSubview(item);
      // listen to controlTouchDown so that we can select the cell
      item.addEventListener(ADControlTouchDownEvent, this, false);
    }
  }

  // save to private ivar
  this._items = items;

  // finally, check if the selected item is still there, otherwise
  // select the first item
  if ((this.selectedItem == null || this.selectedItem.superview !== this || this.selectedItem.needsRemoval) && items.length > 0) {
    this.selectedItem = items[0];
  }

};

/* ==================== Items layout ==================== */

ADTabBar.prototype.updateLayout = function () {
  // do nothing if we don't have items
  if (this._items.length > 0) {
    this.setItemsAnimated(this._items, false);
  }
};

/* ==================== Event Handling ==================== */

ADTabBar.prototype.handleEvent = function (event) {
  this.callSuper(event);
  //
  switch (event.type) {
    // an item was selected
    case ADControlTouchDownEvent : 
      this.selectedItem = event.control;
      break;
    // an item may have faded out
    case 'webkitTransitionEnd' :
      this.removeItemIfNeeded(event.target._control);
      break;
  }
};

ADTabBar.prototype.removeItemIfNeeded = function (item) {
  if (item.needsRemoval) {
    // remove item from view
    item.removeFromSuperview();
    // reset it for future use
    item.needsRemoval = false;
    item.transitionsEnabled = false;
    item.transform = 'scale(1)';
    item.opacity = 1;
  }
};

/* ==================== Customizing Tab Bars ==================== */

ADTabBar.prototype.beginCustomizingItems = function (items) {
  // tell our delegate we're about to start customizing
  if (ADUtils.objectHasMethod(this.delegate, ADTabBarWillBeginCustomizingItems)) {
    this.delegate[ADTabBarWillBeginCustomizingItems](this, items);
  }
  // track that we are customizing
  this.isCustomizing = true;
  // show the customizing modal view
  this.itemsPriorToCustomization = this._items;
  this.customizingItems = items;
  // tell our delegate we're about to start customizing
  if (ADUtils.objectHasMethod(this.delegate, ADTabBarDidBeginCustomizingItems)) {
    this.delegate[ADTabBarDidBeginCustomizingItems](this, items);
  }
};

ADTabBar.prototype.endCustomizingAnimated = function (animated) {
  // figure out if we changed anything?
  // this.itemsPriorToCustomization
  var changed = true;
  // tell our delegate we're about to finish customizing
  if (ADUtils.objectHasMethod(this.delegate, ADTabBarWillEndCustomizingItemsChanged)) {
    this.delegate[ADTabBarWillEndCustomizingItemsChanged](this, this.customizingItems, changed);
  }
  // track that we are no longer customizing
  this.isCustomizing = false;
  // dismiss the customization panel

  // tell our delegate we've finished customizing
  if (ADUtils.objectHasMethod(this.delegate, ADTabBarDidEndCustomizingItemsChanged)) {
    this.delegate[ADTabBarDidEndCustomizingItemsChanged](this, this.customizingItems, changed);
  }
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADTabBar);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  Height of a {@link ADToolbar}.
 *  @constant
 *  @type int
 */
const ADToolbarHeight = 44;
const ADToolbarEdgeMargin = 6;
const ADToolbarItemMargin = 10;

/**
 *  Default style for a {@link ADToolbar}.
 *  @constant
 *  @type String
 */
const ADToolbarStyleDefault = 'default';
/**
 *  Black style for a {@link ADToolbar}.
 *  @constant
 *  @type String
 */
const ADToolbarStyleBlack = 'black';
/**
 *  Translucent black style for a {@link ADToolbar}.
 *  @constant
 *  @type String
 */
const ADToolbarStyleBlackTranslucent = 'black-translucent';

/**
 *  Duration in seconds of the animation as items are changed.
 *  @constant
 *  @type float
 */
const ADToolbarAnimationDuration = 0.2;

// the transition we use for items that are not needed in the toolbar anymore
const ADToolbarFadeOutTransition = {
  properties : ['transform', 'opacity'],
  to : ['scale(0.01)', 0]
};

/* ==================== ADToolbar ==================== */

ADToolbar.inherits = ADView;
ADToolbar.synthetizes = ['items', 'style'];

/**
 *  @class
 *
 *  <p>The ADToolbar class implements a control for selecting one of many buttons, called toolbar items.
 *  A toolbar momentarily highlights or does not change the appearance of an item when tapped. Use the
 *  {@link ADTabBar} class if you need a radio button style control.</p>
 *
 *  <p>Use the {@link ADBarButtonItem} class to create items and the {@link #items} array to display them
 *  within the toolbar. Note that you can use the {@link #setItemsAnimated} method to add items to a
 *  toolbar with an optional animation.</p>
 *
 *  @extends ADView
 *  @since AdLib 0.1
 */
function ADToolbar () {
  // synthesize read-write properties
  this._items = [];
  this._style = '';
  //
  this.callSuper();
  // track transition completions
  this.layer.addEventListener('webkitTransitionEnd', this, false);
  // set up base type
  this.style = ADToolbarStyleDefault;
  // make sure we clip contents
  this.clipsToBounds = true;
};

/* ==================== ADView APIs ==================== */

ADToolbar.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-toolbar');
  //
  this.glow = this.layer.appendChild(document.createElement('div'));
  this.glow.className = 'glow';
};

ADToolbar.prototype.setPosition = function (aPoint) {
  this.callSuper(aPoint);
  // update the toolbar's background if we're in the document
  if (this.layerIsInDocument) {
    // need to do this after the style update has been resolved
    // <rdar://problem/7589997> webkitConvertPointFromNodeToPage doesn't take into account most recent transform style (34023)
    this.callMethodNameAfterDelay('updateBackgroundWithPosition', 0);
  }
};

ADToolbar.prototype.setSize = function (aSize) {
  aSize.height = ADToolbarHeight;
  this.callSuper(aSize);
  this.updateLayout();
};

ADToolbar.prototype.willMoveToSuperview = function (superview) {
  // force bar width to be that of superview if not hard set before
  if (superview !== null && this._size.width == 0) {
    this.size = new ADSize(superview.size.width, ADToolbarHeight);
  }
};

ADToolbar.prototype.layerWasInsertedIntoDocument = function () {
  this.updateBackgroundWithPosition();
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADToolbar.prototype
 *  @property {String} style The toolbar style that specifies its appearance. The default value is <code>ADToolbarStyleDefault</code>.
 *  @property {Array} items The items, instances of {@link ADBarButtonItem}, that are visible on the toolbar in the order they
 *  appear in this array. Any changes to this property are not animated. Use the {@link setItemsAnimated} method to animate changes.
 */

ADToolbar.prototype.setStyle = function (style) {
  this.layer.removeClassName(this._style);
  this.layer.addClassName(style);
  this._style = style;
};

ADToolbar.prototype.setItems = function (items) {
  this.setItemsAnimated(items, false);
};

/**
 *  Sets the items on the toolbar by animating the changes. If animated is <code>true</code>, the changes are dissolved or the
 *  reordering is animated. For example, removed items fade out and new items fade in. This method also adjusts the spacing between items.
 *
 *  @param {Array} items The items to display on the toolbar.
 *  @param {bool} animated If set to <code>true</code>, animates the transition to the items; otherwise, does not.
 */
ADToolbar.prototype.setItemsAnimated = function (items, animated) {

  ADTransaction.begin();
  ADTransaction.defaults.duration = animated ? ADToolbarAnimationDuration : 0;

  // first, go through current items and fade out any that
  // won't be needed anymore
  for (var i = 0; i < this._items.length; i++) {
    var item = this._items[i];
    var view = item.view;
    if (items.indexOf(item) == -1) {
      if (!animated) {
        view.removeFromSuperview();
      }
      else {
        view.needsRemoval = true;
        view.applyTransition(ADToolbarFadeOutTransition);
      }
    }
  }

  // first, add new items to the tree
  for (var i = 0; i < items.length; i++) {
    item = items[i];
    view = item.view;
    if (view.superview !== this) {
      item._newItem = true;
      this.addSubview(view);
      // add observer to track any change in size
      view.addPropertyObserver('size', this);
      // add event handler to track touches on plain items so that we have a glow
      if (item.type == ADBarButtonItemTypePlain) {
        view.addEventListener(ADControlTouchStateChangeEvent, this, false);
      }
    }
  }

  // go through each of the new items and see how many flexible spaces we got
  var number_of_flexible_spaces = 0;
  var collective_item_width = 0;
  for (var i = 0; i < items.length; i++) {
    item = items[i];
    if (item.type == ADBarButtonItemTypeFlexibleSpace) {
      number_of_flexible_spaces++;
    }
    else {
      collective_item_width += item.view.size.width;
    }
  }

  // now go through items and position them
  var available_space = this.size.width - collective_item_width - ADToolbarItemMargin * (items.length - 1) - ADToolbarEdgeMargin * 2;
  var flexible_space_width = (number_of_flexible_spaces > 0) ? (available_space / number_of_flexible_spaces) : 0;
  var next_item_x = ADToolbarEdgeMargin;
  var margin;

  for (var i = 0; i < items.length; i++) {
    margin = ADToolbarItemMargin;
    item = items[i];
    view = item.view;
    // check if we're dealing with a flexible space
    if (item.type == ADBarButtonItemTypeFlexibleSpace) {
      next_item_x += flexible_space_width;
    }
    // or a fixed space
    else if (item.type == ADBarButtonItemTypeFixedSpace) {
      next_item_x += view.size.width;
      // remove the margin if we had a fixed space given that the user likely is sizing
      // things just the way he wants them sized
      margin = 0;
    }
    // otherwise layout the more regular item
    else {
      var position = new ADPoint(next_item_x, (this.size.height - view.size.height) / 2);
      // fade in items that weren't already there
      if (item._newItem) {
        view.position = position;
        item._newItem = false;
        if (animated) {
          view.applyTransition(ADViewTransitionDissolveIn);
        }
      }
      // and animate position of those that aren't new
      else {
        view.applyTransition({
          properties : ['position'],
          to : [position]
        })
      }
      next_item_x += view.size.width;
    }
    // add the margin
    next_item_x += margin;
  }

  ADTransaction.commit();

  // save to private ivar
  this._items = items;
  
};

/* ==================== Event Handling ==================== */

ADToolbar.prototype.updateLayout = function () {
  // do nothing if we don't have items
  if (this._items.length > 0) {
    this.setItemsAnimated(this._items, false);
  }
};

ADToolbar.prototype.updateBackgroundWithPosition = function () {
  var origin = window.webkitConvertPointFromNodeToPage(this.layer, new WebKitPoint(0, 0));
  this.layer[origin.y == 0 ? 'addClassName' : 'removeClassName']('top');
};

/* ==================== Event Handling ==================== */

ADToolbar.prototype.handleEvent = function (event) {
  this.callSuper(event);
  //
  if (event.type == 'webkitTransitionEnd') {
    // nothing to do if we had a transition complete on the layer
    if (event.target === this.layer) {
      return;
    }
    else if (event.target !== this.glow) {
      this.removeItemViewIfNeeded(event.target._control);
    }
    else if (this.glow.style.opacity == 0) {
      this.glow.style.display = 'none';
    }
  }
  else if (event.type == ADControlTouchStateChangeEvent) {
    var item_view = event.control;
    this.glow.style.webkitTransform = ADUtils.t(item_view.position.x + item_view.size.width / 2 - 50, 0);
    this.glow.style.opacity = item_view.touchInside ? 1 : 0;
    this.glow.style.display = 'block';
  }
};

ADToolbar.prototype.removeItemViewIfNeeded = function (view) {
  if (view.needsRemoval) {
    // remove item from view
    view.removeFromSuperview();
    // reset it for future use
    view.needsRemoval = false;
    view.transitionsEnabled = false;
    view.transform = 'scale(1)';
    view.opacity = 1;
  }
};

ADToolbar.prototype.handlePropertyChange = function (object, propertyName) {
  this.setItemsAnimated(this._items, false);
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADToolbar);

/* ==================== CLASS CONSTANTS ==================== */

ADToolbar.init = function () {
  ADUtils.preloadImageAsset('bar/UINavigationBarDefaultBackground.png');
  ADUtils.preloadImageAsset('bar/UINavigationBarBlackOpaqueBackground.png');
  ADUtils.preloadImageAsset('bar/UINavigationBarBlackTranslucentBackground.png');
  ADUtils.preloadImageAsset('bar/toolbar_glow.png');
};

window.addEventListener('load', ADToolbar.init, false);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  The standard {@link ADProgressView} style. This is the default.
 *  @constant
 *  @type String
 */
const ADProgressViewStyleDefault = 'default';
/**
 *  The style of {@link ADProgressView} that is used in a toolbar.
 *  @constant
 *  @type String
 */
const ADProgressViewStyleBar = 'bar';

/**
 *  Height of a {@link ADProgressView} with the default style.
 *  @constant
 *  @type int
 */
const ADProgressViewDefaultHeight = 9;
/**
 *  Height of a {@link ADProgressView} with the bar style.
 *  @constant
 *  @type int
 */
const ADProgressViewBarHeight = 11;

/* ==================== ADProgressView ==================== */

ADProgressView.inherits = ADView;
ADProgressView.synthetizes = ['progress', 'style', 'width'];

/**
 *  @class
 *
 *  <p>You use the ADProgressView class to depict the progress of a task over time. An example of a progress bar
 *  is the one shown at the bottom of the Mail application when it's downloading messages.</p>
 *
 *  <p>The ADProgressView class provides properties for managing the style of the progress bar and for getting
 *  and setting values that are pinned to the progress of a task.</p>
 *
 *  @extends ADView
 *  @since AdLib 0.1
 *
 *  @param {String} style A constant that specifies the style of the object to be created.
 */
function ADProgressView (style) {
  this.callSuper();
  // synthesize read-write properties
  this._progress = 0;
  this._style = '';
  // set up base type
  this.style = (style != null) ? style : ADProgressViewStyleDefault;
};

/* ==================== ADView APIs ==================== */

ADProgressView.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-progress-view');
  //
  this.bar = this.layer.appendChild(document.createElement('div'));
};

ADProgressView.prototype.setSize = function (aSize) {
  aSize.height = (this._style == ADProgressViewStyleDefault) ? ADProgressViewDefaultHeight : ADProgressViewBarHeight;
  this.callSuper(aSize);
  //
  this.updateBar();
};

ADProgressView.prototype.willMoveToSuperview = function (superview) {
  // force bar style if we're moving to a ADToolbar
  if (superview instanceof ADToolbar) {
    this.style = ADProgressViewStyleBar;
  }
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADProgressView.prototype
 *  @property {float} progress The current progress is represented by a floating-point value between 0.0 and 1.0,
 *  inclusive, where 1.0 indicates the completion of the task. The default value is 0.0. Values less than 0.0 and
 *  greater than 1.0 are pinned to those limits.
 *  @property {String} style The value of this property is a constant that specifies the style of the progress view.
 *  The default style is ADProgressViewStyleDefault.
 *  @property {int} width The width in pixels for the progress view.
 */

ADProgressView.prototype.getWidth = function () {
  return this.size.width;
};

ADProgressView.prototype.setWidth = function (width) {
  this.size = new ADSize(width, 0);
};

ADProgressView.prototype.setProgress = function (progress) {
  // clip the value between 0 and 1
  progress = Math.max(Math.min(progress, 1), 0);
  // do nothing if we already have this value
  if (progress == this._progress) {
    return;
  }
  // save to ivar
  this._progress = progress;
  // update the bar
  this.updateBar();
};

ADProgressView.prototype.setStyle = function (style) {
  // do nothing if we already have this style
  if (style == this._style) {
    return;
  }
  //
  this.layer.removeClassName(this._style);
  this.layer.addClassName(style);
  this._style = style;
  // update bounds
  this.size = new ADSize(this.size.width, 0);
};

/* ==================== Drawing ==================== */

ADProgressView.prototype.updateBar = function () {
  this.bar.style.width = ADUtils.px(this.size.width * this._progress);
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADProgressView);

ADProgressView.init = function () {
  ADUtils.preloadImageAsset('progress/default_outer.png');
  ADUtils.preloadImageAsset('progress/default_inner.png');
  ADUtils.preloadImageAsset('progress/bar_outer.png');
  ADUtils.preloadImageAsset('progress/bar_inner.png');
};

window.addEventListener('load', ADProgressView.init, false);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

// Events constants
/**
 *  A touch-down event in the {@link ADControl}.
 *  @constant
 *  @type String
 */
const ADControlTouchDownEvent         = 'controlTouchDown';
//const ADControlTouchDownRepeatEvent   = 'controlTouchDownRepeat';
/**
 *  An event where a finger is dragged inside the bounds of the {@link ADControl}.
 *  @constant
 *  @type String
 */
const ADControlTouchDragInsideEvent   = 'controlTouchDragInside';
/**
 *  An event where a finger is dragged just outside the bounds of the {@link ADControl}.
 *  @constant
 *  @type String
 */
const ADControlTouchDragOutsideEvent  = 'controlTouchDragOutside';
 /**
  *  An event where a finger is dragged into the bounds of the {@link ADControl}.
  *  @constant
  *  @type String
  */
const ADControlTouchDragEnterEvent    = 'controlTouchDragEnter';
/**
 *  An event where a finger is dragged from within a {@link ADControl} to outside its bounds.
 *  @constant
 *  @type String
 */
const ADControlTouchDragExitEvent     = 'controlTouchDragExit';
/**
 *  A touch-up event in the control where the finger is inside the bounds of the {@link ADControl}.
 *  @constant
 *  @type String
 */
const ADControlTouchUpInsideEvent     = 'controlTouchUpInside';
/**
 *  A touch-up event in the control where the finger is outside the bounds of the {@link ADControl}.
 *  @constant
 *  @type String
 */
const ADControlTouchUpOutsideEvent    = 'controlTouchUpOutside';
/**
 *  A system event canceling the current touches for the {@link ADControl}.
 *  @constant
 *  @type String
 */
const ADControlTouchCancelEvent       = 'controlTouchCancel';
/**
 *  A touch dragging or otherwise manipulating a {@link ADControl}, causing it to emit a series of different values.
 *  @constant
 *  @type String
 */
const ADControlValueChangeEvent       = 'controlValueChange';

/**
 *  A higher-level touch event that notifies that the touch state of the {@link ADControl} has changed.
 *  @constant
 *  @type String
 */
const ADControlTouchStateChangeEvent = 'controlTouchStateChange';

// State constants
/**
 *  The normal, or default state of a {@link ADControl}  that is, enabled but neither selected or highlighted.
 *  @constant
 *  @type String
 */
const ADControlStateNormal          = 0;
const ADControlStateNormalCSS       = 'normal';
/**
 *  Highlighted state of a {@link ADControl}. A control enters this state when a touch enters and exits during tracking
 *  and and when there is a touch up. You can retrieve and set this value through the {@link ADControl#highlighted} property.
 *  @constant
 *  @type String
 */
const ADControlStateHighlighted     = 1 << 0;
const ADControlStateHighlightedCSS  = 'highlighted';
/**
 *  Disabled state of a {@link ADControl}. This state indicates that the control is currently disabled. You can retrieve
 *  and set this value through the {@link ADControl#enabled} property.
 *  @constant
 *  @type String
 */
const ADControlStateDisabled        = 1 << 1;
const ADControlStateDisabledCSS     = 'disabled';
/**
 *  Selected state of a {@link ADControl}. For many controls, this state has no effect on behavior or appearance. You can
 *  retrieve and set this value through the {@link ADControl#selected} property.
 *  @constant
 *  @type String
 */
const ADControlStateSelected        = 1 << 2;
const ADControlStateSelectedCSS     = 'selected';
/**
 *  @class
 *  @name ADControlEvent
 *  @extends Event
 *  @since AdLib 0.1
 */

/**
 *  @name ADControlEvent.prototype
 *  @property {ADControl} control The control that triggered this event.
 */

/* ==================== ADControl ==================== */

ADControl.inherits = ADView;
ADControl.includes = [ADEventTarget];
ADControl.synthetizes = ['state', 'enabled', 'selected', 'highlighted', 'touchLayer'];

/**
 *  @class
 *
 *  <p>The ADControl class is the base class for all the AdLib controls. A ADControl is a view and as such
 *  can be positioned, sized and transformed exactly as any other {@link ADView} subclass can. Controls are
 *  interactive by nature and acting upon touch events is as easy as providing a {@link #touchLayer} during the
 *  {@link ADView#createLayer} implementation. From that point, touches can be tracked by overriding the
 *  {@link #touchesBegan}, {@link #touchesMoved}, {@link #touchesEnded} and {@link #touchesCancelled} methods.</p>
 *
 *  <p>Instead of using delegation protocols to provide notifications of state changes or interactivity, a control
 *  leverages the DOM Events mechanisms and dispatch their own set of control events. A {@link ADControlEvent} is simply
 *  a DOM Event with a custom type and potentially custom information. All control events triggered by a control
 *  have a {@link ADControlEvent#control} property which contains a pointer to the control which triggered the event.</p>
 *
 *  <p>In order to further customize the appearance of a control, the user can leverage CSS style rules to match against
 *  the current state of the button. Indeed, as a control state changes, CSS classes are added or removed to the {@link #layer}
 *  DOM element. The properties are as follows:</p>
 *
 *  <ul>
 *    <li>{@link ADControlStateNormal}: <code>normal</code> (or no specific class)</li>
 *    <li>{@link ADControlStateHighlighted}: <code>highlighted</code></li>
 *    <li>{@link ADControlStateDisabled}: <code>disabled</code></li>
 *    <li>{@link ADControlStateSelected}: <code>selected</code></li>
 *  </ul>
 *
 *  @extends ADView
 *  @since AdLib 0.1
 */
function ADControl () {
  /**
   *  The controls's tag, an application-supplied integer that you can use to identify controls in your application.
   *  @type int 
   */
  this.tag = 0;
  // synthesize read-write properties
  this._enabled = true;
  this._selected = false;
  this._highlighted = false;
  this._touchLayer = null;
  //
  this.callSuper();
  // read-only properties
  /**
   *  Indicates whether the control is currently tracking touches.
   *  @type bool 
   */
  this.tracking = false;
  /**
   *  Indicates whether a touch is inside the control's bounds.
   *  @type bool 
   */
  this.touchInside = false;
  // add a way to refer back to the ADControl instance from the layer
  this.layer._control = this;
  // set up the event target so that the ADEventTarget method just work
  this.eventTarget = this.layer;
  //
  this.tracksTouchesOnceTouchesBegan = true;
  // make sure we track events on the .touchLayer
  this.layer.removeEventListener(ADStartEvent, this, false);
  this.touchLayer.addEventListener(ADStartEvent, this, false);
};

/* ==================== ADView APIs ==================== */

ADControl.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-control');
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADControl.prototype
 *  @property {int} state One or more <code>ADControlState</code> bit-mask constants that specify the state of the ADControl object.
 *  Note that the control can be in more than one state, for example, both disabled and selected (<code>ADControlStateDisabled | ADControlStateSelected</code>).
 *  @property {bool} enabled Determines the control's enabled state.
 *  @property {bool} selected Determines the control's selected state.
 *  @property {bool} highlighted Determines whether the control is highlighted.
 *  @property {HTMLElement} touchLayer The element on which the <code>touchstart</code> event will be
 *  registered if event handling is desired. This has to be provided by subclasses in the
 *  {@link ADView#createLayer} method. If not provided, the touchLayer is simply the {@link ADView#layer}.
 */

ADControl.prototype.getState = function () {
  return (  ADControlStateNormal |
            (this._highlighted ? ADControlStateHighlighted : 0) |
            (this._enabled ? 0 : ADControlStateDisabled) |
            (this._selected ? ADControlStateSelected : 0)
         );
};

ADControl.prototype.setEnabled = function (enabled) {
  // do nothing if we're not changing state
  if (enabled == this._enabled) {
    return;
  }
  this.layer[enabled ? 'removeClassName' : 'addClassName'](ADControlStateDisabledCSS);
  this._enabled = enabled;
  this.tracksTouchesOnceTouchesBegan = enabled;
  //
  this.notifyPropertyChange('state');
};

ADControl.prototype.setSelected = function (selected) {
  // do nothing if we're not changing state
  if (selected == this._selected) {
    return;
  }
  this.layer[selected ? 'addClassName' : 'removeClassName'](ADControlStateSelectedCSS);
  this._selected = selected;
  //
  this.notifyPropertyChange('state');
};

ADControl.prototype.setHighlighted = function (highlighted) {
  // do nothing if we're not changing state
  if (highlighted == this._highlighted) {
    return;
  }
  this.layer[highlighted ? 'addClassName' : 'removeClassName'](ADControlStateHighlightedCSS);
  this._highlighted = highlighted;
  //
  this.notifyPropertyChange('state');
};

/* ==================== DEFAULT EVENT HANDLING ==================== */

ADControl.prototype.getTouchLayer = function () {
  return (this._touchLayer != null) ? this._touchLayer : this.layer;
};

ADControl.prototype.touchesBegan = function (event) {
  this.callSuper(event);
  // do nothing if the control is disabled
  if (!this._enabled) {
    return;
  }
  // and prevent the event from bubbling up
  event.stopPropagation();
  // and make sure we kill the event default behavior as well
  // as controls should capture all touches from then on
  event.preventDefault();
  // track that the control is touched
  this.touchInside = true;
  this.highlighted = true;
  // throw event for start of interaction
  this.dispatchEvent(this.createUIEvent(ADControlTouchDownEvent, event));
  // throw event to say we have touch inside too
  this.dispatchEvent(this.createEvent(ADControlTouchStateChangeEvent));
  // archive this event
  this.lastProcessedEvent = event;
};

ADControl.prototype.touchesMoved = function (event) {
  this.callSuper(event);
  // we now are tracking events
  this.tracking = true;
  // now figure out if we're in or outside the element
  var inside = this.pointInside(ADPoint.fromEventInElement(event, this.layer));
  // by default, throw a regular drag event
  var drag_type = inside ? ADControlTouchDragInsideEvent : ADControlTouchDragOutsideEvent;
  // but check if we need to send an enter or exit drag event instead
  if (inside != this.touchInside) {
    this.touchInside = inside;
    this.highlighted = inside;
    drag_type = inside ? ADControlTouchDragEnterEvent : ADControlTouchDragExitEvent;
    // throw event to say we have touch state changed too
    this.dispatchEvent(this.createEvent(ADControlTouchStateChangeEvent));
  }
  // dispatch the event
  this.dispatchEvent(this.createUIEvent(drag_type, event));
  // archive this event
  this.lastProcessedEvent = event;
};

ADControl.prototype.touchesEnded = function (event) {
  this.callSuper(event);
  // we're no longer dragging
  this.tracking = false;
  // no longer highlighted
  this.highlighted = false;
  // throw an event about the touch up
  var up_type = this.touchInside ? ADControlTouchUpInsideEvent : ADControlTouchUpOutsideEvent;
  this.dispatchEvent(this.createUIEvent(up_type, this.lastProcessedEvent));
  // track that the control is not touched anymore
  this.touchInside = false;
  // throw event to say we have touch state changed too
  this.dispatchEvent(this.createEvent(ADControlTouchStateChangeEvent));
};

ADControl.prototype.touchesCancelled = function (event) {
  this.callSuper(event);
  // throw event to notify control lost touches
  // XXX: need to create with mouse event as well for platform independence
  this.dispatchEvent(this.createUIEvent(ADControlTouchCancelEvent, event));
};

/**
 *  Creates a control event of a given type to be dispatched by the control with the {@link #dispatchEvent} method.
 *
 *  @param {String} eventType The event type as a string.
 *  @returns {Event} The event.
 */
ADControl.prototype.createEvent = function (eventType) {
  // create a new type of touch event
  var event = document.createEvent('Event');
  // init it with our custom type
  event.initEvent(eventType, true, false);
  // add a custom field pointing to our control instance
  event.control = this;
  // return our custom event
  return event;
};

/**
 *  Creates a control event, based on an existing touch event, of a given type to be dispatched by the control with the {@link #dispatchEvent} method.
 *
 *  @param {String} eventType The event type as a string.
 *  @param {TouchEvent} baseTouchEvent The base touch event that is being extended.
 *  @returns {TouchEvent} The event.
 */
ADControl.prototype.createUIEvent = function (eventType, event) {
  var ui_event = ADUtils.createUIEvent(eventType, event);
  ui_event.control = this;
  return ui_event;
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADControl);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

const ADBarItemDisabledCSS = 'disabled';

/* ==================== ADBarItem ==================== */

ADBarItem.inherits = ADObject;
ADBarItem.synthetizes = ['enabled', 'image', 'imageOffset', 'title'];

function ADBarItem () {
  this.callSuper();
  // synthesize read-write properties
  this._enabled = true;
  this._image = null;
  this._imageOffset = null;
  this._title = '';
  // regular public properties
  this.tag = 0;
  // private properties
  this.view = null;
};

/**
 *  @name ADBarButtonItem.prototype
 *  @property {bool} enabled Whether the item is enabled.
 *  @property {String} title The button's title.
 *  @property {ADImage} image The button's icon.
 *  @property {ADPoint} imageOffset The button's icon's offset from the top left origin of the button.
 */

ADClass(ADBarItem);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  Height of a {@link ADBarButton}.
 *  @constant
 *  @type int
 */
const ADBarButtonHeight = 30;

// Default x offset for the pointy buttons.
const ADBarButtonPointyXOffset = 3;

/* ==================== ADBarButton ==================== */

ADBarButton.inherits = ADControl;
ADBarButton.synthetizes = ['maxWidth', 'width', 'type', 'style', 'image', 'imageOffset', 'title'];

/**
 *  @class
 *
 *  <p>The ADBarButton class is a {@link ADControl} subclass that implements a button on the touch screen. A button
 *  intercepts touch events and sends a <code>ADControlTouchUpInsideEvent</code> when it's tapped. Methods for
 *  registering an event listener are inherited from {@link ADControl}. This class provides methods for setting the
 *  title, image, and other appearance properties of a button.</p>
 *
 *  @extends ADControl
 *  @since AdLib 0.1
 *  @private
 *
 *  @param {String} type The optional button type, <code>ADBarButtonItemTypeSquare</code> by default.
 */
function ADBarButton (type) {
  this.callSuper();
  // synthesize read-write properties
  this._maxWidth = 0;
  this._width = 0;
  this._type = '';
  this._style = '';
  this._image = null;
  this._imageOffset = null;
  this._title = '';
  // set up base type
  this.type = (type != null) ? type : ADBarButtonItemTypeSquare;
  this.style = ADBarButtonItemStyleDefault;
  // tracks whether we have a user-set image offset
  this.usesAutomaticImageOffset = true;
};

/* ==================== ADView APIs ==================== */

ADBarButton.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-bar-button');
  this.layer.setAttribute('role', 'button');
  // add the background
  this.background = this.layer.appendChild(document.createElement('div'));
  // add the placeholder for the image
  this.icon = this.layer.appendChild(document.createElement('img'));
};

ADBarButton.prototype.layerWasInsertedIntoDocument = function () {
  this.callSuper();
  // if we have a title and a 0-width, attempt to autosize
  if (this._title != '' && this._width == 0) {
    this.autoSizeTitle();
  }
};

ADBarButton.prototype.getSize = function () {
  var width = (this._maxWidth == 0) ? this._size.width : Math.min(this._maxWidth, this._size.width);
  return new ADSize(width, this._size.height);
};

ADBarButton.prototype.setSize = function (aSize) {
  aSize.height = ADBarButtonHeight;
  this.callSuper(aSize);
};

ADBarButton.prototype.touchesBegan = function (event) {
  // do not process events for space buttons
  if (this.type == ADBarButtonItemTypeFlexibleSpace || this.type == ADBarButtonItemTypeFixedSpace) {
    event.preventDefault();
    return;
  }
  this.callSuper(event);
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADBarButton.prototype
 *  @property {String} type The button type, <code>ADBarButtonItemTypeSquare</code> by default.
 *  @property {String} style The button style that specifies its appearance. The default value is <code>ADBarButtonItemStyleDefault</code>.
 *  @property {int} maxWidth The maximum width, in pixels, ensuring that the title is trimmed in case it would overflow that maximum
 *  size. The default value, <code>0</code> means that there is no maximum width and that the {@link #width} value is always used to render the button.
 *  @property {int} width The button width, in pixels, which can be set to <code>0</code> (the default) to let the button auto-size
 *  itself based on its textual or image content. The {@link #maxWidth} property can be set to set a maximum size for the button, even when auto-sizing is on.
 *  @property {String} title The button's title.
 *  @property {ADImage} image The button's icon.
 *  @property {ADPoint} imageOffset The button's icon's offset from the top left origin of the button.
 */

ADBarButton.prototype.setMaxWidth = function (width) {
  this.background.style.maxWidth = (width == 0) ? 'inherit' : ADUtils.px(width);
  this._maxWidth = width;
  // make sure we update the button's size if we have auto-sizing on
  if (this._width == 0) {
    this.autoSizeTitle();
  }
};

ADBarButton.prototype.setWidth = function (width) {
  // If this property value is positive, the width of the combined image and title are fixed.
  // If the value is 0.0 or negative, the item sets the width of the combined image and title to fit.
  // This property is ignored if the style uses radio mode. The default value is 0.0.  
  if (width == 0) {
    this.autoSizeTitle();
    this.autoSizeImage();
    this.positionImage();
  }
  else {
    this._width = width;
    this.size = new ADSize(width, ADBarButtonHeight);
    this.positionImage();
  }
};

ADBarButton.prototype.autoSizeTitle = function () {
  if (this._title != '' && this.layerIsInDocument) {
    this.layer.style.width = 'auto';
    var new_width = parseInt(window.getComputedStyle(this.layer).width);
    this.size = new ADSize(new_width, this._size.height);
  }
};

ADBarButton.prototype.autoSizeImage = function () {
  if (this._image !== null) {
    var x_offsets = this.getBorderXOffsets();
    var new_width = this._image.width + x_offsets.left + x_offsets.right;
    this.size = new ADSize(new_width, this._size.height);
  }
};

ADBarButton.prototype.setTitle = function (title) {
  this.background.innerText = title;
  this._title = title;
  // make sure we update the button's size if we have auto-sizing on
  if (this._width == 0) {
    this.autoSizeTitle();
  }
};

ADBarButton.prototype.setImage = function (image) {
  this.icon.src = image.url;
  this._image = image;
  // if image is already loaded, lay it out right away
  if (image.loaded) {
    this.positionImage();
  }
  // otherwise, monitor loaded state
  else {
    image.addPropertyObserver('loaded', this);
  }
};

ADBarButton.prototype.getImageOffset = function () {
  return new ADPoint(parseInt(this.icon.style.left), parseInt(this.icon.style.top));
};

ADBarButton.prototype.setImageOffset = function (point) {
  this.icon.style.left = ADUtils.px(point.x);
  this.icon.style.top = ADUtils.px(point.y);
  this.usesAutomaticImageOffset = false;
};

ADBarButton.prototype.positionImage = function () {
  // do nothing if we don't have an image (yet) or we have an explicit image offset
  if (this._image === null || !this.usesAutomaticImageOffset) {
    return;
  }
  // see if we need to shrink the image
  var image_height = Math.min(this._image.height, this.size.height);
  // make sure we take into account border-width
  this.icon.height = image_height;
  var offset = new ADPoint(
    (this._size.width - this.icon.width) / 2,
    (ADBarButtonHeight - this.icon.height) / 2
  );
  if (this._type == ADBarButtonItemTypeBack) {
    offset.x += ADBarButtonPointyXOffset;
  }
  if (this._type == ADBarButtonItemTypeForward) {
    offset.x -= ADBarButtonPointyXOffset;
  }
  this.imageOffset = offset;
  this.usesAutomaticImageOffset = true;
};

ADBarButton.prototype.setType = function (type) {
  this.layer.removeClassName(this._type);
  this.layer.addClassName(type);
  this._type = type;
};

ADBarButton.prototype.setStyle = function (style) {
  this.layer.removeClassName(this._style);
  this.layer.addClassName(style);
  this._style = style;
};

/* ==================== Utility ==================== */

/**
 *  Returns a pair of offsets for left and right based on border-width
 */
ADBarButton.prototype.getBorderXOffsets = function () {
  var style = window.getComputedStyle(this.background);
  var left = style.getPropertyCSSValue('border-left-width');
  var right = style.getPropertyCSSValue('border-right-width');
  return {
    left: (left !== null) ? left.getFloatValue(CSSPrimitiveValue.CSS_PX) : 0,
    right: (right !== null) ? right.getFloatValue(CSSPrimitiveValue.CSS_PX) : 0
  };
};

/* ==================== Event Handling ==================== */

ADBarButton.prototype.handlePropertyChange = function (object, propertyName) {
  // auto-size if we have 0 width
  if (this._width == 0) {
    this.autoSizeImage();
  }
  this.positionImage();
};

/* ==================== Event Target APIs ==================== */

ADBarButton.prototype.dispatchEvent = function (event) {
  event.barButtonItem = this.barButtonItem;
  this.callSuper(event);
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADBarButton);

ADBarButton.init = function () {
  var images = [
    'UINavigationBarDoneButtonPressed',
    'UINavigationBarDoneButton',
    'UINavigationBarDefaultForwardPressed',
    'UINavigationBarDefaultForward',
    'UINavigationBarDefaultButtonPressed',
    'UINavigationBarDefaultButton',
    'UINavigationBarDefaultBackPressed',
    'UINavigationBarDefaultBack',
    'UINavigationBarBlackTranslucentForward',
    'UINavigationBarBlackTranslucentButton',
    'UINavigationBarBlackTranslucentBack',
    'UINavigationBarBlackOpaqueForward',
    'UINavigationBarBlackOpaqueButton',
    'UINavigationBarBlackOpaqueBack',
    'UINavigationBarBlackForwardPressed',
    'UINavigationBarBlackButtonPressed',
    'UINavigationBarBlackBackPressed'
  ];
  for (var i = 0; i < images.length; i++) {
    ADUtils.preloadImageAsset('button/' + images[i] + '.png');
  }
};

window.addEventListener('load', ADBarButton.init, false);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  Plain type for a {@link ADBarButton}, no background or borders, just a plain text label.
 *  @constant
 *  @type String
 */
const ADBarButtonItemTypePlain = 'plain';
/**
 *  Square, default type for a {@link ADBarButton}.
 *  @constant
 *  @type String
 */
const ADBarButtonItemTypeSquare = 'square';
/**
 *  Back type for a {@link ADBarButton}.
 *  @constant
 *  @type String
 */
const ADBarButtonItemTypeBack = 'back';
/**
 *  Forward type for a {@link ADBarButton}.
 *  @constant
 *  @type String
 */
const ADBarButtonItemTypeForward = 'forward';
/**
 *  Flexible space type for a {@link ADBarButton}.
 *  @constant
 *  @type String
 */
const ADBarButtonItemTypeFlexibleSpace = 'flexible-space';
/**
 *  Fixed space type for a {@link ADBarButton}.
 *  @constant
 *  @type String
 */
const ADBarButtonItemTypeFixedSpace = 'fixed-space';

/**
 *  Black style for a {@link ADBarButton}.
 *  @constant
 *  @type String
 */
const ADBarButtonItemStyleBlack = 'black';
/**
 *  Default blue style for a {@link ADBarButton}.
 *  @constant
 *  @type String
 */
const ADBarButtonItemStyleDefault = 'default';
/**
 *  Light blue style for a {@link ADBarButton}.
 *  @constant
 *  @type String
 */
const ADBarButtonItemStyleLightBlue = 'lightblue';

/* ==================== ADBarButtonItem ==================== */

ADBarButtonItem.inherits = ADBarItem;
ADBarButtonItem.includes = [ADEventTarget];
ADBarButtonItem.synthetizes = ['maxWidth', 'customView', 'width', 'type', 'style'];

/**
 *  @class
 *
 *  <p>The ADBarButtonItem class is a {@link ADControl} subclass that implements a button on the touch screen. A button
 *  intercepts touch events and sends a <code>ADControlTouchUpInsideEvent</code> when it's tapped. Methods for
 *  registering an event listener are inherited from {@link ADControl}. This class provides methods for setting the
 *  title, image, and other appearance properties of a button.</p>
 *
 *  @extends ADControl
 *  @since AdLib 0.1
 *
 *  @param {String} type The optional button type, <code>ADBarButtonItemTypeSquare</code> by default.
 */
function ADBarButtonItem (type) {
  this.callSuper();
  // set up the default view
  this.view = new ADBarButton(type);
  this.view.isCustomView = false;
  this.view.barButtonItem = this;
  // set up the event target
  this.eventTarget = this.view;
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADBarButtonItem.prototype
 *  @property {String} type The button type, <code>ADBarButtonItemTypeSquare</code> by default.
 *  @property {String} style The button style that specifies its appearance. The default value is <code>ADBarButtonItemStyleDefault</code>.
 *  @property {int} maxWidth The maximum width, in pixels, ensuring that the title is trimmed in case it would overflow that maximum
 *  size. The default value, <code>0</code> means that there is no maximum width and that the {@link #width} value is always used to render the button.
 *  @property {int} width The button width, in pixels, which can be set to <code>0</code> (the default) to let the button auto-size
 *  itself based on its textual or image content. The {@link #maxWidth} property can be set to set a maximum size for the button, even when auto-sizing is on.
 *  @property {String} title The button's title.
 *  @property {ADImage} image The button's icon.
 *  @property {ADPoint} imageOffset The button's icon's offset from the top left origin of the button.
 *  @property {ADView} customView A custom view representing the item.
 */

ADBarButtonItem.prototype.getEnabled = function () {
  return (!this.view.isCustomView) ? this.view.enabled : this.view.userInteractionEnabled;
};

ADBarButtonItem.prototype.setEnabled = function (enabled) {
  if (this.view.isCustomView) {
    this.view.userInteractionEnabled = enabled;
  }
  else {
    this.view.enabled = enabled;
  }
};

ADBarButtonItem.prototype.getMaxWidth = function () {
  return (!this.view.isCustomView) ? this.view.maxWidth : 0;
};

ADBarButtonItem.prototype.setMaxWidth = function (width) {
  if (!this.view.isCustomView) {
    this.view.maxWidth = width;
  }
};

ADBarButtonItem.prototype.getWidth = function () {
  return (!this.view.isCustomView) ? this.view.width : this.view.size.width;
};

ADBarButtonItem.prototype.setWidth = function (width) {
  if (this.view.isCustomView) {
    this.view.size = new ADSize(width, this.view.size.height);
  }
  else {
    this.view.width = width;
  }
};

ADBarButtonItem.prototype.getTitle = function () {
  return (!this.view.isCustomView) ? this.view.title : '';
};

ADBarButtonItem.prototype.setTitle = function (title) {
  if (!this.view.isCustomView) {
    this.view.title = title;
  }
};

ADBarButtonItem.prototype.getImage = function () {
  return (!this.view.isCustomView) ? this.view.image : null;
};

ADBarButtonItem.prototype.setImage = function (image) {
  if (!this.view.isCustomView) {
    this.view.image = image;
  }
};

ADBarButtonItem.prototype.getImageOffset = function () {
  return (!this.view.isCustomView) ? this.view.imageOffset : null;
};

ADBarButtonItem.prototype.setImageOffset = function (point) {
  if (!this.view.isCustomView) {
    this.view.imageOffset = point;
  }
};

ADBarButtonItem.prototype.getType = function () {
  return (!this.view.isCustomView) ? this.view.type : '';
};

ADBarButtonItem.prototype.setType = function (type) {
  if (!this.view.isCustomView) {
    this.view.type = type;
  }
};

ADBarButtonItem.prototype.getStyle = function (style) {
  return (!this.view.isCustomView) ? this.view.style : '';
};

ADBarButtonItem.prototype.setStyle = function (style) {
  if (!this.view.isCustomView) {
    this.view.style = style;
  }
};

ADBarButtonItem.prototype.getCustomView = function () {
  return (this.view.isCustomView) ? this.view : null;
};

ADBarButtonItem.prototype.setCustomView = function (customView) {
  // nothing to do if it's actually the same view
  if (customView === this.view) {
    return;
  }
  // can the previous view if we had one and it wasn't custom
  if (!this.view.isCustomView) {
    this.view = undefined;
  }
  // track that the view is custom
  customView.isCustomView = true;
  // and its bar button item
  customView.barButtonItem = this;
  // and remember it
  this.view = customView;
};

ADClass(ADBarButtonItem);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

const ADSliderThumbWidth = 23;
/**
 *  The constant height of a {@link ADSlider}.
 *  @constant
 *  @type int
 */
const ADSliderHeight = 23;

const ADSliderTouchedCSS = 'touched';

/* ==================== ADSlider ==================== */

ADSlider.inherits = ADControl;
ADSlider.synthetizes = ['minimumValue', 'maximumValue', 'value'];

/**
 *  @class
 *
 *  <p>A ADSlider object is a visual control used to select a single value from a continuous range of values. Sliders
 *  are always displayed as horizontal bars. An indicator, or thumb, notes the current value of the slider and can be
 *  moved by the user to change the setting.</p>
 *
 *  <p>When the user interacts with the slider control, the control sends the <code>ADControlValueChangeEvent</code>
 *  event for handling by the delegate as its {@link #value} changes, if the {@link #continuous} property is set to
 *  <code>true</code>. Otherwise, the value change event is only triggered once the user has released the thumb.</p>
 *
 *  @extends ADControl
 *  @since AdLib 0.1
 */
function ADSlider () {
  /**
   *  Indicates whether changes in the sliders value generate continuous update events. If <code>true</code>, the slider
   *  triggers update events continuously. If <code>false</code>, the slider only sends an action when the user releases
   *  the slider's thumb control to set the final value.</code>
   *  @type bool 
   */
  this.continuous = true;
  // synthesize read-write properties
  this._minimumValue = 0;
  this._maximumValue = 1;
  this._value = 0;
  this.callSuper();
};

/* ==================== ADView APIs ==================== */

ADSlider.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-slider');
  // gutter fill
  this.fill = this.layer.appendChild(document.createElement('div'));
  this.fill.className = 'ad-slider-fill';
  this.leftFill = this.fill.appendChild(document.createElement('div'));
  this.middleFill = this.fill.appendChild(document.createElement('div'));
  this.rightFill = this.fill.appendChild(document.createElement('div'));
  // create the thumb
  this.thumb = this.layer.appendChild(document.createElement('div'));
  this.thumb.className = 'ad-slider-thumb';
  // make the thumb the touch layer
  this.touchLayer = this.thumb;
};

ADSlider.prototype.setSize = function (aSize) {
  aSize.height = ADSliderHeight;
  this.callSuper(aSize);
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADSlider.prototype
 *  @property {float} minimumValue The slider's minimum value. If you change the value of this property,
 *  and the current value of the slider is below the new minimum, the current value is adjusted to match
 *  the new minimum value automatically.
 *  @property {float} maximumValue The slider's maximum value. If you change the value of this property,
 *  and the current value of the slider is above the new maximum, the current value is adjusted to match
 *  the new maximum value automatically.
 *  @property {float} value The slider's current value. Setting this property causes the slider to redraw
 *  itself using the new value. To render an animated transition from the current value to the new value,
 *  you should use the {@link setValueWithAnimation} method instead. If you try to set a value that is below
 *  the minimum or above the maximum value, the minimum or maximum value is set instead. The default value
 *  of this property is <code>0</code>.
 */

ADSlider.prototype.setMinimumValue = function (minimumValue) {
  // do nothing if we're trying to set minimum equal to or larger than maximum
  if (minimumValue >= this.maximumValue) {
    return;
  }
  // update ivar
  this._minimumValue = minimumValue;
  // If you change the value of this property, and the current value of the receiver is below
  // the new minimum, the current value is adjusted to match the new minimum value automatically.
  if (this.value < minimumValue) {
    this.value = minimumValue;
  }
  else {
    this.updateThumb();
  }
};

ADSlider.prototype.setMaximumValue = function (maximumValue) {
  // do nothing if we're trying to set maximum equal to or smaller than minimum
  if (maximumValue <= this.minimumValue) {
    return;
  }
  // update ivar
  this._maximumValue = maximumValue;
  // If you change the value of this property, and the current value of the receiver is above
  // the new maximum, the current value is adjusted to match the new maximum value automatically.
  if (this.value > maximumValue) {
    this.value = maximumValue;
  }
  else {
    this.updateThumb();
  }
};

ADSlider.prototype.setValue = function (value) {
  this.setValueWithAnimation(value, false);
};

/**
 *  Sets the slider's current value, allowing you to animate the change visually. If you try to set a value
 *  that is below the minimum or above the maximum value, the minimum or maximum value is set instead.
 *  The default value of this property is <code>0</code>.
 *
 *  @param {float} value The new value to assign to the {@link #value} property.
 *  @param {bool} animated Specify <code>true</code> to animate the change in value when the receiver is redrawn;
 *  otherwise, specify <code>false</code> to draw the slider with the new value only.
 */
ADSlider.prototype.setValueWithAnimation = function (value, animated) {
  // first, clip the value
  value = Math.max(Math.min(value, this.maximumValue), this.minimumValue);
  // do nothing if we already had this value
  if (value == this._value) {
    return;
  }
  // turn animations on or off
  var duration = animated ? '0.25s' : '0s';
  this.thumb.style.webkitTransitionDuration = duration;
  this.middleFill.style.webkitTransitionDuration = duration;
  this.rightFill.style.webkitTransitionDuration = duration;
  // record the value
  this._value = value;
  // update the thumb value
  this.updateThumb();
  // notify that we have a value change
  if (this.continuous) {
    this.notifyValueChanged();
  }
};

ADSlider.prototype.updateThumb = function (x) {
  // compute the location for the thumb
  var offset = (this.value - this.minimumValue) / (this.maximumValue - this.minimumValue);
  var x = offset * (this.size.width + 2 - ADSliderThumbWidth);
  // update gutter fill
  this.updateGutterWithOffset(offset);
  // update thumb position
  this.thumb.style.webkitTransform = ADUtils.t(x, 0);
};

ADSlider.prototype.updateGutterWithOffset = function (offset) {
  var width = offset * (this.size.width - 16);
  this.middleFill.style.webkitTransform = 'scale(' + (width + 1) + ', 1)';
  this.rightFill.style.webkitTransform = ADUtils.t(width, 0);
};

ADSlider.prototype.notifyValueChanged = function () {
  this.dispatchEvent(this.createEvent(ADControlValueChangeEvent));
};

/* ==================== EVENT HANDLING ==================== */

ADSlider.prototype.touchesBegan = function (event) {
  this.callSuper(event);
  //
  this.maxX = this.size.width - ADSliderThumbWidth;
  this.internalX = ADPoint.fromEventInElement(event, this.touchLayer).x;
  this.startValue = this.value;
  this.valueRange = this.maximumValue - this.minimumValue;
  //
  this.thumb.addClassName(ADSliderTouchedCSS);
};

ADSlider.prototype.touchesMoved = function (event) {
  this.callSuper(event);
  // get local interaction point
  var touch_x = ADPoint.fromEventInElement(event, this.layer).x - this.internalX;
  this.value = this.minimumValue + (touch_x / this.maxX) * this.valueRange;
};

ADSlider.prototype.touchesEnded = function (event) {
  this.callSuper(event);
  // notify of value change if we landed on a different value than the
  // start and we're not continuous
  if (!this.continuous && this.startValue != this.value) {
    this.notifyValueChanged();
  }
  //
  this.thumb.removeClassName(ADSliderTouchedCSS);
};

ADSlider.prototype.touchesCancelled = function (event) {
  this.callSuper(event);
  // reset the original value as interaction failed to complete
  this.setValue(this.startValue, true);
  // XXX: need to notify value changed here too
  this.thumb.removeClassName(ADSliderTouchedCSS);
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADSlider);

ADSlider.init = function () {
  ADUtils.preloadImageAsset('slider/thumb_touched.png');
};

window.addEventListener('load', ADSlider.init, false);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

const ADTabBarItemHeight = 44;

const ADTabBarItemCanvasHeight = 34;

/* ==================== ADTabBarItem ==================== */

ADTabBarItem.inherits = ADControl;
ADTabBarItem.synthetizes = ['title', 'image', 'badgeValue'];

/**
 *  @class
 *
 *  <p>The ADTabBarItem class implements an item on a {@link ADTabBar}. A tab bar operates strictly in radio mode,
 *  where one item is selected at a timetapping a tab bar item toggles the view above the tab bar. You can also
 *  specify a {@link #badgeValue} on the tab bar item for adding additional visual information  for example, the
 *  Phone application uses a badge on the item to show the number of new messages.</p>
 *
 *  <p>Use the {@link ADTabBarItem.initWithTitleImageAndTag} class method to create a custom item with the specified title,
 *  image and tag.</p>
 *
 *  @extends ADControl
 *  @since AdLib 0.1
 */
function ADTabBarItem (title, image, tag) {
  // synthesize read-write properties
  this._title = '';
  this._image = '';
  this._badgeValue = 0;
  //
  this.callSuper();
};

/* ==================== ADView APIs ==================== */

ADTabBarItem.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-tab-bar-item');
  //
  this.canvas = this.layer.appendChild(document.createElement('canvas'));
  this.label = this.layer.appendChild(document.createElement('div'));
  this.badge = this.layer.appendChild(document.createElement('div'));
  this.badgeContent = this.badge.appendChild(document.createElement('span'));
  //
  this.canvas.height = ADTabBarItemCanvasHeight;
};

ADTabBarItem.prototype.setSize = function (aSize) {
  this.callSuper(aSize);
  // update the max width for the badge value
  this.badgeContent.style.maxWidth = ADUtils.px(this.size.width - 12);
  // also update the badge position
  this.updateBadgePosition();
  // size the canvas
  this.canvas.width = this.size.width;
  // and redraw the image  
  this.drawImage();
};

ADTabBarItem.prototype.didMoveToSuperview = function () {
  this.callSuper();
  // now call the update once CSS styles have updated
  var _this = this;
  setTimeout(function () {
    _this.updateBadgePosition();
  }, 0)
};

/* ==================== ADControl APIs ==================== */

ADTabBarItem.prototype.setSelected = function (selected) {
  this.callSuper(selected);
  this.drawImage();
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADTabBarItem.prototype
 *  @property {String} title The item's title. If <code>null</code>, a title is not displayed.
 *  @property {ADImage} image The item's image. If <code>null</code>, an image is not displayed.
 *  The images displayed on the tab bar are derived from this image. If this image is too large
 *  to fit on the tab bar, it is scaled to fit. The size of a tab bar image is typically 30x30
 *  pixels. The alpha values in the source image are used to create the unselected and selected
 *  images  opaque values are ignored.
 *  @property {int} badgeValue Text that is displayed in the upper-right corner of the item with
 *  a surrounding red oval. When the value is <code>0</code>, its default, the badge is not displayed.
 */

ADTabBarItem.prototype.setTitle = function (title) {
  this.label.textContent = title;
  this._title = title;
};

ADTabBarItem.prototype.setImage = function (image) {
  this._image = image;
  // if image is already loaded, lay it out right away
  if (image.loaded) {
    this.drawImage();
  }
  // otherwise, monitor loaded state
  else {
    image.addPropertyObserver('loaded', this);
  }
};

ADTabBarItem.prototype.setBadgeValue = function (badgeValue) {
  this._badgeValue = badgeValue;
  if (badgeValue > 0) {
    this.badge.style.display = 'block';
    this.badgeContent.textContent = badgeValue;
    this.updateBadgePosition();
  }
  else {
    this.badge.style.display = 'none';
  }
};

ADTabBarItem.prototype.updateBadgePosition = function () {
  // do nothing if we got removed from the tree
  // or aren't in the tree yet
  if (this.superview == null) {
    return;
  }
  // check the badge width and make sure we lay it out towards the center,
  // or the right if it'd overflow in the center
  if (this.badge.offsetWidth < this.size.width / 2) {
    this.badge.style.left = ADUtils.px(Math.round(this.size.width / 2));
  }
  else {
    this.badge.style.right = ADUtils.px(-4);
  }
};

/* ==================== Image Drawing ==================== */

ADTabBarItem.prototype.drawImage = function () {
  // do nothing if we have no image
  if (this._image === null || !this._image.loaded) {
    return;
  }
  // get the image canvas
  var ctx = this.canvas.getContext('2d');
  // start clear
  ctx.clearRect(0, 0, this.size.width, this.size.height);
  // draw the icon
  var y = (ADTabBarItemCanvasHeight - this._image.height) / 2;
  ctx.globalCompositeOperation = 'source-over';
  ctx.drawImage(this._image.element, (this.size.width - this._image.width) / 2, y);
  if (this.selected) {
    // draw the mask
    ctx.globalCompositeOperation = 'source-in';
    ctx.drawImage(ADTabBarItem.prototype.maskImage.element, (this.size.width - ADTabBarItem.prototype.maskImage.width) / 2, y);
  }
};

ADTabBarItem.prototype.handlePropertyChange = function (object, propertyName) {
  this.drawImage();
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADTabBarItem);

ADTabBarItem.init = function () {
  ADTabBarItem.prototype.maskImage = new ADImage(ADUtils.assetsPath + 'bar/tab_bar_blue_gradient.png');
};

/**
 *  Creates and returns a new item using the specified properties.
 *
 *  @param {String} title The optional item title.
 *  @param {ADImage} image The optional item image.
 *  @param {int} tag The item's tag, an integer that you can use to identify bar item objects in your application.
 *  @returns {ADTabBarItem} Newly initialized item with the specified properties.
 */
ADTabBarItem.initWithTitleImageAndTag = function (title, image, tag) {
  var item = new ADTabBarItem();
  item.title = title;
  item.image = image;
  item.tag = tag;
  return item;
};

window.addEventListener('load', ADTabBarItem.init, false);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

const ADRatingControlStarWidth = 21;
/**
 *  The contant height for a {@link ADRatingControl}.
 *  @constant
 *  @type int
 */
const ADRatingControlStarHeight = 21;
const ADRatingControlMinValue = 0;
const ADRatingControlMaxValue = 5;
const ADRatingControlMinMargin = 10;

const ADRatingControlFilledCSS = 'filled';

/* ==================== ADRatingControl ==================== */

ADRatingControl.inherits = ADControl;
ADRatingControl.synthetizes = ['width', 'value'];

/**
 *  @class
 *
 *  <p>You use the ADRatingControl class to provide a star-based rating control so that the user can set a grade
 *  for an item on-screen. For instance, the App Store application lets users rate an application and displays
 *  the rating on each application page.</p>
 *
 *  <p>When a user flicks through or taps on a rating control, the control sends the <code>ADControlValueChangeEvent</code>
 *  event for handling by the delegate as its {@link #value} changes.</p>
 *
 *  @extends ADControl
 *  @since AdLib 0.1
 */
function ADRatingControl () {
  // set up default values for private variables
  this.margin = ADRatingControlMinMargin;
  this.numStars = ADRatingControlMaxValue - ADRatingControlMinValue;
  this.minWidth = this.numStars * ADRatingControlStarWidth + (this.numStars - 1) * ADRatingControlMinMargin;
  this.stars = [];
  // synthesize read-write properties
  this._value = 0;
  //
  this.callSuper();
  // ensure we already are using the best size to display
  this.size = new ADSize();
};

/* ==================== ADView APIs ==================== */

ADRatingControl.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-rating-control');
  // build the stars
  for (var i = 0; i < this.numStars; i++) {
    this.stars.push(this.layer.appendChild(document.createElement('div')));
  }
};

ADRatingControl.prototype.setSize = function (aSize) {
  aSize.width = Math.max(this.minWidth, aSize.width);
  aSize.height = ADRatingControlStarHeight;
  this.callSuper(aSize);
  // figure out margin between items
  this.margin = (this.size.width - (this.numStars * ADRatingControlStarWidth)) / (this.numStars - 1);
  for (var i = 0; i < this.numStars - 1; i++) {
    this.stars[i].style.marginRight = ADUtils.px(this.margin);
  }
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADRatingControl.prototype
 *  @property {int} width The control's width in pixels.
 *  @property {int} value The control's value ranging between <code>0</code> and <code>5</code>.
 */

ADRatingControl.prototype.getWidth = function () {
  return this.size.width;
}

ADRatingControl.prototype.setWidth = function (width) {
  this.size = new ADSize(width, 0);
};

ADRatingControl.prototype.setValue = function (value) {
  // first, clip the value
  value = Math.max(Math.min(value, ADRatingControlMaxValue), ADRatingControlMinValue);
  // do nothing if we already had this value
  if (value == this._value) {
    return;
  }
  // record the value
  this._value = value;
  // update the stars display
  for (var i = 0; i < this.numStars; i++) {
    this.stars[i].className = (i >= value - ADRatingControlMinValue) ? '' : ADRatingControlFilledCSS;
  }
  // notify that we have a value change
  this.dispatchEvent(this.createEvent(ADControlValueChangeEvent));
};

/* ==================== EVENT HANDLING ==================== */

ADRatingControl.prototype.touchesBegan = function (event) {
  this.callSuper(event);
  //
  this.startX = ADPoint.fromEventInElement(event, this.layer).x;
  this.startValue = this.value;
};

ADRatingControl.prototype.touchesMoved = function (event) {
  this.callSuper(event);
  //
  this.setValueFromX(ADPoint.fromEventInElement(event, this.layer).x);
};

ADRatingControl.prototype.touchesEnded = function (event) {
  if (!this.tracking) {
    this.setValueFromX(this.startX);
  }
  //
  this.callSuper(event);
};

ADRatingControl.prototype.touchesCancelled = function (event) {
  this.callSuper(event);
  // reset the original value as interaction failed to complete
  this.value = this.startValue;
};

ADRatingControl.prototype.setValueFromX = function (x) {
  this.value = Math.floor((x + ADRatingControlStarWidth + this.margin + this.margin / 2) / (ADRatingControlStarWidth + this.margin));
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADRatingControl);

ADRatingControl.init = function () {
  ADUtils.preloadImageAsset('rating/star_empty.png');
  ADUtils.preloadImageAsset('rating/star_filled.png');
};

window.addEventListener('load', ADRatingControl.init, false);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

const ADPageControlSize = 6;
const ADPageControlMargin = 10;

const ADPageControlSelectedCSS = 'selected';

/* ==================== ADPageControl ==================== */

ADPageControl.inherits = ADControl;
ADPageControl.synthetizes = ['currentPage', 'numberOfPages', 'hidesForSinglePage'];

/**
 *  @class
 *
 *  <p>You use the ADPageControl class to create and manage page controls. A page control is a succession of
 *  dots centered in the control. Each dot corresponds to a page in the application's document (or other
 *  data-model entity), with the white dot indicating the currently viewed page.</p>
 *
 *  <p>For an example of a page control, see the Weather application (with a number of locations configured)
 *  or Safari (with a number of tab views set).</p>
 *
 *  <p>When a user taps a page control to move to the next or previous page, the control sends the
 *  <code>ADControlValueChangeEvent</code> event for handling by the delegate. The delegate can then evaluate the
 *  {@link #currentPage} property to determine the page to display. The page control advances only one page
 *  in either direction.</p>
 *
 *  @extends ADControl
 *  @since AdLib 0.1
 */
function ADPageControl () {
  this.dots = [];
  //
  /**
   *  Set the value of this property to <code>true</code> so that, when the user taps the control to go to a new page,
   *  the class defers updating the page indicator until it calls {@link #updateCurrentPageDisplay}. Set the value to
   *  <code>false</code> (the default) to have the page indicator updated immediately.
   *  @type bool 
   */
  this.defersCurrentPageDisplay = false;
  //
  this._currentPage = 0;
  this._numberOfPages = 0;
  this._hidesForSinglePage = false;
  //
  this.callSuper();
  //
  this.layer.addEventListener(ADControlTouchUpInsideEvent, this, false);
};

/* ==================== ADView APIs ==================== */

/**
 *  @name ADPageControl.prototype
 *  @property {int} currentPage The index for the current page shown minus one; thus a value of <code>0</code> 
 *  (the default) indicates the first page. A page control shows the current page as a white dot. Values outside
 *  the possible range are pinned to either <code>0</code> or {@link #numberOfPages} minus <code>1</code>.
 *  @property {int} numberOfPages The value of the property is the number of pages for the page control to
 *  show as dots. The default value is <code>0</code>.
 *  @property {bool} hidesForSinglePage Assign a value of <code>true</code> to hide the page indicator when
 *  there is only one page; assign <code>false</code> (the default) to show the page indicator if there is only one page.
 */

ADPageControl.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-page-control');
};

ADPageControl.prototype.setSize = function (aSize) {
  // ensure the size can fit at least a page
  aSize.width = Math.max(aSize.width, ADPageControlSize);
  aSize.height = Math.max(aSize.height, ADPageControlSize);
  // size the layer
  this.callSuper(aSize);
  // update the padding to center the page items
  this.updateLayout();
};

/* ==================== Synthetized APIs ==================== */

ADPageControl.prototype.setCurrentPage = function (currentPage) {
  // do nothing if we already have this page as current
  // or there is no such page
  if (this._currentPage == currentPage || currentPage < 0 || currentPage >= this.numberOfPages) {
    return;
  }
  // save to ivar
  this._currentPage = currentPage;
  //
  if (!this.defersCurrentPageDisplay) {
    this.updateCurrentPageDisplay();
  }
  // notify that we have a value change
  this.dispatchEvent(this.createEvent(ADControlValueChangeEvent));
};

ADPageControl.prototype.setNumberOfPages = function (numberOfPages) {
  // do nothing if we already have this number of pages or the number is too low
  if (this._numberOfPages == numberOfPages || numberOfPages < 0) {
    return;
  }
  // remove dots in case we have less than before
  if (this.dots.length > numberOfPages) {
    while (this.dots.length > numberOfPages) {
      this.layer.removeChild(this.dots.pop());
    }
  }
  // add dots if we need more
  else {
    while (this.dots.length < numberOfPages) {
      this.dots.push(this.layer.appendChild(document.createElement('div')));
    }
  }
  // save to ivar
  this._numberOfPages = numberOfPages;
  // update current page display
  this.updateCurrentPageDisplay();
  // update layout and visibility
  this.updateLayout();
  this.updateVisibility();
};

ADPageControl.prototype.setHidesForSinglePage = function (hidesForSinglePage) {
  // save to ivar
  this._hidesForSinglePage = hidesForSinglePage;
  //
  this.updateVisibility();
};

/* ==================== Public Methods ==================== */

/**
 *  This method updates the page indicator so that the current page (the white dot) matches the value
 *  returned from {@link #currentPage}. The class ignores this method if the value of
 *  {@link #defersPageIndicatorUpdate} is <code>false</code>. Setting the {@link #currentPage} value
 *  directly updates the indicator immediately.
 */
ADPageControl.prototype.updateCurrentPageDisplay = function () {
  // first, check if any dot is already selected
  var selected_dot = this.layer.querySelector('.' + ADPageControlSelectedCSS);
  if (selected_dot) {
    selected_dot.removeClassName(ADPageControlSelectedCSS);
  }
  // now, select the dot for the current page if in range
  if (this.currentPage >= 0 && this.currentPage < this.dots.length) {
    this.dots[this.currentPage].addClassName(ADPageControlSelectedCSS);
  }
};

/**
 *  Returns the size the receiver's bounds should be to accommodate the given number of pages. Subclasses
 *  that customize the appearance of the page control can use this method to resize the page control when
 *  the page count changes.
 *
 *  @param {int} pageCount The number of pages to fit in the receiver's bounds.
 *  @returns {int} The minimum size required to display dots for the page count.
 */
ADPageControl.prototype.sizeForNumberOfPages = function (pageCount) {
  return pageCount * (ADPageControlSize + ADPageControlMargin) - ADPageControlMargin;
};

/* ==================== Private Methods ==================== */

ADPageControl.prototype.updateLayout = function (aSize) {
  this.layer.style.paddingLeft = ADUtils.px((this.size.width - this.sizeForNumberOfPages(this._numberOfPages)) / 2);
  this.layer.style.paddingTop = ADUtils.px((this.size.height - ADPageControlSize) / 2);
};

ADPageControl.prototype.updateVisibility = function () {
  this.opacity = (this.numberOfPages == 1 && this.hidesForSinglePage) ? 0 : 1;
};

/* ==================== EVENT HANDLING ==================== */

ADPageControl.prototype.handleEvent = function (event) {
  this.callSuper(event);
  // process touchup event
  if (event.type == ADControlTouchUpInsideEvent) {
    // do nothing if we have a single page on display
    if (this.numberOfPages <= 1) {
      return;
    }
    // figure out if we tapped to the left or the right of the dot
    var tap_x = ADPoint.fromEventInElement(event, this.layer).x;
    // XXX: might be a better way to tell where this is
    var dot_x = this.dots[this.currentPage].offsetLeft;
    // tap was to the left of the dot
    if (tap_x <= dot_x - ADPageControlMargin / 2) {
      this.currentPage--;
    }
    // tap was to the right of the dot
    else if (tap_x >= dot_x + ADPageControlSize + ADPageControlMargin / 2) {
      this.currentPage++;
    }
  }
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADPageControl);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  A segment index value indicating that there is no selected segment in a {@link ADSegmentedControl}.
 *  See {@link ADSegmentedControl#selectedSegment} for further information.
 *  @constant
 *  @type int
 */
const ADSegmentedControlNoSegment = -1;

/**
 *  The large plain style for a {@link ADSegmentedControl}. This style is the default.
 *  @constant
 *  @type String
 */
const ADSegmentedControlStylePlain = 'plain';
/**
 *  The large bordered style for a {@link ADSegmentedControl}.
 *  @constant
 *  @type String
 */
const ADSegmentedControlStyleBordered = 'bordered';
/**
 *  The small toolbar style for a {@link ADSegmentedControl}.
 *  @constant
 *  @type String
 */
const ADSegmentedControlStyleBar = 'bar';
/**
 *  The small black toolbar style for a {@link ADSegmentedControl}.
 *  @constant
 *  @type String
 */
const ADSegmentedControlStyleBarBlack = 'bar black';

const ADSegmentedControlSegmentSelectedCSS = 'selected';

const ADSegmentedControlSegmentMomentarySelectionTimer = 200;

/* ==================== ADSegmentedControl Constructor ==================== */

ADSegmentedControl.inherits = ADControl;
ADSegmentedControl.synthetizes = ['width', 'style', 'numberOfSegments', 'selectedSegmentIndex', 'momentary'];

/**
 *  @class
 *
 *  <p>A ADSegmentedControl object is a horizontal control made of multiple segments, each segment functioning
 *  as a discrete button. A segmented control affords a compact means to group together a number of controls.</p>
 *
 *  <p>A segmented control can display a title or an image. The ADSegmentedControl object automatically resizes
 *  segments to fit proportionally within the control unless they have a specific width set.</p>
 *
 *  <p>When a user taps taps a segment, the control sends the <code>ADControlValueChangeEvent</code>
 *  event for handling by the delegate as its {@link #selectedSegmentIndex} changes.</p>
 *
 *  @extends ADControl
 *  @since AdLib 0.1
 *
 *  @param {Array} items An array of <code>String</code> objects (for segment titles) or {@type ADImage} objects (for segment images).
 */
function ADSegmentedControl (items) {
  this.callSuper();
  // synthesize read-write properties
  this._width = 0;
  this._style = '';
  this._numberOfSegments = 0;
  this._selectedSegmentIndex = ADSegmentedControlNoSegment;
  this._momentary = false;
  // base style
  this.style = ADSegmentedControlStylePlain;
  // check if we have a pre-defined list of items
  if (!ADUtils.objectIsUndefined(items) && ADUtils.objectIsArray(items)) {
    this.populateWithItems(items);
  }
};

/* ==================== Population and Layout ==================== */

ADSegmentedControl.prototype.populateWithItems = function (items) {
  for (var i = 0; i < items.length; i++) {
    var segment = items[i];
    // check if we are dealing with an image
    if (segment instanceof ADImage) {
      this.insertSegmentWithImageAtIndexWithAnimation(segment, this.numberOfSegments, false);
    }
    // or if it's a title
    else if (ADUtils.objectIsString(segment)) {
      this.insertSegmentWithTitleAtIndexWithAnimation(segment, this.numberOfSegments, false);
    }
  }
  // update sizes
  this.updateLayout();
};

ADSegmentedControl.prototype.updateLayout = function () {
  // do nothing if we have a 0 width
  if (this.size.width <= 0) {
    return;
  }
  // first check out how many items are auto-sized
  var cumulative_segments_width = 0;
  var auto_sized_segments = 0;
  for (var i = 0; i < this.subviews.length; i++) {
    var segment = this.subviews[i];
    if (segment.width > 0) {
      cumulative_segments_width += segment.width;
    }
    else {
      auto_sized_segments++;
    }
  }
  // now figure out the size of auto-sized items
  var auto_sized_width = Math.round((auto_sized_segments > 0) ? (this.size.width - cumulative_segments_width) / auto_sized_segments : 0);
  // now go through segments again and size them
  var radius = (this._style == ADSegmentedControlStyleBar || this._style == ADSegmentedControlStyleBarBlack) ? 6 : 10;
  var next_item_x = -radius;
  for (var i = 0; i < this.subviews.length; i++) {
    var segment = this.subviews[i];
    var width = (segment.width == 0) ? auto_sized_width : segment.width;
    segment.layer.style.left = ADUtils.px(next_item_x);
    // check that last item will fill the view for sure
    if (i == this.subviews.length - 1) {
      width = this.size.width - next_item_x - radius;
    }
    this.setActualWidthForSegmentAtIndex(width, i);
    next_item_x += width;
  }
};

/* ==================== ADView APIs ==================== */

ADSegmentedControl.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-segmented-control');
};

ADSegmentedControl.prototype.setSize = function (aSize) {
  aSize.height = (this._style == ADSegmentedControlStyleBar || this._style == ADSegmentedControlStyleBarBlack) ? 30 : 44;
  this.callSuper(aSize);
  // update layout
  this.updateLayout();
};

ADSegmentedControl.prototype.didAddSubview = function (subview) {
  // update the layout of segments
  this.updateLayout();
  // the index of the selected segment may now have changed
  if (this._selectedSegmentIndex != ADSegmentedControlNoSegment) {
    if (subview._indexInSuperviewSubviews <= this._selectedSegmentIndex) {
      this.selectedSegmentIndex++;
    }
  }
};

ADSegmentedControl.prototype.willMoveToSuperview = function (superview) {
  // force bar style if we're moving to a ADToolbar
  if (superview instanceof ADToolbar) {
    this.style = (superview.style === ADToolbarStyleDefault) ? ADSegmentedControlStyleBar : ADSegmentedControlStyleBarBlack;
  }
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADSegmentedControl.prototype
 *  @property {int} width The control's width in pixels.
 *  @property {String} style The control's style. The default style is ADSegmentedControlStylePlain.
 *  @property {int} numberOfSegments Returns the number of segments the receiver has.
 *  @property {int} selectedSegmentIndex The index number identifying the selected segment (that is, the last segment touched).
 *  The default value is <code>ADSegmentedControlNoSegment</code> (no segment selected) until the user touches a segment.
 *  Set this property to <code>ADSegmentedControlNoSegment</code> to turn off the current selection.
 *  @property {bool} momentary A Boolean value that determines whether segments in the receiver show selected state. The default value of
 *  this property is <code>false</code>. If it is set to <code>true</code>, segments in the control do not show selected state and do not
 *  update the value of {@link #selectedSegment} after tracking ends.
 */

ADSegmentedControl.prototype.getWidth = function () {
  return this.size.width;
};

ADSegmentedControl.prototype.setWidth = function (width) {
  this.size = new ADSize(width, 0);
};

ADSegmentedControl.prototype.setStyle = function (style) {
  this.layer.removeClassName(this._style);
  this.layer.addClassName(style);
  this._style = style;
  //
  this.width = this.width;
};

ADSegmentedControl.prototype.getNumberOfSegments = function () {
  return this.subviews.length;
};

ADSegmentedControl.prototype.setSelectedSegmentIndex = function (index) {
  // do nothing if we have no such segment or we already have this index selected
  if (index < -1 || index >= this.subviews.length || index == this._selectedSegmentIndex) {
    return;
  }
  // remove selected state from previously selected segment
  if (this._selectedSegmentIndex != ADSegmentedControlNoSegment && this._selectedSegmentIndex >= 0 && this._selectedSegmentIndex < this.subviews.length) {
    this.subviews[this._selectedSegmentIndex].layer.removeClassName(ADSegmentedControlSegmentSelectedCSS);
  }
  // if we have a visible index, update its selected state
  if (index != ADSegmentedControlNoSegment) {
    this.subviews[index].layer.addClassName(ADSegmentedControlSegmentSelectedCSS);
  }
  // save to ivar
  this._selectedSegmentIndex = index;
  // notify unless momentary and -1
  if (!this._momentary || index != ADSegmentedControlNoSegment) {
    this.dispatchEvent(this.createEvent(ADControlValueChangeEvent));
  }
};

ADSegmentedControl.prototype.setMomentary = function (momentary) {
  // do nothing if we were in that state already
  if (this._momentary == momentary) {
    return;
  }
  // archive this value
  this._momentary = momentary;
  // make sure no item is selected
  this.selectedSegmentIndex = ADSegmentedControlNoSegment;
};

/* ==================== Managing Segment Content ==================== */

/**
 *  Sets the content of a segment to a given image. A segment can only have an image or a title; it can't have both. There is no default image.
 *
 *  @param {ADImage} image An image object to display in the segment.
 *  @param {int} index An index number identifying a segment in the control. It must be a number between <code>0</code>
 *  and the {@link #numberOfSegments} minus <code>1</code>; values exceeding this upper range are pinned to it.
 */
ADSegmentedControl.prototype.setImageForSegmentAtIndex = function (image, index) {
  index = Math.max(Math.min(index, this.subviews.length - 1), 0);  
  this.subviews[index].image = image;
};

/**
 *  Returns the image for a specific segment.
 *
 *  @param {int} index An index number identifying a segment in the control. It must be a number between <code>0</code>
 *  and the {@link #numberOfSegments} minus <code>1</code>; values exceeding this upper range are pinned to it.
 *  @returns {ADImage} The image assigned to the segment as content. If no image has been set, it returns <code>null</code>.
 */
ADSegmentedControl.prototype.imageForSegmentAtIndex = function (index) {
  index = Math.min(index, this.subviews.length - 1);
  return (index >= 0) ? this.subviews[index].image : null;
};

/**
 *  Sets the content of a segment to a given title. A segment can only have an image or a title; it can't have both. There is no default title.
 *
 *  @param {String} title A string to display in the segment as its title.
 *  @param {int} index An index number identifying a segment in the control. It must be a number between <code>0</code>
 *  and the {@link #numberOfSegments} minus <code>1</code>; values exceeding this upper range are pinned to it.
 */
ADSegmentedControl.prototype.setTitleForSegmentAtIndex = function (title, index) {
  index = Math.max(Math.min(index, this.subviews.length - 1), 0);
  this.subviews[index].title = title;
};

/**
 *  Returns the title of the specified segment.
 *
 *  @param {int} index An index number identifying a segment in the control. It must be a number between <code>0</code>
 *  and the {@link #numberOfSegments} minus <code>1</code>; values exceeding this upper range are pinned to it.
 *  @returns {String} title The string (title) assigned to the receiver as content. If no title has been set, it returns <code>null</code>.
 */
ADSegmentedControl.prototype.titleForSegmentAtIndex = function (index) {
  index = Math.min(index, this.subviews.length - 1);
  return (index >= 0) ? this.subviews[index].title : null;
};

/* ==================== Managing Segments ==================== */

/**
 *  Inserts a segment at a specified position in the receiver and gives it an image as content.
 *
 *  @param {ADImage} image An image object to use as the content of the segment.
 *  @param {int} index An index number identifying a segment in the control. It must be a number between <code>0</code>
 *  and the {@link #numberOfSegments} minus <code>1</code>; values exceeding this upper range are
 *  pinned to it. The new segment is inserted just before the designated one.
 *  @param {bool} animated <code>true</code> if the insertion of the new segment should be animated, otherwise <code>false</code>.
 *  Currently, animations are not supported.
 */
ADSegmentedControl.prototype.insertSegmentWithImageAtIndexWithAnimation = function (image, index, animated) {
  index = Math.max(Math.min(index, this.subviews.length), 0);
  //
  var segment = new ADSegment();
  this.insertSubviewAtIndex(segment, index);
  this.setImageForSegmentAtIndex(image, index);
};

/**
 *  Inserts a segment at a specific position in the receiver and gives it a title as content.
 *
 *  @param {String} title A string to use as the segment's title.
 *  @param {int} index An index number identifying a segment in the control. It must be a number between <code>0</code>
 *  and the {@link #numberOfSegments} minus <code>1</code>; values exceeding this upper range are
 *  pinned to it. The new segment is inserted just before the designated one.
 *  @param {bool} animated <code>true</code> if the insertion of the new segment should be animated, otherwise <code>false</code>.
 *  Currently, animations are not supported.
 */
ADSegmentedControl.prototype.insertSegmentWithTitleAtIndexWithAnimation = function (title, index, animated) {
  index = Math.max(Math.min(index, this.subviews.length), 0);
  //
  var segment = new ADSegment();
  this.insertSubviewAtIndex(segment, index);
  this.setTitleForSegmentAtIndex(title, index);
};

/**
 *  Removes the specified segment from the control, optionally animating the transition.
 *
 *  @param {index} index An index number identifying a segment in the control. It must be a number between <code>0</code>
 *  and the {@link #numberOfSegments} minus <code>1</code>; values exceeding this upper range are pinned to it.
 *  @param {bool} animated <code>true</code> if the removal of the new segment should be animated, otherwise <code>false</code>.
 *  Currently, animations are not supported.
 */
ADSegmentedControl.prototype.removeSegmentAtIndexWithAnimation = function (index, animated) {
  index = Math.max(Math.min(index, this.subviews.length - 1), 0);
  //
  this.subviews[index].removeFromSuperview();
  //
  this.updateLayout();
  // the index of the selected segment may now have changed
  if (this._selectedSegmentIndex != ADSegmentedControlNoSegment) {
    if (index < this._selectedSegmentIndex) {
      this.selectedSegmentIndex--;
    }
    else if (index == this._selectedSegmentIndex) {
      this.selectedSegmentIndex = ADSegmentedControlNoSegment;
    }
  }
};

/**
 *  Removes all segments in the control.
 */
ADSegmentedControl.prototype.removeAllSegments = function () {
  while (this.subviews.length) {
    this.subviews[0].removeFromSuperview();
  }
};

/* ==================== Managing Segment Behavior and Appearance ==================== */

// ADSegmentedControl.prototype.setEnabledForSegmentAtIndex = function (enabled, index) {
// };
// 
// ADSegmentedControl.prototype.isEnabledForSegmentAtIndex = function (index) {
//   return false;
// };
// 
// ADSegmentedControl.prototype.setContentOffsetForSegmentAtIndex = function (offset, index) {
//   
// };
// 
// ADSegmentedControl.prototype.contentOffsetForSegmentAtIndex = function (index) {
//   return null;
// };

/**
 *  Sets the width of the specified segment in the control.
 *
 *  @param {int} width A value specifying the width, in pixels, of the segment. The default value is <code>0</code>, which tells
 *  the control to automatically size the segment.
 *  @param {index} index An index number identifying a segment in the control. It must be a number between <code>0</code>
 *  and the {@link #numberOfSegments} minus <code>1</code>; values exceeding this upper range are pinned to it.
 */
ADSegmentedControl.prototype.setWidthForSegmentAtIndex = function (width, index) {
  index = Math.max(Math.min(index, this.subviews.length - 1), 0);
  // record width
  this.subviews[index].width = width;
  //
  this.setActualWidthForSegmentAtIndex(width, index);
  //
  this.updateLayout();
};

/**
 *  Returns the width of the indicated segment in the control
 *
 *  @param {index} index An index number identifying a segment in the control. It must be a number between <code>0</code>
 *  and the {@link #numberOfSegments} minus <code>1</code>; values exceeding this upper range are pinned to it.
 *  @returns A value specifying the width, in pixels, of the segment. If the value is <code>0</code>, which means the control
 *  automatically sizes the segment.
 */
ADSegmentedControl.prototype.widthForSegmentAtIndex = function (index) {
  index = Math.max(Math.min(index, this.subviews.length - 1), 0);
  return this.subviews[index].width;
};

ADSegmentedControl.prototype.setActualWidthForSegmentAtIndex = function (width, index) {
  index = Math.max(Math.min(index, this.subviews.length - 1), 0);
  this.subviews[index].size = new ADSize(width, this.size.height);
};

/* ==================== EVENT HANDLING ==================== */

ADSegmentedControl.prototype.touchesBegan = function (event) {
  this.callSuper(event);
  // update the selected index
  this.selectedSegmentIndex = event.target.getNearestView()._indexInSuperviewSubviews;
  // revert in case we're momentary
  if (this._momentary) {
    var _this = this;
    setTimeout(function () {
      _this.selectedSegmentIndex = ADSegmentedControlNoSegment;
    }, ADSegmentedControlSegmentMomentarySelectionTimer)
  }
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADSegmentedControl);

/* ==================== CLASS CONSTANTS ==================== */

ADSegmentedControl.init = function () {
  ADUtils.preloadImageAsset('segments/bar/button_highlighted_fill.png');
  ADUtils.preloadImageAsset('segments/bar/button_highlighted_left_cap.png');
  ADUtils.preloadImageAsset('segments/bar/button_highlighted_right_cap.png');
  ADUtils.preloadImageAsset('segments/bar/button.png');
  ADUtils.preloadImageAsset('segments/bar/divider_highlighted.png');
  ADUtils.preloadImageAsset('segments/bar/divider.png');
  ADUtils.preloadImageAsset('segments/plain/button_highlighted_fill.png');
  ADUtils.preloadImageAsset('segments/plain/button_highlighted_left_cap.png');
  ADUtils.preloadImageAsset('segments/plain/button_highlighted_right_cap.png');
  ADUtils.preloadImageAsset('segments/plain/button.png');
  ADUtils.preloadImageAsset('segments/plain/divider_hilighted_left.png');
  ADUtils.preloadImageAsset('segments/plain/divider_hilighted_right.png');
  ADUtils.preloadImageAsset('segments/plain/divider.png');
  ADUtils.preloadImageAsset('segments/bordered/button_highlighted_fill.png');
  ADUtils.preloadImageAsset('segments/bordered/button_highlighted_left_cap.png');
  ADUtils.preloadImageAsset('segments/bordered/button_highlighted_right_cap.png');
  ADUtils.preloadImageAsset('segments/bordered/button.png');
  ADUtils.preloadImageAsset('segments/bordered/divider_hilighted_left.png');
  ADUtils.preloadImageAsset('segments/bordered/divider_hilighted_right.png');
  ADUtils.preloadImageAsset('segments/bordered/divider.png');
};

window.addEventListener('load', ADSegmentedControl.init, false);

/* ==================== ADSegment ==================== */

ADSegment.inherits = ADView;
ADSegment.synthetizes = ['title', 'image'];

function ADSegment (items) {
  this.callSuper();
  //
  this.width = 0;
  //
  this._title = null;
  this._image = null;
};

/* ==================== ADView APIs ==================== */

ADSegment.prototype.createLayer = function () {
  this.callSuper(); 
  //
  this.layer.addClassName('ad-segment');
  // add backgrounds
  this.layer.appendChild(document.createElement('div'));
  this.layer.appendChild(document.createElement('div'));
  // add label
  this.label = this.layer.appendChild(document.createElement('span'));
  this.icon =  this.layer.appendChild(document.createElement('img'));
};

/* ==================== ADView APIs ==================== */

ADSegment.prototype.setSize = function (aSize) {
  this.callSuper(aSize);
  if (this._image !== null) {
    this.positionImage();
  }
};

/* ==================== Synthetized APIs ==================== */

ADSegment.prototype.setTitle = function (title) {
  this.label.textContent = title;
  this._title = title;
};

ADSegment.prototype.setImage = function (image) {
  this.icon.src = image.url;
  // save image to ivar
  this._image = image;
  // if image is already loaded, lay it out right away
  if (image.loaded) {
    this.positionImage();
  }
  // otherwise, monitor loaded state
  else {
    image.addPropertyObserver('loaded', this);
  }
};

ADSegment.prototype.handlePropertyChange = function (object, propertyName) {
  // image has finished loading, lay it out
  this.positionImage();
};

ADSegment.prototype.positionImage = function () {
  this.icon.style.left = ADUtils.px((this.size.width - this._image.width) / 2);
  this.icon.style.top = ADUtils.px((this.size.height - this._image.height) / 2);
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADSegment);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

const ADSwitchOffLabel = 'OFF';
const ADSwitchOnLabel = 'ON';

const ADSwitchOffCSS = 'off';
const ADSwitchOnCSS = 'on';

const ADSwitchDownCSS = 'dragging';
const ADSwitchSnapCSS = 'snap';

/**
 *  The constant width of a {@link ADSwitch}.
 *  @constant
 *  @type int
 */
const ADSwitchWidth = 93;
/**
 *  The constant height of a {@link ADSwitch}.
 *  @constant
 *  @type int
 */
const ADSwitchHeight = 27;

const ADSwitchDragIdle = 0;
const ADSwitchCanDrag = 1;
const ADSwitchIsDragging = 2;

const ADSwitchSlideDistance = 54;
const ADSwitchButtonWidth = 40;

const ADSwitchDragMax = 70;
const ADSwitchDragMin = 16;
const ADSwitchDragGutter = 5;
const ADSwitchCapWidth = 4;

const ADSwitchLeftCapTrackCSS = 'left-cap track-cap';
const ADSwitchLeftCapButtonCSS = 'left-cap button-cap';
const ADSwitchRightCapTrackCSS = 'right-cap track-cap';
const ADSwitchRightCapButtonCSS = 'right-cap button-cap';

/* ==================== ADSwitch ==================== */

ADSwitch.inherits = ADControl;
ADSwitch.synthetizes = ['on'];

/**
 *  @class
 *
 *  <p>You use the ADSwitch class to create and manage the On/Off buttons you see, for example, in the preferences
 *  (Settings) for such services as Airplane Mode. These objects are known as switches. As with {@link ADSlider}, when
 *  the user manipulates the switch control ("flips" it) a <code>ADControlValueChangeEvent</code> event is dispatched.</p>
 *
 *  @extends ADControl
 *  @since AdLib 0.1
 */
function ADSwitch () {
  this.callSuper();
  this.dragOffsetX = ADSwitchDragMax;
  // set up size
  this.size = new ADSize();
  this.on = false;
};

/* ==================== ADView APIs ==================== */

ADSwitch.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-switch');
  
  this.leftCap = this.layer.appendChild(document.createElement('div'));
  this.leftCap.className = ADSwitchLeftCapTrackCSS;
  
  this.rightCap = this.layer.appendChild(document.createElement('div'));
  this.rightCap.className = ADSwitchRightCapButtonCSS;
  
  this.slideContainer = this.layer.appendChild(document.createElement('div'));
  this.slideContainer.className = 'slide-container';
  
  this.slider = this.slideContainer.appendChild(document.createElement('div'));
  this.slider.className = 'slider';
  this.slider.addEventListener('webkitTransitionEnd', this, false);

  this.button = this.slider.appendChild(document.createElement('div'));
  this.button.className = 'button';
  
  var offLabel = this.slider.appendChild(document.createElement('div'));
  offLabel.textContent = ADSwitchOffLabel;
  offLabel.className = 'label-off';
  
  var onLabel = this.slider.appendChild(document.createElement('div'));
  onLabel.textContent = ADSwitchOnLabel;
  onLabel.className = 'label-on';
};

ADSwitch.prototype.setSize = function (aSize) {
  aSize.width = ADSwitchWidth;
  aSize.height = ADSwitchHeight;
  this.callSuper(aSize);
};

/* ==================== PUBLIC APIs ==================== */

/**
 *  Set the state of the switch to On or Off, optionally animating the transition.
 *
 *  @param {bool} on <code>true</code> if the switch should be turn to the On position; <code>false</code> if it
 *  should be turned to the Off position. If the switch is already in the designated position, nothing happens.
 *  @param {bool} animated <code>true</code> to animate the "flipping" of the switch; otherwise <code>false</code>.
 */
ADSwitch.prototype.setOnWithAnimation = function (on, animated) {
  if (animated) {
    if (!((on && this.dragOffsetX == ADSwitchDragMax) || (!on && this.dragOffsetX == ADSwitchDragMin))) {
      this.leftCap.className = ADSwitchLeftCapTrackCSS;
      this.rightCap.className = ADSwitchRightCapTrackCSS;
    }
    this.slider.addClassName(ADSwitchSnapCSS);
  } else {
    if (on) {
      this.leftCap.className = ADSwitchLeftCapTrackCSS;
      this.rightCap.className = ADSwitchRightCapButtonCSS;
    } else {
      this.leftCap.className = ADSwitchLeftCapButtonCSS;
      this.rightCap.className = ADSwitchRightCapTrackCSS;
    }
    this.slider.removeClassName(ADSwitchSnapCSS);
  }
  
  if (on) {
    this.slider.style.webkitTransform = ADUtils.t(-ADSwitchCapWidth, 0);
    this.dragOffsetX = ADSwitchDragMax;
  } else {
    this.slider.style.webkitTransform = ADUtils.t(-ADSwitchSlideDistance - ADSwitchCapWidth, 0);
    this.dragOffsetX = ADSwitchDragMin;
  }
  this._on = on;
  this.dispatchEvent(this.createEvent(ADControlValueChangeEvent));
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADSwitch.prototype
 *  @property {bool} on Determines the off/on state of the switch.
 */

ADSwitch.prototype.setOn = function (on) {
  this.setOnWithAnimation(on, false);
};

/* ==================== EVENT HANDLING ==================== */

ADSwitch.prototype.handleEvent = function (event) {
  this.callSuper(event);
  //
  if (event.type == 'webkitTransitionEnd') {
    this.transitionEnded();
  }
};

ADSwitch.prototype.transitionEnded = function () {
  this.slider.removeClassName(ADSwitchSnapCSS);
  if (this._on) {
    this.slider.style.webkitTransform = ADUtils.t(-ADSwitchCapWidth, 0);
    this.rightCap.className = ADSwitchRightCapButtonCSS;
  } else {
    this.slider.style.webkitTransform = ADUtils.t(-ADSwitchSlideDistance - ADSwitchCapWidth, 0);
    this.leftCap.className = ADSwitchLeftCapButtonCSS;
  }
};

/* 
  When we first get a touch, we check to see if it's on the button.
  If so, the control can be dragged.

  TODO: Start a timer that moves from CanDrag to IsDragging.
  FIXME: Should only start the input sequence if the very first touch is in the element, not when the first touch enters the element.
*/
ADSwitch.prototype.touchesBegan = function (event) {
  this.callSuper(event);
  
  var point = ADPoint.fromEventInElement(event, this.layer);
  if ((!this._on && point.x < ADSwitchButtonWidth) || (this.on && point.x > ADSwitchSlideDistance)) {
    this.layer.addClassName(ADSwitchDownCSS);
    this.dragState = ADSwitchCanDrag;
  }
  
};

/*
  If we can drag (i.e. the first touch was over the button), do so.
*/
ADSwitch.prototype.touchesMoved = function (event) {
  this.callSuper(event);
  
  if (this.dragState >= ADSwitchCanDrag) {
    this.slider.removeClassName(ADSwitchSnapCSS);
    this.dragState = ADSwitchIsDragging;
    
    this.rightCap.className = ADSwitchRightCapTrackCSS;
    this.leftCap.className = ADSwitchLeftCapTrackCSS;
    
    var point = ADPoint.fromEventInElement(event, this.layer);
    var minX = ADSwitchButtonWidth / 2;
    var maxX = ADSwitchWidth - minX;
    this.dragOffsetX = point.x;
    
    if (this.dragOffsetX > ADSwitchDragMax - ADSwitchDragGutter) {
      if (this.dragOffsetX >= ADSwitchDragMax) {
        this.dragOffsetX = ADSwitchDragMax;
        this.rightCap.className = ADSwitchRightCapButtonCSS;
      } else {
        this.dragOffsetX = ADSwitchDragMax - ADSwitchDragGutter;
      }
    } else if (this.dragOffsetX < ADSwitchDragMin + ADSwitchDragGutter) {
      if (this.dragOffsetX < ADSwitchDragMin) {
        this.dragOffsetX = ADSwitchDragMin;
        this.leftCap.className = ADSwitchLeftCapButtonCSS;
      } else {
        this.dragOffsetX = ADSwitchDragMin + ADSwitchDragGutter;
      }
    }
    
    this.slider.style.webkitTransform = ADUtils.t(-ADSwitchSlideDistance - (minX - this.dragOffsetX), 0);
  }
};

/*
  If the button was dragged at all, snap back to the closest
  side. If there was no dragging, toggle the state of the
  button.
*/
ADSwitch.prototype.touchesEnded = function (event) {
  this.callSuper(event);
  if (this.dragState === ADSwitchIsDragging) {
    if (this.dragOffsetX < (ADSwitchWidth / 2)) {
      this.setOnWithAnimation(false, true);
    } else if (this.dragOffsetX > (ADSwitchWidth / 2)) {
      this.setOnWithAnimation(true, true);
    } else {
      this.setOnWithAnimation(this.on, true);
    }
  } else {
    this.setOnWithAnimation(!this._on, true);
  }
  
  this.layer.removeClassName(ADSwitchDownCSS);
  this.dragState = ADSwitchDragIdle;
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADSwitch);

/* ==================== CLASS CONSTANTS ==================== */

ADSwitch.init = function () {
  ADUtils.preloadImageAsset('switch/background_precomposed.png');
  ADUtils.preloadImageAsset('switch/button_precomposed.png');
  ADUtils.preloadImageAsset('switch/left_cap_button.png');
  ADUtils.preloadImageAsset('switch/left_cap_track.png');
  ADUtils.preloadImageAsset('switch/right_cap_button.png');
  ADUtils.preloadImageAsset('switch/right_cap_track.png');
  ADUtils.preloadImageAsset('switch/right_cap_button_down.png');
  ADUtils.preloadImageAsset('switch/left_cap_button_down.png');
};

window.addEventListener('load', ADSwitch.init, false);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  Height of a {@link ADSearchBar}.
 *  @constant
 *  @type int
 */
const ADSearchBarHeight = 44;

/**
 *  Default style for a {@link ADSearchBar}.
 *  @constant
 *  @type String
 */
const ADSearchBarStyleDefault = 'default';
/**
 *  Black style for a {@link ADSearchBar}.
 *  @constant
 *  @type String
 */
const ADSearchBarStyleBlack = 'black';
/**
 *  Translucent black style for a {@link ADSearchBar}.
 *  @constant
 *  @type String
 */
const ADSearchBarStyleBlackTranslucent = 'black-translucent';

/**
 *  @class
 *  @name ADSearchBarDelegate
 *  @since AdLib 0.1
 */

/**
 *  Invoked after the user changed the search text or when the text is cleared from the search field.
 *
 *  @name searchBarTextDidChange
 *  @function
 *
 *  @param {ADSearchBar} searchBar The search bar that is being edited.
 *  @param {String} searchText The current text in the search text field.
 *  @memberOf ADSearchBarDelegate.prototype
 */
const ADSearchBarTextDidChange        = 'searchBarTextDidChange';
//const ADSearchBarShouldBeginEditing   = 'searchBarShouldBeginEditing';
/**
 *  Invoked when the user begins editing the search text.
 *
 *  @name searchBarTextDidBeginEditing
 *  @function
 *
 *  @param {ADSearchBar} searchBar The search bar that is being edited.
 *  @memberOf ADSearchBarDelegate.prototype
 */
const ADSearchBarTextDidBeginEditing  = 'searchBarTextDidBeginEditing';
//const ADSearchBarShouldEndEditing     = 'searchBarShouldEndEditing';
/**
 *  Invoked when the user finishes editing the search text.
 *
 *  @name searchBarTextDidEndEditing
 *  @function
 *
 *  @param {ADSearchBar} searchBar The search bar that was being edited.
 *  @memberOf ADSearchBarDelegate.prototype
 */
const ADSearchBarTextDidEndEditing    = 'searchBarTextDidEndEditing';
/**
 *  Invoked when the cancel button is tapped. Typically, the developer implements this method to dismiss the search bar.
 *
 *  @name searchBarCancelButtonClicked
 *  @function
 *
 *  @param {ADSearchBar} searchBar The search bar that is being edited.
 *  @memberOf ADSearchBarDelegate.prototype
 */
const ADSearchBarCancelButtonClicked  = 'searchBarCancelButtonClicked';

const ADSearchBarShowsCancelButtonCSS = 'shows-cancel-button';
const ADSearchBarDisplaysPlaceholder = 'displays-placeholder';

/* ==================== ADSearchBar ==================== */

ADSearchBar.inherits = ADView;
ADSearchBar.includes = [ADEventTriage];
ADSearchBar.synthetizes = ['style', 'placeholder', 'text', 'showsCancelButton', 'editing'];

/**
 *  @class
 *
 *  <p>The ADSearchBar class implements a text field control for text-based searches. The view provides a
 *  text field for entering text and an optional cancel button. You use a {@link #delegate}, an object conforming
 *  to the {@link ADSearchBarDelegate} protocol, to implement the actions when text is entered and buttons are clicked.
 *  The ADSearchBar object does not actually perform any searches and it's the developer's job to display results
 *  as the user interacts with the search bar.</p>
 *
 *  @extends ADView
 *  @since AdLib 0.1
 */
function ADSearchBar () {
  /**
   *  The delegate should conform to the {@link ADSearchBarDelegate} protocol. Set this property to further modify
   *  the customizing behavior. The default value is <code>null</code>.
   *  @type Object 
   */
  this.delegate = null;
  // default values for synthetized properties
  this._style = ADSearchBarStyleDefault;
  this._placeholder = '';
  this._text = '';
  this._showsCancelButton = false;
  this._editing = false;
  // private field to track whether we ever were explicitely sized
  this.hasBeenExplicitelySized = false;
  //
  this.callSuper();
  // add event listeners so we can sync the real label and trigger delegate methods
  this.field.addEventListener('focus', this, false);
  this.field.addEventListener('blur', this, false);
  this.field.addEventListener('input', this, false);
  this.field.parentNode.addEventListener('submit', this, false);
  // add event listeners to track activation of the buttons
  this.button.addEventListener(ADControlTouchUpInsideEvent, this, false);
  this.emptyButton.addEventListener(ADControlTouchUpInsideEvent, this, false);
  this.button.addPropertyObserver('size', this, 'updateLayout');
  // turn off auto-sizing, we need to manage sizing manually
  this.autoresizesSubviews = false;
};

/* ==================== ADView APIs ==================== */

ADSearchBar.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-search-bar');
  //
  var field_container = this.layer.appendChild(document.createElement('form'));
  //
  this.label = field_container.appendChild(document.createElement('div'));
  this.field = field_container.appendChild(document.createElement('input'));
  this.field.type = 'text';
  //
  this.button = new ADBarButtonItem();
  this.button.title = 'Cancel';
  this.button.view.addPropertyObserver('size', this, 'updateLayout');
  this.addSubview(this.button.view);
  //
  this.emptyButton = this.addSubview(new ADButton(ADButtonTypeCustom));
};

ADSearchBar.prototype.setSize = function (aSize) {
  aSize.height = ADSearchBarHeight;
  this.callSuper(aSize);
  // track that we've been explicitely sized
  this.hasBeenExplicitelySized = true;
  //
  this.updateLayout();
};

ADSearchBar.prototype.updateLayout = function () {
  // only update layout once we have useful metrics
  // XXX: might not be necessary due to ADBarButtonItem only changing
  // its .size once it is in the tree
  if (!this.layerIsInDocument) {
    return;
  }
  // size contents
  var button_width = this.showsCancelButton ? (this.button.view.size.width + 5) : 0;
  this.field.parentNode.style.right = ADUtils.px(button_width + 5);
  this.emptyButton.layer.style.right = ADUtils.px(button_width + 10);
};

ADSearchBar.prototype.didMoveToSuperview = function () {
  this.callSuper();
  // do nothing if we have been explicitely sized or we've been
  // removed from the tree altogether
  if (this.hasBeenExplicitelySized || this.superview === null) {
    return;
  }
  // otherwise, make us the width of our parent
  this.size = new ADSize(this.superview.size.width, ADSearchBarHeight);
  // track that the sizing was not explicit
  this.hasBeenExplicitelySized = false;
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADSearchBar.prototype
 *  @property {String} style The toolbar style that specifies its appearance. The default value is <code>ADSearchBarStyleDefault</code>.
 *  @property {String} placeholder The string that is displayed when there is no other text in the text field, the empty string by default.
 *  @property {String} text The current or starting search text, the empty string by default.
 *  @property {bool} editing This property is set to <code>true</code> when the user begins editing text in the search bar's text field,
 *  and it is set to <code>false</code> again when editing ends. Notifications about when editing begins and ends are sent to the search bar delegate.
 *  This property can be used to explicitely place focus on the text field or remove it, although placing focus is limited to code exectued as 
 *  a result of a user action (<code>touchend</code> or <code>click</code> event handler for instance).
 *  @property {bool} showsCancelButton A Boolean value indicating whether the cancel button is displayed, <code>false</code> by default.
 */

ADSearchBar.prototype.setStyle = function (style) {
  this.layer.removeClassName(this._style);
  this.layer.addClassName(style);
  this._style = style;
};

ADSearchBar.prototype.setPlaceholder = function (placeholder) {
  this._placeholder = placeholder;
  // maybe we need to display it right away
  this.checkForPlaceholderDisplay();
};

ADSearchBar.prototype.getText = function (text) {
  return this.field.value;
};

ADSearchBar.prototype.setText = function (text) {
  this.label.innerText = text;
  this.field.value = text;
  // tell our delegate we've got some new text
  if (ADUtils.objectHasMethod(this.delegate, ADSearchBarTextDidChange)) {
    this.delegate[ADSearchBarTextDidChange](this, text);
  }
  //
  this.checkForPlaceholderDisplay();
};

ADSearchBar.prototype.setShowsCancelButton = function (showsCancelButton) {
  // nothing to do if we're not changing value here
  if (this._showsCancelButton == showsCancelButton) {
    return;
  }
  //
  this.layer[showsCancelButton ? 'addClassName' : 'removeClassName'](ADSearchBarShowsCancelButtonCSS);
  this._showsCancelButton = showsCancelButton;
  this.updateLayout();
};

ADSearchBar.prototype.checkForPlaceholderDisplay = function () {
  this.layer[this.text == '' ? 'addClassName' : 'removeClassName'](ADSearchBarDisplaysPlaceholder);
  if (this.text == '') {
    this.label.innerText = this._placeholder;
  }
};

ADSearchBar.prototype.setEditing = function (editing) {
  this._editing = editing;
  this.field[editing ? 'focus' : 'blur']();
};

/* ==================== Event Handling ==================== */

ADSearchBar.prototype.handleFocus = function (event) {
  /**

  XXX: we can't actually gracefully prevent focus from happening at this point
       so we don't support the ADSearchBarShouldBeginEditing delegate method

  var should_begin_editing = true;
  // ask our if we ought to begin editing
  if (ADUtils.objectHasMethod(this.delegate, ADSearchBarShouldBeginEditing)) {
    should_begin_editing = this.delegate[ADSearchBarShouldBeginEditing](this);
  }
  // if not, blur
  if (!should_begin_editing) {
    this.field.blur();
    return;
  }
  */
  // otherwise, tell our delegate we began editing
  if (ADUtils.objectHasMethod(this.delegate, ADSearchBarTextDidBeginEditing)) {
    this.delegate[ADSearchBarTextDidBeginEditing](this);
  }
  //
  this.editing = true;
};

ADSearchBar.prototype.handleBlur = function (event) {
  /**

  XXX: we can't actually prevent blur from happening at this point
       so we don't support the ADSearchBarShouldEndEditing delegate method

  var should_end_editing = true;
  // ask our if we ought to keep editing
  if (ADUtils.objectHasMethod(this.delegate, ADSearchBarShouldEndEditing)) {
    should_end_editing = this.delegate[ADSearchBarShouldEndEditing](this);
  }
  // if not, retain focus
  if (!should_end_editing) {
    event.preventDefault();
    return;
  }
  */
  // otherwise, tell our delegate we're done editing
  if (ADUtils.objectHasMethod(this.delegate, ADSearchBarTextDidEndEditing)) {
    this.delegate[ADSearchBarTextDidEndEditing](this);
  }
  // and update our label to reflect the final field value
  this.label.innerText = this.field.value;
  // check for placeholder display
  this.checkForPlaceholderDisplay();
  //
  this.editing = false;
};

ADSearchBar.prototype.handleInput = function (event) {
  // check on need to display placeholder
  this.checkForPlaceholderDisplay();
  // tell our delegate we've got some new text
  if (ADUtils.objectHasMethod(this.delegate, ADSearchBarTextDidChange)) {
    this.delegate[ADSearchBarTextDidChange](this, this.field.value);
  }
};

ADSearchBar.prototype.handleSubmit = function (event) {
  // ensure we don't reload the page
  event.preventDefault();
  // and blur the field as we're done 
  this.editing = false;
};

ADSearchBar.prototype.handleControlTouchUpInside = function (event) {
  // cross button was activated
  if (event.control === this.emptyButton) {
    this.text = '';
    this.checkForPlaceholderDisplay();
  }
  // cancel button was tapped, tell our delegate about it
  else if (ADUtils.objectHasMethod(this.delegate, ADSearchBarCancelButtonClicked)) {
    this.delegate[ADSearchBarCancelButtonClicked](this);
  }
};

/* ==================== Class Creation ==================== */

ADClass(ADSearchBar);

/* ==================== Class Methods ==================== */

ADSearchBar.init = function () {
  ADUtils.preloadImageAsset('search/background_default.png');
  ADUtils.preloadImageAsset('search/cancel_touched.png');
};

window.addEventListener('load', ADSearchBar.init, false);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== ADPickerComponent ==================== */

ADPickerComponent.inherits = ADScrollView;

function ADPickerComponent (index, numberOfRows) {
  this.callSuper();
  // set us up as delegate to react to end of scrolling
  this.delegate = this;
  //
  this.index = index;
  this.rows = numberOfRows;
  // base style
  this.clipsToBounds = false;
  this.horizontalScrollEnabled = false;
  this.showsVerticalScrollIndicator = false;
};

/* ==================== Public APIs ==================== */

ADPickerComponent.prototype.getSelectedRow = function () {
  var current_index = (this.contentOffset.y + ADPickerViewComponentTopMargin - (ADPickerViewDefaultRowHeight * 2.25)) / ADPickerViewDefaultRowHeight;
  this._selectedRow = Math.min(Math.max(Math.round(current_index), 0), this.rows - 1);
  return this._selectedRow;
};

/* ==================== ADScrollView APIs ==================== */

ADPickerComponent.prototype.beginTracking = function (event) {
  this.callSuper(event);
  // record our y coordinate to figure out where we started scrolling
  var row = Math.floor((ADPoint.fromEventInElement(event, this.hostingLayer).y - ADPickerViewComponentTopMargin) / ADPickerViewDefaultRowHeight);
  this.touchedRow = (row >= 0 && row <= this.rows - 1) ? row : -1;
};

ADPickerComponent.prototype.touchesEnded = function (event) {
  var was_dragging = this.dragging;
  //
  this.callSuper(event);
  // only act at bubble phase and if we're not animating
  if (event.eventPhase == Event.BUBBLING_PHASE || this.decelerating) {
    return;
  }
  // if we were dragging, snap to nearest row
  if (was_dragging) {
    this.superview.selectRowInComponentAnimated(this.getSelectedRow(), this.index, true);
  }  
  // otherwise, see if we tapped a row that we can animate to
  else if (this.touchedRow != -1) {
    this.superview.selectRowInComponentAnimated(this.touchedRow, this.index, true);
  }
};

/* ==================== ADScrollViewDelegate APIs ==================== */

ADPickerComponent.prototype.scrollViewDidEndDecelerating = function (theScrollView) {
  this.superview.selectRowInComponentAnimated(this.getSelectedRow(), this.index, true);
};

ADClass(ADPickerComponent);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== ADPickerViewDataSource ==================== */

/**
 *  @class
 *
 *  <p>The ADPickerViewDataSource protocol is adopted by an object that mediates the application's data model for a
 *  {@link ADPickerView} object. The data source provides the <code>ADPickerView</code> object with the number of rows
 *  and components it needs to display the data in the picker view.</p>
 *
 *  @name ADPickerViewDataSource
 *  @since AdLib 0.1
 */

/**
 *  Asks the data source to return the number of components in the given picker view. Required.
 *
 *  @name numberOfComponentsInPickerView
 *  @function
 *
 *  @param {ADPickerView} pickerView An object representing the picker view requesting the data.
 *  @returns {int} The number of components (or "columns") that the picker view should display.
 *  @memberOf ADPickerViewDataSource.prototype
 */

/**
 *  Asks the data source to return the number of rows in the given component. Required.
 *
 *  @name pickerViewNumberOfRowsInComponent
 *  @function
 *
 *  @param {ADPickerView} pickerView An object representing the picker view requesting the data.
 *  @param {int} component An index number identifying a component of <code>pickerView</code>.
 *  @returns {int} The number of rows displayed by <code>component</code>.
 *  @memberOf ADPickerViewDataSource.prototype
 */

const ADPickerViewNumberOfComponents = 'numberOfComponentsInPickerView';
const ADPickerViewNumberOfRowsInComponent = 'pickerViewNumberOfRowsInComponent';

/* ==================== ADPickerViewDelegate ==================== */

/**
 *  @class
 *
 *  <p>The delegate of a {@link ADPickerView} object must adopt this protocol and implement at least some of its methods
 *  to provide the picker view with the data it needs to construct itself. The delegate implements the required methods of
 *  the protocol to return height, width, row title, and the view content for the rows in each component. It must also provide
 *  the content for each component's row, either as a string or a view. Typically the delegate implements other optional methods
 *  to respond to new selections or deselections of component rows.</p>
 *
 *  <p>Although they are marked optional, you must implement either {@link #pickerViewTitleForRowForComponent} or 
 *  {@link #pickerViewViewForRowForComponent} to provide the content of component rows.</p>
 *
 *  @name ADPickerViewDelegate
 *  @since AdLib 0.1
 */

/**
 *  Asks the delegate to return the row height to use for drawing row content. Optional.
 *
 *  @name pickerViewRowHeightForComponent
 *  @function
 *
 *  @param {ADPickerView} pickerView The picker view requesting this information.
 *  @param {int} component An index number identifying a component of <code>pickerView</code>.
 *  @returns {int} A value indicating the height of the row in pixels.
 *  @memberOf ADPickerViewDelegate.prototype
 */

/**
 *  Asks the delegate to return the row width to use for drawing row content. Optional.
 *
 *  @name pickerViewWidthForComponent
 *  @function
 *
 *  @param {ADPickerView} pickerView The picker view requesting this information.
 *  @param {int} component An index number identifying a component of <code>pickerView</code>.
 *  @returns {int} A value indicating the width of the row in pixels.
 *  @memberOf ADPickerViewDelegate.prototype
 */

/**
 *  Asks the delegate to return the title to use for a given row in a given component. Optional.
 *
 *  @name pickerViewTitleForRowForComponent
 *  @function
 *
 *  @param {ADPickerView} pickerView An object representing the picker view requesting the data.
 *  @param {int} row An index number identifying a row of <code>component</code>.
 *  @param {int} component An index number identifying a component of <code>pickerView</code>.
 *  @returns {String} The string to use as the title of the indicated component row.
 *  @memberOf ADPickerViewDelegate.prototype
 */

/**
 *  Asks the delegate to return the view to use for a given row in a given component. Optional.
 *
 *  @name pickerViewViewForRowForComponent
 *  @function
 *
 *  @param {ADPickerView} pickerView An object representing the picker view requesting the data.
 *  @param {int} row An index number identifying a row of <code>component</code>.
 *  @param {int} component An index number identifying a component of <code>pickerView</code>.
 *  @returns {ADView} A view object to use as the content of <code>row</code>.
 *  @memberOf ADPickerViewDelegate.prototype
 */

/**
 *  Asks the delegate to respond to the user selecting a specific row in a component of a picker view. Optional.
 *
 *  @name pickerViewDidSelectRowInComponent
 *  @function
 *
 *  @param {ADPickerView} pickerView The picker view in which a row just got selected.
 *  @param {int} row An index number identifying a row of <code>component</code>.
 *  @param {int} component An index number identifying a component of <code>pickerView</code>.
 *  @memberOf ADPickerViewDelegate.prototype
 */

const ADPickerViewRowHeightForComponent = 'pickerViewRowHeightForComponent';
const ADPickerViewWidthForComponent = 'pickerViewWidthForComponent';  
const ADPickerViewTitleForRowForComponent = 'pickerViewTitleForRowForComponent';  
const ADPickerViewViewForRowForComponent = 'pickerViewViewForRowForComponent';  
const ADPickerViewDidSelectRowInComponent = 'pickerViewDidSelectRowInComponent';

/* ==================== Constants ==================== */

/**
 *  The fixed width for a {@link ADPickerView}.
 *  @constant
 *  @type int
 */
ADPickerViewWidth = 320;
/**
 *  The fixed height for a {@link ADPickerView}.
 *  @constant
 *  @type int
 */
ADPickerViewHeight = 216;

ADPickerViewDefaultRowHeight = 39;
ADPickerViewComponentTopMargin = 89;
ADPickerViewComponentBottomMargin = 99;

/* ==================== ADPickerView ==================== */

ADPickerView.inherits = ADView;
ADPickerView.synthetizes = ['showsSelectionIndicator'];

/**
 *  @class
 *
 *  <p>The <code>ADPickerView</code> class implements objects that use a spinning-wheel or slot-machine motif to show one or more sets of
 *  values. Users select values by rotating the wheels until a row of values is aligned with a selection indicator.</p>
 *
 *  <p>A <code>ADPickerView</code> object is a potentially multidimensional user-interface element consisting of rows and components. A
 *  component is a wheel, which has a series of items (rows) at indexed locations on the wheel. Each component also has an indexed location
 *  (left to right) in a picker view. Each row on a component has content, which is either a string or a view object.</p>
 *
 *  <p>A <code>ADPickerView</code> object requires the cooperation of a delegate for constructing its components and a data source for providing
 *  the number of rows and components. The delegate must adopt the <code>{@link ADPickerViewDelegate}</code> protocol and implement the required
 *  methods to return the drawing rectangle for rows in each component. It also provides the content for each component's row, either as a string
 *  or a view, and it typically responds to new selections or deselections. The data source must adopt the <code>{@link ADPickerViewDataSource}</code>
 *  protocol and implement the required methods to return number of components and the number of rows in each component.</p>
 *
 *  <p>You can dynamically change the rows of one or all components by calling the {@link #reloadComponent} or {@link #reloadAllComponents} methods,   
 *  respectively. These methods cause <code>ADPickerView</code> to request the {@link #delegate} to provide new component and row data. You reload a
 *  picker view when a selected value in one component changes the set of values in another component; for example, changing a row value from February
 *  to March in one component would change a related component representing the days of the month.</p>
 *  
 *  @extends ADView
 *  @since AdLib 0.1
 */
function ADPickerView () {
  this.callSuper();
  /**
   *  The number of components displayed by the picker view, fetched from the {@link #delegate} and and cached. The default value is <code>0</code>.
   *  @type int
   */
  this.numberOfComponents = 0;
  /**
   *  The delegate for the picker view. The delegate must adopt the {@link ADPickerViewDelegate} protocol.
   *  @type Object
   */
  this.delegate = null;
  /**
   *  The data source for the receiver. The data source must adopt the {@link ADPickerViewDataSource} protocol.
   *  @type Object
   */
  this.dataSource = null;
  // private properties
  this._showsSelectionIndicator = false;
  this.components = [];
  this._size = new ADSize(ADPickerViewWidth, ADPickerViewHeight);
  //
  this.clipsToBounds = true;
};

/* ==================== ADView APIs ==================== */

ADPickerView.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-picker-view');
  //
  var frame = this.layer.appendChild(document.createElement('div'));
  frame.className = 'frame';
  // add background
  this.componentsBackground = frame.appendChild(document.createElement('div'));
  // add selection indicator background
  this.indicatorsBackground = frame.appendChild(document.createElement('div'));
  // add hosting layer
  this.hostingLayer = frame.appendChild(document.createElement('div'));
  // add shadow overlay
  this.shadowOverlay = frame.appendChild(document.createElement('div'));
  // add frame overlay
  this.frameOverlay = frame.appendChild(document.createElement('div'));
  // add selection
  this.indicatorsOverlay = frame.appendChild(document.createElement('div'));
};

ADPickerView.prototype.setSize = function (aSize) {
  // force default metrics
  aSize.width = ADPickerViewWidth;
  aSize.height = ADPickerViewHeight;
  this.callSuper(aSize);
};

/* ==================== Synthetized Properties ==================== */

/**
 *  @name ADPickerView.prototype
 *  @property {book} showsSelectionIndicator A Boolean value that determines whether the selection indicator is displayed. If the value of
 *  the property is <code>true</code>, the picker view shows a clear overlay across the current row. The default value of this property is
 *  <code>false</code>.
 */

ADPickerView.prototype.setShowsSelectionIndicator = function (showsSelectionIndicator) {
  this.layer.toggleClassName('shows-selection-indicator');
  this._showsSelectionIndicator = showsSelectionIndicator;
};

/* ==================== Getting Dimensions ==================== */

/**
 *  Returns the number of rows for a component. <code>ADPickerView</code> fetches the value of this property from the delegate and and
 *  caches it. The default value is <code>0</code>.
 *
 *  @param {int} component An index number identifying a component.
 *  @returns {int} The number of rows in the given component.
 */
ADPickerView.prototype.numberOfRowsInComponent = function (component) {
  if (component < 0 || component >= this.numberOfComponents) {
    throw(new Error('ADPickerView.numberOfRowsInComponent - no component ' + component));
  }
  return this.components[component].rows;
};

// ADPickerView.prototype.rowSizeForComponent = function (component) {
//   
// };

/* ==================== Reloading ==================== */

ADPickerView.prototype.updateNumberOfComponents = function () {
  if (ADUtils.objectHasMethod(this.dataSource, ADPickerViewNumberOfComponents)) {
    this.numberOfComponents = this.dataSource[ADPickerViewNumberOfComponents](this);
    if (this.numberOfComponents < 1) {
      throw(new Error('ADPickerView must have at least one component'));
    }
  }
  else {
    throw(new Error('the ADPickerView dataSource must implement ' + ADPickerViewNumberOfComponents));
  }
};

/**
 *  Reloads all components of the receiver. Calling this method causes the receiver to query the delegate for new data for all components.
 */
ADPickerView.prototype.reloadAllComponents = function () {
  this.updateNumberOfComponents();
  // remove old components
  for (var old_component = this.numberOfComponents; old_component < this.components.length; old_component++) {
    if (this.components[old_component] !== undefined) {
      this.components[old_component].removeFromSuperview();
      this.components[old_component] = undefined;
    }
  }
  // now update all components
  for (var component = 0; component < this.numberOfComponents; component++) {
    this.reloadComponent(component);
  }
  // now update layout of components
  this.updateComponentsLayout();
};

/**
 *  Reloads a particular component of the receiver. Calling this method causes the receiver to query the delegate for new data for the given component.
 *
 *  @param {int} component An index number identifying a component of the receiver.
 */
ADPickerView.prototype.reloadComponent = function (component) {
  // in case we don't have such a component yet, update our component count
  if (component > this.numberOfComponents) {
    this.updateNumberOfComponents();
  }
  // check that we have such a component
  if (component < 0 || component >= this.numberOfComponents) {
    throw(new Error('ADPickerView.reloadComponent was called with an invalid component : ' + component));
  }
  // check that we're told how many rows for this component
  if (ADUtils.objectHasMethod(this.dataSource, ADPickerViewNumberOfRowsInComponent)) {
    var rows = this.dataSource[ADPickerViewNumberOfRowsInComponent](this, component);
  }
  else {
    throw(new Error('the ADPickerView dataSource must implement ' + ADPickerViewNumberOfRowsInComponent));
  }
  // check that we have the right methods to populate this component
  var population_delegate = null;
  if (ADUtils.objectHasMethod(this.delegate, ADPickerViewTitleForRowForComponent)) {
    population_delegate = this.delegate[ADPickerViewTitleForRowForComponent];
  }
  else if (ADUtils.objectHasMethod(this.delegate, ADPickerViewViewForRowForComponent)) {
    population_delegate = this.delegate[ADPickerViewViewForRowForComponent];
  }
  else {
    throw(new Error('the ADPickerView delegate must implement either ' + ADPickerViewTitleForRowForComponent + ' or ' + ADPickerViewViewForRowForComponent));
  }
  // add this component if it does not exist yet
  if (this.components[component] === undefined) {
    this.components[component] = this.addSubview(new ADPickerComponent(component, rows));
  }
  // get a pointer to this component's view
  var view = this.components[component];
  // remove all content in this view
  view.hostingLayer.innerText = '';
  view.subviews = 0;
  // now add rows
  for (var i = 0; i < rows; i++) {
    var row = population_delegate.call(this.delegate, this, i, component);
    // dealing with a view
    if (row instanceof ADView) {
      view.addSubview(row);
    }
    // dealing with a string
    else {
      var element = document.createElement('div');
      element.className = 'title-row';
      element.innerText = row;
      view.hostingLayer.appendChild(element);
    }
  }
};

ADPickerView.prototype.updateComponentsLayout = function () {
  // first off, remove all backgrounds
  this.componentsBackground.innerText = '';
  this.indicatorsBackground.innerText = '';
  // now lay each component out and update the background sizes
  var base_width = 300 / this.numberOfComponents;
  var uses_custom_width = ADUtils.objectHasMethod(this.delegate, ADPickerViewWidthForComponent);
  var combined_width = 0;
  for (var i = 0; i < this.numberOfComponents; i++) {
    var view = this.components[i];
    var width = uses_custom_width ? this.delegate[ADPickerViewWidthForComponent](this, i) : base_width;
    view.position = new ADPoint(combined_width, 0);
    view.size = new ADSize(width, ADPickerViewHeight);
    view.contentSize = new ADSize(width, this.components[i].rows * ADPickerViewDefaultRowHeight + ADPickerViewComponentTopMargin + ADPickerViewComponentBottomMargin);
    combined_width += width;
    // add background and indicator
    this.componentsBackground.appendChild(document.createElement('div')).style.width = ADUtils.px(width);
    this.indicatorsBackground.appendChild(document.createElement('div')).style.width = ADUtils.px(width);
  }
  // now resize the containing frame
  this.frameOverlay.parentNode.style.width = ADUtils.px(combined_width + 20);
  this.frameOverlay.parentNode.style.left = ADUtils.px((320 - combined_width - 20) / 2);
};

/* ==================== Selecting Rows ==================== */

/**
 *  Selects a row in a specified component of the receiver.
 *
 *  @param {int} row An index number identifying a row of component.
 *  @param {int} component An index number identifying a component of the picker view.
 *  @param {bool} animated <code>true</code> to animate the selection by spinning the wheel (component) to the new value; if you specify
 *  <code>false</code>, the new selection is shown immediately.
 */
ADPickerView.prototype.selectRowInComponentAnimated = function (row, component, animated) {
  if (component < 0 || component >= this.numberOfComponents) {
    throw(new Error('ADPickerView.selectRowInComponentAnimated - no component ' + component));
  }
  if (row < 0 || row >= this.numberOfRowsInComponent(component)) {
    throw(new Error('ADPickerView.selectRowInComponentAnimated - no row ' + row + ' in component ' + component));
  }
  this.components[component].setContentOffsetWithAnimation(new ADPoint(0, row * ADPickerViewDefaultRowHeight), animated);
  // notify our delegate
  if (ADUtils.objectHasMethod(this.delegate, ADPickerViewDidSelectRowInComponent)) {
    this.delegate[ADPickerViewDidSelectRowInComponent](this, row, component);
  }
};

/**
 *  Returns the index of the selected row in a given component.
 *
 *  @param {int} component An index number identifying a component of the picker view.
 *  @returns {int} An index number identifying the selected row, or -1 if no row is selected.
 */
ADPickerView.prototype.selectedRowInComponent = function (component) {
  if (component < 0 || component >= this.numberOfComponents) {
    throw(new Error('ADPickerView.selectedRowInComponent - no component ' + component));
  }
  return this.components[component].getSelectedRow();
};

/* ==================== Accessing Rows ==================== */

ADPickerView.prototype.viewForRowForComponent = function () {
  
};

ADClass(ADPickerView);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  No button style.
 *  @constant
 *  @type String
 */
const ADButtonTypeCustom = 'custom-type';
/**
 *  A rounded-rectangle style button.
 *  @constant
 *  @type String
 */
const ADButtonTypeRoundedRect = 'rounded-rect-type';
/**
 *  A detail disclosure button.
 *  @constant
 *  @type String
 */
const ADButtonTypeDetailDisclosure = 'detail-disclosure-type';
/**
 *  An information button that has a light background.
 *  @constant
 *  @type String
 */
const ADButtonTypeInfoLight = 'info-light-type';
/**
 *  An information button that has a dark background.
 *  @constant
 *  @type String
 */
const ADButtonTypeInfoDark = 'info-dark-type';
/**
 *  A contact add button.
 *  @constant
 *  @type String
 */
const ADButtonTypeContactAdd = 'contact-add-type';

/**
 *  Default height in pixels of a {@link ADButton}.
 *  @constant
 *  @type int
 */
const ADButtonDefaultHeight = 37;

/* ==================== ADButton ==================== */

ADButton.inherits = ADControl;
ADButton.synthetizes = ['currentTitle', 'autosized'];

/**
 *  @class
 *
 *  <p>The ADButton class is a {@link ADControl} subclass that implements a button on the touch screen. A button
 *  intercepts touch events and sends a <code>ADControlTouchUpInsideEvent</code> when it's tapped. Methods for
 *  registering an event listener are inherited from {@link ADControl}. This class provides methods for setting the
 *  title in the various states of the control.</p>
 *
 *  <p>In order to further customize the appearance of a button, the user will use CSS style rules to control the
 *  appearance of borders, backgrounds, text sizes, padding, etc. See the {@link ADControl} documentation to learn
 *  more about reflecting control state visually by addressing custom CSS style rules.</p>
 *
 *  @extends ADControl
 *  @since AdLib 0.1
 *
 *  @param {String} type The button type from the <code>ADButtonType</code> family of constants.
 */
function ADButton (type) {
  this.type = type;
  //
  this.callSuper();
  //
  this._autosized = true;
  this.titles = [''];
  // sync the title to the state
  this.syncTitleToState();
  // and observe further state changes to sync the label
  this.addPropertyObserver('state', this, 'syncTitleToState');
};

/* ==================== ADView APIs ==================== */

ADButton.prototype.createLayer = function () {
  this.callSuper();
  //
  this.layer.addClassName('ad-button ' + this.type);
};

ADButton.prototype.getSize = function () {
  var style = window.getComputedStyle(this.layer);
  return new ADSize(parseInt(style.width), parseInt(style.height));
};

ADButton.prototype.setSize = function (aSize) {
  // we're no longer auto-sized
  this._autoSized = false;
  // perform default sizing
  this.callSuper(aSize);
};

/* ==================== Synthetized APIs ==================== */

/**
 *  @name ADButton.prototype
 *  @property {String} currentTitle The current title that is displayed on the button.
 *  @property {bool} autoSized Indicates whether the button sizes itself dynamically to best fit its title. Defaults to <code>true</code>.
 */

ADButton.prototype.getCurrentTitle = function () {
  return this.layer.innerText;
};

ADButton.prototype.setAutosized = function (autosized) {
  if (autosized) {
    this.layer.style.width = 'auto';
    this.layer.style.height = ADUtils.px(ADButtonDefaultHeight);
  }
  this._autosized = autosized
};

/* ==================== Public Methods ==================== */

/**
 *  Returns the title used for a state.
 *
 *  @param {int} state The state that uses the title. Possible values are the <code>ADControlState</code> set of constants.
 *  @returns {String} The title for the specified state..
 */
ADButton.prototype.titleForState = function (state) {
  return this.titles[state] || null;
};

/**
 *  Sets the title to use for the specified state. In general, if a property is not specified for a state, the default is to use the
 *  {@link ADControlStateNormal} value. If the value for {@link ADControlStateNormal} is not set, then the property defaults to a system value.
 *  Therefore, at a minimum, you should set the value for the normal state.
 *
 *  @param {String} title The title to use for the specified state.
 *  @param {int} state The state that uses the specified title. The values are described in the <code>ADControlState</code> set of constants.
 */
ADButton.prototype.setTitleForState = function (title, state) {
  this.titles[state] = title;
  // if we're already in that state, udpate the title
  if (state == this.state) {
    this.syncTitleToState();
  }
};

/* ==================== Private Methods ==================== */

ADButton.prototype.syncTitleToState = function () {
  if (!(this.type === ADButtonTypeRoundedRect || this.type === ADButtonTypeCustom)) {
    return;
  }
  this.layer.innerText = this.titles[this.state] || this.titles[ADControlStateNormal];
};

/* ==================== CLASS CREATION ==================== */

ADClass(ADButton);

/* ==================== IMAGE PRELOAD ==================== */

ADButton.init = function () {
  ADUtils.preloadImageAsset('tableview/UITableSelection.png');
};

window.addEventListener('load', ADButton.init, false);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

ADInterfaceOrientationPortrait           = 0;
ADInterfaceOrientationPortraitUpsideDown = 1;
ADInterfaceOrientationLandscapeLeft      = 2;
ADInterfaceOrientationLandscapeRight     = 3;

/**
 *  Default transition for a {@link ADViewController} that is being pushed onto the navigation stack.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewControllerTransitionInFromRight = {
  properties : ['transform'],
  from : ['translateX($width)'],
  to : ['translateX(0)']
};
/**
 *  Default transition for a {@link ADViewController} that was the back item of a navigation stack.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewControllerTransitionInFromLeft = {
  properties : ['transform'],
  from : ['translateX(-$width)'],
  to : ['translateX(0)']
};
/**
 *  Default transition for a {@link ADViewController} that is being popped off a navigation stack.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewControllerTransitionOutToRight = {
  properties : ['transform'],
  from : ['translateX(0)'],
  to : ['translateX($width)']
};
/**
 *  Default transition for a {@link ADViewController} that is becoming the back item of a navigation stack.
 *  @constant
 *  @type ADTransitionDefinition
 */
const ADViewControllerTransitionOutToLeft = {
  properties : ['transform'],
  from : ['translateX(0)'],
  to : ['translateX(-$width)']
};

/* ==================== ADViewController ==================== */

ADViewController.inherits = ADObject;
ADViewController.synthetizes = ['view', 'title', 'navigationItem', 'toolbarItems', 'navigationController', 'tabBarItem', 'tabBarController'];

/**
 *  @class
 *
 *  <p>The ADViewController class is very similar to UIViewController, more information forthcoming.</p>
 *
 *  @extends ADObject
 *  @since AdLib 0.2
 */
function ADViewController (viewArchiveURL) {
  this.callSuper();
  //
  this.viewArchiveURL = viewArchiveURL;
  // public properties
  this._view = null;
  this._title = '';
  this._tabBarItem = null;
  this._navigationItem = null;
  this._toolbarItems = null;
  this.hidesBottomBarWhenPushed = false;
  /**
   *  The controller for the active modal viewthat is, the view that is temporarily displayed on top of the view managed by the receiver. Typically, a
   *  modal view is used to present an edit page or additional details of a model object. The modal view is optionally displayed using a vertical sheet
   *  transition.
   *  @type ADViewController
   *  @private
   */
  this.modalViewController = null;
  /**
   *  The parent of the current view controller. Parent view controllers are relevant in navigation, tab bar, and modal view controller hierarchies. In each
   *  of these hierarchies, the parent is the object responsible for displaying the current view controller. If you are using a view controller as a
   *  standalone object  that is, not as part of a view controller hierarchy  the value in this property is <code>null</code>.
   *  @type ADViewController
   */
  this.parentViewController = null;
  /**
   *  The transition style to use when presenting the current view controller modally. This property affects the way the current view controller is presented
   *  when it is presented using the {@link #presentModalViewControllerAnimated} method. To change the transition type, you must set this property before
   *  presenting the view controller. The default value for this property is {@link ADModalTransitionStyleCoverVertical}.
   *  @type ADViewController
   *  @private
   */
  this.modalTransitionStyle = null;
  /**
   *  The search display controller associated with the view controller. If you create your search display controller programmatically, this property is set
   *  automatically by the search display controller when it is initialized.
   *  @type ADSearchDisplayController
   *  @private
   */
  this.searchDisplayController = null;
  // transitions
  // XXX: needs cleaning up, the names are too specific to ADNavigationController operations
  /**
   *  The transition to use when this item was the back item and is being transitioned to being the top item.
   *  Using the <code>null</code> value will prevent any animation from happening.
   *  Defaults to {@link ADViewControllerTransitionInFromLeft}.
   *  @type ADTransitionDefinition
   */
  this.wasBackItemTransition = ADViewControllerTransitionInFromLeft;
  /**
   *  The transition to use when this item was the top item and is being transitioned to being the back item.
   *  Using the <code>null</code> value will prevent any animation from happening.
   *  Defaults to {@link ADViewControllerTransitionOutToLeft}.
   *  @type ADTransitionDefinition
   */
  this.becomesBackItemTransition = ADViewControllerTransitionOutToLeft;
  /**
   *  The transition to use when this item was the top item and is being pushed off the navigation stack.
   *  Using the <code>null</code> value will prevent any animation from happening.
   *  Defaults to {@link ADViewControllerTransitionOutToRight}.
   *  @type ADTransitionDefinition
   */
  this.wasTopItemTransition = ADViewControllerTransitionOutToRight;
  /**
   *  The transition to use when this item is being pushed onto the navigation stack.
   *  Using the <code>null</code> value will prevent any animation from happening.
   *  Defaults to {@link ADViewControllerTransitionInFromRight}.
   *  @type ADTransitionDefinition
   */
  this.becomesTopItemTransition = ADViewControllerTransitionInFromRight;  //
  //
  this.interfaceOrientation = ADInterfaceOrientationPortrait;
};

/* ==================== Synthetized Properties ==================== */

ADViewController.prototype.parentControllerOfKind = function (kind) {
  var controller = this;
  var controller_of_kind = null;
  while (controller.parentViewController !== null) {
    if (controller instanceof kind) {
      controller_of_kind = controller;
      break;
    }
    controller = controller.parentViewController;
  }
  return controller_of_kind;
}

/**
 *  @name ADViewController.prototype
 *  @property {ADNavigationController} navigationController A parent or ancestor that is a navigation controller. Only returns a navigation controller if
 *  the view controller is in its stack. This property is <code>null</code> if a navigation controller cannot be found.
 */
ADViewController.prototype.getNavigationController = function () {
  return this.parentControllerOfKind(ADNavigationController);
};

/**
 *  @name ADViewController.prototype
 *  @property {ADTabBarController} tabBarController A parent or ancestor that is a tab bar controller. If the receiver is added to a tab bar controller,
 *  this property is the tab bar controller. If the receiver's navigation controller is added to a tab bar controller, this property is the navigation
 *  controller's tab bar controller. If no tab bar is present or the receiver is a modal view, this property is <code>null</code>.
 */
ADViewController.prototype.getTabBarController = function () {
  return this.parentControllerOfKind(ADTabBarController);
};

/* ==================== Managing the View ==================== */

/**
 *  @name ADViewController.prototype
 *  @property {ADView} view The view that the controller manages.
 *  
 *  <p>The view stored in this property represents the root view for the view controller's view hierarchy. Whenever you present the view controller on screen 
 *  (either modally or as part of view controllerbased interface), this view is retrieved and displayed in the application window. The default value of this
 *  property is <code>null</code>.</p>
 *
 *  <p>If you access this property and its value is currently <code>null</code>, the view controller automatically calls the {@link #loadView} method and
 *  returns the resulting view. You should override the {@link #loadView} method and use it to create the root view and all of its subviews.</p>
 *
 *  <p>Each view controller object is the sole owner of its view. You must not associate the same view object with multiple view controller objects.</p>
 *
 *  <p>Because accessing this property can cause the view to be loaded automatically, you can use the {@link #isViewLoaded} method to determine if the view
 *  is currently in memory. Unlike this property, the {@link #isViewLoaded} method does not force the loading of the view if it is not currently in
 *  memory.</p>
 */
ADViewController.prototype.getView = function () {
  // create the view if it's not set yet
  if (this._view === null) {
    this.loadView();
  }
  return this._view;
};

/**
 *  Creates the view that the controller manages.
 *
 *  <p>You should never call this method directly. The view controller calls this method when the view property is requested but is currently <code>null</code>.    
 *  If you create your views manually, you must override this method and use it to create your views.</p>
 *  
 *  <p>If you override this method in order to create your views manually, you should do so and assign the root view of your hierarchy to the view property. 
 *  (The views you create should be unique instances and should not be shared with any other view controller object.) Your custom implementation of this method
 *  should not {@link ADObject#callSuper}.</p>
 */
ADViewController.prototype.loadView = function () {
  // load HTML view if we have one set up
  if (this.viewArchiveURL) {
    this._view = new ADView();
    var loader = new ADHTMLFragmentLoader(this.viewArchiveURL, this);
  }
  else {
    // otherwise create a default view
    this.createDefaultView();
  }
};

ADViewController.prototype.createDefaultView = function () {
  this._view = new ADView();
  this.viewDidLoad();
};

/**
 *  Returns a Boolean value indicating whether the view is currently loaded into memory. Calling this method simply reports whether the view is loaded.
 *  Unlike the {@link #view} property, it does not attempt to load the view if it is not already in memory.
 *
 *  @returns {bool} Whether the view is currently loaded into memory
 */
ADViewController.prototype.isViewLoaded = function () {
  return (this._view != null);
};

ADViewController.prototype.viewDidLoad = function () {};

ADViewController.prototype.viewDidUnload = function () {};

/**
 *  @name ADViewController.prototype
 *  @property {String} title A string that represents the view that this controller manages. Subclasses should set the title to a human-readable string
 *  that represents the view to the user. If the receiver is a navigation controller, the default value is the top view controller's title.
 */
ADViewController.prototype.setTitle = function (title) {
  // record in ivar
  this._title = title;
  // see if we're managed by a navigation controller
  if (this.parentViewController instanceof ADTabBarController) {
    this.tabBarItem.title = title;
  }
  else if (this.parentViewController instanceof ADNavigationController) {
    this.navigationItem.title = title;
  }
};

/* ==================== Responding to View Events ==================== */

/**
 *  Notifies the view controller that its view is about to be added to a window.
 *
 *  <p>This method is called in response to a view being added either directly or indirectly to a window. You add views to a window directly by using the
 *  {@link ADView#addSubview} method to add them to the window's view hierarchy. You can also add a view indirectly in several ways, including by pushing it
 *  onto a navigation stack or by presenting it modally (using the {@link #presentModalViewControllerAnimated} method). This method is called before the view
 *  is actually added to the window and before any animations are configured.</p>
 *
 *  <p>You can override this method to perform custom tasks associated with presenting the view. For example, you might use this method to change the
 *  orientation or style of the status bar to coordinate with the orientation or style of the view being presented. If you override this method, you must
 *  call super at some point in your implementation.</p>
 *
 *  @param {bool} animated If <code>true</code>, the view is being added to the window using an animation.
 */
ADViewController.prototype.viewWillAppear = function (animated) {};

/**
 *  Notifies the view controller that its view was added to a window. You can override this method to perform additional tasks associated with presenting
 *  the view. If you override this method, you must call super at some point in your implementation.
 *
 *  @param {bool} animated If <code>true</code>, the view was added to the window using an animation.
 */
ADViewController.prototype.viewDidAppear = function (animated) {};

/**
 *  Notifies the view controller that its view is about to be dismissed, covered, or otherwise hidden from view.
 *
 *  <p>This method is called in response to a view being removed from its window or covered by another view. This method is called before the view is
 *  actually removed or covered and before any animations are configured.</p>
 *
 *  <p>Subclasses can override this method and use it to commit editing changes or perform other relevant tasks. For example, you might use this method to
 *  revert changes to the orientation or style of the status bar that were made in the {@link #viewDidDisappear} method when the view was first presented.
 *  If you override this method, you must call super at some point in your implementation.</p>
 *
 *  @param {bool} animated If <code>true</code>, the disappearance of the view is being animated.
 */
ADViewController.prototype.viewWillDisappear = function (animated) {};

/**
 *  Notifies the view controller that its view was dismissed, covered, or otherwise hidden from view. You can override this method to perform additional tasks
 *  associated with dismissing or hiding the view. If you override this method, you must call super at some point in your implementation.
 *
 *  @param {bool} animated If <code>true</code>, the disappearance of the view was animated.
 */
ADViewController.prototype.viewDidDisappear = function (animated) {};

/* ==================== Responding to View Rotation Events ==================== */

ADViewController.prototype.willRotateToInterfaceOrientation = function (orientation) {};

ADViewController.prototype.didRotateFromInterfaceOrientation = function (orientation) {};

/* ==================== Presenting Modal Views ==================== */

/**
 *  Presents a modal view managed by the given view controller to the user. Sets the {@link #modalViewController} property to the specified view controller.
 *  Resizes its view and attaches it to the view hierarchy. The view is animated according to the transition style specified in the
 *  {@link #modalTransitionStyle} property of the controller in the <code>modalViewController</code> parameter.
 *
 *  @param {ADViewController} modalViewController The view controller that manages the modal view.
 *  @param {bool} animated If <code>true</code>, animates the view as it's presented; otherwise, does not.
 */
ADViewController.prototype.presentModalViewControllerAnimated = function (modalViewController, animated) {
  this.modalViewController = modalViewController;
};

/**
 *  Dismisses the modal view controller that was presented by the receiver.
 *
 *  <p>The parent view controller is responsible for dismissing the modal view controller it presented using the {@link presentModalViewControllerAnimated}
 *  method. If you call this method on the modal view controller itself, however, the modal view controller automatically forwards the message to its parent
 *  view controller.</p>
 *
 *  <p>If you present several modal view controllers in succession, and thus build a stack of modal view controllers, calling this method on a view controller
 *  lower in the stack dismisses its immediate child view controller and all view controllers above that child on the stack. When this happens, only the
 *  top-most view is dismissed in an animated fashion; any intermediate view controllers are simply removed from the stack. The top-most view is dismissed
 *  using its modal transition style, which may differ from the styles used by other view controllers lower in the stack.</p>
 *
 *  <p>If you want to retain a reference to the receiver's modal view controller, get the value in the {@link #modalViewController} property before calling
 *  this method.</p>
 *
 *  @param {bool} animated If <code>true</code>, this method animates the view as it's dismissed; otherwise, it does not. The style of animation is determined
 *  by the value in the {@link #modalTransitionStyle} property of the view controller being dismissed.
 */
ADViewController.prototype.dismissModalViewControllerAnimated = function (animated) {
  this.modalViewController = null;
};

/* ==================== Configuring Navigation Items ==================== */

/**
 *  @name ADViewController.prototype
 *  @property {ADNavigationItem} navigationItem The navigation item used to represent the view controller. This is a unique instance of {@link ADNavigationItem}
 *  created to represent the view controller when it is pushed onto a navigation bar. The first time you access this property, the {@link ADNavigationItem} is
 *  created. Therefore, you shouldn't access this property if you are not using a navigation controller.
 */
ADViewController.prototype.getNavigationItem = function () {
  if (this._navigationItem === null) {
    this._navigationItem = new ADNavigationItem(this.title);
    this._navigationItem.viewController = this;
  }
  return this._navigationItem;
};

/* ==================== Configuring Tab Bar Items ==================== */

/**
 *  @name ADViewController.prototype
 *  @property {ADTabBarItem} tabBarItem The tab bar item that represents the view controller when added to a tab bar controller. The default value is a tab bar
 *  item that displays the view controller's title. The first time you access this property, the {@link ADTabBarItem} is created. Therefore, you shouldn't 
 *  access this property if you are not using a tab bar controller.
 */
ADViewController.prototype.getTabBarItem = function () {
  if (this._tabBarItem === null) {
    this._tabBarItem = new ADTabBarItem();
    this._tabBarItem.title = this.title;
    this._tabBarItem.viewController = this;
  }
  return this._tabBarItem;
};

/* ==================== Configuring the Navigation Controller's Toolbar ==================== */

/**
 *  @name ADViewController.prototype
 *  @property {Array} toolbarItems The toolbar items associated with the view controller. This property contains an array of {@link ADBarButtonItem} objects
 *  and works in conjunction with a {@link ADNavigationController} object. If this view controller is embedded inside a navigation controller interface, and
 *  the navigation controller displays a toolbar, this property identifies the items to display in that toolbar. You can set the value of this property
 *  explicitly or use the {@link #setToolbarItemsAnimated} method to animate changes to the visible set of toolbar items.
 */
ADViewController.prototype.setToolbarItems = function (items) {
  this.setToolbarItemsAnimated(items, false);
};

/**
 *  Sets the toolbar items to be displayed along with the view controller. View controllers that are managed by a navigation controller can use this method
 *  to specify toolbar items for the navigation controller's built-in toolbar. You can set the toolbar items for your view controller before your view
 *  controller is displayed or after it is already visible.
 *
 *  @param {Array} toolbarItems The toolbar items to display in a built-in toolbar.
 *  @param {bool} animated If <code>true</code>, animate the change of items in the toolbar.
 */
ADViewController.prototype.setToolbarItemsAnimated = function (items, animated) {
  this._toolbarItems = items;
  // update the toolbar items in our navigation controller's bar if we're managed by one
  var parent_controller = this.parentViewController;
  if (parent_controller !== null && parent_controller instanceof ADNavigationController) {
    parent_controller.toolbar.setItemsAnimated(items, animated);
  }
};

/* ==================== ADHTMLFragmentLoaderDelegate Implementation ==================== */

ADViewController.prototype.htmlFragmentLoaderDidLoad = function (loader) {
  var new_view = new ADContentView(loader.fragment);
  // let's see if a temporary layer was inserted while we were loading the content
  // and swap the newly-loaded one while sharing the style
  if (this._view !== null && this._view.superview) {
    var old_layer = this._view.layer;
    var new_layer = new_view.layer;
    new_layer.setAttribute('style', new_view.layer.style.cssText + old_layer.style.cssText);
    old_layer.parentNode.replaceChild(new_layer, old_layer);
  }
  this._view.layer = new_view.layer;
  new_view = null;
  //
  this.viewDidLoad();
};

ADViewController.prototype.htmlFragmentLoaderDidFail = function (loader) {
  console.warn('ADViewController  could not load view archive at URL ' + loader.url);
  this.createDefaultView();
};

ADClass(ADViewController);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

/**
 *  @class
 *
 *  The ADNavigationControllerDelegate protocol defines methods a navigation controller delegate can implement to change the behavior when view controllers
 *  are pushed and popped from the stack of a navigation controller.
 *
 *  @name ADNavigationControllerDelegate
 *  @since AdLib 0.2
 */

/**
 *  Sent to the receiver just before the navigation controller displays a view controller's view and navigation item properties.
 *
 *  @name navigationControllerWillShowViewControllerAnimated
 *  @function
 *
 *  @param {ADNavigationController} navigationController The navigation controller that is showing the properties of a view controller.
 *  @param {ADViewController} viewController The view controller whose view and navigation item properties are being shown.
 *  @param {bool} animated Indicates whether the view controller will become visible with an animation.
 *  @memberOf ADNavigationControllerDelegate.prototype
 */
const ADNavigationControllerWillShowViewController = 'navigationControllerWillShowViewControllerAnimated';  

/**
 *  Sent to the receiver just after the navigation controller displays a view controllers view and navigation item properties.
 *
 *  @name navigationControllerDidShowViewControllerAnimated
 *  @function
 *
 *  @param {ADNavigationController} navigationController The navigation controller that is showing the properties of a view controller.
 *  @param {ADViewController} viewController The view controller whose view and navigation item properties are being shown.
 *  @param {bool} animated Indicates whether the view controller became visible with an animation.
 *  @memberOf ADNavigationControllerDelegate.prototype
 */
const ADNavigationControllerDidShowViewController = 'navigationControllerDidShowViewControllerAnimated';  

/**
 *  Duration in seconds of the animation as the navigation bar is hidden or shown.
 *  @constant
 *  @type float
 */
const ADNavigationControllerHideShowBarDuration = 0.2;

/* ==================== ADNavigationController ==================== */

ADNavigationController.inherits = ADViewController;
ADNavigationController.synthetizes = ['viewControllers', 'topViewController', 'visibleViewController', 'navigationBarHidden', 'toolbarHidden'];

/**
 *  @class
 *
 *  <p>The ADNavigationController class is very similar to UINavigationController, more information forthcoming.</p>
 *
 *  @extends ADViewController
 *  @since AdLib 0.2
 */
function ADNavigationController (viewArchiveURL, rootViewController) {
  this.callSuper(viewArchiveURL);
  // public properties
  this.delegate = null;
  this._viewControllers = [];
  this._navigationBarHidden = false;
  this._toolbarHidden = true;
  this.navigationBar = new ADNavigationBar();
  this.toolbar = null;
  // private properties
  this.busy = false;
  this.requiresDeferredHostViewSizeUpdate = false;
  this.previousController = null;
  //
  if (rootViewController !== undefined) {
    this.viewControllers = [rootViewController];
  }
};

/* ==================== ADViewController APIs ==================== */

ADNavigationController.prototype.loadView = function () {
  if (this.isViewLoaded()) {
    return;
  }
  // container
  this._view = new ADNavigationView(this);
  this._view.layer.addClassName('ad-navigation-controller-view');
  this._view.clipsToBounds = true;
  // nav bar
  this._view.addSubview(this.navigationBar);
  this.navigationBar.autoresizingMask = ADViewAutoresizingFlexibleWidth;
  // host view
  this.hostView = this._view.addSubview(new ADView());
  this.hostView.layer.addClassName('ad-navigation-controller-host-view');
  // host view
  this.toolbar = this._view.addSubview(new ADToolbar());
  this.toolbar.autoresizingMask = ADViewAutoresizingFlexibleWidth;
  // size the container view
  this._view.addPropertyObserver('size', this, 'sizeChanged');
  this._view.size = new ADSize(window.innerWidth, window.innerHeight);
  this._view.autoresizingMask = ADViewAutoresizingFlexibleWidth | ADViewAutoresizingFlexibleHeight;
};

ADNavigationController.prototype.viewMovedToNewSuperview = function () {
  var top_item = this.navigationBar.topItem;
  if (top_item !== null) {
    top_item.updateLayoutIfTopItem();
  }
};

/* ==================== Sizing ==================== */

ADNavigationController.prototype.sizeChanged = function () {
  // compute position for bars and host view
  var navigation_bar_y = -ADNavigationBarHeight;
  var host_view_y = 0;
  var toolbar_y = this._view.size.height;
  if (!this._navigationBarHidden) {
    navigation_bar_y = 0;
    host_view_y += ADNavigationBarHeight;
  }
  if (!this._toolbarHidden) {
    toolbar_y -= ADToolbarHeight;
  }
  // position
  this.navigationBar.position = new ADPoint(0, navigation_bar_y);
  this.toolbar.position = new ADPoint(0, toolbar_y);
  this.hostView.position = new ADPoint(0, host_view_y);
  // size the host view
  this.updateHostViewSize();
};

ADNavigationController.prototype.updateHostViewSize = function () {
  var height = this._view.size.height;
  if (!this._navigationBarHidden) height -= ADNavigationBarHeight;
  if (!this._toolbarHidden) height -= ADToolbarHeight;
  this.hostView.size = new ADSize(this._view.size.width, height);
};

/* ==================== Accessing Items on the Navigation Stack ==================== */

/**
 *  @name ADNavigationController.prototype
 *  @property {ADViewController} topViewController The view controller at the top of the navigation stack.
 */
ADNavigationController.prototype.getTopViewController = function () {
  return (this._viewControllers.length > 0) ? this._viewControllers[this._viewControllers.length - 1] : null;
};

/**
 *  @name ADNavigationController.prototype
 *  @property {ADViewController} visibleViewController The view controller associated with the currently visible view in the navigation interface. The currently
 *  visible view can belong either to the view controller at the top of the navigation stack or to a view controller that was presented modally.
 */
ADNavigationController.prototype.getVisibleViewController = function () {
  var top_view_controller = this.topViewController;
  return top_view_controller.modalViewController || top_view_controller;
};

/**
 *  @name ADNavigationController.prototype
 *  @property {Array} viewControllers The view controllers currently on the navigation stack. The root view controller is at index <code>0</code> in the array,
 *  the back view controller is at index <code>n-2</code>, and the top controller is at index <code>n-1</code>, where <code>n</code> is the number of items in
 *  the array. Assigning a new array of view controllers to this property is equivalent to calling the {@link #setViewControllersAnimated} method with the
 *  <code>animated</code> parameter set to <code>false</code>.
 */
ADNavigationController.prototype.setViewControllers = function (controllers) {
  this.setViewControllersAnimated(controllers, false);
};

/**
 *  <p>Replaces the view controllers currently managed by the navigation controller with the specified items.</p>
 *
 *  <p>You can use this method to update or replace the current view controller stack without pushing or popping each controller explicitly. In addition,
 *  this method lets you update the set of controllers without animating the changes, which might be appropriate at launch time when you want to return the
 *  navigation controller to a previous state.</p>
 *
 *  <p>If animations are enabled, this method decides which type of transition to perform based on whether the last item in the items array is already in the
 *  navigation stack. If the view controller is currently in the stack, but is not the topmost item, this method uses a pop transition; if it is the topmost
 *  item, no transition is performed. If the view controller is not on the stack, this method uses a push transition. Only one transition is performed, but when
 *  that transition finishes, the entire contents of the stack are replaced with the new view controllers. For example, if controllers A, B, and C are on the
 *  stack and you set controllers D, A, and B, this method uses a pop transition and the resulting stack contains the controllers D, A, and B.</p>
 *
 *  @param {Array} viewControllers The view controllers to place in the stack. The front-to-back order of the controllers in this array represents the
 *  new bottom-to-top order of the controllers in the navigation stack. Thus, the last item added to the array becomes the top item of the navigation stack.
 *  @param {bool} animated If <code>true</code>, animate the pushing or popping of the top view controller. If <code>false</code>, replace the view
 *  controllers without any animations.
 */
ADNavigationController.prototype.setViewControllersAnimated = function (controllers, animated) {
  
  // do nothing if we're busy or are served an empty array
  if (this.busy || controllers.length == 0) {
    return;
  }

  // start a new transaction
  ADTransaction.begin();

  // make sure the view is loaded
  this.loadView();

  // get pointers to the relevant items
  var previous_controller = this.topViewController;
  var next_controller = controllers[controllers.length - 1];
  var controller_was_in_stack = (this._viewControllers.indexOf(next_controller) != -1)

  // trigger notifications and delegates
  if (previous_controller !== null) {
    previous_controller.viewWillDisappear(animated);
  }
  next_controller.viewWillAppear(animated);

  // dissociate the navigation controller from controllers in the current stack 
  for (var i = 0; i < this._viewControllers.length; i++) {
    this._viewControllers[i].parentViewController = null;
  }
  // and associate the navigation controller to controllers in the new stack 
  for (var i = 0; i < controllers.length; i++) {
    controllers[i].parentViewController = this;
  }

  // size the new view and keep it 100% of the hostView
  next_controller.view.size = this.hostView.size.copy();
  next_controller.view.autoresizingMask = ADViewAutoresizingFlexibleWidth | ADViewAutoresizingFlexibleHeight;
  // add it to the tree
  this.hostView.addSubview(next_controller.view);

  // there was no previous item, in this case we don't use a transition
  if (previous_controller === null) {
    // fire delegate letting us know we're moving to a new controller
    if (ADUtils.objectHasMethod(this.delegate, ADNavigationControllerWillShowViewController)) {
      this.delegate[ADNavigationControllerWillShowViewController](this, next_controller, animated);
    }
    // and notify that the view did appear
    next_controller.viewDidAppear();
  }
  // otherwise, we'll need to transition views
  else if (previous_controller !== next_controller) {
    // fire delegate letting us know we're moving to a new controller
    if (ADUtils.objectHasMethod(this.delegate, ADNavigationControllerWillShowViewController)) {
      this.delegate[ADNavigationControllerWillShowViewController](this, next_controller, animated);
    }
    // trigger the transition
    this.transitionToController(next_controller, previous_controller, !controller_was_in_stack, animated);
  }

  // remember our new stack of controllers
  this._viewControllers = controllers;

  // update the navigation bar
  var items = [];
  for (var i = 0; i < controllers.length; i++) {
    items.push(controllers[i].navigationItem);
  }
  this.navigationBar.setItemsAnimated(items, animated);

  // update the toolbar items with those from the newly added controller
  var toolbar_items = next_controller.toolbarItems;
  if (toolbar_items !== null) {
    this.toolbar.setItemsAnimated(toolbar_items, animated);
  }

  // animate things that need to animate
  ADTransaction.commit();
};

/* ==================== Pushing and Popping Stack Items ==================== */

/**
 *  Pushes a view controller onto the receivers stack and updates the display.
 *
 *  <p>The object in the <code>viewController</code> parameter becomes the top view controller on the navigation stack. Pushing a view controller results in
 *  the display of the view it manages. How that view is displayed is determined by the <code>animated</code> parameter. If the <code>animated</code> parameter
 *  is <code>true</code>, the view is animated into position; otherwise, the view is simply displayed in place. The view is automatically resized to fit between
 *  the navigation bar and toolbar (if present) before it is displayed.</p>
 *
 *  <p>In addition to displaying the view associated with the new view controller at the top of the stack, this method also updates the navigation bar and tool
 *  bar accordingly. The contents of the built-in navigation toolbar are updated to reflect the toolbar items of the new view controller.</p>
 *
 *  @param {ADViewController} viewController The view controller that is pushed onto the stack. This object cannot be an instance of tab bar controller and it
 *  must not already be on the navigation stack.
 *  @param {bool} animated Specify <code>true</code> to animate the transition or <code>false</code> if you do not want the transition to be animated. You might
 *  specify <code>false</code> if you are setting up the navigation controller at launch time.
 */
ADNavigationController.prototype.pushViewControllerAnimated = function (controller, animated) {
  this.setViewControllersAnimated(this._viewControllers.concat([controller]), animated);
};

/**
 *  Pops the top view controller from the navigation stack and updates the display.
 *
 *  <p>This method removes the top view controller from the stack and makes the new top of the stack the active view controller. If the view controller at the
 *  top of the stack is the root view controller, this method does nothing. In other words, you cannot pop the last item on the stack.</p>
 *
 *  <p>In addition to displaying the view associated with the new view controller at the top of the stack, this method also updates the navigation bar and tool
 *  bar accordingly. The contents of the built-in navigation toolbar are updated to reflect the toolbar items of the new view controller.</p>
 *
 *  @param {bool} animated Set this value to <code>true</code> to animate the transition. Pass <code>false</code> if you are setting up a navigation controller
 *  before its view is displayed.
 *
 *  @returns {ADViewController} The view controller that was popped from the stack.
 */
ADNavigationController.prototype.popViewControllerAnimated = function (animated) {
  // do nothing if we don't have enough controllers to pop back to
  if (this._viewControllers.length > 1) {
    var old_top_view_controller = this.topViewController;
    this.setViewControllersAnimated(this._viewControllers.slice(0, this._viewControllers.length - 1), animated);
    return old_top_view_controller;
  }
  return null;
};

/**
 *  Pops all the view controllers on the stack except the root view controller and updates the display.
 *
 *  @param {bool} animated Set this value to <code>true</code> to animate the transition. Pass <code>false</code> if you are setting up a navigation controller
 *  before its view is displayed.
 *
 *  @returns {Array} An array of view controllers that are popped from the stack.
 */
ADNavigationController.prototype.popToRootViewControllerAnimated = function (animated) {
  return this.popToViewControllerAnimated(this._viewControllers[0], animated);
};

/**
 *  Pops view controllers until the specified view controller is at the top of the navigation stack.
 *
 *  @param {ADViewController} viewController The view controller that you want to be at the top of the stack.
 *  @param {bool} animated Set this value to <code>true</code> to animate the transition. Pass <code>false</code> if you are setting up a navigation controller
 *  before its view is displayed.
 *
 *  @returns {Array} An array containing the view controllers that were popped from the stack.
 */
ADNavigationController.prototype.popToViewControllerAnimated = function (controller, animated) {
  // do nothing if we are busy transitioning already
  // or if the top controller is not in the stack
  var controller_index = this._viewControllers.indexOf(controller);
  // nothing to pop if we're dealing either with an unknown controller or the top controller
  if (controller_index < 0 || controller_index >= this._viewControllers.length - 1) {
    return [];
  }
  //
  var popped_view_controllers = this._viewControllers.slice(controller_index + 1);
  this.setViewControllersAnimated(this._viewControllers.slice(0, controller_index + 1), animated);
  return popped_view_controllers;
};

/* ==================== Transition ==================== */

ADNavigationController.prototype.transitionToController = function (top_controller, previous_controller, going_forward, animated) {
  // mark that a transition is now in progress
  this.busy = animated;
  // record some parameters that we will need at the end of the transition
  this.previousController = previous_controller;
  // apply them, will do nothing if they turn out to be null
  ADTransaction.defaults.duration = animated ? ADNavigationBarAnimationDuration : 0;
  if (previous_controller !== null) {
    previous_controller.view.applyTransition(going_forward ? previous_controller.becomesBackItemTransition : previous_controller.wasTopItemTransition, false);
  }
  top_controller.view.applyTransition(going_forward ? top_controller.becomesTopItemTransition : top_controller.wasBackItemTransition, false);
};

/* ==================== ADNavigationBarDelegate APIs ==================== */

ADNavigationController.prototype.transitionsEnded = function () {
  // see if we were animating
  var was_animated_transition = this.busy;
  // send notifications and tidy up
  if (this.previousController !== null) {
    // remove the previous controller from the view
    this.previousController.view.removeFromSuperview();
    this.previousController.viewDidDisappear(was_animated_transition);
  }
  this.topViewController.viewDidAppear(was_animated_transition);
  // mark that we're not busy anymore
  this.busy = false;
  // tell the delegate we moved to a new view controller
  if (ADUtils.objectHasMethod(this.delegate, ADNavigationControllerDidShowViewController)) {
    this.delegate[ADNavigationControllerDidShowViewController](this, this.topViewController, was_animated_transition);
  }
};

/* ==================== Configuring Navigation Bars ==================== */

ADNavigationController.prototype.setNavigationBarHidden = function (hidden) {
  this.setNavigationBarHiddenAnimated(hidden, false);
};

ADNavigationController.prototype.setNavigationBarHiddenAnimated = function (hidden, animated) {
  // nothing to do if we were already in that state
  if (this._navigationBarHidden == hidden) {
    return;
  }
  // save the new value to our ivar
  this._navigationBarHidden = hidden;
  // do nothing if the view is not loaded yet
  if (!this.isViewLoaded()) {
    return;
  }
  // create a new transaction
  ADTransaction.begin();
  ADTransaction.defaults.duration = animated ? ADNavigationControllerHideShowBarDuration : 0;
  ADTransaction.defaults.properties = ['position'];
  // animate the navigation bar
  new ADTransition({
    target : this.navigationBar,
    to : [new ADPoint(0, hidden ? -ADNavigationBarHeight : 0)]
  }).start();
  // animate the hosting view
  new ADTransition({
    target : this.hostView,
    to : [new ADPoint(0, hidden ? 0 : ADNavigationBarHeight)],
    delegate : this
  }).start();
  // run the transitions
  ADTransaction.commit();
  // and update size of the host view directly if we're hiding or we're not animated
  if (!animated || hidden) {
    this.updateHostViewSize();
  }
  else {
    this.requiresDeferredHostViewSizeUpdate = true;
  }
};

// called upon transition completion on .hostView in .setNavigationBarHiddenAnimated
// or .setToolbarHiddenAnimated ADTransitionDelegate API
ADNavigationController.prototype.transitionDidComplete = function (transition) {
  if (this.requiresDeferredHostViewSizeUpdate) {
    this.requiresDeferredHostViewSizeUpdate = false;
    this.updateHostViewSize();
  }
};

/* ==================== Configuring Custom Toolbars ==================== */

/**
 *  @name ADNavigationController.prototype
 *  @property {bool} toolbarHidden A Boolean indicating whether the navigation controllers built-in toolbar is visible.
 */
ADNavigationController.prototype.setToolbarHidden = function (hidden) {
  this.setToolbarHiddenAnimated(hidden, false);
};

/**
 *  Changes the visibility of the navigation controllers built-in toolbar. You can use this method to animate changes to the visibility of the built-in
 *  toolbar. Calling this method with the animated parameter set to <code>false</code> is equivalent to setting the value of the {@link #toolbarHidden} property
 *  directly. The toolbar simply appears or disappears depending on the value in the hidden parameter.
 *
 *  @param {bool} hidden Specify <code>true</code> to hide the toolbar or <code>false</code> to show it.
 *  @param {bool} animated Specify <code>true</code> if you want the toolbar to be animated on or off the screen.
 */
ADNavigationController.prototype.setToolbarHiddenAnimated = function (hidden, animated) {
  // nothing to do if we were already in that state
  if (this._toolbarHidden == hidden) {
    return;
  }
  // save the new value to our ivar
  this._toolbarHidden = hidden;
  // do nothing if the view is not loaded yet
  if (!this.isViewLoaded()) {
    return;
  }
  // animate the toolbar
  new ADTransition({
    target : this.toolbar,
    properties : ['position'],
    to : [new ADPoint(0, this._view.size.height - (hidden ? 0 : ADNavigationBarHeight))],
    duration : animated ? ADNavigationControllerHideShowBarDuration : 0,
    delegate : this
  }).start();
  // and update size of the host view directly if we're hiding or we're not animated
  if (!animated || hidden) {
    this.updateHostViewSize();
  }
  else {
    this.requiresDeferredHostViewSizeUpdate = true;
  }
};

ADClass(ADNavigationController);

/* ==================== ADNavigationView ==================== */

ADNavigationView.inherits = ADView;

function ADNavigationView (viewController) {
  this.callSuper();
  this.viewController = viewController;
};

ADNavigationView.prototype.didMoveToSuperview = function (superview) {
  this.callSuper(superview);
  if (superview !== null) {
    this.viewController.viewMovedToNewSuperview();
  }
};

ADClass(ADNavigationView);

/**
*
* Copyright  2009, 2010 Apple Inc.  All rights reserved.
*
**/

/* ==================== Constants ==================== */

const ADTabBarControllerShouldSelectViewController = 'tabBarControllerShouldSelectViewController';  
const ADTabBarControllerDidSelectViewController = 'tabBarControllerDidSelectViewController';  

const ADTabBarControllerMaxItems = 5;

/* ==================== ADTabBarController ==================== */

ADTabBarController.inherits = ADViewController;
ADTabBarController.synthetizes = ['viewControllers', 'selectedViewController', 'selectedIndex'];

function ADTabBarController (viewArchiveURL) {
  this.callSuper(viewArchiveURL);
  //
  this.delegate = null;
  this.tabBar = new ADTabBar();
  this._viewControllers = [];
  this.customizableViewControllers = [];
  this.moreNavigationController = null;
  // make ourselves the tab bar's delegate
  this.tabBar.delegate = this;
};

/* ==================== ADViewController APIs ==================== */

ADTabBarController.prototype.loadView = function () {
  if (this.isViewLoaded()) {
    return;
  }
  // container
  this._view = new ADView(this);
  this._view.layer.addClassName('ad-tab-bar-controller-view');
  this._view.size = new ADSize(window.innerWidth, window.innerHeight);
  this._view.autoresizingMask = ADViewAutoresizingFlexibleWidth | ADViewAutoresizingFlexibleHeight;
  // host view
  this.hostView = this._view.addSubview(new ADView());
  this.hostView.layer.addClassName('ad-tab-bar-controller-host-view');
  this.hostView.size = new ADSize(this._view.size.width, this._view.size.height - ADTabBarHeight);
  this.hostView.autoresizingMask = ADViewAutoresizingFlexibleWidth | ADViewAutoresizingFlexibleHeight;
  // tab bar
  this._view.addSubview(this.tabBar);
  this.tabBar.autoresizingMask = ADViewAutoresizingFlexibleWidth | ADViewAutoresizingFlexibleTopMargin;
};

/* ==================== ADTabBarDelegate APIs ==================== */

ADTabBarController.prototype.tabBarDidSelectItem = function (theTabBar, theSelectedItem) {
  // remove any controller already in place
  while (this.hostView.subviews.length) {
    this.hostView.subviews[0].removeFromSuperview();
  }
  // get a pointer to the view controller by the newly selected view controller
  var controller = this.selectedViewController;
  var view = controller.view;
  // send notification we're about to show this view
  controller.viewWillAppear(false);
  // size the new view and keep it 100% of its parent view
  view.size = this.hostView.size.copy();
  view.autoresizingMask = ADViewAutoresizingFlexibleWidth | ADViewAutoresizingFlexibleHeight;
  // add it to our host view
  this.hostView.addSubview(view);
  // notify our delegate
  if (ADUtils.objectHasMethod(this.delegate, ADTabBarControllerDidSelectViewController)) {
    this.delegate[ADTabBarControllerDidSelectViewController](this, controller);
  }
  // send notification we've shown this view
  controller.viewDidAppear(false);
};

/* ==================== Managing the View Controllers ==================== */

ADTabBarController.prototype.setViewControllers = function (controllers) {
  this.setViewControllersAnimated(controllers, false);
};

ADTabBarController.prototype.setViewControllersAnimated = function (controllers, animated) {
  // see if we need a More item in there
  if (controllers.length > ADTabBarControllerMaxItems) {
    
  }
  //
  this._viewControllers = controllers;
  // update the tab bar
  var items = [];
  for (var i = 0; i < controllers.length; i++) {
    items.push(controllers[i].tabBarItem);
  }
  this.tabBar.setItemsAnimated(items, animated);
};

/* ==================== Managing the Selected Tab ==================== */

ADTabBarController.prototype.getSelectedViewController = function () {
  var controller = null;
  var index_in_tab_bar = this.selectedIndex;
  // check that there is such an index in our list of view controllers
  if (index_in_tab_bar >= 0 && index_in_tab_bar < this._viewControllers.length) {
    controller = this._viewControllers[index_in_tab_bar];
  }
  return controller;
};

ADTabBarController.prototype.setSelectedViewController = function (controller) {
  // check that we have this controller within our list of view controllers
  if (this._viewControllers.indexOf(controller) != -1) {
    this.tabBar.selectedItem = controller.tabBarItem;
  }
};

ADTabBarController.prototype.getSelectedIndex = function () {
  return this.tabBar.items.indexOf(this.tabBar.selectedItem);
};

ADTabBarController.prototype.setSelectedIndex = function (index) {
  // check that there is such an index in our list of view controllers
  if (index >= 0 && index < this._viewControllers.length) {
    this.tabBar.selectedItem = this._viewControllers[index];
  }
};

ADClass(ADTabBarController);